

/*
  cc -g -static -o ../bin/raedchen ../src/raedchen.c
*/

/* This is actually the program version of cd_backup_planer.c  */
#define PROG_VERSION "0.8.8"

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/select.h>

/*
*/
#define Raedchen_with_fifO 1


static char raedchen[]= {"|/-\\"};

#define Raedchen_bufsizE 4096

static int abort_run= 0;

void Cleanup_handler_pipe()
{
 abort_run= 1;
 return;
}


double Scanf_io_size(text,flag)
char *text;
int flag;
/*
 bit0= default value -1 rather than 0
*/
{
 int c;
 double ret= 0.0;

 if(flag&1)
   ret= -1.0;
 if(text[0]==0)
   return(ret);
 sscanf(text,"%lf",&ret);
 c= text[strlen(text)-1];
 if(c=='k' || c=='K') ret*= 1024.0;
 if(c=='m' || c=='M') ret*= 1024.0*1024.0;
 if(c=='g' || c=='G') ret*= 1024.0*1024.0*1024.0;
 if(c=='t' || c=='T') ret*= 1024.0*1024.0*1024.0*1024.0;
 if(c=='p' || c=='P') ret*= 1024.0*1024.0*1024.0*1024.0*1024.0;
 if(c=='e' || c=='E') ret*= 1024.0*1024.0*1024.0*1024.0*1024.0*1024.0;
 return(ret);
}


int Pipe_close(zielfp,command,flag)
FILE **zielfp;
char *command;
int flag;
/*
flag:
 bit0= execute command even if pipe seems to have failed
return:
 0= ok
 1= pclose() returned failure indicating exit status
 2= system() returned failure indicating exit status
 3= both indicate failure
*/
{
 int pret= 0,sret= 0;

 if((*zielfp)==NULL)
   return(0);
 fflush(*zielfp);
 pret= pclose(*zielfp);
 *zielfp= NULL;
 if(pret==0 || flag&1) {
   if(command[0]!=0 && !(command[0]=='-' && command[1]==0))
     sret= system(command);
 }
 return((pret!=0)|((sret!=0)<<1));
}


int Sfile_scale(value,result,siz,thresh,flag)
double value;
char *result;
int siz;
double thresh;
int flag;
/*
 bit0= eventually ommit 'b'
 bit1= make text as short as possible
 bit2= no fraction (if it would fit at all)
*/
{
 char scale_c,scales[6],form[80];
 int i,dec_siz= 0,avail_siz= 1;

 strcpy(scales,"bkmgtp");
 scale_c= scales[0];
 for(i=1;scales[i]!=0;i++) {
   if(value<thresh) 
 break;
   value/= 1024;
   scale_c= scales[i];
 } 
 if(scale_c!='b' && !(flag&4)) { /* is there room for fractional part ? */
   avail_siz= siz-1;
   sprintf(form,"%%.lf");
   sprintf(result,"%.lf",value);
   if(strlen(result)<avail_siz-2) 
     dec_siz= 1;                                  /* we are very modest */
 }
 if(scale_c=='b' && (flag&1)) {
   if(flag&2)
     sprintf(form,"%%.lf");
   else
     sprintf(form,"%%%d.lf",siz);
   sprintf(result,form,value);
 } else {
   if(flag&2)
     sprintf(form,"%%.lf%%c");
   else if(dec_siz>0) 
     sprintf(form,"%%%d.%dlf%%c",avail_siz,dec_siz);
   else
     sprintf(form,"%%%d.lf%%c",siz-1);
   sprintf(result,form,value,scale_c);
 }
 return(1);
}


/* --------------------------------------------------------------------- */
/* CdrfifO is a spinoff of cdrskin, another project of scdbackup@gmx.net */


/* Macro for creation of arrays of objects (or single objects) */
#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));


#define Cdrfifo_buffer_chunK 2048

/* 1= enable , 0= disable status messages to stderr */
static int Cdrfifo_debuG= 0;


struct CdrfifO {
 int chunk_size;

 int source_fd;
 double in_counter;

 char *buffer;
 int buffer_size;
 int buffer_is_full;
 int write_idx;
 int read_idx;
  
 int dest_fd;
 double out_counter;

 struct timeval start_time;
 double speed_limit;

 /* statistics */
 double interval_counter;
 struct timeval interval_start_time;
 double interval_start_counter;
 int total_min_fill;
 int interval_min_fill;

 double put_counter;
 double get_counter;
 double empty_counter;
 double full_counter;
};


/** Create a fifo object.
    @param ff Returns the address of the new object.
    @param source_fd Filedescriptor opened to a readable data stream.
    @param destination_fd Filedescriptor opened to a writable data stream.
                          To work with libburn, it needs to be attached to a
                          struct burn_source object.
    @param chunk_size size of buffer block for a single transaction (0=default)
    @param buffer_size size of fifo buffer
    @param flag unused yet
    @return 1 on success, <=0 on failure
*/
int Cdrfifo_new(ff,source_fd,destination_fd,chunk_size,buffer_size,flag)
struct CdrfifO **ff;
int source_fd;
int destination_fd;
int chunk_size;
int buffer_size;
int flag;
{
 struct CdrfifO *o;
 struct timezone tz;

 (*ff)= o= TSOB_FELD(struct CdrfifO,1);
 if(o==NULL)
   return(-1);
 if(chunk_size<=0)
   chunk_size= Cdrfifo_buffer_chunK;
 o->chunk_size= chunk_size;
 if(buffer_size%chunk_size)
   buffer_size+= chunk_size-(buffer_size%chunk_size);
 o->source_fd= source_fd;
 o->in_counter= 0.0;
 o->buffer= NULL;
 o->buffer_is_full= 0;
 o->buffer_size= buffer_size;
 o->write_idx= 0;
 o->read_idx= 0;
 o->dest_fd= destination_fd;
 o->out_counter= 0.0;
 memset(&(o->start_time),0,sizeof(o->start_time));
 gettimeofday(&(o->start_time),&tz);
 o->speed_limit= 0.0;
 o->interval_counter= 0.0;
 memset(&(o->interval_start_time),0,sizeof(o->interval_start_time));
 gettimeofday(&(o->interval_start_time),&tz);
 o->interval_start_counter= 0.0;
 o->total_min_fill= buffer_size;
 o->interval_min_fill= buffer_size;
 o->put_counter= 0.0;
 o->get_counter= 0.0;
 o->empty_counter= 0.0;
 o->full_counter= 0.0;
 o->buffer= TSOB_FELD(char,buffer_size);
 if(o->buffer==NULL)
   goto failed;
 return(1);
failed:;
 Cdrfifo_destroy(ff,0);
 return(-1);
}


/** Release from memory a fifo object previously created by Cdrfifo_new().
    @param ff the victim (gets returned as NULL, call can stand *ff==NULL))
    @param flag Bitfield for control purposes: 
                bit0= do not close destination fd
*/
int Cdrfifo_destroy(ff,flag)
struct CdrfifO **ff;
int flag;
/* flag
 bit0= do not close destination fd
*/
{
 struct CdrfifO *o;

 o= *ff;
 if(o==NULL)
   return(0);
 if(o->dest_fd!=-1 && !(flag&1))
   close(o->dest_fd);
 if(o->buffer!=NULL)
   free((char *) o->buffer);
 free((char *) o);
 (*ff)= NULL;
 return(1);
}


int Cdrfifo_get_sizes(o,chunk_size,buffer_size,flag)
struct CdrfifO *o;
int *chunk_size;
int *buffer_size;
int flag;
{
 *chunk_size= o->chunk_size;
 *buffer_size= o->buffer_size;
 return(1);
}


int Cdrfifo_set_speed_limit(o,bytes_per_second,flag)
struct CdrfifO *o;
double bytes_per_second;
int flag;
{
 o->speed_limit= bytes_per_second;
 return(1);
}


int Cdrfifo_set_fds(o,source_fd,dest_fd,flag)
struct CdrfifO *o;
int source_fd;
int dest_fd;
int flag;
{
 o->source_fd= source_fd;
 o->dest_fd= dest_fd;
 return(1);
}


int Cdrfifo_get_fds(o,source_fd,dest_fd,flag)
struct CdrfifO *o;
int *source_fd;
int *dest_fd;
int flag;
{
 *source_fd= o->source_fd;
 *dest_fd= o->dest_fd;
 return(1);
}


static int Cdrfifo_tell_buffer_space(o,flag)
struct CdrfifO *o;
int flag;
{
 if(o->buffer_is_full)
   return(0);
 if(o->write_idx>=o->read_idx)
   return((o->buffer_size - o->write_idx) + o->read_idx);
 return(o->read_idx - o->write_idx);
}


/** Obtain buffer state.
    @param o The buffer object
    @param fill Returns the number of pending payload bytes in the buffer
    @param space Returns the number of unused buffer bytes
    @param flag unused yet
    @return -1=error , 0=inactive , 1=reading and writing ,
            2=reading ended (but still writing)
*/
int Cdrfifo_get_buffer_state(o,fill,space,flag)
struct CdrfifO *o;
int *fill;
int *space;
int flag;
/* return : 
 -1=error
  0=inactive  
  1=reading and writing
  2=reading ended, still writing
*/
{
 *space= Cdrfifo_tell_buffer_space(o,0);
 *fill= o->buffer_size-(*space);
 if(o->dest_fd==-1)
   return(0);
 if(o->source_fd<0)
   return(2);
 return(1);
}


int Cdrfifo_get_counters(o,in_counter,out_counter,flag)
struct CdrfifO *o;
double *in_counter;
double *out_counter;
int flag;
{
 *in_counter= o->in_counter;
 *out_counter= o->out_counter;
 return(1);
}


/** reads min_fill and begins measurement interval for next min_fill */
int Cdrfifo_next_interval(o,min_fill,flag)
struct CdrfifO *o;
int *min_fill;
int flag;
{
 struct timezone tz;
 
 o->interval_counter++;
 gettimeofday(&(o->interval_start_time),&tz);
 o->interval_start_counter= o->out_counter;
 *min_fill= o->interval_min_fill;
 o->interval_min_fill= o->buffer_size - Cdrfifo_tell_buffer_space(o,0);
 return(1);
}


int Cdrfifo_get_min_fill(o,total_min_fill,interval_min_fill,flag)
struct CdrfifO *o;
int *total_min_fill;
int *interval_min_fill;
int flag;
{
 *total_min_fill= o->total_min_fill;
 *interval_min_fill= o->interval_min_fill;
 return(1);
}


int Cdrfifo_get_cdr_counters(o,put_counter,get_counter,empty_counter,
                             full_counter,flag)
struct CdrfifO *o;
double *put_counter;
double *get_counter;
double *empty_counter;
double *full_counter;
int flag;
{
 *put_counter= o->put_counter;;
 *get_counter= o->get_counter;
 *empty_counter= o->empty_counter;
 *full_counter= o->full_counter;
 return(1);
}


/** Check for pending data at the fifo's source file descriptor and wether the
    fifo is ready to take them. Simultaneously check the buffer for existing
    data and the destination fd for readiness to accept some. If so, a small
    chunk of data is transfered to and/or from the fifo. 
    The check and transactions are repeated until a given timespan has elapsed.
    Applications call this function in the burn loop instead of sleep().
    @param o The fifo object
    @param wait_usec The time in microseconds after which the function shall
                     return.
    @param reply_buffer with bit2: returns write-ready buffer chunk and must
                                   be able to take at least chunk_size bytes
    @param reply_count with bit2: returns number of writeable bytes in reply_pt
    @param flag Bitfield for control purposes: 
                bit0= enable debug pacifier (same with Cdrfifo_debuG)
                bit1= do not write, just fill buffer
                bit2= rather than writing a chunk: return it and its size
                bit3= with bit2: do not check destination fd for readiness
    @return <0 = error , 0 = idle , 1 = did some work , 2 = all work is done
*/
int Cdrfifo_try_to_work(o,wait_usec,reply_buffer,reply_count,flag)
struct CdrfifO *o;
int wait_usec; 
char *reply_buffer;
int *reply_count;
int flag;
/* flag:
 bit0= enable debug pacifier (same with Cdrfifo_debuG)
 bit1= do not write, just fill buffer
 bit2= rather than writing a chunk: return it and its size in reply_*
 bit3= with bit2: do not check destination fd for readiness
return: <0 = error , 0 = idle , 1 = did some work , 2 = all work is done
*/
{
 struct timeval wt,start_tv,current_tv;
 struct timezone tz;
 fd_set rds,wts,exs;
 int ready,ret,max_fd= -1,buffer_space,buffer_fill,can_write,can_read;
 int did_work= 0,elapsed,still_to_wait,speed_limiter= 0,ready_to_write;
 int dummy;
 double start_out_counter,diff_time,min_wait_time,diff_counter,limit;

 start_out_counter= o->out_counter;
 gettimeofday(&start_tv,&tz);
 still_to_wait= wait_usec;
 if(flag&4)
   *reply_count= 0;

try_again:; 
 if(o->dest_fd==-1)
   return(2);
 FD_ZERO(&rds);
 FD_ZERO(&wts);
 FD_ZERO(&exs);
 buffer_space= Cdrfifo_tell_buffer_space(o,0);
 buffer_fill= o->buffer_size-buffer_space;
 if(buffer_space>=o->buffer_size && o->source_fd==-1) { /* work is done */
   if(o->dest_fd!=-1 && !(flag&4))
     close(o->dest_fd);
   o->dest_fd= -1;
   if((flag&1)||Cdrfifo_debuG)
     fprintf(stderr,
            "\ncdrfifo_debug:  w=%d r=%d | b=%d s=%d | i=%.lf o=%.lf (done)\n",
            o->write_idx,o->read_idx,buffer_fill,buffer_space,
            o->in_counter,o->out_counter);
   return(2);
 }
 if(o->interval_counter>0) {
   if(o->total_min_fill>buffer_fill && o->source_fd>=0)
     o->total_min_fill= buffer_fill;
   if(o->interval_min_fill>buffer_fill)
     o->interval_min_fill= buffer_fill;
 }
 if(o->speed_limit!=0) {
   speed_limiter= 0;
   gettimeofday(&current_tv,&tz);
   if(o->speed_limit>0) {
     diff_time= ((double) current_tv.tv_sec)-((double) o->start_time.tv_sec)+ 
         (((double) current_tv.tv_usec)-((double) o->start_time.tv_usec))*1e-6;
     diff_counter= o->out_counter;
     limit= o->speed_limit;
   } else if(flag&4) {
     if(o->interval_start_time.tv_sec==0)
       o->interval_start_time= start_tv;
     diff_time= ((double) current_tv.tv_sec) 
              - ((double) o->interval_start_time.tv_sec)
              + (((double) current_tv.tv_usec) 
                 -((double) o->interval_start_time.tv_usec))*1e-6;
     diff_counter= o->out_counter - o->interval_start_counter;
     limit= -o->speed_limit;
   } else {
     diff_time= ((double) current_tv.tv_sec) - ((double) start_tv.tv_sec)
              + (((double) current_tv.tv_usec) 
                 -((double)start_tv.tv_usec))*1e-6;
     diff_counter= o->out_counter - start_out_counter;
     limit= -o->speed_limit;
   }
   if(diff_time>0.0)
     if(diff_counter/diff_time>limit) {
       min_wait_time= (diff_counter/limit - diff_time)*1.0e6;
       if(min_wait_time<still_to_wait)
         still_to_wait= min_wait_time;
       if(still_to_wait>0)
         speed_limiter= 1;
     }
 }

 max_fd= -1; 
 if(o->source_fd>=0) {
   if(buffer_space>0) {
     FD_SET((o->source_fd),&rds);
     if(max_fd<o->source_fd)
       max_fd= o->source_fd;
   } else if(o->interval_counter>0)
     o->full_counter++;
 }
 ready_to_write= 0;
 if(o->dest_fd>=0 && (!(flag&2)) && !speed_limiter) {
   if(buffer_fill>=o->chunk_size || o->source_fd<0) {
     if((flag&(4|8))==(4|8)) {
       still_to_wait= 0;
       ready_to_write= 1;
     } else {
       FD_SET((o->dest_fd),&wts);
       if(max_fd<o->dest_fd)
         max_fd= o->dest_fd;
     }
   } else if(o->interval_counter>0)
     o->empty_counter++;
 }

 if(still_to_wait>0 || max_fd>=0) {
   wt.tv_sec=  still_to_wait/1000000;
   wt.tv_usec= still_to_wait%1000000;
   ready= select(max_fd+1,&rds,&wts,&exs,&wt);
 } else
   ready= 0;
 if(ready<=0) {
   if(!ready_to_write)
     goto check_wether_done;
   FD_ZERO(&rds);
 }
 if(ready_to_write)
   FD_SET((o->dest_fd),&wts);
 if(o->dest_fd>=0) if(FD_ISSET((o->dest_fd),&wts)) {
   can_write= o->buffer_size - buffer_space;
   if(can_write>o->chunk_size)
     can_write= o->chunk_size;
   if(o->read_idx+can_write > o->buffer_size)
     can_write= o->buffer_size - o->read_idx;
   if(flag&4) {
     memcpy(reply_buffer,o->buffer+o->read_idx,can_write);
     *reply_count= ret= can_write;
   } else {
     ret= write(o->dest_fd,o->buffer+o->read_idx,can_write);
   }
   if(ret==-1) {

     /* >>> handle broken pipe */;
     fprintf(stderr,"\ncdrfifo: on write: errno=%d , \"%s\"\n",errno,
                    errno==0?"-no error code available-":strerror(errno));

     if(!(flag&4))
       close(o->dest_fd);
     o->dest_fd= -1;
     {ret= -1; goto ex;}
   }
   did_work= 1;
   o->get_counter++;
   o->out_counter+= ret;
   o->read_idx+= ret;
   if(o->read_idx>=o->buffer_size)
     o->read_idx= 0;
   o->buffer_is_full= 0;
 }
 if(o->source_fd>=0) if(FD_ISSET((o->source_fd),&rds)) {
   can_read= o->buffer_size - o->write_idx;
   if(can_read>o->chunk_size)
     can_read= o->chunk_size;
   if(o->write_idx<o->read_idx && o->write_idx+can_read > o->read_idx)
     can_read= o->read_idx - o->write_idx;
   ret= read(o->source_fd,o->buffer+o->write_idx,can_read);
   if(ret==-1) {

     /* >>> handle input error */;
     fprintf(stderr,"\ncdrfifo: on read: errno=%d , \"%s\"\n",errno,
                    errno==0?"-no error code available-":strerror(errno));

     o->source_fd= -1;
   } else if(ret==0) { /* eof */
     if(Cdrfifo_debuG)
       fprintf(stderr,"\ncdrfifo: on read(%d,buffer,%d): eof\n",
               o->source_fd,can_read);
     o->source_fd= -1;
   } else {
     did_work= 1;
     o->put_counter++;
     o->in_counter+= ret;
     o->write_idx+= ret;
     if(o->write_idx>=o->buffer_size)
       o->write_idx= 0;
     if(o->write_idx==o->read_idx)
       o->buffer_is_full= 1;
   }
 }

check_wether_done:;
 if((flag&4) && *reply_count>0)
   {ret= 1; goto ex;}
 gettimeofday(&current_tv,&tz);
 elapsed= (current_tv.tv_sec-start_tv.tv_sec)*1000000 + 
          (((int) current_tv.tv_usec) - ((int) start_tv.tv_usec));
 still_to_wait= wait_usec-elapsed;
 if(still_to_wait>0)
   goto try_again;
 ret= !!did_work;
ex:;
 if(flag&4) {
   gettimeofday(&current_tv,&tz);
   elapsed= (current_tv.tv_sec - o->interval_start_time.tv_sec)*1000000
          + (((int) current_tv.tv_usec) 
              - ((int) o->interval_start_time.tv_usec));
 } else
   elapsed= wait_usec;
 if(elapsed>=wait_usec) {
   if((flag&1)||Cdrfifo_debuG) {
     buffer_space= Cdrfifo_tell_buffer_space(o,0);
     fprintf(stderr,
             "\ncdrfifo_debug:  w=%d r=%d | b=%d s=%d | i=%.lf o=%.lf\n",
             o->write_idx,o->read_idx,
             o->buffer_size-buffer_space,buffer_space,
             o->in_counter,o->out_counter);
   }
   if(flag&4)
     Cdrfifo_next_interval(o,&dummy,0);
 }
 return(ret);
}


/** Fill the fifo as far as possible without writing to destination fd */
int Cdrfifo_fill(o,flag)
struct CdrfifO *o;
int flag;
{
 int ret,fill= 0,space,state;

 while(1) {
   state= Cdrfifo_get_buffer_state(o,&fill,&space,0);
   if(state==-1) {

     /* >>> handle error */;

     return(0);
   } else if(state!=1)
 break;
   if(space<=0)
 break;
   ret= Cdrfifo_try_to_work(o,100000,NULL,NULL,2);
   if(ret<0) {

     /* >>> handle error */;

     return(0);
   }
   if(ret==2)
 break;
 }
 o->total_min_fill= fill;
 o->interval_min_fill= fill;
 return(1);
}


int Raedchen_read_fifo(ff,fifo_interval,buffer,size,gelesen,flag)
struct CdrfifO *ff;
double fifo_interval;
char *buffer;
int size;
int *gelesen;
int flag;
/*
 bit0= fifo with verbous messages
 bit3= do not check destination fd for readiness
*/
{
 static char changebuf[Raedchen_bufsizE];
 static int changebuf_fill= 0, changebuf_pos= 0;
 int can_copy,readin,fifo_bs,fifo_size,ret;
 char *rbuf;

 *gelesen= 0;
 if(changebuf_fill>changebuf_pos) {
read_changebuf:;
   can_copy= size;
   if(changebuf_fill-changebuf_pos<can_copy)
     can_copy= changebuf_fill-changebuf_pos;
   memcpy(buffer,changebuf+changebuf_pos,can_copy);
   *gelesen= can_copy;
   changebuf_pos+= can_copy;
   return(1);
 }
 Cdrfifo_get_sizes(ff,&fifo_bs,&fifo_size,0);
 if(size<fifo_bs) {
   rbuf= changebuf;
   changebuf_fill= changebuf_pos= 0;
 } else {
   rbuf= buffer;
 }
 while(1) {
   ret= Cdrfifo_try_to_work(ff,(int) (fifo_interval*1e6),rbuf,&readin,
                            (flag&(1|8))|4);
   if(ret<0) {
      /* >>> error */;
   } else if(ret==2) {
      readin= 0;
 break;
   } else if(readin>0) {
 break;
   }
 }
 if(readin<=0)
   return(2);
 if(size<fifo_bs) {
   changebuf_fill= readin;
   goto read_changebuf;
 }
 *gelesen= readin;
 return(1);
}


int Raedchen_pad(progname,fp,pad_size,flag)
char *progname;
FILE *fp;
double pad_size;
int flag;
/*
 bit0= do not complain about broken pipe error
*/
{
 double count;
 int step= 1,i,ret;
 char buffer[Raedchen_bufsizE];

 if(fp==NULL)
   return(2);
 step= Raedchen_bufsizE;
 for(i=0;i<step;i++)
   buffer[i]= raedchen[i%4];
 i= 0;
 for(count=0; count<pad_size; count+= step) {
   if(pad_size-count<Raedchen_bufsizE)
     step= 1;
   if(step==1) {
     if(fputc(raedchen[i],fp)==EOF) {
       perror(progname);
       { ret= 0; goto ex; }
     }
     if(++i>=4)
       i= 0;
   } else {
     if(fwrite(buffer,1,step,fp)!=step){
       if(errno!=EPIPE || !(flag&1))
         perror(progname);
       { ret= 0; goto ex; }
     }
   }
 }
 ret= 1;
ex:;
 return(ret);
}


main(argc,argv)
int argc;
char **argv;
{
 static char post_text[4096]={"bytes"};
 static char pipe_text[4096]={""};
 static char post_pipe_text[4096]={""};
 static char intermediate_pipe_text[4096]={""};
 static char read_file[4096]={""};
 static char prompt_text[4096]={"Prepare next piece and press ENTER"};
 static double pipe_stueck_gr= 1.0e99;
 static int single_piece= 0;
 static char overflow_mark[4096]={""};
 static FILE *zielfp;
 static FILE *readfp;
 static int quiet= 0;
 static int do_flush= 0;
 static double baudlimit;
 char buffer[Raedchen_bufsizE],inp[321],num[81];
 double zaehler= 0.0,pipe_zaehler= 0.0,pipe_w_count= 0.0,pipes_opened= 0;
 int pipe_command_not_started= 1,compact_num= -1,no_prompt= 0;
 int i,radpos=0,drehrest=0,dreh=64*1024,c,ret= 0,bufgr,gelesen,no_epipe= 0;
 int zahlzeigen=1, cursor_home=0, schlucken=0, einzel_char=0, schlusszahl=0;
 int rret,pret,sret;
 double pad_size= 0;
 int pad_to_pipe_size= 0;
 FILE *fpin,*fphelp;
 time_t start_time;
 double elapsed_time;
#ifdef Raedchen_with_fifO
 struct CdrfifO *ff= NULL;
#endif /* Raedchen_with_fifO */
 int fifo_size= 0,fifo_bs= 2048,fifo_fill= 0,fifo_verbous= 0,ziel_fd,source_fd;
 double fifo_speed_limit= 0.0,fifo_interval= 1.0;

 bufgr= sizeof(buffer);
 zielfp= stdout;
 readfp= stdin;

 for(i=1;i<argc;i++){
   if(strcmp(argv[i],"-count")==0) {
     zahlzeigen= 1;
   } else if(strcmp(argv[i],"-no_count")==0) {
     zahlzeigen= 0;
   } else if(strcmp(argv[i],"-print_count")==0) {
     schlusszahl= 2;
   } else if(strcmp(argv[i],"-print_count_off")==0) {
     schlusszahl= 0;
   } else if(strcmp(argv[i],"-print_count_stdout")==0) {
     schlusszahl= 1;
   } else if(strcmp(argv[i],"-cursor_home")==0) {
     cursor_home= 1;
   }else if(strcmp(argv[i],"-cursor_end")==0) {
     cursor_home= 0;
   }else if(strcmp(argv[i],"-output")==0) {
     schlucken= 0;
   }else if(strcmp(argv[i],"-no_output")==0) {
     schlucken= 1;
   }else if(strcmp(argv[i],"-epipe")==0) {
     no_epipe= 0;
   }else if(strcmp(argv[i],"-compact_num")==0) {
     compact_num= 1;
   }else if(strcmp(argv[i],"-no_compact_num")==0) {
     compact_num= 0;
   }else if(strcmp(argv[i],"-no_epipe")==0) {
     no_epipe= 1;
   }else if(strcmp(argv[i],"-step")==0 && i+1<argc) {
     i++;
     dreh= Scanf_io_size(argv[i],0);
     if(dreh<=0) {
       quiet= 1;
     } else {
       if(dreh<bufgr)
         einzel_char= 1;
       else
         einzel_char= 0;
     }
   }else if(strcmp(argv[i],"-post_text")==0 && i+1<argc) {
     i++;
     strcpy(post_text,argv[i]);
   }else if(strcmp(argv[i],"-pipe")==0 && i+3<argc) {
     i++;
     strcpy(read_file,argv[i]);
     if(read_file[0]=='-' && read_file[1]==0)
       read_file[0]= 0;
     i++;
     pipe_stueck_gr= Scanf_io_size(argv[i],0);
     i++;
     strcpy(pipe_text,argv[i]);
   }else if(strcmp(argv[i],"-post_pipe_command")==0 && i+1<argc) {
     i++;
     strcpy(post_pipe_text,argv[i]);
   }else if(strcmp(argv[i],"-intermediate_pipe_command")==0 && i+1<argc) {
     i++;
     strcpy(intermediate_pipe_text,argv[i]);
   }else if(strcmp(argv[i],"-no_prompt")==0) {
     no_prompt= 1;
   }else if(strcmp(argv[i],"-prompt")==0 && i+1<argc) {
     i++;
     strcpy(prompt_text,argv[i]);
     no_prompt= 0;
   }else if(strcmp(argv[i],"-single_piece")==0 && i+1<argc) {
     i++;
     strcpy(overflow_mark,argv[i]);
     single_piece= 1;
   }else if(strcmp(argv[i],"-quiet")==0) {
     quiet= 1;
   }else if(strcmp(argv[i],"-loud")==0) {
     quiet= 0;
   }else if(strcmp(argv[i],"-pad")==0 && i+1<argc) {
     i++;
     pad_size= Scanf_io_size(argv[i],0);
     if(pad_size<0)
       pad_size= 0.0;
   }else if(strcmp(argv[i],"-pad_to_pipe_size")==0) {
     pad_to_pipe_size= 1;
   }else if(strcmp(argv[i],"-baudlimit")==0 && i+1<argc) {
     i++;
     baudlimit= Scanf_io_size(argv[i],0);
   }else if(strcmp(argv[i],"-flush")==0) {
     do_flush= 1;
   } else if(strcmp(argv[i],"-size_to_bytecount_and_exit")==0 && i+1<argc) {
     i++;
     baudlimit= Scanf_io_size(argv[i],0);
     printf("%.lf\n",baudlimit);
     exit(0);

   }else if(strncmp(argv[i],"fs=",3)==0) {
     fifo_size= Scanf_io_size(argv[i]+3,0);
     if(fifo_size>1024.0*1024.0*1024.0) {
       fprintf(stderr,"raedchen: NOTE : fs= is limited to 1g\n");
       fifo_size= 0;
     }
   }else if(strncmp(argv[i],"bs=",3)==0) {
     fifo_bs= Scanf_io_size(argv[i]+3,0);
     if(fifo_bs>bufgr) {
       fprintf(stderr,"raedchen: NOTE : bs= truncated to %d bytes\n",bufgr);
       fifo_bs= bufgr;
     }
   }else if(strncmp(argv[i],"fl=",3)==0) {
     fifo_fill= !! (int) Scanf_io_size(argv[i]+3,0);
   }else if(strncmp(argv[i],"iv=",3)==0) {
     fifo_interval= Scanf_io_size(argv[i]+3,0);
     if(fifo_interval<0.001 || fifo_interval>1000.0)
       fifo_interval= 1;
   }else if(strncmp(argv[i],"sl=",3)==0) {
     fifo_speed_limit= Scanf_io_size(argv[i]+3,0);
   }else if(strncmp(argv[i],"vb=",3)==0) {
     fifo_verbous= Scanf_io_size(argv[i]+3,0);

   }else if(strcmp(argv[i],"-help")==0) {
     ret= 0;
     fphelp= stdout;
     goto usage_text;
   }else {
     fphelp= stderr;
     fprintf(fphelp,"%s : unknown option\n",argv[i]);
     ret= 1;
usage_text:;
     fprintf(fphelp,"usage : %s\n",argv[0]);
     fprintf(fphelp,
             "           -count | -no_count | -cursor_home | -cursor_end |\n");
     fprintf(fphelp,
             "           -output | -no_output | -step number |\n");
     fprintf(fphelp,
             "           -print_count | -print_count_stdout |\n");
     fprintf(fphelp,
             "           -print_count_off | -post_text text |\n");
     fprintf(fphelp,
             "           -pipe source size command | -prompt text | -no_prompt\n");
     fprintf(fphelp,
             "           -post_pipe_command command | -quiet | -loud |\n");
     fprintf(fphelp,
             "           -intermediate_pipe_command command |\n");
     fprintf(fphelp,
             "           -epipe | -no_epipe | -baudlimit size | flush |\n");
     fprintf(fphelp,
             "           -pad size | -pad_to_pipe_size |\n");
     fprintf(fphelp,
             "           -single_piece mark_file_address |\n");
     fprintf(fphelp,
             "           fs=N | bs=N | vb=N | fl=N | iv=N | sl=[-]N\n");
     fprintf(fphelp,"\n");
     fprintf(fphelp,"This program entertains the user while copying\n");
     fprintf(fphelp,"standard input to standard output. It usually shows a\n");
     fprintf(fphelp,"spinning bar and the byte count.\n");
     fprintf(fphelp,"-quiet prevents the output of entertaining messages.\n");
     fprintf(fphelp,"-loud (which is the default) enables this output.\n");
     fprintf(fphelp,"-no_count prevents showing the byte count.\n");
     fprintf(fphelp,"-no_output is equal to >/dev/null but 30 %% faster.\n");
     fprintf(fphelp,"-step adjusts the count interval. -step 0 is -quiet.\n");
     fprintf(fphelp,"(Default is -step 64k, try -step 100000 or -step 1m )\n");
     fprintf(fphelp,"-compact_num enables byte count texts like 105.3m\n");
     fprintf(fphelp,"(This is the default if -step is at least 1024).\n");
     fprintf(fphelp,"-no_compact_num prints flat bytecounts like 110415052.\n");
     fprintf(fphelp,"-print_count_off leaves no visible text after exit.\n");
     fprintf(fphelp,"-print_count prints the final count on standard error.\n");
     fprintf(fphelp,"-print_count_stdout prints this on standard output.\n");
     fprintf(fphelp,"-post_text defines the text to be shown after the\n");
     fprintf(fphelp,"bytecount. (default is -post_text 'bytes' )\n");
     fprintf(fphelp,"\n");
     fprintf(fphelp,"A second purpose of this program is to cut the data\n");
     fprintf(fphelp,"stream into pieces and to feed them into the\n");
     fprintf(fphelp,"standard input of a shell command. This command is\n");
     fprintf(fphelp,"executed at the start of each piece and should end\n");
     fprintf(fphelp,"properly on EOF at standard input.\n");
     fprintf(fphelp,"Between the pieces, the user is prompted to prepare\n");
     fprintf(fphelp,"the next one (e.g. insert next CD)\n");
     fprintf(fphelp,"-pipe source size command \n");
     fprintf(fphelp,"divides the output into pieces of at most the given\n");
     fprintf(fphelp,"size. Each piece is piped into the standard input\n");
     fprintf(fphelp,"of a newly started instance of 'command'.\n");
     fprintf(fphelp,"The command is only started after at least one byte of\n");
     fprintf(fphelp,"input arrived for the piece.\n");
     fprintf(fphelp,"If 'source' is '-' the standard input is read. Any\n");
     fprintf(fphelp,"other text is used as name of a file which is opened\n");
     fprintf(fphelp,"at the start of each piece and closed after its end.\n");
     fprintf(fphelp,"This is suitable for device files. Data files should\n");
     fprintf(fphelp,"be read by piping them into standard input.\n");
     fprintf(fphelp,"-prompt defines the text to be asked between the\n");
     fprintf(fphelp,"pieces. default: 'Prepare next piece and press ENTER'\n");
     fprintf(fphelp,"-no_prompt disables prompting but assumes that the\n");
     fprintf(fphelp,"piece change precautions are done automatically.\n");
     fprintf(fphelp,"The last one of -prompt, -no_prompt gets into effect.\n");
     fprintf(fphelp,"-post_pipe_command defines a shell command which is\n");
     fprintf(fphelp,"executed when a -pipe piece is closed and the -pipe\n");
     fprintf(fphelp,"command has ended.\n");
     fprintf(fphelp,"-intermediate_pipe_command is executed after a prompt\n");
     fprintf(fphelp,"has been answered and before the next pipe starts\n");
     fprintf(fphelp,"(i.e. it is not executed with the first pipe).\n");
     fprintf(fphelp,"-no_epipe disables the error message if the pipe gets\n");
     fprintf(fphelp,"closed by the receiver of output. (-epipe enables)\n");
     fprintf(fphelp,"\n");
     fprintf(fphelp,"Following options influence performance and content:\n");
     fprintf(fphelp,"-baudlimit roughly enforces bandwidth restriction to\n");
     fprintf(fphelp,"the given size per second. Granularity is -step and\n");
     fprintf(fphelp,"full seconds. Set -step and -baudlimit to equal size.\n");
     fprintf(fphelp,"-baudlimit 0 disables bandwidth restriction.\n");
     fprintf(fphelp,"-flush causes output to be fflushed and fsynced after\n");
     fprintf(fphelp,"each -step interval.\n");
     fprintf(fphelp,"-pad writes a trailing piece of data of the given\n");
     fprintf(fphelp,"size. This also happens at the end of any -pipe piece.\n");
     fprintf(fphelp,"-pad_to_pipe_size causes padding of the last -pipe\n");
     fprintf(fphelp,"piece up to the size given with -pipe if it contains\n");
     fprintf(fphelp,"at least one byte.\n");
     fprintf(fphelp,"-single_piece causes the program to abort with error\n");
     fprintf(fphelp,"message and exit value 1 if more than one -pipe piece\n");
     fprintf(fphelp,"emerges. If mark_file_address is not '-' then a text\n");
     fprintf(fphelp,"is written to that file address. The text contains\n");
     fprintf(fphelp,"the word 'Overflow'.\n");
     fprintf(fphelp,"-size_to_bytecount_and_exit prints a valid size text\n");
     fprintf(fphelp,"(like 4.5g) as integer number to stdout and exits.\n");
     fprintf(fphelp,"\n");
     fprintf(fphelp,"Some options of cdrfifo have been implemented:\n");
     fprintf(fphelp,"fs=N enables a fifo buffer of given size N (0=off).\n");
     fprintf(fphelp,"bs=N sets the transaction block size of the fifo.\n");
     fprintf(fphelp,
                "vb=1 causes start and end message. vb=2 is for debugging.\n");
     fprintf(fphelp,"fl=1 causes the fifo to be filled before write start.\n");
     fprintf(fphelp,"iv=N sets the interval for vb=2 and sl= (default 1.0)\n");
     fprintf(fphelp,"sl=N >0 sets a speedlimit in bytes/second. Eventual\n");
     fprintf(fphelp,"     slowdowns are caught up over the total run time.\n");
     fprintf(fphelp,"     <0 sets a speedlimit. Eventual slowdowns are\n");
     fprintf(fphelp,"     caught up only over the durance of iv= .\n");
     fprintf(fphelp,".\n");
     exit(ret);
   }
 }

 signal(SIGPIPE,Cleanup_handler_pipe);

 start_time= time(0);
 if(read_file[0]!=0){
   readfp= fopen(read_file,"r");
   if(readfp==NULL) {
     fprintf(stderr,"Cannot open for reading : %s\n",read_file);
     perror("Reason given ");
     return(1);
   }
 }

#ifdef Raedchen_with_fifO
 if(fifo_size>0) {
   if(fifo_bs>bufgr)
     fifo_bs= bufgr;
   if(fifo_verbous>=1)
     fprintf(stderr,
             "raedchen: bs=%d fl=%d fs=%d iv=%lf sl=%.lf vb=%d\n",
             fifo_bs,fifo_fill,fifo_size,fifo_interval,fifo_speed_limit,
             fifo_verbous);
   ziel_fd= 1;
   if(zielfp!=NULL)
     ziel_fd= fileno(zielfp);
   ret= Cdrfifo_new(&ff,fileno(readfp),ziel_fd,fifo_bs,fifo_size,0);
   if(ret<=0) {
     fprintf(stderr,
           "raedchen: FATAL : creation of fifo object with %d bytes failed\n",
             fifo_size);
     return(1);
   }
   if(fifo_speed_limit!=0)
     Cdrfifo_set_speed_limit(ff,fifo_speed_limit,0);
   if(fifo_fill) {
     ret= Cdrfifo_fill(ff,0);
     if(ret<=0) {
       fprintf(stderr,
               "raedchen: FATAL : initial filling of fifo buffer failed\n");
       return(1);
     }
   }
 }
#endif /* Raedchen_with_fifO */

 if(compact_num==-1 && dreh<1024) 
   compact_num= 0;
 ret= 0;
 while(!abort_run) {
   if(einzel_char){
lies_einzeln:;

#ifdef Raedchen_with_fifO

     if(fifo_size!=0) {
       Raedchen_read_fifo(ff,fifo_interval,buffer,1,&gelesen,0);
     } else {
       if((c= fgetc(readfp))==EOF) 
         gelesen= 0;
       else {
         ((unsigned char *) buffer)[0]= c;
         gelesen= 1;
       }
     }
     if(gelesen==0)
 break;


#else /* Raedchen_with_fifO */

     if((c= fgetc(readfp))==EOF) 
 break;
     ((unsigned char *) buffer)[0]= c;
     gelesen= 1;

#endif /* ! Raedchen_with_fifO */

     drehrest++;
     pipe_zaehler++;
   }else{

     if(pipe_zaehler+bufgr>pipe_stueck_gr && pipe_zaehler<pipe_stueck_gr){
       einzel_char= 1;
       goto lies_einzeln;
     } 

#ifdef Raedchen_with_fifO

     if(fifo_size!=0)
       Raedchen_read_fifo(ff,fifo_interval,buffer,bufgr,&gelesen,0);
     else
       gelesen= fread(buffer,1,bufgr,readfp);

#else /* Raedchen_with_fifO */

     gelesen= fread(buffer,1,bufgr,readfp);

#endif /* ! Raedchen_with_fifO */

     if(gelesen==0) 
 break;
     drehrest+= gelesen;
     pipe_zaehler+= gelesen;
   }

   if(pipe_command_not_started) {
     if(pipes_opened>0 && intermediate_pipe_text[0]!=0) {
       sret= system(intermediate_pipe_text);
       if(sret!=0) {
         fprintf(stderr,"%s : -intermediate_pipe_command failed\n",
                        argv[0]);
         fprintf(stderr,"   %s\n",intermediate_pipe_text);
         return(1);
       }
     }
     if(pipe_text[0]!=0){
       zielfp= popen(pipe_text,"w");
       if(zielfp==NULL) {
         fprintf(stderr,"Cannot open pipe for writing and execute :\n");
         fprintf(stderr,"   %s\n",pipe_text);
         perror("Reason given ");
         return(1);
       }

#ifdef Raedchen_with_fifO
       if(ff!=NULL) {
         Cdrfifo_get_fds(ff,&source_fd,&ziel_fd,0);
#ifdef NIX
         if(ziel_fd>=0)
           close(ziel_fd);
         ziel_fd= dup(fileno(zielfp));
         if(ziel_fd<0)
           goto dup_failed;
         Cdrfifo_set_fds(ff,source_fd,ziel_fd,0);
#else
         Cdrfifo_set_fds(ff,source_fd,fileno(zielfp),0);
#endif

       }
#endif /* ! Raedchen_with_fifO */

     }
     pipe_command_not_started= 0;
     pipes_opened++;
   }

   if(read_file[0]!=0){
     if(!schlucken){
       if(fwrite(buffer,1,gelesen,zielfp)!=gelesen){
         if(errno!=EPIPE || no_epipe==0)
           perror(argv[0]);
         ret= 1;
 break;
       }
     }
     pipe_w_count= pipe_zaehler;
   }

   if(pipe_text[0]!=0 && pipe_w_count>=pipe_stueck_gr){
     if(single_piece) {
       fprintf(stderr,"\n");
       fprintf(stderr,
         "%s : Overflow of limit preset by options -pipe and -single_piece.\n",
         argv[0]);
       fprintf(stderr,"%s : Limit is %.lf\n",argv[0],pipe_stueck_gr);
       fprintf(stderr,"%s : Aborted.\n",argv[0]);
       fprintf(stderr,"\n");
       if(overflow_mark[0]!=0 && strcmp(overflow_mark,"-")!=0) {
         FILE *fp;
         fp= fopen(overflow_mark,"w");
         if(fp!=NULL) {
           fprintf(fp,"%s : Overflow %.lf\n",argv[0],pipe_stueck_gr);
           fclose(fp);
         }
       }
       ret= 1;
 break;
     }
     if(read_file[0]!=0){
       if(readfp!=stdin)
         fclose(readfp);
     }

     rret= Raedchen_pad(argv[0],zielfp,pad_size,no_epipe);
     if(rret<=0) {
       ret= 1;
 break;
     }

     pret= Pipe_close(&zielfp,post_pipe_text,0);
     if(pret!=0) {
       if(pret&1) {
         fprintf(stderr,"%s : -pipe failed\n",argv[0]);
         fprintf(stderr,"+   %s\n",pipe_text);
       }
       if(pret&2) {
         fprintf(stderr,"%s : -post_pipe_command failed\n",argv[0]);
         fprintf(stderr,"+   %s\n",post_pipe_text);
       }
       ret= 1;
 break;
     }
     pipe_command_not_started= 1;

     if(no_prompt) {
       fprintf(stderr,"\n");
       fprintf(stderr,"Note: raedchen -no_prompt : Will write next part immediately.\n");
       fprintf(stderr,"\n");
     } else {
       fprintf(stderr,"\n");
       for(i=0;i<strlen(prompt_text); i++) fprintf(stderr,"-");
       fprintf(stderr,"\n");
       fprintf(stderr,"%s\n",prompt_text);
       for(i=0;i<strlen(prompt_text); i++) fprintf(stderr,"-");
       fprintf(stderr,"\n");
       fpin= fopen("/dev/tty","r");
       fgets(inp,sizeof(inp),fpin);
       fclose(fpin);
     }
     fprintf(stderr,"... continuing ...\n");

     if(read_file[0]!=0){
       readfp= fopen(read_file,"r");
       if(readfp==NULL) {
         fprintf(stderr,"Cannot open for reading : %s\n",read_file);
         perror("Reason given ");
         return(1);
       }
     }
     pipe_zaehler-= pipe_w_count;
     if(dreh<bufgr || pipe_zaehler+bufgr>pipe_stueck_gr) 
       einzel_char= 1;
     else
       einzel_char= 0;
     /* there are data waiting : restart pipe command */
     if(pipes_opened>0 && intermediate_pipe_text[0]!=0) {
       sret= system(intermediate_pipe_text);
       if(sret!=0) {
         fprintf(stderr,"%s : -intermediate_pipe_command failed\n",
                        argv[0]);
         fprintf(stderr,"   %s\n",intermediate_pipe_text);
         return(1);
       }
     }
     if(pipe_text[0]!=0){
       zielfp= popen(pipe_text,"w");
       if(zielfp==NULL) {
         fprintf(stderr,"Cannot open pipe for writing and execute :\n");
         fprintf(stderr,"   %s\n",pipe_text);
         perror("Reason given ");
         return(1);
       }

#ifdef Raedchen_with_fifO
       if(ff!=NULL) {
         Cdrfifo_get_fds(ff,&source_fd,&ziel_fd,0);
#ifdef NIX
         if(ziel_fd>=0)
           close(ziel_fd);
         ziel_fd= dup(fileno(zielfp));
         if(ziel_fd<0)
           goto dup_failed;
         Cdrfifo_set_fds(ff,source_fd,ziel_fd,0);
#else
         Cdrfifo_set_fds(ff,source_fd,fileno(zielfp),0);
#endif
       }
#endif /* ! Raedchen_with_fifO */

     }
     pipe_command_not_started= 0;
     pipes_opened++;
   } 

   if(read_file[0]==0){
     if(!schlucken){
       if(fwrite(buffer,1,gelesen,zielfp)!=gelesen){
         if(errno!=EPIPE || no_epipe==0)
           perror(argv[0]);
         ret= 1;
 break;
       }
     }
     pipe_w_count= pipe_zaehler;
   }

   while(drehrest>=dreh) {
     zaehler+= dreh;
     drehrest-= dreh;
     if(!quiet){
       if(zahlzeigen) {
         if(compact_num) {
           Sfile_scale(zaehler,num,7,1e5,1);
         } else {
           sprintf(num,"%12.lf",zaehler);
         }
         fprintf(stderr,"\r%c  %s %s ",raedchen[radpos],num,post_text);
       } else
         fprintf(stderr,"\r%c  ",raedchen[radpos]);
       if(cursor_home)
         fprintf(stderr,"\r");
     }
     if(baudlimit>0) {
       elapsed_time= time(0)-start_time;
       if(elapsed_time>0) {
         while(zaehler/elapsed_time>baudlimit) {
           if(!quiet)
             fprintf(stderr,"\r%c*\b",raedchen[radpos]);
           sleep(1);
           elapsed_time= time(0)-start_time;
         }
       } else
         sleep(1);
     }
     radpos++;
     if(radpos>=4) 
       radpos= 0;
     if(do_flush) {
       if(!schlucken){
         fflush(zielfp);
         fsync(fileno(zielfp));
       }
     }
   }

 }

 if(ret==0 && zielfp!=NULL) {
   if(pad_to_pipe_size) 
     if(pipe_w_count>0 && pipe_w_count<pipe_stueck_gr)
       pad_size+= (pipe_stueck_gr-pipe_w_count);
   rret= Raedchen_pad(argv[0],zielfp,pad_size,no_epipe);
   if(rret<=0)
     ret= 1;
 }
 if(zielfp!=NULL)
   fflush(zielfp);

 if(pipe_text[0]!=0) {
   if(readfp!=NULL)
     fclose(readfp);
   readfp= NULL;
   pret= Pipe_close(&zielfp,post_pipe_text,0);
   if(ret==0 && pret!=0)
     ret= 1;
   if(pret&1) {
     fprintf(stderr,"%s : -pipe failed :\n",argv[0]);
     fprintf(stderr,"   %s\n",pipe_text);
   }
   if(pret&2) {
     fprintf(stderr,"%s : -post_pipe_command failed :\n",argv[0]);
     fprintf(stderr,"   %s\n",post_pipe_text);
   }
 }
 
 if(!quiet)
   fprintf(stderr,"\r                                                                           \r");
 if(compact_num) {
   Sfile_scale(zaehler+drehrest,num,7,1e5,1);
 } else {
   sprintf(num,"%12.lf",zaehler+drehrest);
 }
 if(schlusszahl==1){
   fprintf(stdout,"%s %s\n",num,post_text);
 }else if(schlusszahl==2) {
   fprintf(stderr,"\r   %s %s\n",num,post_text);
 }

#ifdef Raedchen_with_fifO
 if(fifo_verbous>=1 && ff!=NULL) {
   double in_counter,out_counter;
   double put_counter,get_counter,empty_counter,full_counter;
   int min_fill,total_min_fill,fifo_percent;

   Cdrfifo_get_counters(ff,&in_counter,&out_counter,0);
   fprintf(stderr,"\nraedchen: fifo : %.lf bytes in , %.lf bytes out\n",
           in_counter,out_counter);
   Cdrfifo_get_min_fill(ff,&total_min_fill,&min_fill,0);
   fifo_percent= 100.0*((double) total_min_fill)/(double) fifo_size;
   if(fifo_percent==0 && total_min_fill>0)
     fifo_percent= 1;
   Cdrfifo_get_cdr_counters(ff,&put_counter,&get_counter,
                            &empty_counter,&full_counter,0);
   fprintf(stderr,"raedchen: fifo had %.lf puts and %.lf gets.\n",
           put_counter,get_counter);
   fprintf(stderr,"raedchen: fifo was %.lf times empty and %.lf times full",
           empty_counter,full_counter);
   if(fifo_fill)
     fprintf(stderr,", min fill was %d%%.\n",fifo_percent);
   else
     fprintf(stderr,".\n");
 }
 Cdrfifo_destroy(&ff,1);
#endif /* Raedchen_with_fifO */

 exit(ret);
}

