/*
    cc -g -DAskme_with_maiN -DAskme_with_regeX -o ../bin/saskme ../src/askme.c

    cc -g -DAskme_with_maiN -DAskme_with_regeX -DAskme_with_readlinE \
          -Wall -D_FILE_OFFSET_BITS=64 -D_LARGEFILE_SOURCE \
          -o bin/saskme src/askme.c -lreadline
*/

#define PROG_VERSION "0.8.8"

#include <ctype.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <dirent.h>
#include <time.h>
#include <utime.h>

#ifdef Askme_with_regeX
#include <regex.h>
#endif /* Askme_with_regeX */

#ifdef Askme_with_readlinE
#ifdef Askme_with_old_readlinE
#include <readline.h>
#include <history.h>
#else /* Askme_with_old_readlinE */
#include <readline/readline.h>
#include <readline/history.h>
#endif /* ! Askme_with_old_readlinE */
#endif /* Askme_with_readlinE */

#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));


/* Diet facility: exclude help texts and german dialog from binaries */
/* This will eventually be redefined to eat up its content */
#define AlN(x) x
/* This shall replace standard german-english dialog if clauses */
#define Askme_de_eN(de,en) \
 if(askme->german_dialog) {\
   de \
 } else { \
   en \
 }


/* There are two stages of diet: Askme_no_germaN and Askme_no_helP */
#ifdef Askme_no_helP
#ifndef Askme_no_germaN
#define Askme_no_germaN
#endif
#endif
#ifdef Askme_no_germaN
#ifndef Askme_no_german_helP
#define Askme_no_german_helP
#endif
#undef AlN
#define AlN(x) 
#undef Askme_de_eN
#define Askme_de_eN(de,en) en
#endif


#include "askme.h"


/* ------------------------------------------------------------------------ */


#ifndef Askme_sfile_externaL

#define Smem_malloC malloc
#define Smem_freE free
#define SfileadrL 4096


char *Sfile_fgets(line,maxl,fp)
char *line;
int maxl;
FILE *fp;
{
 int l;
 char *ret;

 ret= fgets(line,maxl,fp);
 if(ret==NULL)
   return(NULL);
 l= strlen(line);
 if(l>0) if(line[l-1]=='\r') line[--l]= 0;
 if(l>0) if(line[l-1]=='\n') line[--l]= 0;
 if(l>0) if(line[l-1]=='\r') line[--l]= 0;
 return(ret);
}


int Sfile_fclose(fp)
FILE *fp;
{
 if(fp!=stdout && fp!=stdin && fp!=NULL)
   fclose(fp);
 return(1);
}


int Sfile_count_components(path,flag)
char *path;
int flag;
/*
 bit0= do not ignore trailing slash
 bit1= do not ignore empty components (other than the empty root name)
*/
{
 int l,count= 0;
 char *cpt;

 l= strlen(path);
 if(l==0)
   return(0);
 count= 1;
 for(cpt= path+l-1;cpt>=path;cpt--) { 
   if(*cpt=='/') {
     if(*(cpt+1)==0   && !(flag&1))
 continue;
     if(*(cpt+1)=='/' && !(flag&2))
 continue;
     count++;
   }
 }
 return(count);
}


int Sfile_component_pointer(path,sourcept,idx,flag)
char *path;
char **sourcept;
int idx;
int flag;
/*
 bit0= do not ignore trailing slash
 bit1= do not ignore empty components (other than the empty root name)
 bit2= accept 0 as '/'
*/
{
 int count= 0;
 char *spt;

 for(spt= path;*spt!=0 || (flag&4);spt++) {
   if(count>=idx) {
     *sourcept= spt;
     return(1);
   }
   if(*spt=='/' || *spt==0) {
     if(*(spt+1)=='/' && !(flag&2))
 continue;
     if(*(spt+1)==0 && !(flag&1))
 continue;
     count++;
   }
 }
 if((flag&1) && count>=idx)
   return(1);
 return(0);
}


int Sfile_add_to_path(path,addon,flag)
char path[4096];
char *addon;
int flag;
{
 int l;

 l= strlen(path);
 if(l==0) {
   strcpy(path,"/");
   l= 1;
 } else if(path[l-1]!='/') {
   path[l++]= '/';
   path[l]= 0;
 }
 if(l+strlen(addon)>=4096)
   return(0);
 if(addon[0]=='/')
   strcpy(path+l,addon+1);
 else
   strcpy(path+l,addon);
 return(1);
}


int Sfile_being_group_member(stbuf,flag)
struct stat *stbuf;
int flag;
{
 int i, suppl_groups;
 gid_t *suppl_glist;

 if (getegid()==stbuf->st_gid)
   return(1);
 suppl_groups= getgroups(0, NULL);
 suppl_glist= (gid_t *) malloc((suppl_groups + 1) * sizeof(gid_t));
 if (suppl_glist==NULL)
   return(-1);
 suppl_groups= getgroups(suppl_groups+1,suppl_glist);
 for (i= 0; i<suppl_groups; i++) {
   if (suppl_glist[i]==stbuf->st_gid) {
     free((char *) suppl_glist);
     return(1);
   }
 }
 free((char *) suppl_glist);
 return(0);
}


int Sfile_type(filename,flag)
char *filename;
int flag;
/*
 bit0= return -1 if file is missing 
 bit1= return a hardlink with siblings as type 5
 bit2= evaluate eventual link target rather than the link object itself
 bit3= return a socket or a char device as types 7 or 8 rather than 0
*/
/*
 return:
  0=unknown
  1=regular
  2=directory
  3=symbolic link
  4=named pipe
  5=multiple hardlink (with bit1)
  6=block device
  7=socket (with bit3)
  8=character device (with bit3)
*/
{
 struct stat stbuf;
 if(flag&4) {
   if(stat(filename,&stbuf)==-1) {
     if(flag&1) return(-1);
     else       return(0);
   }
 } else {
   if(lstat(filename,&stbuf)==-1) {
     if(flag&1) return(-1);
     else       return(0);
   }
 }
 if(S_ISREG(stbuf.st_mode)) {
   if(flag&2)
     if(stbuf.st_nlink>1)
       return(5);
   return(1);
 }
 if(S_ISDIR(stbuf.st_mode))
   return(2);
 if((stbuf.st_mode&S_IFMT)==S_IFLNK)
   return(3);
 if(S_ISFIFO(stbuf.st_mode))
   return(4);
 if(S_ISBLK(stbuf.st_mode))
   return(6);
 if(flag&8)
   if((stbuf.st_mode&S_IFMT)==S_IFSOCK)
     return(7);
 if(flag&8)
   if(S_ISCHR(stbuf.st_mode))
     return(8);
 return(0);
}


int Sfile_lookup_permissions(fname,perms,st_mode,flag)
char *fname;
int *perms; /* bit0= r , bit1= w , bit2= x */
mode_t *st_mode;
int flag;
/* return: <0 fatal error, 0= file nonexistent, 1=owner, 2=group, 3=other */
{
 struct stat stbuf;

 *perms= 0;
 if(stat(fname,&stbuf)==-1)
   return(0);
 *st_mode= stbuf.st_mode;
 if(geteuid()==stbuf.st_uid || geteuid()==0) {
   if(stbuf.st_mode & S_IRUSR)
     (*perms)|= 1; 
   if(stbuf.st_mode & S_IWUSR)
     (*perms)|= 2; 
   if(stbuf.st_mode & S_IXUSR)
     (*perms)|= 4; 
   return(1);
 }
 /* group membership is a complicated thing */
 if(Sfile_being_group_member(&stbuf,0)>0) {
   if(stbuf.st_mode & S_IRGRP)
     (*perms)|= 1; 
   if(stbuf.st_mode & S_IWGRP)
     (*perms)|= 2; 
   if(stbuf.st_mode & S_IXGRP)
     (*perms)|= 4; 
   return(2);
 }
 if(stbuf.st_mode & S_IROTH)
   (*perms)|= 1; 
 if(stbuf.st_mode & S_IWOTH)
   (*perms)|= 2; 
 if(stbuf.st_mode & S_IXOTH)
   (*perms)|= 4; 
 return(3);
}


int Sfile_select(fds,fdcount,fdidx,microsec,flag)
int *fds;
int fdcount;
int fdidx[3];
unsigned int microsec;
int flag;
/*
 bit0= check for reading (and choose one into fdidx[0])
 bit1= check for writing (and choose one into fdidx[1])
 bit2= check for exception (and choose one into fdidx[2])
 bit3= initialize fdidx
return-bits are set according to matching flag-bits
*/
{
 struct timeval wt;
 fd_set rds,wts,exs,*s;
 int ready,ret,i,j,max_fd= -1;

 if(flag&8)
   for(j=0;j<3;j++)
     fdidx[j]= 0;
 wt.tv_sec=  microsec/1000000;
 wt.tv_usec= microsec%1000000;
 FD_ZERO(&rds);
 FD_ZERO(&wts);
 FD_ZERO(&exs);
 for(i= 0;i<fdcount;i++) {
   if(fds[i]<0)
 continue;
   if(flag&1)
     FD_SET(fds[i],&rds);
   if(flag&2)
     FD_SET(fds[i],&wts);
   if(flag&4)
     FD_SET(fds[i],&exs);
   if(fds[i]>max_fd)
     max_fd= fds[i];
 }
 ready= select(max_fd+1,&rds,&wts,&exs,&wt);
 if(ready<=0)
   return(0);
 ret= 0;
 if(fdcount>0)
   for(j=0;j<2;j++)
     fdidx[j]= (fdidx[j]+1)%fdcount;
 for(i= 0;i<fdcount;i++) {
   for(j=0;j<3;j++) {
     if(j==0)
       s= &rds;
     else if(j==1)
       s= &wts;
     else
       s= &exs;
     if((flag&(1<<j)) && !(ret&(1<<j))) {
       if(fds[fdidx[j]]<0) 
         fdidx[j]= (fdidx[j]+1)%fdcount;
       else if(FD_ISSET(fds[fdidx[j]],s))
         ret|= (1<<j);
       else
         fdidx[j]= (fdidx[j]+1)%fdcount;
     }
   }
 }
 return(ret);
}


int Sfile_identical(file1,file2,flag)
char *file1,*file2;
int flag;
/*
 bit0= check wether it is the same  i-node 
*/
{
 FILE *fp1,*fp2;
 struct stat stbuf;
 int ret,c1,c2;
 ino_t ino1,ino2;
 dev_t dev1,dev2;

 ret= 0;

 if(flag&1){
   if(stat(file1,&stbuf)==-1) return(0);
   ino1= stbuf.st_ino;
   dev1= stbuf.st_dev;
   if(stat(file2,&stbuf)==-1) return(0);
   ino2= stbuf.st_ino;
   dev2= stbuf.st_dev;
   return((ino1==ino2) && (dev1==dev2));
 }

 fp1= fopen(file1,"rb");
 fp2= fopen(file2,"rb");
 if(fp1==NULL||fp2==NULL) goto fertig;
 while(1){
   c1= fgetc(fp1);
   c2= fgetc(fp2);
   if(c1==EOF && c2==EOF){
     ret= 1;
 break;
   }
   if(c1!=c2)
 break;
 }
fertig:;
 if(fp1!=NULL) fclose(fp1);
 if(fp2!=NULL) fclose(fp2);
 return(ret);
}


char *Sfile_datestr(tim,flag)
time_t tim;
short int flag;
/*
 bit0=with hours+minutes  
 bit1=with seconds

 bit8= local time rather than UTC
*/
{
 static char zeitcode[80]={"000000"};
 char puff[80];
 struct tm *azt;

 if(flag&256) 
   azt = localtime(&tim);
 else
   azt = gmtime(&tim);
 
 if(azt->tm_year>99)
   sprintf(zeitcode,"%c%1.1d%2.2d%2.2d",
           'A'+(azt->tm_year-100)/10,azt->tm_year%10,
           azt->tm_mon+1,azt->tm_mday);
 else
   sprintf(zeitcode,"%2.2d%2.2d%2.2d",
           azt->tm_year,azt->tm_mon+1,azt->tm_mday);
 if(flag&1){
   sprintf(puff,".%2.2d%2.2d",azt->tm_hour,azt->tm_min);
   strcat(zeitcode,puff);
 }
 if(flag&2){
   sprintf(puff,"%2.2d",azt->tm_sec);
   strcat(zeitcode,puff);
 }

 return(zeitcode);
}


int Sfile_scale(value,result,siz,thresh,flag)
double value;
char *result;
int siz;
double thresh;
int flag;
/*
 bit0= eventually ommit 'b'
 bit1= make text as short as possible
 bit2= no fraction (if it would fit at all)
*/
{
 char scale_c,scales[6],form[80];
 int i,dec_siz= 0,avail_siz= 1;

 strcpy(scales,"bkmgtp");
 scale_c= scales[0];
 for(i=1;scales[i]!=0;i++) {
   if(value<thresh-0.5) 
 break;
   value/= 1024.0;
   scale_c= scales[i];
 } 
 if(scale_c!='b' && !(flag&4)) { /* is there room for fractional part ? */
   avail_siz= siz-1;
   sprintf(form,"%%.lf");
   sprintf(result,"%.lf",value);
   if(strlen(result)<=avail_siz-2) 
     dec_siz= 1;                                  /* we are very modest */
 }
 if(scale_c=='b' && (flag&1)) {
   if(flag&2)
     sprintf(form,"%%.lf");
   else
     sprintf(form,"%%%d.lf",siz);
   sprintf(result,form,value);
 } else {
   if(flag&2)
     sprintf(form,"%%.lf%%c");
   else if(dec_siz>0) 
     sprintf(form,"%%%d.%dlf%%c",avail_siz,dec_siz);
   else
     sprintf(form,"%%%d.lf%%c",siz-1);
   sprintf(result,form,value,scale_c);
 }
 return(1);
}


int Sfile_make_argv(progname,line,argc,argv,flag)
char *progname;
char *line;
int *argc;
char ***argv;
int flag;
/*
 bit0= read progname as first argument from line
 bit1= just release argument list argv and return
 bit2= abort with return(0) if incomplete quotes are found
*/
{
 int i,pass,maxl=0,l,argzaehl=0,bufl;
 char *cpt,*start;
 char buf[SfileadrL];

 Sfile_destroy_argv(argc,argv,0);
 if(flag&2) return(1);

 for(pass=0;pass<2;pass++) {
   cpt= line-1;
   if(!(flag&1)){
     argzaehl= 1;
     if(pass==0)
       maxl= strlen(progname);
     else
       strcpy((*argv)[0],progname);
   } else {
     argzaehl= 0;
     if(pass==0) maxl= 0;
   }
   while(*(++cpt)!=0){
   if(isspace(*cpt)) continue;
     start= cpt;
     buf[0]= 0;
     cpt--;
     while(*(++cpt)!=0) {
     if(isspace(*cpt)) break;
       if(*cpt=='"'){
         l= cpt-start; bufl= strlen(buf);
         if(l>0) {strncat(buf,start,l);buf[bufl+l]= 0;}
         l= strlen(buf);
         start= cpt+1;
         while(*(++cpt)!=0) if(*cpt=='"') break;
         if((flag&4) && *cpt==0)
           return(0);
         l= cpt-start; bufl= strlen(buf);
         if(l>0) {strncat(buf,start,l);buf[bufl+l]= 0;}
         start= cpt+1;
       }else if(*cpt=='\''){
         l= cpt-start; bufl= strlen(buf);
         if(l>0) {strncat(buf,start,l);buf[bufl+l]= 0;}
         l= strlen(buf);
         start= cpt+1;
         while(*(++cpt)!=0) if(*cpt=='\'') break;
         if((flag&4) && *cpt==0)
           return(0);
         l= cpt-start; bufl= strlen(buf);
         if(l>0) {strncat(buf,start,l);buf[bufl+l]= 0;}
         start= cpt+1;
       }
     if(*cpt==0) break;
     }
     l= cpt-start;
     bufl= strlen(buf);
     if(l>0) {strncat(buf,start,l);buf[bufl+l]= 0;}
     l= strlen(buf);
     if(pass==0){
       if(l>maxl) maxl= l;
     }else{
       strcpy((*argv)[argzaehl],buf);
     }
     argzaehl++;
   if(*cpt==0) break;
   }
   if(pass==0){
     *argc= argzaehl;
     if(argzaehl>0) {
         *argv= (char **) Smem_malloC(argzaehl*sizeof(char *));
       if(*argv==NULL)
         return(-1);
     }
     for(i=0;i<*argc;i++) {
       (*argv)[i]= (char *) Smem_malloC((maxl+1));
       if((*argv)[i]==NULL)
         return(-1);
     }
   }
 }
 return(1);
}


int Sfile_destroy_argv(argc,argv,flag)
int *argc;
char ***argv;
int flag;
{
 int i;

 if(*argc>0 && *argv!=NULL){
   for(i=0;i<*argc;i++){
     if((*argv)[i]!=NULL)
       Smem_freE((*argv)[i]);
   }
   Smem_freE((char *) *argv);
 }
 *argc= 0;
 *argv= NULL;
 return(1);
}


#endif /*  Askme_sfile_externaL */


/* --------------------------------- misc --------------------------------- */


int Write_to_channel(text,channel_no,flag)
char *text;
int channel_no;
int flag;
/*
 bit0= packet write disabled, write to stdin resp. stderr
 bit1= text is the name of the log file for the given channel 
 bit2= text is the name of the consolidated packet log file for all channels 
bit15= with bit1 or bit2: close depicted log file
*/
{
 char *rpt,*npt,ret= 1;
 char prefix[16];
 static int num_channels= 4;
 static char channel_prefixes[4][4]= {"-","R","I","M"};
 static FILE *logfile_fp[4]= {NULL,NULL,NULL,NULL};
 static FILE *pktlog_fp= NULL;

 if(channel_no<0 || channel_no>=num_channels)
   return(-1);

 /* Logfiles */
 if((flag&2) && logfile_fp[channel_no]!=NULL) {
   fprintf(logfile_fp[channel_no],
      "! end ! end ! end ! end ! end ! end ! end ! end saskme log : %s : %s\n",
           channel_prefixes[channel_no],Sfile_datestr(time(0),1|2|256));
   fclose(logfile_fp[channel_no]);
   logfile_fp[channel_no]= NULL;
 }
 if((flag&4) && pktlog_fp!=NULL) {
   fprintf(pktlog_fp,
  "I:1:! end ! end ! end ! end ! end ! end ! end ! end saskme log : %s : %s\n",
           channel_prefixes[channel_no],Sfile_datestr(time(0),1|2|256));
   fclose(pktlog_fp);
   pktlog_fp= NULL;
 }
 if(flag&(1<<15))
   return(1);
 if((flag&2)) {
   logfile_fp[channel_no]= fopen(text,"a");
   if(logfile_fp[channel_no]==NULL)
     return(0);
   fprintf(logfile_fp[channel_no],
      "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! saskme log : %s : %s\n",
           channel_prefixes[channel_no],Sfile_datestr(time(0),1|2|256));
   fflush(logfile_fp[channel_no]);
 }
 if((flag&4)) {
   pktlog_fp= fopen(text,"a");
   if(pktlog_fp==NULL)
     return(0);
   fprintf(pktlog_fp,
   "I:1:!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! saskme log : . : %s\n",
           Sfile_datestr(time(0),1|2|256));
   fflush(pktlog_fp);
 }
 if(flag&(2|4))
   return(1);
 if(flag&1) {
   if(channel_no==1 || channel_no==3)
     printf("%s",text);
   if(channel_no==2 || channel_no==3)
     fprintf(stderr,"%s",text);
   if(logfile_fp[channel_no]!=NULL) {
     fprintf(logfile_fp[channel_no],"%s",text);
     fflush(logfile_fp[channel_no]);
   }
   if(pktlog_fp!=NULL)
     return(1);
 }
 rpt= text;
 sprintf(prefix,"%s:x: ",channel_prefixes[channel_no]);
 while(*rpt!=0) {
   npt= strchr(rpt,'\n');
   if(npt==NULL)
     prefix[2]= '0';
   else
     prefix[2]= '1';
   if(!(flag&1)) {
     ret= fwrite(prefix,5,1,stdout);
     if(ret<=0)
       return(0);
   }
   if(pktlog_fp!=NULL) {
     ret= fwrite(prefix,5,1,pktlog_fp);
     if(ret<=0)
       return(0);
   }
   if(npt==NULL) {
     if(!(flag&1)) {
       ret= fwrite(rpt,strlen(rpt),1,stdout);
       if(ret<=0)
         return(0);
       ret= fwrite("\n",1,1,stdout);
       if(ret<=0)
         return(0);
     }
     if(pktlog_fp!=NULL) {
       ret= fwrite(rpt,strlen(rpt),1,pktlog_fp);
       if(ret<=0)
         return(0);
       ret= fwrite("\n",1,1,pktlog_fp);
       if(ret<=0)
         return(0);
     }
 break;
   } else {
     if(!(flag&1)) {
       ret= fwrite(rpt,npt+1-rpt,1,stdout);
       if(ret<=0)
         return(0);
     }
     if(pktlog_fp!=NULL) {
       ret= fwrite(rpt,npt+1-rpt,1,pktlog_fp);
       if(ret<=0)
         return(0);
     }
   }
   rpt= npt+1;
 }
 if(!(flag&1))
   fflush(stdout);
 if(pktlog_fp!=NULL)
   fflush(pktlog_fp);
 return(1);
}


int Exec_cmd(prog,cmd,simple_arg,argv,errmsg,flag)
char *prog;
char *cmd;
char *simple_arg;
char **argv;
char errmsg[4096];
int flag;
/*
 bit0= packade mode (!bit0 of Write_to_channel())
*/
{
 int wait_status,is_dead= 0,pipes_open= 0,ret,channel_no;
 int stdout_pipe[2],stderr_pipe[2],fds[2],fdidx[3];
 char **exec_argv,*simple_argv[3],buffer[256];
 pid_t pid,dead_pid;
 static int debug= 0;

 errmsg[0]= 0;
 if(debug)
   fprintf(stderr,"Exec_cmd():fork\n");

 /* create pipes */
 if(pipe(stdout_pipe)==-1) {
cannot_make_pipe:;
   sprintf(errmsg,"%s : cannot create two pipes\n   reason given:   %s\n",
                  prog,strerror(errno));
   return(-1);
 }
 if(pipe(stderr_pipe)==-1) {
   close(stdout_pipe[0]);
   close(stdout_pipe[1]);
   goto cannot_make_pipe;
 }
 pipes_open= 1;

 pid= fork();
 if(pid==-1) {
   sprintf(errmsg,"%s : cannot fork a new process\n   reason given:   %s\n",
                  prog,strerror(errno));
   {ret= -1; goto father_ex;}
 } else if(pid>0) { /* original process */
   if(debug)
     fprintf(stderr,"Exec_cmd():father\n");
   /* close unused pipe ends so we register EOF by the child process */
   close(stdout_pipe[1]);
   close(stderr_pipe[1]);
   fds[0]= stdout_pipe[0];
   fds[1]= stderr_pipe[0];
   is_dead= 0;
   while(1) {
     if((fds[0]==-1 && fds[1]==-1)) { /* not before both pipes have closed */
       dead_pid= wait3(&wait_status,WNOHANG,NULL);
       if(dead_pid==pid) {
         is_dead= 1;
   break;
       } else if(dead_pid<0) {
         sprintf(errmsg,
                 "%s : error while waiting for end of sub process\n   %s\n",
                 prog,strerror(errno));
         {ret= -1; goto father_ex;}
       }
     }

     /* check for input at both pipes */;
     ret= Sfile_select(fds,2,fdidx,100000,1|8);
     if(ret<=0 || fdidx[0]<0 || fdidx[0]>1)
   continue;
     ret= read(fds[fdidx[0]],buffer,sizeof(buffer)-3);
     if(ret<0) {
       sprintf(errmsg,"--- error %d on read from %s pipe : %s\n",
                      errno,(fdidx[0]?"stdout":"stderr"),
                      (errno>0?strerror(errno):"-unknown error-"));
       {ret= -1; goto father_ex;}
     }
     if(ret==0) {
       fds[fdidx[0]]= -1;
   continue;
     }  
     buffer[ret]= 0;
     channel_no= fdidx[0]+1;
     ret= Write_to_channel(buffer,channel_no,!(flag&1));
     if(ret<=0) {
       sprintf(errmsg,"--- error %d on write to %s : %s\n",
                      errno,(fdidx[0]?"stdout":"stderr"),
                      (errno>0?strerror(errno):"-unknown error-"));
       {ret= -1; goto father_ex;}
     }
   }
   ret= 127;
   if(WIFEXITED(wait_status))
     ret= WEXITSTATUS(wait_status);
father_ex:;
   if(pipes_open) {
     close(stdout_pipe[0]);
     close(stderr_pipe[0]);
   }
   return(ret);
 }

 if(debug)
   fprintf(stderr,"Exec_cmd():child\n");

 /* Connect stdout and stderr to the appropriate ends of the pipes. */
 close(1);
 ret= dup(stdout_pipe[1]);
 if(ret==-1) {
cannot_dup_pipe:;
   sprintf(errmsg,
         "%s : cannot redirect standard i/o to pipes\n   reason given:   %s\n",
           prog,strerror(errno));
   Write_to_channel(errmsg,2,!(flag&1));
   exit(1);
 }
 close(2);
 ret= dup(stderr_pipe[1]);
 if(ret==-1)
   goto cannot_dup_pipe;
 /* close pipes so we register EOF by the father process */
 close(stdout_pipe[0]);
 close(stdout_pipe[1]);
 close(stderr_pipe[0]);
 close(stderr_pipe[1]);

 /* switch process to desired program */
 if(simple_arg!=NULL) {
   simple_argv[0]= cmd;
   simple_argv[1]= simple_arg;
   simple_argv[2]= NULL;
   exec_argv= simple_argv;
 } else {
   exec_argv= argv;
 }
 if(debug) { int i;
   fprintf(stderr,"debug: cmd='%s'\n",cmd);
   for(i=0;exec_argv[i]!=NULL;i++)
     fprintf(stderr,"debug: argv[%d]='%s'\n",i,exec_argv[i]);
 }
 if(strchr(cmd,'/')!=NULL)
   execv(cmd,exec_argv);
 else
   execvp(cmd,exec_argv);
 sprintf(errmsg,"%s : cannot start program %s\n   reason given:   %s\n",
                prog,cmd,strerror(errno));
 Write_to_channel(errmsg,2,!(flag&1));
 exit(127); 
}


int Exec_cmd_line(prog,cmd_line,extra_argc,extra_argv,errmsg,flag)
char *prog;
char *cmd_line;
int extra_argc;
char **extra_argv;
char *errmsg;
int flag;
/*
 bit0= packade mode (!bit0 of Write_to_channel())
*/
{
 int ret,cmd_argc= 0,i,w;
 char **cmd_argv= NULL,**argv= NULL;

 ret= Sfile_make_argv("",cmd_line,&cmd_argc,&cmd_argv,1);
 if(ret<=0)
   {ret= -1; goto ex;}
 argv= TSOB_FELD(char *,cmd_argc+extra_argc+1);
 if(argv==NULL)
   {ret= -1; goto ex;}
 w= 0;
 for(i=0;i<cmd_argc;i++)
   argv[w++]= cmd_argv[i];
 for(i=0;i<extra_argc;i++)
   argv[w++]= extra_argv[i];
 argv[w]= NULL;
 ret= Exec_cmd(prog,argv[0],NULL,argv,errmsg,flag&1);
ex:;
 if(argv!=NULL)
   free((char *) argv);
 Sfile_make_argv("",cmd_line,&cmd_argc,&cmd_argv,2);
 return(ret);
}


int Strcmp(pt1,pt2)
char **pt1,**pt2;
{
 return(strcmp(*pt1,*pt2));
}


int Sort_argv(argc,argv,flag)
int argc;
char **argv;
int flag;
{
 if(argc<=0)
   return(2);
 qsort(argv,(size_t) argc,sizeof(char *),Strcmp);
 return(1);
}


int Set_bit(data,idx,flag)
int *data;
int idx;
int flag;
/*
 bit0= clear bit
*/
{
 if(idx<0 || idx>31)
   return(0);
 if(flag&1)
   (*data)&= ~(1<<idx);
 else
   (*data)|= (1<<idx);
 return(1);
}


FILE *Afile_fopen(filename,mode,flag)
char *filename;
char *mode;
int flag;
/*
 bit0= do not print error message on failure
 bit6= write packeted error messages (see Write_to_channel())
*/
{
 FILE *fp= NULL;
 char errmsg[2*4096];

 if(strcmp(filename,"-")==0) {
   if(mode[0]=='a' || mode[0]=='w' ||
      (mode[0]=='r' && mode[1]=='+') ||
      (mode[0]=='r' && mode[1]=='b' && mode[2]=='+')) 
     fp= stdout;
   else
     fp= stdin;
 } else if(strncmp(filename,"tcp:",4)==0){
   sprintf(errmsg,"sorry - TCP/IP service isn't implemented yet.\n");
   Write_to_channel(errmsg,2,!(flag&64));
 } else if(strncmp(filename,"file:",5)==0){
   fp= fopen(filename+5,mode);
 } else {
   fp= fopen(filename,mode);
 }
 if(fp==NULL){
   if(!(flag&1)) {
     sprintf(errmsg,"failed to open file '%s' in %s mode\n",filename,mode);
     if(errno>0)
       sprintf(errmsg+strlen(errmsg),"   %s\n",strerror(errno));
     Write_to_channel(errmsg,2,!(flag&64));
   }
   return(NULL);
 }
 return(fp);
}


/* ------------------------------------------------------------------------ */


#ifndef Askme_sregex_externaL

#define Smem_malloC malloc
#define Smem_freE free


int Sregex_string_cut(handle,text,len,flag)
char **handle;
char *text;
int len;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int l=0;
 char *old_handle;

 if((flag&1)&&*handle!=NULL)
   l+= strlen(*handle);
 old_handle= *handle;
 if(text!=NULL) {
   l+= len;
   *handle= TSOB_FELD(char,l+1);
   if(*handle==NULL) {
     *handle= old_handle;
     return(0);
   }
   if((flag&1) && old_handle!=NULL)
     strcpy(*handle,old_handle);
   else
     (*handle)[0]= 0;
   if(len>0)
     strncat(*handle,text,len);
 } else {
   *handle= NULL;
 }
 if(old_handle!=NULL)
   Smem_freE(old_handle);
 return(1);
}


int Sregex_string(handle,text,flag)
char **handle;
char *text;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int ret,l=0;

 if(text!=NULL)
   l= strlen(text);
 
/* #define Sregex_looking_for_contenT 1 */
#ifdef Sregex_looking_for_contenT
 /* a debugging point if a certain text content has to be caught */
 if(text!=NULL)
   if(strcmp(text,"clear")==0)
     ret= 0;
#endif

 ret= Sregex_string_cut(handle,text,l,flag&1);
 return(ret);
}


/*
  from Juergen Helbing's  yencode-c.txt 
  found at  www.yenc.org  , Feb 4,  2002
  license : "It is public domain - as well as the yEnc specification"
  some alterations made to adapt it to Sregex_crc32() 

  seems to be compatible with SFV/CSV files
*/                       

int Sregex_crc32_yenc(data,len,crc,flag)
char *data;
int len;
unsigned long *crc;
int flag;
/*
 bit0= start of data polynom
*/
{
 /* X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0 */
 static unsigned long crc_tab[256] = {
   0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 
   0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
   0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
   0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
   0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
   0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
   0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,
   0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
   0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
   0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
   0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940,
   0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
   0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116,
   0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
   0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
   0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
   0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a,
   0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
   0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818,
   0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
   0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
   0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
   0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c,
   0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
   0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
   0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
   0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
   0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
   0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086,
   0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
   0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4,
   0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
   0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
   0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
   0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
   0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
   0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe,
   0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
   0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
   0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
   0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252,
   0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
   0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60,
   0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
   0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
   0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
   0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04,
   0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
   0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a,
   0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
   0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
   0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
   0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e,
   0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
   0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
   0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
   0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
   0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
   0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0,
   0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
   0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6,
   0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
   0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
   0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};

 int k;
 unsigned long crc_val,ch1,ch2,cc;

 if(flag&1)
   crc_val = ~0;
 else
   crc_val = ~*crc;

 for(k=0;k<len;k++) {
   cc= data[k] & 0xffL;
   ch1= (crc_val ^ cc) & 0xffL;
   ch1= crc_tab[ch1];
   ch2= (crc_val>>8L) & 0xffffffL;
   crc_val= ch1 ^ ch2;
 }
 *crc= ~crc_val;
 return(1);
}

#endif /*  Askme_sregex_externaL */


#ifndef Askme_text_shellsafe_externaL 

char *Text_shellsafe(in_text,out_text,flag)
char *in_text,*out_text;
int flag;
{
 int l,i,w=0;

 /* enclose everything by hard quotes */
 l= strlen(in_text);
 out_text[w++]= '\'';
 for(i=0;i<l;i++){
   if(in_text[i]=='\''){
     /* escape hard quote within the text */
     out_text[w++]= '\'';
     out_text[w++]= '"';
     out_text[w++]= '\'';
     out_text[w++]= '"';
     out_text[w++]= '\'';
   } else {
     out_text[w++]= in_text[i];
   }
 }
 out_text[w++]= '\'';
 out_text[w++]= 0;
 return(out_text);
}


#endif /* ! Askme_text_shellsafe_externaL */


int A_open_listfile(fppt,adr,mode,flag)
FILE **fppt;
char *adr;
char *mode;
int flag;
/*
 bit0= pipe through gzip/gunzip if adr indicates so (*.gz and *.GZ) 
 bit1= mode w : unlink but do not create new file nor return file pointer
 bit2= do not issues error messages on non-existent file in mode "r"
 bit6= write packeted error messages (see Write_to_channel())
return:
 <=0 = error
 1 = Afile_fopen() successful
 2 = popen() successful
*/
{
 int ret= 0,adr_has_dotgz= 0,l;
 char cmd[4096+80],shell_adr[3*4096],dotgz[4],errmsg[2*4096];
 FILE *Afile_fopen();
 struct stat stbuf;

 /* Remove dead children if needed */
 wait3(NULL,WNOHANG,NULL);

 *fppt= NULL;
 l= strlen(adr);
 if(l>3)
   if(strcmp(adr+(l-3),".gz")==0 || strcmp(adr+(l-3),".GZ")==0)
     adr_has_dotgz= 1;
 if(strcmp(mode,"w")==0) {
   unlink(adr);
   if(stat(adr,&stbuf)!=-1) {
     sprintf(errmsg,"--- A_open_listfile( %s ,w) : obstinate file\n",adr);
     Write_to_channel(errmsg,2,!(flag&64));
     ret= 0; goto ex;
   }
   if(flag&2)
     return(1);
   if((flag&1) && adr_has_dotgz) {
     sprintf(cmd,"gzip >%s",Text_shellsafe(adr,shell_adr,0));
     *fppt= popen(cmd,"w");
     if(*fppt==NULL) {
       sprintf(errmsg,"--- popen(\"%s\",\"w\") failed\n",cmd);
       Write_to_channel(errmsg,2,!(flag&64));
       sprintf(errmsg,"--- errno= %d , '%s'\n",errno,strerror(errno));
       Write_to_channel(errmsg,2,!(flag&64));
       ret= 0;
     } else
       ret= 2;
   } else {
     *fppt= Afile_fopen(adr,"wb",flag&64);
     ret= 1;
   }
 } else if(strcmp(mode,"r")==0) {
   dotgz[0]= 0;
   shell_adr[0]= 0;
   if((flag&1) && adr_has_dotgz)
     goto try_popen_gunzip;
   *fppt= Afile_fopen(adr,"rb",1|(flag&64));
   if(*fppt==NULL) {
     /* temporarily store the error message of fopen in case there is no .gz */
     sprintf(shell_adr,"--- errno= %d , '%s'\n",errno,strerror(errno));
     /* try to find old_complete_list_adr.gz */
     strcpy(dotgz,".gz");
try_popen_gunzip:;
     sprintf(cmd,"%s%s",adr,dotgz);
     *fppt=  Afile_fopen(cmd,"rb",1|(flag&64));
     if(*fppt!=NULL) {
       Sfile_fclose(*fppt);
       sprintf(cmd,"gunzip <%s%s",Text_shellsafe(adr,shell_adr,0),dotgz);
       *fppt=  popen(cmd,"r");
       if(*fppt==NULL) {
         sprintf(errmsg,"--- popen(\"%s\",\"r\") failed\n",cmd);
         Write_to_channel(errmsg,2,!(flag&64));
         sprintf(errmsg,"--- errno= %d , '%s'\n",errno,strerror(errno));
         Write_to_channel(errmsg,2,!(flag&64));
         ret= 0;
       } else
         ret= 2;
     } else {
       if(shell_adr[0]==0)
         sprintf(shell_adr,"--- errno= %d , '%s'\n",errno,strerror(errno));
       if(!(flag&4)) {
         sprintf(errmsg,"failed to open file '%s' in read mode\n",adr);
         Write_to_channel(errmsg,2,!(flag&64));
         /* stored error message of fopen */
         Write_to_channel(shell_adr,2,!(flag&64));
       }
       ret= 0;
     }
   } else 
     ret= 1;
 } else {
   sprintf(errmsg,"--- A_open_listfile : mode %s not supported\n",mode);
   Write_to_channel(errmsg,2,!(flag&64));
 }
 if(*fppt==NULL)
   ret= 0;
ex:;
 return(ret);
}


int A_youngest_listfile(adr,youngest_adr,flag)
char *adr;
char youngest_adr[4096];
int flag;
{
 int ret,nsl,has_suffix= 0,max_date_valid= 0,dirl;
 char dir_adr[4096],name_start[4096],suffix[4096],sub_name[4096],try_adr[4096];
 char *spt,*npt;
 struct DirseQ *dirseq= NULL;
 time_t max_date= 0;
 struct stat stbuf;

 /* determine dir_adr , name_start , suffix */
 spt= strrchr(adr,'/');
 if(spt==NULL) {
   strcpy(dir_adr,".");
   npt= adr;
 } else {
   strncpy(dir_adr,adr,spt-adr);
   dir_adr[spt-adr]= 0;
   npt= spt+1;
 }
 spt= strchr(npt,'.');
 if(spt==NULL) {
   strcpy(name_start,npt);
   suffix[0]= 0;
 } else {
   strncpy(name_start,npt,spt-npt);
   name_start[spt-npt]= 0;
   strcpy(suffix,spt+1);
   has_suffix= 1;
 }
 /* scan directory for matching files */
 dirl= strlen(dir_adr);
 nsl= strlen(name_start);
 ret= Dirseq_new(&dirseq,dir_adr,0);
 if(ret<=0)
   goto ex;
 while(1) {
   ret= Dirseq_next_adr(dirseq,sub_name,0);
   if(ret<0)
     goto ex;
   if(ret==0)
 break;
   if(strncmp(name_start,sub_name,nsl)!=0)
 continue;
   spt= strchr(sub_name,'.');
   if(spt==NULL) {
     if(has_suffix)
 continue;
   } else {
     if(!has_suffix)
 continue;
     if(strcmp(spt+1,suffix)!=0)
 continue;
   }
   /* filter out youngest */
   if(dirl+strlen(sub_name)+1>=sizeof(try_adr))
 continue;
   sprintf(try_adr,"%s/%s",dir_adr,sub_name);
   if(stat(try_adr,&stbuf)==-1)
 continue;
   if(max_date_valid)
     if(stbuf.st_mtime<=max_date)
 continue;
   max_date= stbuf.st_mtime;
   max_date_valid= 1;
   strcpy(youngest_adr,try_adr);
 }
 ret= !!max_date_valid;
ex:;
 Dirseq_destroy(&dirseq,0);
 return(ret);
}

 
#ifndef Askme_fileliste_externaL 

int Fileliste__target_source_limit(line,sep,limit_pt,flag)
char *line;
char sep;
char **limit_pt;
int flag;
{
 char *npt;

 for(npt= line;*npt!=0;npt++) {
   if(*npt=='\\') {
     if(*(npt+1)!=0)
       npt++;
 continue;
   }
   if(*npt=='=')
 break;
 }
 if(*npt==0)
   npt= NULL;
 (*limit_pt)= npt;
 return(npt!=NULL);
}


int Fileliste__unescape_target(line,endpt,sep,flag)
char *line;
char *endpt;
char sep;
int flag;
{
 char *wpt,*rpt;

 if(endpt==NULL)
   endpt= line+strlen(line);

 /* unescape sep and '\\' */
 wpt= line;
 for(rpt= line; rpt<endpt; rpt++) {
   if(*rpt=='\\') {
     if(*(rpt+1)==sep || *(rpt+1)=='\\')
       rpt++;
   }
   *(wpt++)= *rpt;
 }
 *wpt= 0;
 return(1);
} 


int Fileliste__escape_target(target,flag)
char *target;
int flag;
{
 int i,offset= 0;
 /* escape '=' and '\\' in target address */
 for(i= strlen(target)-1;i>=0;i--)
   if(target[i]=='=' || target[i]=='\\')
     offset++;
 target[strlen(target)+offset]= 0;
 for(i= strlen(target)-1;i>=0;i--) {
   if(target[i]=='=' || target[i]=='\\') {
     target[i+offset]= target[i];
     offset--;
     target[i+offset]= '\\';
   } else {
     target[i+offset]= target[i];
   }
 }
 return(1);
}

#endif /* ! Askme_fileliste_externaL */


#ifndef Askme_cleanup_externaL

#include <signal.h>
typedef void (*sighandler_t)(int);


/* Signals to be caught */
static int signal_list[]= {
 SIGHUP, SIGINT, SIGQUIT, SIGILL, SIGABRT,
 SIGFPE, SIGSEGV, SIGPIPE, SIGALRM, SIGTERM,
 SIGUSR1, SIGUSR2, SIGXCPU, SIGTSTP, SIGTTIN,
 SIGTTOU,
 SIGBUS, SIGPOLL, SIGPROF, SIGSYS, SIGTRAP,
 SIGVTALRM, SIGXCPU, SIGXFSZ, -1
};
static char *signal_name_list[]= {
 "SIGHUP", "SIGINT", "SIGQUIT", "SIGILL", "SIGABRT",
 "SIGFPE", "SIGSEGV", "SIGPIPE", "SIGALRM", "SIGTERM",
 "SIGUSR1", "SIGUSR2", "SIGXCPU", "SIGTSTP", "SIGTTIN",
 "SIGTTOU",
 "SIGBUS", "SIGPOLL", "SIGPROF", "SIGSYS", "SIGTRAP",
 "SIGVTALRM", "SIGXCPU", "SIGXFSZ", "@"
};
static int signal_list_count= 24;


/* Signals not to be caught */
static int non_signal_list[]= {
  SIGKILL, SIGCHLD, SIGSTOP, SIGURG, -1
};
static int non_signal_list_count= 4;


/* run time dynamic part */
static char cleanup_msg[4096]= {""};
static int cleanup_exiting= 0;

static void *cleanup_app_handle= NULL;
static Cleanup_app_handler_T cleanup_app_handler= NULL;
static int cleanup_perform_app_handler_first= 0;


static int Cleanup_handler_exit(int exit_value, int signum, int flag)
{
 int ret;

 if(cleanup_perform_app_handler_first)
   if(cleanup_app_handler!=NULL) {
     ret= (*cleanup_app_handler)(cleanup_app_handle,signum,0);
     if(ret==2)
       return(2);
   }
 if(cleanup_exiting) {
   if(cleanup_msg[0]!=0)
     fprintf(stderr,"%s\n",cleanup_msg);
   fprintf(stderr,"cleanup: ABORT : repeat by pid=%d, signum=%d\n",
           getpid(),signum);
   return(0);
 }
 cleanup_exiting= 1;
 if(cleanup_msg[0]!=0)
   fprintf(stderr,"%s\n",cleanup_msg);
 alarm(0);
 if(!cleanup_perform_app_handler_first)
   if(cleanup_app_handler!=NULL)
     (*cleanup_app_handler)(cleanup_app_handle,signum,0);
 exit(exit_value);
}


static void Cleanup_handler_generic(int signum)
{
 int i;

 sprintf(cleanup_msg,"UNIX-SIGNAL caught:  %d  errno= %d",signum,errno);
 for(i= 0; i<signal_list_count; i++)
   if(signum==signal_list[i]) {
     sprintf(cleanup_msg,"UNIX-SIGNAL:  %s  errno= %d",
             signal_name_list[i],errno);
 break;
   }
 Cleanup_handler_exit(1,signum,0);
}


int Cleanup_set_handlers(void *handle, Cleanup_app_handler_T handler, int flag)
/*
 bit0= set to default handlers
 bit1= set to ignore
 bit2= set cleanup_perform_app_handler_first
 bit3= set SIGABRT to handler (makes sense with bits 0 or 1)
*/
{
 int i,j,max_sig= -1,min_sig= 0x7fffffff;
 sighandler_t sig_handler;

 cleanup_msg[0]= 0;
 cleanup_app_handle= handle;
 cleanup_app_handler= handler;
 if(flag&4)
   cleanup_perform_app_handler_first= 1;
 if(flag&1)
   sig_handler= SIG_DFL;
 else if(flag&2)
   sig_handler= SIG_IGN;
 else
   sig_handler= Cleanup_handler_generic;
 /* set all signal numbers between the lowest and highest in the list
    except those in the non-signal list */
 for(i= 0; i<signal_list_count; i++) {
   if(signal_list[i]>max_sig)
     max_sig= signal_list[i];
   if(signal_list[i]<min_sig)
     min_sig= signal_list[i];
 }
 for(i= min_sig; i<=max_sig; i++) {
   for(j= 0; j<non_signal_list_count; j++)
     if(i==non_signal_list[j])
   break;
   if(j>=non_signal_list_count) {
     if(i==SIGABRT && (flag&8))
       signal(i,Cleanup_handler_generic);
     else
       signal(i,sig_handler);
   }
 }
 return(1);
}


#endif /* ! Askme_cleanup_externaL */


#ifndef Askme_dirseq_externaL

/* ------------------------------------------------------------------------ */
/* DirseQ : crawl along a directory's content list */

static int Dirseq_buffer_sizE= 1000;

struct DirseQ {
  DIR *dirpt;
  int count;  
  char **buffer;
  int buffer_size;
  int buffer_fill;
  int buffer_rpt;
};


int Dirseq_new(o,dir,flag)
struct DirseQ **o;
char *dir;
int flag;
/*
 bit0= do not complain about failed opendir()
*/
{
 int ret,i;
 struct DirseQ *m;

 m= *o= TSOB_FELD(struct DirseQ,1);
 if(m==NULL)
   return(-1);
 m->dirpt= NULL;
 m->count= 0;
 m->buffer= NULL;
 m->buffer_size= 0;
 m->buffer_fill= 0;
 m->buffer_rpt= 0;
 m->buffer= TSOB_FELD(char *,Dirseq_buffer_sizE);
 if(m->buffer==NULL)
   {ret= -1; goto failed;}
 m->buffer_size= Dirseq_buffer_sizE;
 for(i= 0;i<m->buffer_size;i++)
   m->buffer[i]= NULL;
 if(dir[0]==0)
   m->dirpt= opendir(".");
 else
   m->dirpt= opendir(dir);
 if(m->dirpt==NULL) {
   if(!(flag&1))
     fprintf(stderr,"opendir(%s) failed : %s\n",dir,strerror(errno));
   ret= 0; goto failed;
 }
 return(1);
failed:;
 Dirseq_destroy(o,0);
 return(ret);
}


int Dirseq_destroy(o,flag)
struct DirseQ **o;
int flag;
{
 int i;

 if(*o==NULL)
   return(0);
 if((*o)->dirpt!=NULL)
   closedir((*o)->dirpt);
 if((*o)->buffer!=NULL) {
   for(i=0;i<(*o)->buffer_size;i++)
     if((*o)->buffer[i]!=NULL)
       free((*o)->buffer[i]);
   free((char *) (*o)->buffer);
 }
 free((char *) *o);
 (*o)= NULL;
 return(1);
}


int Dirseq_rewind(o,flag)
struct DirseQ *o;
int flag;
{
 rewinddir(o->dirpt);
 return(1);
}


int Dirseq_next_adr(o,reply,flag)
struct DirseQ *o;
char reply[4096];
int flag;
/*
flag:
 bit0= permission to use buffer
 bit1= do not increment counter
 bit2= ignore buffer in any case
 bit3= do not exclude '.' and '..'
 bit4= sort buffer
 bit5= sort only incomplete last buffer
return:
 <0 error
 0= no more entries available
 1= ok, reply is valid
*/
{
 int ret;
 struct dirent *entry;
 char *name;

 static int override_flag_0= 0,override_flag_1= 32;
 flag= (flag&~override_flag_0)|override_flag_1;

 if((flag&1) && o->buffer_rpt>=o->buffer_fill) {
   /* permission to buffer and buffer empty : load a buffer */
   ret= Dirseq_next_adrblock(o,o->buffer,&(o->buffer_fill),
                             o->buffer_size,2|4|(flag&16));
   if(ret<=0)
     return(ret);
   o->buffer_rpt= 0;
   if((flag&32) && o->buffer_fill<o->buffer_size && o->buffer_fill>0) 
     Sort_argv(o->buffer_fill,o->buffer,0);
 }
 if(o->buffer_rpt<o->buffer_fill && !(flag&4)) {
   strcpy(reply,o->buffer[o->buffer_rpt]);
   Sregex_string(&(o->buffer[o->buffer_rpt]),NULL,0);
   (o->buffer_rpt)++;
   if(!(flag&2))
     o->count++;
   return(1);
 }
 do {
   entry= readdir(o->dirpt);
   if(entry==NULL) {
     /* >>> how to distinguish error from EOF , do i need a (FILE *) ? */
     return(0);
   }
   if(strlen(entry->d_name)>4095) {
     fprintf(stderr,"--- oversized directory entry (number %d) :\n    %s",
                    o->count+1,entry->d_name);
     return(-1);
   }
   name= entry->d_name;
   if(flag&8)
 break;
   /* skip "." and ".." */
 } while(name[0]=='.' && ((name[1]=='.' && name[2]==0) || name[1]==0));
 strcpy(reply,name);
 if(!(flag&2))
   o->count++;
 return(1);
}


int Dirseq_next_adrblock(o,replies,reply_count,max_replies,flag)
struct DirseQ *o;
char *replies[];                       /* a vector of Sregex_string pointers */
int *reply_count;
int max_replies;
int flag;
/*
flag:
 bit0= permission to use buffer
 bit1= do not increment counter
 bit2= ignore buffer in any case
 bit4= sort replies
return:
 <0 error
 0= no more entries available
 1= ok, reply is valid
*/
{
 int i,ret;
 char reply[4096];

 *reply_count= 0;
 for(i=0;i<max_replies;i++) {
   ret= Dirseq_next_adr(o,reply,flag&(1|2|4));
   if(ret<0)
     return(ret);
   if(ret==0)
 break;
   if(Sregex_string(&(replies[i]),reply,0)<=0)
     return(-1);
   (*reply_count)++;
 }
 if((*reply_count)==0)
   return(0);
 if(flag&16) 
   Sort_argv(*reply_count,replies,0);
 return(1);
}


#endif /* ! Askme_dirseq_externaL */


/* ------------------------------------------------------------------------ */


int Bitmap2d_new(bmap,width,length,flag)
struct Bitmap2D **bmap;
int width;
int length;
int flag;
{
 int byte_width,size;
 struct Bitmap2D *m;

 byte_width= width/8+1; 
 size= byte_width*length;
 if(size<=0) {
   fprintf(stderr,
           "Programm error: attempt to create bitmap of size %d x %d\n",
           width,length);
   return(0);
 }
 *bmap= m= TSOB_FELD(struct Bitmap2D,1);
 if(m==NULL)
   return(-1);
 m->width= byte_width;
 m->size= size;
 m->map= NULL;

 m->map= TSOB_FELD(char,size);
 if(m->map==NULL)
   goto failed;
 memset(m->map,0,(size_t) m->size);
 return(1);
failed:;
 Bitmap2d_destroy(bmap,0);
 return(-1);
}


int Bitmap2d_destroy(bmap,flag)
struct Bitmap2D **bmap;
int flag;
{
 if(*bmap==NULL)
   return(0);
 if((*bmap)->map!=NULL)
   free((*bmap)->map);
 free((char *) *bmap);
 *bmap= NULL;
 return(1);
}


int Bitmap2d_init(o,flag)
struct Bitmap2D *o;
int flag;
/*
 bit0-7: value for memset
*/
{
 memset(o->map,(flag&255),(size_t) o->size);
 return(1);
}


int Bitmap2d_idx(o,x,y,flag)
struct Bitmap2D *o;
int x;
int y;
int flag;
/*
 bit1= do not report program error
*/
{
 int idx;

 idx= y * o->width + x/8;
 if(idx<0 || idx>=o->size) {
   if(!(flag&2))
     fprintf(stderr,"Programm error: idx=%d out of bitmap range [0,%d]\n",
                    idx,o->size-1);
   return(-1);
 }
 return(idx);
}


int Bitmap2d_is_set(o,x,y,flag)
struct Bitmap2D *o;
int x;
int y;
int flag;
/*
 bit0= set bit if not already set
 bit1= do not report program error
*/
{
 int idx;

 idx= Bitmap2d_idx(o,x,y,flag&2);
 if(idx<0)
   return(-1);
 if((o->map[idx]>>(x%8))&1)
   return(1);
 if(flag&1)
   o->map[idx]|= 1<<(x%8);
 return(0);
}


int Bitmap2d_set(o,x,y,flag)
struct Bitmap2D *o;
int x;
int y;
int flag;
/*
 bit1= do not report program error
*/
{
 int idx;

 idx= Bitmap2d_idx(o,x,y,flag&2);
 if(idx<0)
   return(-1);
 o->map[idx]|= 1<<(x%8);
 return(1);
}


int Bitmap2d_op(o,op1,op2,flag)
struct Bitmap2D *o;
struct Bitmap2D *op1;
struct Bitmap2D *op2;
int flag;
/*
 bit0-3 : logical operation   
  2=  op1-op2
 14=  op1|op2
*/
{
 int op,i,j,value;

 if(o->size!=op1->size || o->size!=op2->size)
   return(-1);
 op= flag&15;
 if(op==2) {
   for(i= 0;i<o->size;i++)
     o->map[i]= op1->map[i] & ~op2->map[i];
 } else if(op==14) {
   for(i= 0;i<o->size;i++)
     o->map[i]= op1->map[i] | op2->map[i];
 } else { /* general encoding : op bits correspond to truth table of op */
   for(i= 0;i<o->size;i++)
     for(j=0;j<8;j++) {
       value= (op>>( ((op1->map[i]>>j)&1) | (((op2->map[i]>>j)&1)<<1) ))&1;
       o->map[i]= (o->map[i]&~(1<<j))|(value<<j);
     }
 }
 return(1);
}


int Bitmap2d_count(o,flag)
struct Bitmap2D *o;
int flag;
{
 int i,j,count= 0;

 for(i=0;i<o->size;i++) {
   if(o->map[i]==0)
 continue;
   for(j=0;j<8;j++)
     if(o->map[i]&(1<<j))
       count++;
 }
 return(count);
}


/* ----------------------------------------------------------------------- */


int Lreader_new(reader,adr,fp,flag)
struct LreadeR **reader;
char *adr;
FILE *fp;
int flag;
{
 int i;
 struct LreadeR *o;

 *reader= o= TSOB_FELD(struct LreadeR,1);
 if(o==NULL)
   return(-1);
 strcpy(o->adr,adr);
 o->fp= fp;
 o->line[0]= 0;
 o->linecount= 0;
 o->errcode= -1;

 o->list_type[0]= 0;
 o->has_wd= -1;
 o->initial_wd[0]= 0;
 o->id[0]= 0;
 o->date[0]= 0;
 o->volumes_start_at_0= -1;
 o->layout= 8;
 o->highest_volume= -1;

 o->level= 0;
 o->younger_than[0]= 0;
 o->volume= -1;

 o->valid= 0;
 o->line_source= o->absolute_source;
 o->line_target= o->absolute_target;
 o->absolute_source[0]= 0;
 o->absolute_target[0]= 0;
 o->type_string[0]= 0;
 o->file_type= 0;
 o->part_status= 0;
 o->part_start= 0;
 o->part_size= 0;
 o->part_full_size= 0;
 o->file_size= 0;
 o->crc32= 0;
 for(i=0;i<16;i++)
   o->md5[i]= 0;
 o->permissions= 0;
 o->owner[0]= 0;
 o->group[0]= 0;
 o->timestamp[0]= 0;
 o->mtime= 0;
 o->ctime= 0;

 o->source_root= NULL;
 o->target_root= NULL;
 return(1);
}


int Lreader_destroy(reader,flag)
struct LreadeR **reader;
int flag;
{
 if((*reader)==NULL)
   return(0);
 free((char *) *reader);
 (*reader)= NULL;
 return(1);
}


int Lreader_set_errcode(o,errcode,flag)
struct LreadeR *o;
int errcode;
int flag;
{
 o->errcode= errcode;
 return(1);
}


int Lreader_get_errcode(o,flag)
struct LreadeR *o;
int flag;
{
 return(o->errcode);
}


char *Lreader_strerror(o,flag)
struct LreadeR *o;
int flag;
{
 static char msg_data[3*4096],*msg;
 sprintf(msg_data,"While reading line %.lf of '%s' :\n",
                  (double) o->linecount,o->adr);
 if(o->errcode<0) {
   strcpy(msg_data+strlen(msg_data),
 "(Readability and format of this line was ok. Its meaning obviously was not.)"
         );
   return(msg_data);
 }
 strcat(msg_data,"*** ");
 msg= msg_data+strlen(msg_data);
 switch(o->errcode) {
 case 0:
   strcpy(msg,"Unexpected end of input");
 break;case 1:
   sprintf(msg,"Read failed");
   if(errno!=0)
     sprintf(msg+strlen(msg)," : %s",strerror(errno));
 break;case 2:
   strcpy(msg,"Line component is too long to be accepted");
 break;case 3:
   strcpy(msg,"Volume number is out of its expected range");
 break;case 4:
   strcpy(msg,"Wrong list type");
 break;case 5:
   strcpy(msg,"Unknown layout code");
 break;case 6:
   strcpy(msg,"Violation of announced layout");
 break;case 7: 
   strcpy(msg,"Relative address without absolute cursor position");
 break;case 8: 
   strcpy(msg,"Relative address leads to oversized absolute address");
 break;case 9: 
   strcpy(msg,"Absolute address without target");
 break;case 10: 
   strcpy(msg,"Address component not valid");
 break;case 11:
   strcpy(msg,"Program error: no file pointer");
 break;case 12:
   strcpy(msg,"Unexpected kind of line in header part");
 break; default:
   sprintf(msg,"Program error: unknown error code %d",o->errcode);
 }
 strcat(msg," ***");
 msg= msg_data;
 return(msg);
}


int Lreader_read_line(o,flag)
struct LreadeR *o;
int flag;
/*
 >>> bit0= permission to issue pacifier messages
*/
/*
 return:
 <0= error, especially:
  0= eof 
 -1= error on Sfile_fgets()
 -2= line component is too long to be accepted
 -3= volume number is out of its expected range
 -4= wrong list type
 -5= unknown layout code
 -6= violation of announced layout
 -7= relative without absolute cursor (e.g. relative file after absolute file)
 -8= relative address leads to oversized absolute address
 -9= absolute address without target
-10= address component not valid
-11= program error: no file pointer
-12= unexpected kind of line in header part

  1= /x/x/x/   absolute directory address
  2= /x/x/x    absolute file address
  3= ./x/      descent into directory x
  4= ./x       file x in current directory
  5= ..        one directory up
  6= ../       to start directory
  7=           unexpected or unidentified line type (e.g. empty line)
  8= #=# xxxx  additional information (unspecified meaning)
  9= list type (e.g. "#=# backup content list")
 10= hostname  (e.g. "#=# hostname ts4")
 11= level     (e.g. "#=# level    2  ( -younger_than A60309.121653 )") 
 12= date      (e.g. "#=# date     A60325.211935")
 13= layout    (e.g. "#=# line layout : type size adr"
 14= pre-wd    (e.g. "#=# next line tells current working directory")
 15= wd        (e.g. "/home/thomas/")
 16= volume    (e.g. "#=# volume 1")
 17= end       (e.g. "#=# end")
 18= id        (e.g. "#=# id  bfacae8d94d574b8a4cd0418fd72e9ef")
 19= /x/x/x    part info: absolute file address file part ("p")
 20= /x/x/x    part info: absolute file address full original of part ("P")
*/
{
 int ret= 7,x,i;
 double num= -1.0;
 char *cpt,*npt,buf[3];

 o->errcode= -1;
 if(o->fp==NULL) 
   {ret= -11; goto ex;}

 /* >>> 0520 : concatenate addresses with newline chars */

 if(Sfile_fgets(o->line,sizeof(o->line)-1,o->fp)==NULL) {
   if(ferror(o->fp))
     {ret= -1; goto ex;}
   {ret= 0; goto ex;}
 }
 o->linecount++;

 /* >>> classify, parse and interpret line */;
 if(o->linecount==1){
   if(strcmp(o->line,"#=# complete backup area file list")==0 ||
      strcmp(o->line,"#=# backup content list")==0 ||
      strcmp(o->line,"#=# complete backup area directory list")==0 ||
      strcmp(o->line,"#=# vanished file objects list")==0) {
     strcpy(o->list_type,o->line+4);
     ret= 9;
   } else
     {ret= -4; goto ex;}

 } else if(o->has_wd<0) {
   if(strncmp(o->line,"#=# hostname ",13)==0) {
     ret= 10;
   } else if(strncmp(o->line,"#=# id ",7)==0) {
     /* #=# id     bfacae8d94d574b8a4cd0418fd72e9ef */
     if(strlen(o->line)>80+7)
       {ret= -2; goto ex;}
     sscanf(o->line+7,"%s",o->id);
     ret= 18;
   } else if(strncmp(o->line,"#=# level ",10)==0) {
     /* #=# level    1  ( -younger_than A60329.161456 ) */
     sscanf(o->line+10,"%lf",&num);
     if(num<0)
       {ret= -3; goto ex;}
     o->level= num;
     cpt= strstr(o->line,"-younger_than ");
     if(cpt==NULL)
       o->younger_than[0]= 0;
     else {
       if(strlen(cpt)>80)
         {ret= -2; goto ex;}
       sscanf(cpt+14,"%s",o->younger_than);
     }  
     ret= 11;
   } else if(strncmp(o->line,"#=# date ",9)==0) {
     /* #=# date     A60329.165455 */
     if(strlen(o->line)>80)
       {ret= -2; goto ex;}
     sscanf(o->line+9,"%s",o->date);
     ret= 12;
   } else if(strncmp(o->line,"#=# line layout : ",18)==0) {
     /* #=# line layout : type size md5 adr */
     o->layout= 0;
     for(cpt= o->line+18;(*cpt)!=0;cpt++) {
       if((*cpt)==' ')
     continue;
       if(strncmp(cpt,"type ",5)==0) {
         o->layout|= 1;
         cpt+= 4;
       } else if(strncmp(cpt,"chmod ",6)==0) {
         o->layout|= 2;
         cpt+= 5;
       } else if(strncmp(cpt,"chown ",6)==0) {
         o->layout|= 4;
         cpt+= 5;
       } else if(strncmp(cpt,"chgrp ",6)==0) {
         o->layout|= 8;
         cpt+= 5;
       } else if(strncmp(cpt,"date ",5)==0) {
         o->layout|= 16;
         cpt+= 4;
       } else if(strncmp(cpt,"size ",5)==0) {
         o->layout|= 32;
         cpt+= 4;
       } else if(strncmp(cpt,"crc ",4)==0 ||
                 strncmp(cpt,"md5 ",4)==0 ||
                 strncmp(cpt,"chk ",4)==0) {
         o->layout|= 64;
         cpt+= 3;
       } else if(strncmp(cpt,"adr",3)==0 && (cpt[3]==' ' || cpt[3]==0)) {
         /* actually layout adr is redundant and for human eye only */
         cpt+= 3;
       } else {
         ret= -5; goto ex;
       }
     }
     ret= 13;
   } else if(strcmp(o->line,"#=# next line tells current working directory")
             ==0) {
     o->has_wd= 0;
     ret= 14;
   } else if(o->line[0]=='#' && o->line[1]=='=' && o->line[2]=='#' 
             && o->line[3]==' ') {
     ret= 8;
   } else {
     ret= -12; goto ex;
   } 

 } else if(o->has_wd==0) {
   if(strlen(o->line)>=4096)
     {ret= -2; goto ex;}
   strcpy(o->initial_wd,o->line);
   if(o->initial_wd[0]==0)
     {ret= -12; goto ex;}
   if(o->initial_wd[strlen(o->initial_wd)-1]!='/')
     {ret= -12; goto ex;}
   strcpy(o->absolute_source,o->initial_wd);
   o->line_source= o->absolute_source+strlen(o->absolute_source);
   o->source_cursor= o->absolute_source+strlen(o->absolute_source);
   strcpy(o->absolute_target,o->initial_wd);
   o->line_target= o->absolute_target+strlen(o->absolute_target);
   o->target_cursor= o->absolute_target+strlen(o->absolute_target);
   o->has_wd= 1;
   ret= 15;

 } else if(o->line[0]=='#' && o->line[1]=='=' && o->line[2]=='#' 
           && o->line[3]==' ') {
   if(strncmp(o->line,"#=# volume ",11)==0) {
     /* #=# volume 1 */
     sscanf(o->line+11,"%lf",&num);
     if(num<0)
       {ret= -3; goto ex;}
     if(o->volumes_start_at_0==-1) {
       if(num==0)
         o->volumes_start_at_0= 1;
       else if(num==1)
         o->volumes_start_at_0= 0;
       else
         {ret= -3; goto ex;}
     }
     o->volume= num+o->volumes_start_at_0;    
     if(o->highest_volume<0 || o->volume>o->highest_volume)
       o->highest_volume= o->volume;
     ret= 16;
   } else if(strcmp(o->line,"#=# end")==0) {
     ret= 17;
   } else
     ret= 8;

 } else if(o->line[0]=='.' && o->line[1]=='.') {
   cpt= o->line;
   if(cpt[2]==0) {
     /* relative hop-up */

/* 0571 A80308 : this could lead to string overflow of the elsewise
                 fewly used variables o->absolute_source, o->absolute_target
*/
#ifdef Reenable_bug_0571
     npt= o->absolute_source+strlen(o->absolute_source);
#else
     npt= o->source_cursor-2;
#endif


     if(npt-o->absolute_source>2) {
       for(npt= npt-2;npt>o->absolute_source;npt--)
         if(*npt=='/')
       break;
         npt++;
     } else
       npt= o->absolute_source+1;
     *npt= 0;
     o->source_cursor= o->line_source= npt;
     npt= o->absolute_target+strlen(o->absolute_target);
     if(npt-o->absolute_target>2) {
       for(npt= npt-2;npt>o->absolute_target;npt--)
         if(*npt=='/')
       break;
       npt++;
     } else
       npt= o->absolute_target+1;
     *npt= 0;
     o->target_cursor= o->line_target= npt;
     ret= 5;
   } else if(strcmp(cpt,"../")==0) {
     /* full hop-up */
     strcpy(o->absolute_source,o->initial_wd);
     o->line_source= o->source_cursor= 
                                 o->absolute_source+strlen(o->absolute_source);
     strcpy(o->absolute_target,o->initial_wd);
     o->line_target= o->target_cursor=
                                 o->absolute_target+strlen(o->absolute_target);
     ret= 6;
   } else {
     ret= -10; goto ex;
   }

 } else {

   /* >>> decode into  typestring, length, checksum, address */;
   o->valid= 0;
   o->part_status= 0;
   cpt= o->line;
   if(o->layout&1) {
     npt= strchr(cpt,' ');
     if(npt==NULL || npt-cpt>=4096)
       {ret= -6; goto ex;}
     strncpy(o->type_string,cpt,npt-cpt);
     o->type_string[npt-cpt]= 0;

     /* interpret typestring */
     if(*cpt=='0')
       o->file_type= 0;
     if(*cpt=='1' || *cpt=='f' || *cpt=='F' || *cpt=='-')
       o->file_type= 1;
     else if(*cpt=='2' || *cpt=='d' || *cpt=='D')
       o->file_type= 2;
     else if(*cpt=='3' || *cpt=='l' || *cpt=='L')
       o->file_type= 3;
     else if(*cpt=='4' || *cpt=='p' || *cpt=='P')
       o->file_type= 4;
     else if(*cpt=='5' || *cpt=='m' || *cpt=='M')
       o->file_type= 5;
     else if(*cpt=='6' || *cpt=='b' || *cpt=='B')
       o->file_type= 6;
     else
       o->file_type= 0;
     o->part_status= 0;
     while(*(cpt+1)==',') {
       if(*(cpt+2)=='c') {
         /* no portable scanf() formatters for time_t, ino_t, dev_t, off_t */
         o->ctime= 0;
         for(i=3;cpt[i]>='0' && cpt[i]<='9';i++)
           o->ctime= (o->ctime*10)+(cpt[i]-'0');
         o->valid|= 1<<16;
       } else if(*(cpt+2)=='d') {
         o->device= 0;
         for(i=3;cpt[i]>='0' && cpt[i]<='9';i++)
           o->device= (o->device*10)+(cpt[i]-'0');
         o->valid|= 1<<14;
       } else if(*(cpt+2)=='i') {
         o->inode= 0;
         for(i=3;cpt[i]>='0' && cpt[i]<='9';i++)
           o->inode= (o->inode*10)+(cpt[i]-'0');
         o->valid|= 1<<13;

       } else if(*(cpt+2)=='P') { /* read startbyte and bytecount */
         o->part_start= 0;
         for(i=3;cpt[i]>='0' && cpt[i]<='9';i++)
           o->part_start= (o->part_start*10)+(cpt[i]-'0');
         o->part_size= 0;
         if(cpt[i]!=',' && cpt[i]!=' ' && cpt[i]!=0) 
           for(i++;cpt[i]>='0' && cpt[i]<='9';i++)
             o->part_size= (o->part_size*10)+(cpt[i]-'0');
         o->part_status= 2;

       } else if(*(cpt+2)=='p') {
                               /* read startbyte and original_full_bytecount */
         o->part_start= 0;
         for(i=3;cpt[i]>='0' && cpt[i]<='9';i++)
           o->part_start= (o->part_start*10)+(cpt[i]-'0');
         o->part_full_size= 0;
         if(cpt[i]!=',' && cpt[1]!=' ' && cpt[i]!=0) 
           for(i++;cpt[i]>='0' && cpt[i]<='9';i++)
             o->part_full_size= (o->part_full_size*10)+(cpt[i]-'0');

         o->part_status= 1;
       } else if(*(cpt+2)=='t') {
         o->mtime= 0;
         for(i=3;cpt[i]>='0' && cpt[i]<='9';i++)
           o->mtime= (o->mtime*10)+(cpt[i]-'0');
         o->valid|= 1<<15;
       } else {
         /* unknown extra info type : ignore and advance */;
       }
       cpt= strchr(cpt+2,',');
       if(cpt==NULL || cpt>npt)
     break;
       cpt--;
     }
     o->valid|= 1;
     cpt= npt;
   }

   if(o->layout&2) { /* chmod */
     unsigned int uint_value;
     for(; *cpt==' '; cpt++);
     if(*cpt==0)
       {ret= -6; goto ex;}
     sscanf(cpt,"%o",&uint_value);
     o->permissions= uint_value&0xfff;
     o->permissions|= (1<<31);
     o->valid|= 2;
     for(; *cpt!=0 && *cpt!=' '; cpt++);
     if(*cpt==0)
       {ret= -6; goto ex;}
   }

   if(o->layout&4) { /* owner */
     for(; *cpt==' '; cpt++);
     if(*cpt==0)
       {ret= -6; goto ex;}
     for(npt= cpt; *npt!=0 && *npt!=' ';npt++);
     if(*npt==0 || npt-cpt>=sizeof(o->owner))
       {ret= -6; goto ex;}
     strncpy(o->owner,cpt,npt-cpt);
     o->owner[npt-cpt]= 0;
     o->valid|= 4;
     cpt= npt;
   }

   if(o->layout&8) { /* group */
     for(; *cpt==' '; cpt++);
     if(*cpt==0)
       {ret= -6; goto ex;}
     for(npt= cpt; *npt!=0 && *npt!=' ';npt++);
     if(*npt==0 || npt-cpt>=sizeof(o->group))
       {ret= -6; goto ex;}
     strncpy(o->group,cpt,npt-cpt);
     o->group[npt-cpt]= 0;
     o->valid|= 8;
     cpt= npt;
   }

   if(o->layout&16) { /* date */
     for(; *cpt==' '; cpt++);
     if(*cpt==0)
       {ret= -6; goto ex;}
     for(npt= cpt; *npt!=0 && *npt!=' ';npt++);
     if(*npt==0 || npt-cpt>=sizeof(o->timestamp))
       {ret= -6; goto ex;}
     strncpy(o->timestamp,cpt,npt-cpt);
     o->timestamp[npt-cpt]= 0;
     o->valid|= 16;
     cpt= npt;
   }

   if(o->layout&32) { /* size */
     for(; *cpt==' '; cpt++);
     if(*cpt==0)
       {ret= -6; goto ex;}
     o->file_size= 0;
     if(*cpt!='-') {
       for(i=0;cpt[i]>='0' && cpt[i]<='9';i++)
         o->file_size= (o->file_size*10)+(cpt[i]-'0');
       o->valid|= 32;
     }
     for(; *cpt!=0 && *cpt!=' ';cpt++);
     if(*cpt==0)
       {ret= -6; goto ex;}
   }

   if(o->layout&64) { /* checksum */
     for(; *cpt==' '; cpt++);
     if(*cpt==0)
       {ret= -6; goto ex;}
     if(strcmp(o->list_type,"complete backup area directory list")==0) {
       /* There was a bug in 0.8.6 about directory lists layout announcing crc
          but not delivering it */
       if(*cpt=='.' || *cpt=='/')
         goto skip_crc;
     }
     for(npt= cpt; *npt!=0 && *npt!=' ';npt++)
     if(*npt==0)
       {ret= -6; goto ex;}
     if(npt-cpt==2) {
       ;
     } else if(npt-cpt==8) {
       sscanf(cpt,"%lx",&(o->crc32));
     } else if(npt-cpt==32){
       o->valid|= 1<<12;
       buf[2]= 0;
       for(i=0;i<16;i++) {
         buf[0]= cpt[2*i];
         buf[1]= cpt[2*i+1];
         sscanf(buf,"%x",&x);
         o->md5[i]= x;
       }
     } else
       {ret= -6; goto ex;}
     o->valid|= 64;
     for(cpt= npt; *cpt!=0 && *cpt!=' '; cpt++);
   }

skip_crc:;
   for(; *cpt==' '; cpt++);
   /* evaluate kind of address , set ret */;
   strcpy(o->linebuffer,cpt);
   cpt= o->linebuffer;
   if(*cpt!='.' && *cpt!='/')
     {ret= 7; goto ex;}
   npt= NULL;
   if(cpt[0]=='/') {
     /* absolute target=source address */
     Fileliste__target_source_limit(cpt,'=',&npt,0);
     if(npt!=NULL) {
       *npt= 0;
       Fileliste__unescape_target(cpt,npt,'=',0);
       if(strlen(cpt)>=4096)
         {ret= -2; goto ex;}
       strcpy(o->absolute_target,cpt);
       strcpy(o->absolute_source,npt+1);
       o->line_source= o->absolute_source;
       o->line_target= o->absolute_target;
       if(o->line_source[0]==0)
         {ret= -6; goto ex;}
       if(o->line_source[strlen(o->line_source)-1]=='/' ||
          o->file_type==2) {
         o->source_cursor= o->absolute_source+strlen(o->absolute_source);
         o->target_cursor= o->absolute_target+strlen(o->absolute_target);
         ret= 1;
       } else {
         o->source_cursor= NULL;
         o->target_cursor= NULL;
         ret= 2;
       }
     } else {
       ret= -9; goto ex;
     }
   } else if(cpt[1]==0) {
     {ret= -10; goto ex;}
   } else if(cpt[0]=='.' && cpt[1]=='/') {
     /* relative source address (with implicitily same relative target) */
     if(o->source_cursor==NULL || o->target_cursor==NULL)
       {ret= -7; goto ex;}
     if((o->source_cursor-o->absolute_source)+strlen(cpt)-2>=4096)
       {ret= -8; goto ex;}
     if((o->target_cursor-o->absolute_target)+strlen(cpt)-2>=4096)
       {ret= -8; goto ex;}
     strcpy(o->source_cursor,cpt+2);
     strcpy(o->target_cursor,cpt+2);
     o->line_source= o->source_cursor;
     o->line_target= o->target_cursor;
     ret= 4;
     if(o->line_source[strlen(o->line_source)-1]=='/') {
       o->source_cursor= o->line_source+strlen(o->line_source);
       o->target_cursor= o->line_target+strlen(o->line_target);
       ret= 3;
     }

   } else {
     {ret= -10; goto ex;}
   }
 }
ex:;
 if(ret<=0)
   o->errcode= -ret;
 if(ret==2 && o->part_status==1)
   ret= 19;
 if(ret==2 && o->part_status==2)
   ret= 20;
 return(ret);
}


int Lreader_get_line(o,linept,count,flag)
struct LreadeR *o;
char **linept;
off_t *count;
int flag;
{
 *linept= o->line;
 *count= o->linecount;
 return(1);
}


int Lreader_get_list_type(o,typept,flag)
struct LreadeR *o;
char **typept;
int flag;
{
 *typept= o->list_type;
 return(1);
}


int Lreader_get_id(o,idpt,flag)
struct LreadeR *o;
char **idpt;
int flag;
{
 *idpt= o->id;
 return(1);
}


int Lreader_get_src_target(o,sourcept,targetpt,flag)
struct LreadeR *o;
char **sourcept;
char **targetpt;
int flag;
/*
 bit0= get eventual relative address
*/
{
 if((flag&1)&&(o->line_source!=NULL)) {
   *sourcept= o->line_source;
   *targetpt= o->line_target;
 } else {
   *sourcept= o->absolute_source;
   *targetpt= o->absolute_target;
 }
 return(1); 
}


int Lreader_get_nodedata(o,sourcept,level,volume,file_type,flag)
struct LreadeR *o;
char **sourcept;
int *level;
int *volume;
int *file_type;
int flag;
/*
 bit0= get eventual relative address
*/
{
 if((flag&1)&&(o->line_source!=NULL))
   *sourcept= o->line_source;
 else
   *sourcept= o->absolute_source;
 *level= o->level;
 *volume= o->volume;
 *file_type= o->file_type;
 return(1); 
}


int Lreader_get_dir_data(o,device,flag)
struct LreadeR *o;
dev_t *device;
int flag;
{
 *device= o->device;
 if((o->valid)&(1<<14))
   return(0);
 return(1);
}


int Lreader_get_info_data(o,size,crc32,md5pt,inode,timestamp,mtime,ctime,
                          valid,flag)
struct LreadeR *o;
off_t *size;
unsigned long *crc32;
char **md5pt;
ino_t *inode;
char timestamp[81];
time_t *mtime;
time_t *ctime;
int *valid;
int flag;
/*
 return:
  bit0= size valid
  bit1= crc32 valid
  bit2= md5pt valid
  bit3= inode valid
  bit4= timestamp valid
  bit5= mtime valid
  bit6= ctime valid
*/
{
 int ret= 0;

 *size= o->file_size;
 *crc32= o->crc32;
 *md5pt= o->md5;
 *inode= o->inode;
 strcpy(timestamp,o->timestamp);
 *mtime= o->mtime;
 *ctime= o->ctime;
 *valid= o->valid;
 if(o->valid&32)
   ret|= 1;
 if((o->valid)&64) {
   if((o->valid)&(1<<12))
     ret|= 4;
   else
     ret|= 2;
 }
 if((o->valid)&(1<<13))
   ret|= 8;
 if((o->valid)&16)
   ret|= 16;
 if((o->valid)&(1<<15))
   ret|= 32;
 if((o->valid)&(1<<16))
   ret|= 64;
 return(ret);
}


int Lreader_get_ownership(o,permissions,owner,group,valid,flag)
struct LreadeR *o;
int *permissions; /* see struct FsinfO.permissions */
char owner[4096];
char group[4096];
int *valid;
int flag;
/*
 return:
  bit0= permissions valid
  bit1= owner valid
  bit2= group valid
*/
{
 int ret= 0;

 *permissions= o->permissions;
 strcpy(owner,o->owner);
 strcpy(group,o->group);
 *valid= o->valid;
 if((o->permissions) & (1<<31))
   ret|= 1;
 if(o->valid&4)
   ret|= 2;
 if(o->valid&8)
   ret|= 4;
 return(1);
}


int Lreader_get_extra_data(o,targetpt,flag)
struct LreadeR *o;
char **targetpt;
int flag;
/*
 bit0= get eventual relative address
*/
{
 if((flag&1)&&(o->line_target!=NULL))
   *targetpt= o->line_target;
 else
   *targetpt= o->absolute_target;
 return(1);
}


int Lreader_get_highest_volume(o,highest_volume,flag)
struct LreadeR *o;
int *highest_volume;
int flag;
{
 *highest_volume= o->highest_volume;
 return(1);
}


int Lreader_get_id_info(o,datept,level,younger_than_pt,num_volumes,flag)
struct LreadeR *o;
char **datept;
int *level;
char **younger_than_pt;
int *num_volumes;
int flag;
{
 *datept= o->date;
 *level= o->level;
 *younger_than_pt= o->younger_than;
 *num_volumes= o->highest_volume;
 return(1);
}


int Lreader_get_initial_wd(o,initial_wd,flag)
struct LreadeR *o;
char initial_wd[4096];
int flag;
{
 strcpy(initial_wd,o->initial_wd);
 return(1);
}


int Lreader_set_root_nodes(o,source_root,target_root,flag)
struct LreadeR *o;
struct FsnodE *source_root;
struct FsnodE *target_root;
int flag;
{
 o->source_root= source_root;
 o->target_root= target_root;
 return(1);
}


int Lreader_get_root_nodes(o,source_root,target_root,flag)
struct LreadeR *o;
struct FsnodE **source_root;
struct FsnodE **target_root;
int flag;
{
 *source_root= o->source_root;
 *target_root= o->target_root;
 return(1);
}


int Lreader_get_part_info(o,part_status,part_start,part_size,part_full_size,
                          flag)
struct LreadeR *o;
int *part_status;
off_t *part_start;
off_t *part_size;
off_t *part_full_size;
int flag;
{
 *part_status= o->part_status;
 *part_start= o->part_start;
 *part_size= o->part_size;
 *part_full_size= o->part_full_size;
 return(1);
}


/* ----------------------------------------------------------------------- */


int Fsrptmap_new(rpt,name,width,length,next,flag)
struct FsrptmaP **rpt;
char *name;
int width;
int length;
struct FsrptmaP *next;
int flag;
{
 int ret;
 struct FsrptmaP *o;

 *rpt= o= TSOB_FELD(struct FsrptmaP,1);
 if(o==NULL)
   return(-1);
 o->name= NULL;
 o->bitmap= NULL;
 o->next= next;
 if(Sregex_string(&(o->name),name,0)<=0)
   goto failed;
 ret= Bitmap2d_new(&(o->bitmap),width,length,0);
 if(ret<=0)
   goto failed;
 return(1);
failed:;
 Fsrptmap_destroy(rpt,0);
 return(-1);
}


int Fsrptmap_destroy(rpt,flag)
struct FsrptmaP **rpt;
int flag;
/*
 bit0= destroy complete chain
*/
{
 struct FsrptmaP *o,*next_o;

 o= *rpt;
 if(o==NULL)
   return(0);
once_again:;
 next_o= o->next;
 Sregex_string(&(o->name),NULL,0);
 Bitmap2d_destroy(&(o->bitmap),0);
 free((char *) o);
 if(flag&1) {
   o= next_o;
   if(o!=NULL)
     goto once_again;
 }
 return(1);
}


/* ----------------------------------------------------------------------- */

int Fsinfo_new(info,size,crc32,md5,inode,flag)
struct FsinfO **info;
off_t size;
unsigned long crc32;
char *md5;
ino_t inode;
int flag;
{
 struct FsinfO *o;

 *info= o= TSOB_FELD(struct FsinfO,1);
 if(o==NULL)
   return(-1);
 o->size= size;
 o->crc32= crc32;
 o->md5= NULL;
 o->inode= inode;
 o->flag= flag;
 if(flag&(1<<12)) {
   o->md5= TSOB_FELD(char,16);
   if(o==NULL)
     goto failed;
   memcpy(o->md5,md5,16);
 }
 return(1);
failed:;
 Fsinfo_destroy(info,0);
 return(-1);
}


int Fsinfo_destroy(info,flag)
struct FsinfO **info;
int flag;
{
 if(*info==NULL)
   return(0);
 if((*info)->md5!=NULL)
   free((*info)->md5);
 free((char *) *info);
 (*info)= NULL;
 return(1);
}


/* ----------------------------------------------------------------------- */


int Fsdirinfo_new(info,device,flag)
struct FsdirinfO **info;
dev_t device;
int flag;
/*
 bit14= device is valid
*/
{
 int i;
 struct FsdirinfO *o;

 *info= o= TSOB_FELD(struct FsdirinfO,1);
 if(o==NULL)
   return(-1);
 o->flag= flag&(1<<14);
 o->device= device;
 o->hashsize= Fsdirinfo_initial_hashsizE;
 if(o->hashsize<=0)
   o->hashsize= 1;
 o->subcount= 0;
 o->sublist= TSOB_FELD(struct FsnodE *,o->hashsize);
 if(o->sublist==NULL)
   goto failed;
 for(i=0;i<o->hashsize;i++)
   o->sublist[i]= NULL;
 return(1);
failed:;
 Fsdirinfo_destroy(info,0);
 return(-1);
}


int Fsdirinfo_destroy(info,flag)
struct FsdirinfO **info;
int flag;
{
 int i;

 if(*info==NULL)
   return(0);
 if((*info)->sublist!=NULL) {
   for(i=0;i<(*info)->hashsize;i++)
     Fsnode_destroy(&((*info)->sublist[i]),1);
   free((char *) (*info)->sublist);
 }
 free((char *) *info);
 (*info)= NULL;
 return(1);
}


int Fsdir_info_hash_idx(o,name,flag)
struct FsdirinfO *o;
char *name;
int flag;
{
 unsigned long crc32;

 Sregex_crc32_yenc(name,strlen(name),&crc32,1);
 return((int) (crc32 % o->hashsize));
}


int Fsdirinfo_get_sublist(o,name,sublist,flag)
struct FsdirinfO *o;
char *name;
struct FsnodE **sublist;
int flag;
{
 *sublist= o->sublist[Fsdir_info_hash_idx(o,name,0)];
 return(1);
}


int Fsdirinfo_get_next_node(o,node,flag)
struct FsdirinfO *o;
struct FsnodE **node;
int flag;
/*
 bit0= do not accept nodes with tlink
 bit1= accept nodes with volume == -1
*/
{
 int idx= 0,valid= 1;

 if((*node)==NULL) {
   idx= 0;
search_next_start_node:;
   for(;idx<o->hashsize;idx++)
     if(o->sublist[idx]!=NULL) {
       *node= o->sublist[idx];
       if(flag&1)
         if((*node)->extra!=NULL)
           if((*node)->extra->tlink!=NULL)
   break;
       if((*node)->volume==-1)
         if(!(flag&2))
   break;
       return(1);
     }
   if(idx>=o->hashsize)
     return(0);
 }
 if((*node)->next==NULL) {
use_next_sublist:;
   idx= Fsdir_info_hash_idx(o,(*node)->name,0)+1;
   goto search_next_start_node;
 }
 while(1) {
   (*node)= (*node)->next;
   valid= 1;
   if(flag&1)
     if((*node)->extra!=NULL)
       if((*node)->extra->tlink!=NULL)
         valid= 0;
   if(valid && ((*node)->volume!=-1 || (flag&2)))
 break;
   if((*node)->next==NULL)
     goto use_next_sublist;
 }
 return(1);
}


int Fsdirinfo_widen_hashsize(o,flag)
struct FsdirinfO *o;
int flag;
/*
 bit0= make sure that there is no need for a new widening soon 
*/
{
 int ret,old_hashsize,new_hashsize,i,idx;
 unsigned long crc32;
 struct FsnodE **new_sublist,*node;

 old_hashsize= o->hashsize;
 new_hashsize= old_hashsize*Fsdirinfo_hash_expansioN;
 if(flag&1)
   if(new_hashsize*Fsdirinfo_avg_hash_lengtH <
      o->subcount*Fsdirinfo_hash_expansioN)
     new_hashsize= (Fsdirinfo_hash_expansioN*o->subcount)/
                    Fsdirinfo_avg_hash_lengtH +1;
 if(new_hashsize>Fsdirinfo_hash_maX)
   new_hashsize= Fsdirinfo_hash_maX;
 if(new_hashsize==old_hashsize)
   return(2); /* may not expand further */
 new_sublist= TSOB_FELD(struct FsnodE *,new_hashsize);
 if(new_sublist==NULL)
   {ret= -1; goto ex;}
 for(i=0;i<new_hashsize;i++)
   new_sublist[i]= NULL;

 for(i=0;i<old_hashsize;i++) {
   while(o->sublist[i]!= NULL) {
     node= o->sublist[i];
     o->sublist[i]= node->next;
     node->next= NULL;
     Sregex_crc32_yenc(node->name,strlen(node->name),&crc32,1);
     idx= (crc32 % new_hashsize);
     node->next= new_sublist[idx];
     new_sublist[idx]= node;
   } 
 }
 free((char *) o->sublist);
 o->sublist= new_sublist;
 o->hashsize= new_hashsize;
 ret= 1;
ex:;
 return(ret);
}


int Fsdirinfo_add_sub(o,node,flag)
struct FsdirinfO *o;
struct FsnodE *node;
int flag;
/*
 bit0= do not widen hashsize automatically
*/
{
 int idx,ret;
 struct FsnodE *list;

 idx= Fsdir_info_hash_idx(o,node->name,0);
 list= o->sublist[idx];
 node->next= list;
 o->sublist[idx]= node;
 (o->subcount)++;
 if(flag&1)
   return(1);
 if(o->subcount > o->hashsize*Fsdirinfo_avg_hash_lengtH) {
   ret= Fsdirinfo_widen_hashsize(o,1);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Fsdirinfo_outdate_sub(o,dir_node,old_node,new_node,flag)
struct FsdirinfO *o;
struct FsnodE *dir_node,*old_node,*new_node;
int flag;
/*
 bit2= do not keep old_node
 bit5= this is a target tree node, capture tlink chain
*/
{
 int ret,idx;
 struct FsnodE *node,*prev_node= NULL;
 struct FstlinK *tlink;

 if(!(flag&4)) {
   if(new_node->extra==NULL) {
     ret= Fsextra_new(&(new_node->extra),dir_node,NULL,NULL,NULL,NULL,NULL,0);
     if(ret<=0)
       return(-1);
   }
 }
 idx= Fsdir_info_hash_idx(o,old_node->name,0);
 for(node= o->sublist[idx]; node!=NULL; node= node->next) {
   if(node==old_node)
 break;
   prev_node= node;
 }
 if(node==NULL)
   return(0);
 if(flag&32) {
   if(new_node->extra==NULL || old_node->extra==NULL) {
     fprintf(stderr,
             "Programm error: wish to manipulate tlink on sparse nodes\n"); 
     return(-1);
   }
   tlink= new_node->extra->tlink;
   if(tlink!=NULL)
     for(;tlink->next_in_target!=NULL;tlink= tlink->next_in_target);
   tlink->next_in_target= old_node->extra->tlink;
   old_node->extra->tlink= NULL;
   for(tlink= tlink->next_in_target;tlink!=NULL;tlink= tlink->next_in_target)
     tlink->target= new_node;
 } else {
   tlink= NULL;
   if(old_node->extra!=NULL)
     if(old_node->extra->tlink!=NULL) {
       tlink= old_node->extra->tlink;
       tlink->status|= 1;
     }
   if(tlink==NULL) {

     /* >>> A60518 check wether new_node will have a different target than
                old_node */;
       /* >>> eventually set outdated tlink */;

   }
 }
 new_node->next= old_node->next;
 old_node->next= NULL;
 if(prev_node==NULL)
   o->sublist[idx]= new_node;
 else
   prev_node->next= new_node;
 if(flag&4) {
   Fsnode_destroy(&old_node,0);
 } else
   new_node->extra->outdated= old_node;
 return(1);
}


int Fsdirinfo_delete_sub(o,dir_node,old_node,flag)
struct FsdirinfO *o;
struct FsnodE *dir_node,*old_node;
int flag;
{
 int idx;
 struct FsnodE *node,*prev_node= NULL;

 idx= Fsdir_info_hash_idx(o,old_node->name,0);
 for(node= o->sublist[idx]; node!=NULL; node= node->next) {
   if(node==old_node)
 break;
   prev_node= node;
 }
 if(node==NULL)
   return(0);
 if(prev_node==NULL)
   o->sublist[idx]= node->next;
 else
   prev_node->next= node->next;
 Fsnode_destroy(&old_node,0);
 o->subcount--;
 return(1);
}


int Fsdirinfo_search_sub_name(o,found,name,flag)
struct FsdirinfO *o;
struct FsnodE **found;
char *name;
int flag;
/*
 bit0= accept nodes which are marked as deleted
*/
{
 int skip_deleted;
 struct FsnodE *node;

 skip_deleted= !(flag&1);
 for(node= o->sublist[Fsdir_info_hash_idx(o,name,0)];
     node!=NULL; node= node->next) {
   if(node->volume==-1 && skip_deleted)
 continue;
   if(strcmp(name,node->name)==0)
 break;
 }
 *found= node;
 return(node!=NULL);
} 


int Fsdirinfo_is_empty(o,flag)
struct FsdirinfO *o;
int flag;
/*
 bit0= accept empty sub directories as reason for return(0)
*/
{
 int i,ret;
 struct FsnodE *node;

 for(i=0;i<o->hashsize;i++) {
   for(node= o->sublist[i];node!=NULL; node= node->next) {
     if(node->volume==-1)
   continue;
     if(flag&1)
       return(0);
     if(node->extra==NULL)
       return(0);
     if(node->extra->dirinfo==NULL)
       return(0);
     ret= Fsdirinfo_is_empty(node->extra->dirinfo,0);
     if(ret<=0)
       return(ret);
   }
 }
 return(1);
}


/* ----------------------------------------------------------------------- */


int Fspartrec_new(o,level,volume,partno,partcount,start_byte,byte_count,flag)
struct FspartreC **o;
int level;
int volume;
int partno;
int partcount;
off_t start_byte;
off_t byte_count;
int flag;
{
 *o= TSOB_FELD(struct FspartreC,1);
 if((*o)==NULL)
   return(-1);
 (*o)->level= level;
 (*o)->volume= volume;
 (*o)->partno= partno;
 (*o)->partcount= partcount;
 (*o)->start_byte= start_byte;
 (*o)->byte_count= byte_count;
 (*o)->tempno= 0;
 (*o)->tempadr= NULL;
 return(1);
}


int Fspartrec_destroy(o,flag)
struct FspartreC **o;
int flag;
{
 if((*o)==NULL)
   return(0);
 if((*o)->tempadr!=NULL)
   free((*o)->tempadr);
 free((char *) *o);
 *o= NULL;
 return(1);
}



/* ----------------------------------------------------------------------- */


int Fstlink_new(lnk,target,target_path,source,source_dir,next_in_target,flag)
struct FstlinK **lnk;
struct FsnodE *target;
char *target_path;
struct FsnodE *source;
struct FsnodE *source_dir;
struct FstlinK *next_in_target;
int flag;
{
 struct FstlinK *o;

 *lnk= o= TSOB_FELD(struct FstlinK,1);
 if(o==NULL)
   return(-1);
 o->status= 0;
 o->target= target;
 o->target_path= NULL;
 o->source= source;
 o->source_dir= source_dir;
 o->partrec= NULL;
 o->next_in_source= NULL;
 o->next_in_target= next_in_target;
 if(Sregex_string(&(o->target_path),target_path,0)<=0)
   return(-1);
 return(1);
}


int Fstlink_destroy(lnk,flag)
struct FstlinK **lnk;
int flag;
/*
 bit0= destroy complete target chain of links
 bit1= destroy complete source chain of links
*/
{
 struct FstlinK *o,*next_in_target,*next_in_source;

 o= *lnk;
 if(o==NULL)
   return(0);
once_again_source:;
 next_in_source= o->next_in_source;
once_again_target:;
 next_in_target= o->next_in_target;
 /* tell target and source that the link is gone */
 if(o->source!=NULL)
   if(o->source->extra!=NULL)
     o->source->extra->tlink= NULL; /* beware: this might be == *lnk */
 if(o->target!=NULL)
   if(o->target->extra!=NULL)
     if(o->target->extra!=NULL)
       if(o->target->extra->tlink==o)
         o->target->extra->tlink= next_in_target;
 Sregex_string(&(o->target_path),NULL,0);
 free((char *) o);
 if((flag&1) && next_in_target!=NULL) {
   *lnk= o= next_in_target;
   goto once_again_target;
 }
 if((flag&2) && next_in_source!=NULL) {
   *lnk= o= next_in_source;
   goto once_again_source;
 }
 (*lnk)= NULL;
 return(1);
}


int Fstlink_sort_cmp(o1,o2)
struct FstlinK **o1;
struct FstlinK **o2;
{
 int ret;
 if((*o1)->source==NULL)
   return(-1);
 if((*o2)->source==NULL)
   return(1);
 ret= strcmp((*o1)->source->name,(*o2)->source->name);
 return(ret);
}


/* ----------------------------------------------------------------------- */


int Fsextra_new(extra,up,target,target_root,dirinfo,nodeinfo,outdated,flag)
struct FsextrA **extra;
struct FsnodE *up;
char *target;
struct FsnodE **target_root;
/* the following three will get managed by the new FsextrA : */
struct FsdirinfO *dirinfo;
struct FsinfO *nodeinfo;
struct FsnodE *outdated;
int flag;
{
 int ret;
 struct FsextrA *o;

 *extra= o= TSOB_FELD(struct FsextrA,1);
 if(o==NULL)
   return(-1);
 o->up= up;
 o->tlink= NULL;
 o->dirinfo= dirinfo;
 o->nodeinfo= nodeinfo;
 o->outdated= outdated;
 if(target!=NULL) {
   ret= Fsextra_set_target(o,target,target_root,NULL,NULL,0);
   if(ret<=0)
     goto failed;
   /* eventual o->tlink->source has to be set by the caller */
 }
 return(1);
failed:;
 Fsextra_destroy(extra,0);
 return(-1);
}


int Fsextra_destroy(extra,flag)
struct FsextrA **extra;
int flag;
{
 if(*extra==NULL)
   return(0);
 Fstlink_destroy(&((*extra)->tlink),0);
 Fsdirinfo_destroy(&((*extra)->dirinfo),0);
 Fsinfo_destroy(&((*extra)->nodeinfo),0);
 Fsnode_destroy(&((*extra)->outdated),1);
 free((char *) *extra);
 (*extra)= NULL;
 return(1);
}


int Fsextra_fwd_tgt_to_outdated(o,target_root,source_node,source_dir,
                                target_path,flag)
struct FsextrA *o;
struct FsnodE **target_root;
struct FsnodE *source_node;
struct FsnodE *source_dir;
char *target_path;
int flag;
{
 int ret;
 struct FsnodE *node;
 
 if(source_node!=NULL) if(source_node->extra!=NULL) {
   for(node= source_node->extra->outdated;node!=NULL;
       node= node->extra->outdated) {
     if(node->volume<0) {
       if(node->extra==NULL)
   break;
   continue;
     }
     if(node->extra!=NULL) {
       if(node->extra->tlink!=NULL)
   continue;
     } else {
       ret= Fsextra_new(&(node->extra),NULL,NULL,NULL,NULL,NULL,NULL,0);
       if(ret<=0)
         return(ret);
     }
     ret= Fsextra_set_target(node->extra,target_path,target_root,node,
                             source_dir,0);
     if(ret<=0)
       return(ret);
   }
 }
 return(1);
}


int Fsextra_forward_target(o,target_root,source_node,source_dir,target_path,
                           flag)
struct FsextrA *o;
struct FsnodE **target_root;
struct FsnodE *source_node;
struct FsnodE *source_dir;
char *target_path;
int flag;
{
 int ret,tl;
 char target[4096];
 struct FsnodE *node;
 struct FsdirinfO *d;

 d= o->dirinfo;
 if(d==NULL || o->tlink==NULL) {
   ret= Fsextra_fwd_tgt_to_outdated(o,target_root,source_node,source_dir,
                                    target_path,0);
   return(ret);
 }

 strcpy(target,target_path);
 tl= strlen(target);
 if(tl>0)
   if(target[tl-1]!='/') {
     target[tl]= '/';
     target[tl+1]= 0;
     tl++;
   }
 node= NULL;
 while(1) {
   ret= Fsdirinfo_get_next_node(d,&node,2);
   if(ret<=0 || node==NULL)
 break;
   if(node->extra==NULL) {
     ret= Fsextra_new(&(node->extra),NULL,NULL,NULL,NULL,NULL,NULL,0);
     if(ret<=0)
       return(ret);
   }
   if(node->extra->tlink!=NULL)
 continue;
   if(node->volume>=0) {
     if(node->extra->dirinfo!=NULL)
       sprintf(target+tl,"%s/",node->name);
     else
       strcpy(target+tl,node->name);
     ret= Fsextra_set_target(node->extra,target,target_root,node,d,0);
     if(ret<=0)
       return(ret);
   }
   ret= Fsextra_fwd_tgt_to_outdated(o,target_root,node,d,target,0);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Fsextra_set_target(o,target,target_root,source_node,source_dir,flag)
struct FsextrA *o;
char *target;
struct FsnodE **target_root;
struct FsnodE *source_node;
struct FsnodE *source_dir;
int flag;
/*
 bit6= write paketed error messages (see Write_to_channel())
*/
{
 int ret,l,old_depth;
 struct FsnodE *target_node;
 char old_target[4096],old_source[4096];

 /* determine current target. If equal to new target : return without action */
 old_target[0]= 0;
 if(source_node!=NULL && source_dir!=NULL) {
   ret= Fsnode_get_adr(source_node,source_dir,old_source,old_target,&old_depth,
                       0);
   if(ret<=0)
     return(ret);
   l= strlen(old_target);
   if(l>0)
     if(strncmp(target,old_target,l)==0)
       if(target[l]==0 || (target[l]=='/' && target[l+1]==0))
         return(2); 
 }
 if(old_target[0]!=0) {
   /* copy old link to all eventual sub nodes which have no own link */
   ret= Fsextra_forward_target(o,target_root,source_node,source_dir,target,0);
   if(ret<=0)
     return(-1);
 }
 if(o->tlink!=NULL)
   Fstlink_destroy(&(o->tlink),0);
 ret= Fsnode_graft_branch(target_root,&target_node,target,NULL,NULL,
                           32|(flag&64));
 if(ret<=0)
   return(-1);
 ret= Fstlink_new(&(o->tlink),target_node,target,
                  source_node,source_dir,target_node->extra->tlink,0);
 if(ret<=0)
   return(-1);
 target_node->extra->tlink= o->tlink;
 return(1);
}


/* ------------------------------------------------------------------------ */


int Fstoutdate_new(fsto,root,tip,next,flag)
struct FstoutdatE **fsto;
struct FsnodE *root;
struct FsnodE *tip;
struct FstoutdatE *next;
int flag;
{
 struct FstoutdatE *o;

 *fsto= o= TSOB_FELD(struct FstoutdatE,1);
 if(o==NULL)
   return(-1);
 o->root= root;
 o->tip= tip;
 o->next= next;
 return(1);
}


int Fstoutdate_destroy(fsto,flag)
struct FstoutdatE **fsto;
int flag;
/*
 bit0= destroy complete chain of nodes
*/
{
 struct FstoutdatE *o,*next_fsto;

 if(*fsto==NULL)
   return(0);
 o= *fsto;
once_again:;
 next_fsto= o->next;
 free((char *) o);
 if((flag&1) && next_fsto!=NULL) {
   *fsto= o= next_fsto;
   goto once_again;
 }
 (*fsto)= NULL;
 return(1);
}


int Fstoutdate_revert(fsto,flag)
struct FstoutdatE **fsto;
int flag;
{
 struct FstoutdatE *o,*n;

 o= *fsto;
 *fsto= NULL;
 while(o!=NULL) {
   n= o->next;
   o->next= *fsto;
   *fsto= o;
   o= n;
 }
 return(1);
}


int Fstoutdate_clone(fsto,from,flag)
struct FstoutdatE **fsto;
struct FstoutdatE *from;
int flag;
{
 int ret;
 struct FstoutdatE *o;

 *fsto= NULL;
 for(o= from;o!=NULL;o= o->next) {
   ret= Fstoutdate_new(fsto,o->root,o->tip,*fsto,0);
   if(ret<=0)
     return(-1);
 }
 Fstoutdate_revert(fsto,0);
 return(1);
}


/* ------------------------------------------------------------------------ */


int Fstraversal_new(trav,root_node,flag)
struct FstraversaL **trav;
struct FsnodE *root_node;
int flag;
/*
 bit5= target tree controlled run
 bit6= permission to follow outdate links
*/
{
 struct FstraversaL *o;

 *trav= o= TSOB_FELD(struct FstraversaL,1);
 if(o==NULL)
   return(-1);
 o->cursor_state= 0;
 if(flag&32)
   o->cursor_state|= 33;
 if(flag&64)
   o->cursor_state|= 64;
 o->node= root_node;
 o->dirnode= NULL;
 o->outdate_stack= NULL;
 o->tlink= NULL;
 o->tlink_node= NULL;
 o->tlink_dirnode= NULL;
 return(1);
}


int Fstraversal_destroy(trav,flag)
struct FstraversaL **trav;
int flag;
{
 if(*trav==NULL)
   return(0);
 Fstoutdate_destroy(&((*trav)->outdate_stack),1);
 free((char *) *trav);
 *trav= NULL;
 return(1);
}


int Fstraversal_copy(from,to,flag)
struct FstraversaL *from;
struct FstraversaL *to;
int flag;
{
 int ret;

 to->cursor_state= from->cursor_state;
 to->node= from->node;
 to->dirnode= from->dirnode;
 to->tlink= from->tlink;
 to->tlink_node= from->tlink_node;
 to->tlink_dirnode= from->tlink_dirnode;
 Fstoutdate_destroy(&(to->outdate_stack),1);
 ret= Fstoutdate_clone(&(to->outdate_stack),from->outdate_stack,0);
 if(ret<=0)
   return(ret);
 return(1);
}


int Fstraversal_is_in_target_tree(o,flag)
struct FstraversaL *o;
int flag;
{
 return((o->cursor_state&33)==33);
}


int Fstraversal_is_outdated(o,flag)
struct FstraversaL *o;
int flag;
{
 return(!!(o->cursor_state&2));
}


int Fstraversal_get_pos(o,node,dirnode,cstate,flag)
struct FstraversaL *o;
struct FsnodE **node;
struct FsnodE **dirnode;
int *cstate;
int flag;
{
 *node= o->node;
 *dirnode= o->dirnode;
 *cstate= o->cursor_state;
 return(1);
}


int Fstraversal_next(trav,end_on_node,next_node,next_dirnode,flag)
struct FstraversaL *trav;
struct FsnodE *end_on_node;
struct FsnodE **next_node;
struct FsnodE **next_dirnode;
int flag;
/*
 bit0= do not dive into next potential directory
 bit1= do not follow next potential tlink from target to source
 bit2= return tlink target tree side rather than source side
 bit4= never dive into directories
 bit5= never follow tlinks
*/
{
 int ret,target_in_source= 0,fresh_in_tlink= 0;
 struct FsnodE *node,*old_node,*dirnode,*outdate_tip;
 struct FsdirinfO *d= NULL;
 struct FstlinK *next_tlink;
 struct FstoutdatE *fsto,*next_outdate;

 static int debug_verbous= 0;
 static int debug_source_count= 0;

get_next_node:;
 *next_node= old_node= trav->node;
 *next_dirnode= dirnode= trav->dirnode;
 if(old_node==NULL)
   return(2);

process_next_tlink:;
 if((trav->cursor_state&33)==33) { /* target traversal in target tree */
   if(flag&(2|32)) {
     flag&= ~2;
   } else {;
     /* explore tlinks first */
     if(trav->tlink!=NULL)
       next_tlink= (trav->tlink)->next_in_target;
     else if(old_node->extra!=NULL)
       next_tlink= old_node->extra->tlink;
     else
       next_tlink= NULL;
     /* skip unusable tlinks */
     for(;next_tlink!=NULL;next_tlink= next_tlink->next_in_target) {
       if((next_tlink->status&1)&&!(trav->cursor_state&64))
     continue;
       if(next_tlink->source==NULL)
     continue;
       if(next_tlink->source->volume==-1)
     continue;

#ifdef NOCHNICHT
       if((flag&4)) {
         /* >>> follow tlink only if directory and permission to dive */
         if(next_tlink->source->extra==NULL || (flag&(1|16)))
     continue;
         if(ext_tlink->source->extra->dirinfo==NULL)
     continue;

       }
#endif /* NOCHNICHT */

     break;
     }
     if(next_tlink!=NULL) { /* dive into  tlink */
       trav->tlink= next_tlink;
       trav->tlink_node= old_node;
       trav->tlink_dirnode= dirnode;
       trav->node= next_tlink->source;
       trav->dirnode= next_tlink->source_dir;
       trav->cursor_state&= ~1;
       fresh_in_tlink= 1;
       *next_node= old_node= trav->node;
       *next_dirnode= dirnode= trav->dirnode;
       if(debug_verbous)  {
         char source[4096]; int depth;
         fprintf(stderr,"Following target link : %s\n",
                        trav->tlink->target_path);
         ret= Fsnode_get_adr(*next_node,*next_dirnode,source,NULL,&depth,4);
         fprintf(stderr,"                   to : %d,%d  %s\n",
                        (*next_node)->level,(*next_node)->volume,source);
       }
       debug_source_count= 0;
       if(!(flag&4)) 
         goto return_found;
     }
   }
 }

/* process_dir_list:; */
 if(old_node->extra!=NULL)
   d= old_node->extra->dirinfo;
 if(flag&(1|16)) {
   flag&= ~1;
 } else if(d!=NULL) {
   /* if not empty : dive into directory */

 /* <<< debugging
       if(strcmp(old_node->name,"fd0")==0) {
         printf("");
       }
 */

   node= NULL;
   ret= Fsdirinfo_get_next_node(d,&node,0);
   if(ret==1) {
     trav->dirnode= *next_dirnode= old_node;
     trav->node= *next_node= node;
     if((trav->cursor_state&33)==33) /* target traversal in target tree */
       trav->tlink= NULL;
     goto return_found;
   }
 }

process_outdated:;
 if((trav->cursor_state&64) && old_node->extra!=NULL) {
                                                /* return next outdated node */
   if(old_node->extra->outdated!=NULL) {

     if(!(trav->cursor_state&2)) { /* newly entering an outdate list */
       for(outdate_tip= old_node->extra->outdated;outdate_tip->extra!=NULL;
           outdate_tip= outdate_tip->extra->outdated)
         if(outdate_tip->extra->outdated==NULL)
       break;
       ret= Fstoutdate_new(&fsto,old_node,outdate_tip,trav->outdate_stack,0);
       if(ret<=0)
         return(-1);
       trav->outdate_stack= fsto;
       trav->cursor_state|= 2;
     }

     *next_node= trav->node= old_node= old_node->extra->outdated;
     if(old_node->volume<0)
       goto process_outdated;
     target_in_source= ((trav->cursor_state&33)==32);
     if(target_in_source) 
       if(old_node->extra!=NULL)
         if(old_node->extra->tlink!=NULL) /* tlink will do its own thing */
            goto process_outdated;
     goto return_found;
   }
 }

 if((trav->cursor_state&66)==66 && trav->outdate_stack!=NULL) {
   if(trav->outdate_stack->tip==old_node) {
/* hop_up_outdated:; */
     *next_node= trav->node= old_node= trav->outdate_stack->root;
     trav->cursor_state&= ~2;
     next_outdate= trav->outdate_stack->next;
     Fstoutdate_destroy(&(trav->outdate_stack),0);
     trav->outdate_stack= next_outdate;
   }
 }

/* process_back_tlink:; */
 target_in_source= ((trav->cursor_state&33)==32);
 if(target_in_source) {
   /* follow eventual tlink back into target tree */
   if(old_node->extra!=NULL)
     if(old_node->extra->tlink!=NULL) {

       if(end_on_node!=NULL && dirnode==end_on_node) {

         /* <<< debugging
         printf("");
         */

         goto it_is_done;
       }

       if(debug_verbous)  {
         char source[4096]; int depth;
         fprintf(stderr,"Hopping up to target  : %s\n",
                         trav->tlink->target_path);
         ret= Fsnode_get_adr(old_node,dirnode, source,NULL,&depth,4);
         fprintf(stderr,"                from  : %d,%d  %s\n",
                        old_node->level,old_node->volume,source);
       }

       if(old_node->extra->tlink!=trav->tlink) {
         fprintf(stderr,
                 "Program error: old_node->extra->tlink!=trav->tlink\n");
         return(-1);
       }
       trav->node= old_node= trav->tlink_node;
       trav->dirnode= dirnode= trav->tlink_dirnode;
       trav->cursor_state|= 1;
       goto process_next_tlink;
     }
 }

/* process_node_chain:; */
 /* return eventual next valid sublist item of dirnode */
 if(dirnode==NULL || dirnode==end_on_node) {
it_is_done:;
   trav->node= *next_node= NULL;
   trav->dirnode= *next_dirnode= NULL;
   if((trav->cursor_state&33)==33) /* target traversal in target tree */
     trav->tlink= NULL;
   goto return_found;
 }
 d= dirnode->extra->dirinfo;
 node= old_node;
                 /* if target_in_source:  do not accept nodes with own tlink */
                 /* if show_outdated : accept nodes with volume -1 */
 ret= Fsdirinfo_get_next_node(d,&node,
                              target_in_source|((trav->cursor_state&64)>>5));
 if(ret==1) {
   old_node= trav->node= *next_node= node;
   trav->dirnode= *next_dirnode= dirnode;
   if((trav->cursor_state&33)==33) /* target traversal in target tree */
     trav->tlink= NULL;
   if(node->volume<0)
     goto process_outdated;
   goto return_found;
 }
 /* no more node found in inode chain */

/* hop_up_directory:; */
 /* move one directory level up */
 old_node= dirnode;
 dirnode= dirnode->extra->up;
 *next_dirnode= dirnode;
 if(dirnode==NULL || dirnode==end_on_node)
   goto it_is_done;
 goto process_outdated; /* now handle old_node like an empty directory */

return_found:;
 if(*next_node!=NULL) {
   if((trav->cursor_state&33)==33 && trav->tlink==NULL) { /* new target node */
     if((*next_node)->extra!=NULL)
       if((*next_node)->extra->tlink!=NULL && !(flag&(4|32))) { 
                            /* sourced targets show up only by their sources */
         old_node= *next_node;
         goto get_next_node;
       }
   }
 }
 debug_source_count++;
 return(1);
}


int Fstraversal_get_upnode(trav,upnode,idx,flag)
struct FstraversaL *trav;
struct FsnodE **upnode;
int idx;
int flag;
{
 int i;
 struct FsnodE *dirnode;

 /* crawl up tree by idx steps */; 
 *upnode= trav->node;
 if((*upnode)==NULL)
   return(0);
 dirnode= trav->dirnode;
 for(i=0;i<idx;i++) {
   if(dirnode==NULL)
     return(0);
   *upnode= dirnode;
   dirnode= dirnode->extra->up;
 }
 return(1);
}


/* ----------------------------------------------------------------------- */


int Fsnode_new(node,name,level,volume,extra,next,flag)
struct FsnodE **node;
char *name;
int level;
int volume;
struct FsextrA *extra;
struct FsnodE *next;
int flag;
/*
bit5= this is a node in the target tree
*/
{
 struct FsnodE *o;

 *node= o= TSOB_FELD(struct FsnodE,1);
 if(o==NULL)
   return(-1);
 o->name= NULL;
 o->level= level;
 o->volume= volume;
#ifdef Askme_fsnode_with_timeS
 o->mtime= 0;
 o->ctime= 0;
#endif /* Askme_fsnode_with_timeS */
 o->extra= extra;
 o->next= next;
 if(name!=NULL) {
   o->name= TSOB_FELD(char, strlen(name)+1);
   if(o->name==NULL)
     goto failed;
   strcpy(o->name,name);
 }
 if(o->extra!=NULL)
   if(o->extra->tlink!=NULL) {
     if(flag&32)
       o->extra->tlink->target= o;
     else
       o->extra->tlink->source= o;
   }

 return(1);
failed:;
 Fsnode_destroy(node,0);
 return(-1);
}


int Fsnode_destroy(node,flag)
struct FsnodE **node;
int flag;
/*
 bit0= destroy complete chain of nodes
*/
{
 struct FsnodE *o,*next_node;

 o= *node;
 if(o==NULL)
   return(0);
once_again:;
 next_node= o->next;
 if(o->name!=NULL)
   free(o->name);
 Fsextra_destroy(&(o->extra),0);
 free((char *) o);
 if((flag&1) && next_node!=NULL) {
   *node= o= next_node;
   goto once_again;
 }
 (*node)= NULL;
 return(1);
}


int Fsnode_set_next(o,next,flag)
struct FsnodE *o;
struct FsnodE *next;
int flag;
{
 struct FsnodE *n;

 n= next;
 if(next==NULL)
   while(n->next!=NULL)
     n= n->next;
 if(n!=NULL)
   n->next= o->next;
 o->next= next;
 return(1);
}


int Fsnode_get_sublist(o,name,sublist,flag)
struct FsnodE *o;
char *name;
struct FsnodE **sublist; /* out: a reference, not a managed object */
int flag;
{
 int ret;
 struct FsextrA *x= NULL;

 x= o->extra; 
 if(x==NULL)
   return(0);
 if(x->dirinfo==NULL)
   return(0);
 ret= Fsdirinfo_get_sublist(x->dirinfo,name,sublist,0);
 return(ret);
}


int Fsnode_get_by_path(start_node,path,found_node,found_dir,found_trav,flag)
struct FsnodE *start_node;
char *path;
struct FsnodE **found_node;
struct FsnodE **found_dir;
struct FstraversaL **found_trav;
int flag;
/*
 bit0= silently return 0 on bad start_node
 bit1= *found_trav already contains a prepared traversal object
       start_node and flag bit5|bit6 are taken from there
 bit2= do not hand over trav but destroy it on end (found_trav==NULL is ok)
 bit5= target tree controlled run
 bit6= permission to follow outdate links
*/
{
 int ret,components,i,cstate,nodive,j,p_len;
 struct FstraversaL *trav= NULL;
 struct FsnodE *node,*dirnode;
 char p[4096],*cpt,*ept;

 if(flag&2) {
   trav= *found_trav;
   flag|= trav->cursor_state&(32|64);
 } else {
   ret= Fstraversal_new(&trav,start_node,flag&(32|64));
   if(ret<=0)
     goto ex;
 }
 Fstraversal_get_pos(trav,&node,&dirnode,&cstate,0);
 if(strcmp(path,"/")==0)
   {ret= 1; goto ex;}
 strcpy(p,path);
 components= Sfile_count_components(p,0);
 if(components==0)
   {ret= 1; goto ex;}
 Sfile_component_pointer(p,&cpt,0,0);
 p_len= strlen(p);
 for(i= 0;i<components;i++) {
   if(i+1>=components)
     ept= p+p_len;
   else
     Sfile_component_pointer(cpt,&ept,1,4);
   for(j= 1;ept-j>=cpt;j++)
     if(*(ept-j)=='/')
       *(ept-j)= 0;
   if(*cpt==0) {
     cpt= ept;
 continue;
   }

   for(nodive= 0;;nodive= 1) {
     ret= Fstraversal_next(trav,NULL,&node,&dirnode,nodive|4);
     if(ret<=0)
       {ret= -1; goto ex;}
     if(node==NULL)  /* cannot find a node with the desired name */
       {ret= 0; goto ex;}
     if(strcmp(node->name,cpt)==0)
   break;
   }

   if(i+1<components)
     cpt= ept;
 }

 ret= 1;
ex:;
 if(ret>0 && trav!=NULL) {
   *found_node= trav->node;
   *found_dir= trav->dirnode;
 }
 if(flag&4)
   Fstraversal_destroy(&trav,0);
 else
   *found_trav= trav;
 return(ret);
}


int Fsnode_evaluate_tlink(o,dirnode,reader,targetpt,flag)
struct FsnodE *o;
struct FsnodE *dirnode;
struct LreadeR *reader;
char **targetpt;
int flag;
/*
 bit1= take absolute target address from reader (must be set to get a targetpt)
 bit5= this is for a node in the target tree (must not be set to get one)
*/
{
 int ret,depth,l,i;
 char target[4096],source[4096],*sourcept;

 *targetpt= NULL;
 if(reader==NULL || (flag&32) || !(flag&2))
   return(2);
 Lreader_get_src_target(reader,&sourcept,targetpt,0);
 if((*targetpt)==NULL)
   return(1);

 ret= Fsnode_get_adr(o,dirnode,source,target,&depth,0);
 if(ret<=0)
   return(0);
 l= strlen(target);
 if(strncmp(target,*targetpt,l)==0) {
   for(i= l;(*targetpt)[i]=='/';i++);
   if((*targetpt)[i]==0) {
     /* current target matches intended target up to some trailing slashes */
     *targetpt= NULL;
     return(1);
   }
 }
 return(1);
}


int Fsnode_make_leaf(branch_end,leaf,outdated,name,reader,flag)
struct FsnodE *branch_end;
struct FsnodE **leaf;
struct FsnodE *outdated;
char *name;
struct LreadeR *reader;
int flag;
/*
 bit0= create all nodes with struct FsinfO 
 bit5= this is a node in the target tree
*/
{
 int ret= 0,level,volume,file_type,info_flag= 0;
 unsigned long crc32;
 char *md5pt;
 dev_t device;
 ino_t inode;
 off_t size;
 time_t mtime, ctime;
 struct FsextrA *extra= NULL;
 struct FsdirinfO *dirinfo= NULL;
 struct FsinfO *nodeinfo= NULL;
 char *sourcept,timestamp[81];

 *leaf= NULL;
 if(reader==NULL) {
   sourcept= "";
   level= -1;
   volume= 0;
   file_type= 2; /* some kind of softlink, actually */
 } else
   Lreader_get_nodedata(reader,&sourcept,&level,&volume,&file_type,1);
 if(file_type==2 || (flag&1) || (flag&2) || (flag&32)) {
   if(file_type==2) {
     if(reader==NULL) {
       ret= 0;
       device= 0;
     } else
       ret= Lreader_get_dir_data(reader,&device,0);
     ret= Fsdirinfo_new(&dirinfo,device,(!!ret)<<14);
     if(ret<=0)
        goto ex;
   }
   if(flag&1) {
     if(reader==NULL) {
       size= 0.0;
       crc32= 0;
       md5pt="0123456789abcdef";
       inode= 0;
       mtime= ctime= 0;
       info_flag= 0;
     } else
       Lreader_get_info_data(reader,&size,&crc32,&md5pt,&inode,timestamp,
                             &mtime,&ctime,&info_flag,0);
     ret= Fsinfo_new(&nodeinfo,size,crc32,md5pt,inode,info_flag<<10);
     if(ret<=0)
       goto ex;
   }
   ret= Fsextra_new(&extra,branch_end,NULL,NULL,dirinfo,nodeinfo,outdated,0);
   if(ret<=0)
     goto ex;
 }
 ret= Fsnode_new(leaf,name,level,volume,extra,NULL,flag&32);
 if(ret<=0)
   goto ex;

#ifdef Askme_fsnode_with_timeS
 if(reader!=NULL && nodeinfo==NULL) /* info was not fetched yet */
   Lreader_get_info_data(reader,&size,&crc32,&md5pt,&inode,timestamp,
                         &mtime,&ctime,&info_flag,0);
 if(info_flag&(1<<15))
   (*leaf)->mtime= mtime;
 if(info_flag&(1<<16))
   (*leaf)->ctime= ctime;
#endif /* Askme_fsnode_with_timeS */

 ret= 1;
ex:;
 if(ret<=0) {
   Fsnode_destroy(leaf,0);
   Fsextra_destroy(&extra,0);
   Fsinfo_destroy(&nodeinfo,0);
   Fsdirinfo_destroy(&dirinfo,0);
 }
 return(ret);
}


int Fsnode_make_impl_dir(branch_end,dir,name,flag)
struct FsnodE *branch_end;
struct FsnodE **dir;
char *name;
int flag;
/*
bit5= this is a node in the target tree
*/
{
 int ret;
 struct FsextrA *extra= NULL;
 struct FsdirinfO *dirinfo= NULL;

 ret= Fsdirinfo_new(&dirinfo,(dev_t) 0,0);
 if(ret<=0)
   goto ex;
 ret= Fsextra_new(&extra,branch_end,NULL,NULL,dirinfo,NULL,NULL,0);
 if(ret<=0)
   goto ex;
 ret= Fsnode_new(dir,name,-1,0,extra,NULL,flag&32);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 if(ret<=0) {
   Fsnode_destroy(dir,0);
   Fsextra_destroy(&extra,0);
   Fsdirinfo_destroy(&dirinfo,0);
 }
 return(ret);
}


int Fsnode_graft_branch(start_node,end_node,path,reader,partrec,flag)
struct FsnodE **start_node; /* might get affected by addresses "/" or "" */ 
struct FsnodE **end_node;
char *path;
struct LreadeR *reader;
struct FspartreC *partrec;
int flag;
/*
 bit0= create all nodes with struct FsinfO 
 bit1= take absolute target address from reader
 bit2= do not keep outdated nodes
 bit3= leaf is a generic directory
 bit5= this is a target tree graft
 bit6= write packeted error messages (see Write_to_channel())
 bit7+8= pickiness : see struct AskmE.read_pickiness
 >>> bit12= add partrec to existing node with same level and volume.
        create such a node if not existing yet.
 bit13= do only update permissions of existing directories 
 >>> bit14= do only update (resp. create) FsinfO of existing nodes 
 bit15= delete files from tree rather than inserting them
        (not useable together with bit5 !)
*/
/* return:
 <=0 error
   1= ok
   2= ok (but no change was necessary in tree)
*/
{
 int ret,is_dir,leaf_exists= 0,pickiness,is_deleted,will_become_dir;
 char name[4096],*cpt,*spt,*npt,*targetpt,*sourcept,errmsg[2*4096];
 struct FsnodE *tcr,*found,*dir_node,*source_root,*target_root;
 struct FsnodE *target_node= NULL,*target_dirnode;
 int permissions,info_flag;
 char owner[4096],group[4096],timestamp[81];
 unsigned long crc32; 
 char *md5pt;
 ino_t inode;
 off_t size;
 time_t mtime= 0,ctime= 0;

 if((flag&32)&&(flag&(1<<15))) {
   fprintf(stderr,"Program error: Fsnode_graft_branch( bit5 && bit15 )\n");
   return(-1);
 }
 pickiness= (flag>>7)&3;
 
 /* follow path of directories through tree below start_node */
   /* create directory nodes as needed */
 if((*start_node)==NULL) {
   ret= Fsnode_make_impl_dir(NULL,start_node,"",0);
   if(ret<=0)
     goto ex;
   if(reader!=NULL) {
     Lreader_get_root_nodes(reader,&source_root,&target_root,0);
     if(flag&32)
       target_root= *start_node;
     else
       source_root= *start_node;
     Lreader_set_root_nodes(reader,source_root,target_root,0);
   }
 }
 cpt= path;
 tcr= *start_node;
 if(tcr->extra==NULL) {
start_node_not_a_dir:;
   fprintf(stderr,"Program error: Start node is not a directory: %s\n",
                  tcr->name);
   ret= -1; goto ex;
 }
 if(tcr->extra->dirinfo==NULL) 
   goto start_node_not_a_dir;
 if(*cpt=='/')
   cpt++;
 while(1) {
   spt= strchr(cpt,'/');
   if(spt==NULL)
 break;
   for(npt= spt+1;*npt=='/';npt++); /* skip eventual multiple slashes */
   if(*npt==0) /* trailing slash marks a directory */
 break;
   strncpy(name,cpt,spt-cpt);
   name[spt-cpt]= 0;
   ret= Fsdirinfo_search_sub_name(tcr->extra->dirinfo,&found,name,1);
   if(ret<0)
     goto ex;
   if(ret==0) {
     if(flag&(3<<13))
       goto not_found_on_infoupdate;
     if(flag&(1<<15))
       goto not_found_on_delete;
     ret= Fsnode_make_impl_dir(tcr,&dir_node,name,0);
     if(ret<=0)
       goto ex;
     ret= Fsdirinfo_add_sub(tcr->extra->dirinfo,dir_node,0);
     if(ret<=0)
       goto ex;
     tcr= dir_node;
   } else {
     /* verify that it is a directory */;
     is_dir= 1;
     is_deleted= 0;
     if(found->volume==-1) {
       is_deleted= 1;
       is_dir= 0;
     } else if(found->extra==NULL)
       is_dir= 0;
     else if(found->extra->dirinfo==NULL)
       is_dir= 0;
     if(!is_dir) {
       if(flag&(3<<13))
         goto not_found_on_infoupdate;
       if(flag&(1<<15)) {
         if(is_deleted)
           goto delete_deleted;
         if(pickiness>1) {
           sprintf(errmsg,
                   "\n--- Deletion path component not a directory : %s\n",
                   name);
           Write_to_channel(errmsg,2,!(flag&64));
           sprintf(errmsg,
                "    ( this would be tolerable with  -read_pickiness:low )\n");
           Write_to_channel(errmsg,2,!(flag&64));
           ret= 0;
         } else {
           ret= 2;
         }
         goto ex;
       }
       if(pickiness>=1 && !is_deleted) {
         sprintf(errmsg,"\n--- Path component not a directory : %s\n",name);
         Write_to_channel(errmsg,2,!(flag&64));
         sprintf(errmsg,
              "    ( this would be tolerable with  -read_pickiness:zero )\n");
         Write_to_channel(errmsg,2,!(flag&64));
         {ret= 0; goto ex;}
       }

       /* eventually outdate and replace by new directory node */;
       ret= Fsnode_make_impl_dir(tcr,&dir_node,name,0);
       if(ret<=0)
         goto ex;
       ret= Fsdirinfo_outdate_sub(tcr->extra->dirinfo,tcr,found,dir_node,
                                  flag&(4|32));
       if(ret<=0)
         goto ex;
       found= dir_node;
     }
     tcr= found;
   }
   cpt= npt;
 }

 /* The empty address points to the start node itself */
 /* >>> this is for future relative grafting */
 if(*cpt==0) {
   if(tcr!=*start_node) {
     fprintf(stderr,
           "Program error: Empty leaf address does not point to start_node\n");
     ret= -1; goto ex;
   }

   /* A61231 : disabled for now */
   ret= 1; goto ex;

   /* <<< A61231 : this seems to be wrong, leads to NULL and then SIGSEGVs */
   tcr= tcr->extra->up;

   if(flag&(1<<15)) {

     /* >>> delete start_node */;

   } else {

     /* create new node with same name as start_node */;
     ret= Fsnode_make_leaf(tcr,end_node,NULL,(*start_node)->name,
                         reader,flag&1);
     if(ret<=0)
       goto ex;

   }
   /* replace start_node by new node */;
   ret= Fsdirinfo_outdate_sub(tcr->extra->dirinfo,
                              tcr,*start_node,*end_node,(flag&(4|32)));
   if(ret<=0)
     goto ex;


   if(flag&2) {
     /* >>> set target here */;
   }

   *start_node= *end_node;
   ret= 1; goto ex;
 }

 /* Finally, the leaf part of the address has been reached */ 
 strcpy(name,cpt);
 cpt= strchr(name,'/');
 if(cpt!=NULL)
   *cpt= 0;
 ret= Fsdirinfo_search_sub_name(tcr->extra->dirinfo,&found,name,1);
 if(ret<0)
   goto ex;
 if(ret>0) {
   leaf_exists= 1;
   is_dir= 1;
   is_deleted= 0;
   if(found->extra==NULL)
     is_dir= 0;
   else if(found->extra->dirinfo==NULL)
     is_dir= 0;
   if(found->volume==-1) { /* deleted */
     is_deleted= 1;
     if(flag&(3<<13))
       goto not_found_on_infoupdate;
     if(flag&(1<<15)) {
delete_deleted:;
       if(pickiness>=2) {
         sprintf(errmsg,"\n--- Attempt to delete a deleted node: %s\n",name);
         Write_to_channel(errmsg,2,!(flag&64));
         sprintf(errmsg,
                "    ( this would be tolerable with  -read_pickiness:low )\n");
         Write_to_channel(errmsg,2,!(flag&64));
         {ret= 0; goto ex;}
       } else
         {ret= 2; goto ex;}
     }
     is_dir= 0;
   }
   will_become_dir= !!(flag&8);
   if(reader==NULL)
     will_become_dir= 1;
   else if(reader->file_type==2)
     will_become_dir= 1;
   if(is_dir) {
     if(flag&(1<<15)) {
       if(found->extra->dirinfo->subcount>0) {
         if(pickiness>=1) {
           sprintf(errmsg,"\n--- Directory not empty on delete: %s\n",name);
           Write_to_channel(errmsg,2,!(flag&64));
           sprintf(errmsg,
               "    ( this would be tolerable with  -read_pickiness:zero )\n");
           Write_to_channel(errmsg,2,!(flag&64));
           {ret= 0; goto ex;}
         } else {
           {ret= 2; goto ex;} /* leave directory undeleted */
         }
       }
     } else if(flag&(3<<13)) {
       if(!will_become_dir) {
type_change_on_infoupdate:;
         if(pickiness>=1) {
           sprintf(errmsg,
                  "\n--- Type change detected on attribute update: %s\n",name);
           Write_to_channel(errmsg,2,!(flag&64));
           sprintf(errmsg,
               "    ( this would be tolerable with  -read_pickiness:zero )\n");
           Write_to_channel(errmsg,2,!(flag&64));
           ret= 0;
         } else {
           ret= 2;
         }
         goto ex;
       }
     } else {
       if(!will_become_dir) {
         if(pickiness>=1) {
           sprintf(errmsg,
                 "\n--- Non-directory trying to replace directory: %s\n",name);
           Write_to_channel(errmsg,2,!(flag&64));
           sprintf(errmsg,
               "    ( this would be tolerable with  -read_pickiness:zero )\n");
           Write_to_channel(errmsg,2,!(flag&64));
           {ret= 0; goto ex;}
         }
       } else {
         *end_node= found;
         {ret= 2; goto ex;} /* nothing to do */
       }
     }
   } else if(flag&(3<<13)) {
     if(will_become_dir)
       goto type_change_on_infoupdate;
     if(flag&(1<<13)) {
       if(pickiness>=1) {
         sprintf(errmsg,
            "\n--- Cannot update directory permissions of non-directory: %s\n",
            name);
         Write_to_channel(errmsg,2,!(flag&64));
         sprintf(errmsg,
               "    ( this would be tolerable with  -read_pickiness:zero )\n");
         Write_to_channel(errmsg,2,!(flag&64));
         ret= 0;
       } else {
         ret= 2;
       }
       goto ex;
     }
   } else if((!is_deleted) && !(flag&(1<<15))) {
     if(pickiness>=3) {
       sprintf(errmsg,
            "\n--- Address already exists and is not a directory : %s\n",name);
       Write_to_channel(errmsg,2,!(flag&64));
       sprintf(errmsg,
               "    ( this would be tolerable with  -read_pickiness:high )\n");
       Write_to_channel(errmsg,2,!(flag&64));
       {ret= 0; goto ex;}
     }
   }
 } else if(flag&(1<<15)) {
not_found_on_delete:;
   if(pickiness>=2) {
     sprintf(errmsg,"\n--- Address not found on delete: %s\n",name);
     Write_to_channel(errmsg,2,!(flag&64));
     sprintf(errmsg,
             "    ( this would be tolerable with  -read_pickiness:low )\n");
     Write_to_channel(errmsg,2,!(flag&64));
     ret= 0;
   } else {
     ret= 2;
   }
   goto ex;
 } else if(flag&(3<<13)) {
not_found_on_infoupdate:;
   if(pickiness>=1) {
     sprintf(errmsg,"\n--- Address not found on attribute update: %s\n",name);
     Write_to_channel(errmsg,2,!(flag&64));
     sprintf(errmsg,
             "    ( this would be tolerable with  -read_pickiness:zero )\n");
     Write_to_channel(errmsg,2,!(flag&64));
     ret= 0;
   } else {
     ret= 2;
   }
   goto ex;
 }
 if(flag&(1<<15)) {
   if(flag&4) {
     ret= Fsdirinfo_delete_sub(tcr->extra->dirinfo,tcr,found,0);
     if(ret<=0)
       goto ex;
   } else {
     ret= Fsnode_make_leaf(tcr,end_node,NULL,name,reader,0);
     if(ret<=0)
       goto ex;
     (*end_node)->volume= -1;
     ret= Fsdirinfo_outdate_sub(tcr->extra->dirinfo,tcr,found,*end_node,
                                flag&32);
     if(ret<=0)
       goto ex;
     tcr->extra->dirinfo->subcount--;
   }
 } else if(flag&(3<<13)) {
   Lreader_get_info_data(reader,&size,&crc32,&md5pt,&inode,timestamp,
                         &mtime,&ctime,&info_flag,0);
   Lreader_get_ownership(reader,&permissions,owner,group,&info_flag,0);
   if(flag&(1<<13)) {
     if(!(permissions&(1<<31))) {
       if(pickiness>=2) {
         sprintf(errmsg,"\n--- No permissions read on update request: %s\n",
                        name);
         Write_to_channel(errmsg,2,!(flag&64));
         sprintf(errmsg,
               "    ( this would be tolerable with  -read_pickiness:low )\n");
         Write_to_channel(errmsg,2,!(flag&64));
         ret= 0;
       } else {
         ret= 2;
       }
     }
     found->extra->dirinfo->permissions= permissions;

     /* care for eventual counterpart in target tree */
     Lreader_get_src_target(reader,&sourcept,&targetpt,0);
     if(targetpt!=NULL) {
       Lreader_get_root_nodes(reader,&source_root,&target_root,0);
       ret= Fsnode_get_by_path(target_root,targetpt,
                               &target_node,&target_dirnode,NULL,1|4|32);
       if(ret<0)
         goto ex;
       if(ret>0 && target_node!=NULL && target_node!=found)
         if(target_node->extra!=NULL)
           if(target_node->extra->dirinfo!=NULL)
             target_node->extra->dirinfo->permissions= permissions;
     }

     *end_node= found;
   }
   if(flag&(1<<14)) {

     /* >>> create or update FsinfO */;

   }
 } else { 
   if(leaf_exists && (flag&32)) { /* Do not outdate. Source will be added. */
     *end_node= found;
     ret= 1; goto ex;
   }
   if((flag&8)) {
     ret= Fsnode_make_impl_dir(tcr,end_node,name,flag&32);
   } else {
     ret= Fsnode_make_leaf(tcr,end_node,NULL,name,reader,flag&(1|32));
   }
   if(ret<=0)
     goto ex;
   if(leaf_exists) {
     ret= Fsdirinfo_outdate_sub(tcr->extra->dirinfo,tcr,found,*end_node,
                                flag&4);
     if(ret<=0)
       goto ex;
   } else {
     ret= Fsdirinfo_add_sub(tcr->extra->dirinfo,*end_node,0);
     if(ret<=0)
       goto ex;
   }

   if(flag&2) {
     ret= Fsnode_evaluate_tlink(*end_node,tcr,reader,&targetpt,flag&(2|32));
     if(ret>0 && targetpt!=NULL) {
       if((*end_node)->extra==NULL) {
         ret= Fsextra_new(&((*end_node)->extra),tcr,NULL,NULL,
                          NULL,NULL,NULL,0);
         if(ret<=0)
           goto ex;
       }
       Lreader_get_root_nodes(reader,&source_root,&target_root,0);
       ret= Fsextra_set_target((*end_node)->extra,targetpt,&target_root,
                               (*end_node),tcr,0);
       if(ret<=0)
         goto ex;
       Lreader_set_root_nodes(reader,source_root,target_root,0);
     }
   }

 }
 ret= 1;
ex:;
 if(ret<=0) {
   sprintf(errmsg,"--- affected path : %s\n",path);
   Write_to_channel(errmsg,2,!(flag&64));
 }
 return(ret);
}


/* <<< is on the way to become  Askme_read_tree()  (for backup_id) */
int Fsnode_read_tree(root_node,target_root,adr,level,returned_reader,flag)
struct FsnodE **root_node;
struct FsnodE **target_root;
char *adr;
int level;
struct LreadeR **returned_reader;
int flag;
/*
 bit0= create all inodes with struct FsinfO 
 bit1= only accept content list or complete list 
       resp. with bit15 vanish list, resp. with bit13 directory list
 bit2= do not issue error message about wrong list type
 bit3= do not destroy LreadeR
 bit4= do not destroy old LreadeR before creating new one
 bit5= do only accept list file of given level
 bit6= write paketed error messages (see Write_to_channel())
 bit7+8= pickiness : see struct AskmE.read_pickiness
 bit12= search for youngest file which matches adr_body*.adr_suffix
 bit13= do only update permissions of existing directories 
 >>> bit14= do only update (resp. create) FsinfO of existing nodes 
 bit15= delete files from tree rather than inserting them
*/
/*
 return:
 >0   highest volume number (at least 1)
 <= 0 error, especially:
 -4   wrong list type with bit1
*/
{
 int popened= 0,ret,is_dir= 0,node_level,dummy,report_error= 1,pickiness;
 FILE *fpin= NULL;
 struct LreadeR *reader= NULL;
 struct FsnodE *tcr,*newtn= NULL,*wd_node= NULL;
 char *linept,*sourcept,*targetpt,*typept,*idpt,errmsg[2*4096];
 char *eff_adr,youngest_adr[4096];
 off_t count;

 pickiness= (flag>>7)&3;

 eff_adr= adr;
 if(flag&(1<<12)) {
   ret= A_youngest_listfile(adr,youngest_adr,0);
   if(ret>0)
     eff_adr= youngest_adr;
 }
 popened= A_open_listfile(&fpin,eff_adr,"r",1|(flag&64));
 if(popened!=2)
   popened= 0;
 if(fpin==NULL) 
   {ret= 0; goto ex;}
 if(!(flag&16))
   Lreader_destroy(returned_reader,0);
 ret= Lreader_new(&reader,eff_adr,fpin,0);
 if(ret<=0) {
   sprintf(errmsg,"Failed to create reader object (out of memory ?)\n");
   Write_to_channel(errmsg,2,!(flag&64));
   goto ex;
 }
 Lreader_set_root_nodes(reader,*root_node,*target_root,0);

 tcr= *root_node;
 while(1) {
   ret= Lreader_read_line(reader,0);
   if(ret<=0) {
     if(ret!= -4)
       ret= 0; 
     goto ex;
   }
   if(ret==1 || ret==2) {
     is_dir= (ret==1);
     ret= Fsnode_graft_branch(root_node,&newtn,reader->absolute_source,reader,
                              NULL,(flag&(1|64|384|(7<<13)))|2);
     if(ret<=0)
       goto ex;
     if(is_dir)
       tcr= newtn;

   } else if (ret==3 || ret==4) {
     is_dir= (ret==3);
     ret= Fsnode_graft_branch(&tcr,&newtn,reader->line_source,reader,
                              NULL,flag&(1|64|384|(7<<13)));
     if(ret<=0)
       goto ex;
     if(is_dir)
       tcr= newtn;

   } else if (ret==5) {
     /* move tree one directory up */;
     tcr= tcr->extra->up;

   } else if (ret==6) {
     /* move to list end of initial wd node */;
     tcr= wd_node; 

   } else if (ret==9) { /* list type determined */
     if(flag&2) {
       Lreader_get_list_type(reader,&typept,0);
       if(flag&(1<<15)) {
         if(strcmp(typept,"vanished file objects list")!=0)
           {ret= -4; goto ex;}
       } else if(flag&(1<<13)) {
         if(strcmp(typept,"complete backup area directory list")!=0)
           {ret= -4; goto ex;}
       } else {
         if(strcmp(typept,"complete backup area file list")!=0 &&
            strcmp(typept,"backup content list")!=0)
           {ret= -4; goto ex;}
       }
     }
     sprintf(errmsg,"reading  %s\n",eff_adr);
     Write_to_channel(errmsg,2,!(flag&64));
 
   } else if(ret==11) { /* level */
     if(flag&32) {
       Lreader_get_nodedata(reader,&sourcept,&node_level,&dummy,&dummy,0);
       if(node_level!=level) {
         sprintf(errmsg,
                 "\n--- unexpected level number  %d . Expected is %d\n\n",
                 node_level,level);
         Write_to_channel(errmsg,2,!(flag&64));
         if(pickiness>0) {
           sprintf(errmsg,
                "   ( this would be tolerable with  -read_pickiness:zero )\n");
           Write_to_channel(errmsg,2,!(flag&64));
           report_error= 0;
           ret= 0; goto ex;
         }
       }
     }

   } else if(ret==15) { /* working directory */
     ret= Fsnode_graft_branch(root_node,&newtn,reader->line,reader,
                              NULL,(flag&(1|64|384))|8);
     if(ret<=0)
       goto ex;
     wd_node= newtn;
     tcr= newtn;

   } else if(ret==17) { /* #=# end */
 break;
   } else if(ret==18) { /* id */

     /* >>> record or compare id */;
     Lreader_get_id(reader,&idpt,0);

   } else {
     /* ignored for now */;
   }
 }
 ret= 1;
ex:
 if(ret>0 && reader!=NULL) {
   Lreader_get_line(reader,&linept,&count,0);
   sprintf(errmsg,"    %12.lf lines read\n",(double) count);
   Write_to_channel(errmsg,2,!(flag&64));
 } else if(ret<=0 && reader!=NULL && report_error) {
   if(ret==-4)
     Lreader_set_errcode(reader,4,0);
   if(!((flag&4) && Lreader_get_errcode(reader,0)==4)) {
     sprintf(errmsg,"\n%s\n",Lreader_strerror(reader,0));
     Write_to_channel(errmsg,2,!(flag&64));
     Lreader_get_line(reader,&linept,&count,0);
     sprintf(errmsg,"Last input: %.lf :  %s\n", (double) count,linept);
     Write_to_channel(errmsg,2,!(flag&64));
     Lreader_get_src_target(reader,&sourcept,&targetpt,0);
     sprintf(errmsg,"Last  addr: %s = %s\n",targetpt,sourcept);
     Write_to_channel(errmsg,2,!(flag&64));
   }
 }
 if(fpin!=NULL){
   if(popened)
     pclose(fpin);
   else
     Sfile_fclose(fpin);
 }
 if(ret>0) {
   Lreader_get_highest_volume(reader,&ret,0);
   if(ret<=0)
     ret= 1;
 }
 if(reader!=NULL)
   Lreader_get_root_nodes(reader,root_node,target_root,0);
 if(!(flag&8)) {
   Lreader_destroy(&reader,0);
   *returned_reader= NULL;
 } else
   *returned_reader= reader;
 return(ret);
}


int Fsnode_get_data(o,name,level,volume,flag)
struct FsnodE *o;
char **name;
int *level;
int *volume;
int flag;
{
 *name= o->name;
 *level= o->level;
 *volume= o->volume;
 return(1);
}


int Fsnode_get_local_target(o,targetpt,flag)
struct FsnodE *o;
char **targetpt;
int flag;
/*
 bit0= set targetpt to leaf name
*/
{
 int ret= 1,l;
 char *spt;

 *targetpt= o->name;
 if(o->extra!=NULL)
  if(o->extra->tlink!=NULL) {
    *targetpt= o->extra->tlink->target_path;
    ret= 2;
  }

 if(flag&1) { /* try to find start of leaf name */
   l= strlen(*targetpt);
   spt= *targetpt+l;
   while(spt>*targetpt) {
     spt--;
     if(*spt!='/')
   break;
   }
   while(spt>*targetpt) {
     if(*(spt-1)=='/')
   break;
     spt--;
   }
   *targetpt= spt;
 }
 return(ret);
}


int Fsnode_get_adr(o,dirnode,source,target,depth,flag)
struct FsnodE *o;
struct FsnodE *dirnode;
char source[4096];
char target[4096];
int *depth; /* number of path components in source */
int flag;
/*
 bit0= *depth shall count target components rather than source
 bit1= do not remove eventual trailing slash of target
 bit2= do not compose target
*/
{
 int l,ltgt;
 char *cpt,*targetpt;
 struct FsnodE *node;

 *depth= 0;
 /* compose absolute source adress */
 l= strlen(o->name);
 for(node= dirnode; node!=NULL; node= node->extra->up)
   l+= strlen(node->name)+1;
 if(l>=4096)
   return(-1);
 cpt= source+(l-strlen(o->name));
 strcpy(cpt,o->name);
 for(node= dirnode; node!=NULL; node= node->extra->up) {
   cpt-= strlen(node->name)+1;
   strcpy(cpt,node->name);
   cpt[strlen(node->name)]= '/';
   (*depth)++;
 }
 if(source[0]==0)
   strcpy(source,"/");

 if(flag&4)
   return(1);

 /* target */
 Fsnode_get_local_target(o,&targetpt,0); 
 l= strlen(targetpt);
 if(targetpt[0]!='/') {
   for(node= dirnode; node!=NULL; node= node->extra->up) {
     Fsnode_get_local_target(node,&targetpt,0); 
     ltgt= strlen(targetpt);
     if(ltgt>0)
       if(targetpt[ltgt-1]=='/')
         ltgt--;
     l+= ltgt+1;
     if(targetpt[0]=='/')
   break;
   }
 }
 if(l>=4096)
   return(-1);
 Fsnode_get_local_target(o,&targetpt,0); 
 cpt= target+(l-strlen(targetpt));
 strcpy(cpt,targetpt);
 if(!(flag&2))
   while(l>cpt-target) {
     if(target[l-1]!='/')
   break;
     target[--l]= 0;
   }
 if(targetpt[0]!='/') {
   for(node= dirnode; node!=NULL; node= node->extra->up) {
     Fsnode_get_local_target(node,&targetpt,0); 
     ltgt= strlen(targetpt);
     if(ltgt>0)
       if(targetpt[ltgt-1]=='/')
         ltgt--;
     cpt-= ltgt+1;
     strncpy(cpt,targetpt,ltgt);
     cpt[ltgt]= '/';
     if(targetpt[0]=='/')
   break;
   }
 } 
 if(cpt!=target) {
   fprintf(stderr,"Programm error with absolute target composition\n");
   return(-1);
 }
 if(flag&1) {
   /* count number of valid slashes */
   *depth= 0;
   /* skip trailing slashes */
   for(cpt= target+strlen(target);cpt>target;cpt--)
     if(*cpt!='/' && *cpt!=0)
   break;
   for(;cpt>=target;cpt--)
     if(*cpt=='/') {
       if(cpt>target) {
         if(*(cpt-1)!='/')
           (*depth)++;
       } else
         (*depth)++;
     }
 }
 if(target[0]==0)
   strcpy(target,"/");
 return(1);
}


int Fsnode_has_explicit_target(o,dirnode,flag)
struct FsnodE *o;
struct FsnodE *dirnode;
int flag;
/* return:
 0= no target defined
 1= some boss directory has a target
 2= o itself has an own target
*/
{
 int ret;
 char *targetpt;
 struct FsnodE *node;

 ret= Fsnode_get_local_target(o,&targetpt,0);
 if(ret==2)
   return(2);
 for(node= dirnode; node!=NULL; node= node->extra->up) {
   ret= Fsnode_get_local_target(node,&targetpt,0);
   if(ret==2)
     return(1);
 }
 return(0);
}


int Fsnode_get_type(o,file_type,flag)
struct FsnodE *o;
int *file_type;
int flag;
{
 *file_type= 0;
 if(o->extra!=NULL) {
   if(o->extra->dirinfo!=NULL)
     *file_type= 2;
   else if(o->extra->nodeinfo!=NULL) {
     *file_type= o->extra->nodeinfo->flag&127;
   }
 }
 return(1);
}


int Fsnode_sort_cmp(o1,o2)
struct FsnodE **o1;
struct FsnodE **o2;
{
 int ret;

 ret= strcmp((*o1)->name,(*o2)->name);
 return(ret);
}


int Fsnode_sort_tree(o,flag)
struct FsnodE *o;
int flag;
/*
 bit0= recursion
*/
{
 int subcount= 0,ret,file_type,i,tcount= 0;
 struct FsdirinfO *dirinfo;
 struct FsnodE *subnode,**pt,*chain_start;
 struct FstlinK *tlink,*tl,**tpt;

 if(o->extra==NULL)
   return(0);

 /* Directory recursion (and counting of directory content) */
 dirinfo= o->extra->dirinfo;
 if(dirinfo==NULL)
   goto tlink_sort;
 subnode= NULL;
 while(1) {
   ret= Fsdirinfo_get_next_node(dirinfo,&subnode,2);
   if(ret==0)
 break;
   subcount++;
   Fsnode_get_type(subnode,&file_type,0);
   if(file_type==2) {
     ret= Fsnode_sort_tree(subnode,flag&1);
     if(ret<=0)
       return(ret);
   }
 }

 /* sort directory sub list */
 pt= TSOB_FELD(struct FsnodE *,subcount+1);
 if(pt==NULL)
   return(-1);
 subnode= NULL;
 subcount= 0;
 while(1) {
   ret= Fsdirinfo_get_next_node(dirinfo,&subnode,2);
   if(ret==0)
 break;
   pt[subcount]= subnode;
   subcount++;
 }
 pt[subcount]= NULL;
 if(subcount>0)
   qsort(pt,subcount,sizeof(struct FsnodE *),Fsnode_sort_cmp);
 for(i=0;i<subcount;i++)
   pt[i]->next= pt[i+1];
 chain_start= pt[0];
 free((char *) pt);
 free((char *) dirinfo->sublist);
 dirinfo->sublist= TSOB_FELD(struct FsnodE *,1);
 if(dirinfo->sublist==NULL)
   return(-1);
 dirinfo->sublist[0]= chain_start;
 dirinfo->hashsize= 1;

tlink_sort:;
 /* sort tlink list */
 tlink= o->extra->tlink;
 if(tlink==NULL)
   return(1);
 if(tlink->next_in_target==NULL)
   return(1);
 for(tl= tlink;tl!=NULL;tl= tl->next_in_target)
   tcount++;
 tpt= TSOB_FELD(struct FstlinK *,tcount+1);
 if(tpt==NULL)
   return(-1);
 i= 0;
 for(tl= tlink;tl!=NULL;tl= tl->next_in_target)
   tpt[i++]= tl;
 tpt[tcount]= NULL;
 qsort(tpt,tcount,sizeof(struct FstlinK *),Fstlink_sort_cmp);
 for(i=0;i<tcount;i++)
   tpt[i]->next_in_target= tpt[i+1];
 o->extra->tlink= tpt[0];
 free((char *) tpt);
 return(1);
}
 

int Fsnode_is_empty_dir(o,flag)
struct FsnodE *o;
int flag;
{
 int ret;

 if(o->extra==NULL)
   return(-1);
 if(o->extra->dirinfo==NULL)
   return(-1);
 ret= Fsdirinfo_is_empty(o->extra->dirinfo,0);
 return(ret);
}


int Fsnode_get_permissions(o,permissions,flag)
struct FsnodE *o;
int *permissions;
int flag;
/*
 bit0= do not reset *permissions
*/
{
 int perms_valid= 0,perms= 0,i;

 if(!(flag&1))
   *permissions= 0;
 if(o->extra==NULL)
   return(0);
 if(o->extra->nodeinfo!=NULL) {
   if(o->extra->nodeinfo->permissions&(1<<31)) {
     perms= 0;
     for(i= 0;i<Fsinfo_modebit_counT;i++)
       if(o->extra->nodeinfo->permissions&(1<<i))
         perms|= Fsinfo_modebitS[i];
     perms_valid= 1;
   }
 }
 if(o->extra->dirinfo!=NULL && !perms_valid) {
   if(o->extra->dirinfo->permissions&(1<<31)) {
     perms= 0;
     for(i= 0;i<Fsinfo_modebit_counT;i++)
       if(o->extra->dirinfo->permissions&(1<<i))
         perms|= Fsinfo_modebitS[i];
     perms_valid= 1;
   }
 }
 if(perms_valid)
   *permissions= perms;
 return(perms_valid);
} 


int Fsnode_get_mctimes(o,mtime,ctime,flag)
struct FsnodE *o;
time_t *mtime,*ctime;
int flag;
{
#ifdef Askme_fsnode_with_timeS
 *mtime= o->mtime;
 *ctime= o->ctime;
  return(1);
#else
 *mtime= *ctime= 0;
  return(0);
#endif /* ! Askme_fsnode_with_timeS */
}


/* ------------------------------------------------------------------------ */


int Fprotector_new(pr,adr,redirection,prev,next,flag)
struct FprotectoR **pr;
char *adr;
char *redirection;
struct FprotectoR *prev;
struct FprotectoR *next;
int flag;
/*
 bit0= do not determine inode number
 bit1= this is a shadow protection
*/
{
 int ret;
 struct FprotectoR *m;

 *pr= m= TSOB_FELD(struct FprotectoR,1);
 if(m==NULL)
   return(-1);
 m->valid= (flag&2)<<2;
 m->adr= NULL;
 m->inode= (ino_t) 0;
 m->device= (dev_t) 0;
 m->redirection= NULL;
 m->last_ok= NULL;
 m->prev= NULL;
 m->next= NULL;
 if(adr!=NULL) {
   if(Sregex_string(&(m->adr),adr,0)<=0)
     goto failed;
   m->valid|= 1;
 }
 if(redirection!=NULL) {
   if(Sregex_string(&(m->redirection),redirection,0)<=0)
     goto failed;
   m->valid|= 4;
 }
 if(!(flag&1)) {
   ret= Fprotector_determine_inode(m,0); /* evtl. appends more FprotectoR */
   if(ret<0)
     goto failed;
 }
 for(;m->prev!=NULL;m= m->prev);
 m->prev= prev;
 if(m->prev!=NULL)
   m->prev->next= m;

 for(;m->next!=NULL;m= m->next);
 m->next= next;
 if(m->next!=NULL)
   m->next->prev= m;
 return(1);
failed:;
 Fprotector_destroy(pr,0);
 return(-1);
}


int Fprotector_destroy(pr,adr,next,flag)
struct FprotectoR **pr;
char *adr;
struct FprotectoR *next;
int flag;
{
 struct FprotectoR *m,*next_m,*prev_m;

 m= *pr;
 if(m==NULL)
   return(0);
destroy_it:;
 prev_m= m->prev;
 next_m= m->next;
 if(prev_m!=NULL)
  prev_m->next= next_m;
 if(next_m!=NULL)
  next_m->prev= prev_m;
 Sregex_string(&(m->adr),NULL,0);
 Sregex_string(&(m->redirection),NULL,0);
 Sregex_string(&(m->last_ok),NULL,0);
 free((char *) m);
 if((flag&1) && next_m!=NULL) {
   m= next_m;
   goto destroy_it;
 }
 *pr= NULL;
 return(1);
}
 

int Fprotector_copy_chain(o,target,flag)
struct FprotectoR *o;
struct FprotectoR **target;
int flag;
{
 int ret;
 struct FprotectoR *m,*n;

 if(o==NULL)
   return(2);
 for(m= o;m->next!=NULL;m= m->next);
 n= *target;
 for(;m!=NULL;m= m->prev) {
   ret= Fprotector_new(&n,m->adr,m->redirection,NULL,n,0);
   if(ret<=0)
     return(-1);
   *target= n;
   n->valid= m->valid;
   n->inode= m->inode;
   n->device= m->device;
   if(Sregex_string(&(n->last_ok),m->last_ok,0)<=0)
     return(-1);
 }
 return(1);
}
 
  
int Fprotector_determine_inode(o,flag)
struct FprotectoR *o;
int flag;
{
 int ret,link_count= 0;
 struct stat stbuf;
 struct FprotectoR *ltp,*l_curr,*o_next;
 char adr[4096],link_target[4096];
 static int link_max= 16;

 if(lstat(o->adr,&stbuf)==-1)
   return(0);
 o->inode= stbuf.st_ino;
 o->device= stbuf.st_dev;
 o->valid|= 2;
 if((stbuf.st_mode&S_IFMT)!=S_IFLNK)
   return(1);
 strcpy(adr,o->adr);
 o_next= o->next;
 l_curr= o;
 while((stbuf.st_mode&S_IFMT)==S_IFLNK) {
   link_count++;
   ret= readlink(adr,link_target,sizeof(link_target));
   if(ret<0)
     return(0);
   link_target[ret]= 0;
   ret= Fprotector_new(&ltp,link_target,o->redirection,l_curr,o_next,
                       1|((o->valid&8)>>2));
   if(ret<=0)
     return(-1);
   ltp->valid|= 16;
   l_curr= ltp;
   if(link_count>=link_max) {
     if(stat(link_target,&stbuf)==-1)
 break;
     ltp->valid|= 32;
   } else {
     if(lstat(link_target,&stbuf)==-1)
 break;
   }
   ltp->inode= stbuf.st_ino;
   ltp->device= stbuf.st_dev;
   ltp->valid|= 2;
   strcpy(adr,link_target);
   if(link_count>=link_max) 
     return(0);
 }
 return(1);
}


int Fprotector_find_inode(o,adr,found,flag)
struct FprotectoR *o;
char *adr;
struct FprotectoR **found;
int flag;
/*
 bit0= do also link target examination
*/
{
 int pass;
 struct stat stbuf;
 struct FprotectoR *m;
 ino_t inode;
 dev_t device;
 
 *found= NULL;
 for(pass=0;pass<1+(flag&1);pass++) {
   if(pass==0) {
     if(lstat(adr,&stbuf)==-1)
 continue;
   } else {
     if(stat(adr,&stbuf)==-1)
 continue;
   }
   inode= stbuf.st_ino;
   device= stbuf.st_dev;
   for(m= o;m!=NULL;m= m->next) {
     if((m->valid&2)!=2)
   continue;
     if(inode!=m->inode)
   continue;
     if(device!=m->device)
   continue;
     *found= m;
     return(1);
   }
   if((stbuf.st_mode&S_IFMT)!=S_IFLNK)
 break;
 }
 return(0);
}


int Fprotector_find(o,adr,found,flag)
struct FprotectoR *o;
char *adr;
struct FprotectoR **found;
int flag;
/*
 bit0= do also link target examination
*/
{
 int ret,l,l_ok;
 char path[4096],*spt;
 struct FprotectoR *m;

 *found= NULL;

 for(m= o;m!=NULL;m= m->next) {
   if(m->adr==NULL || !(m->valid&1))
 continue;
   l= strlen(m->adr);
   if(strncmp(m->adr,adr,l)==0) 
     if(adr[l]==0 || adr[l]=='/') {
       *found= m;
       {ret= 1; goto ex;}
     }
 }
 strcpy(path,adr);
 l_ok= 0;
 if(o->last_ok!=NULL)
   l_ok= strlen(o->last_ok);
 l= strlen(path);
 while(path[0]!=0) {
   if(l_ok>=l) { /* shortcut possible as soon as on the same path as last_ok */
     if(o->last_ok[l]=='/' || o->last_ok[l]==0)
       if(strncmp(o->last_ok,path,l)==0)
         {ret= 0; goto ex;}
   }    
   ret= Fprotector_find_inode(o,path,found,flag&1);
   if(ret!=0)
     return(ret);
   /* clip off last component (double slashes slow down but do not spoil) */;
   for(spt= path+l-1;*spt!='/' && spt!=path;spt--);
   *spt= 0;
   l= spt-path;
 }
 ret= 0;
ex:;
 if(ret==0)
   if(Sregex_string(&(o->last_ok),adr,0)<=0)
     return(-1);
 return(ret);
}


int Fprotector_get_par(o,adr,redirection,is_shadow,link,flag)
struct FprotectoR *o;
char **adr;
char **redirection;
int *is_shadow;
struct FprotectoR **link;
int flag;
/*
 bit0= return prev as *link (else next)
*/
/*
 return: <=0 error, 1=ok , 2=ok, is automatically created link protector
*/
{
 *adr= NULL;
 if(o->valid&1)
   *adr= o->adr;
 *redirection= NULL;
 if(o->valid&4)
   *redirection= o->redirection;
 *is_shadow= !!(o->valid&8);
 if(flag&1)
   *link= o->prev;
 else
   *link= o->next;
 return(1+!!(o->valid&16));
}


int Fprotector_get_boss(o,subordinate,boss,flag)
struct FprotectoR *o;
struct FprotectoR *subordinate;
struct FprotectoR **boss;
int flag;
{
 struct FprotectoR *m;

 *boss= subordinate;
 if(!(subordinate->valid&16))
   return(1);
 for(m= subordinate;m!=NULL;m= m->prev)
   if(!(m->valid&16))
 break;
 if(m==NULL)
   return(0);
 *boss= m;
 return(1);
}


int Fprotector_chain_end(o,found,flag)
struct FprotectoR *o;
struct FprotectoR **found;
int flag;
{
 *found= o;
 if(o==NULL)
   return(2);
 for(;(*found)->next!=NULL; (*found)= (*found)->next);
 return(1);
}


int Fprotector_add(o,adr,redirection,new_o,existing,flag)
struct FprotectoR *o;
char *adr;
char *redirection;
struct FprotectoR **new_o;
struct FprotectoR **existing;
int flag;
/*
 bit0= do not determine inode number
 bit1= this is a shadow protection
*/
/* return: <=0 error , 1=ok, added , 2=did already exist */
{
 int ret;
 struct FprotectoR *m,*o_prev;

 if(existing!=NULL)
   *existing= NULL;
 *new_o= o;
 if(o==NULL)
   m= NULL;
 else
   for(m= o;m->prev!=NULL;m= m->prev);
 for(;m!=NULL;m= m->next) {
   if(m->valid&16)
 continue;
   if((m->valid&8)!=((flag&2)<<2))
 continue;
   if(adr==NULL && (m->valid&1))
 continue;
   if(adr!=NULL) {
     if(!(m->valid&1))
 continue;
     if(strcmp(adr,m->adr)!=0)
 continue;
   }
   if(redirection==NULL && (m->valid&4))
 continue;
   if(redirection!=NULL) {
     if(!(m->valid&4))
 continue;
     if(strcmp(redirection,m->redirection)!=0)
 continue;
   }
 break;
 }
 if(m!=NULL) {
   if(existing!=NULL)
     *existing= m;
   return(2); /* was already recorded */
 }
 o_prev= NULL;
 if(o!=NULL)
   o_prev= o->prev;
 ret= Fprotector_new(new_o,adr,redirection,o_prev,o,flag&3);
 if(ret<=0)
   return(-1);
 return(1);
}


/* ------------------------------------------------------------------------ */


int Askme_prestate_new(pst,flag)
struct Askme_prestatE **pst;
int flag;
{
 struct Askme_prestatE *m;
 
 *pst= m= TSOB_FELD(struct Askme_prestatE,1);
 if(m==NULL)
   return(-1);
 m->language_set= 0;
 m->pkt_set= 0;
 m->picky_set= 0;
 m->logfile_set[0]= m->logfile_set[1]= m->logfile_set[2]= m->logfile_set[3]= 0;
 m->logfile_pkt_set= 0;
 m->media_mode_set= 0;
 m->info_source_enabled= 0;
 m->execute_first_argument= 0;
 m->is_resume_from_file= 0;
 return(1);
}


int Askme_prestate_destroy(pst,flag)
struct Askme_prestatE **pst;
int flag;
{
 if(*pst==NULL)
   return(0);
 free((char *) *pst);
 *pst= NULL;
 return(1);
}


/* ------------------------------------------------------------------------ */


int Askme_rescp_new(rescp,boss,askme,trav,level,volume,backup_adr,restore_adr,
                    flag)
struct Askme_rescP **rescp;
struct Askme_rescP *boss;
struct AskmE *askme;
struct FstraversaL *trav;
int level;
int volume;
char *backup_adr;
char *restore_adr;
int flag;
/*
 bit0= do not push on askme->restore_ctl_stack
*/
{
 struct Askme_rescP *m;

 *rescp= m= TSOB_FELD(struct Askme_rescP,1);
 if(m==NULL)
   return(-1);
 m->boss= boss;
 m->askme= askme;
 m->trav= trav;
 m->level= level;
 m->volume= volume;
 m->backup_adr= backup_adr;
 m->restore_adr= restore_adr;
 m->simulate_action= 0;
 m->use_model_tree= 0;
 m->verbous_action= 0;
 m->verbous_file= 0;
 m->overwrite= 0;
 m->move_to_asylum= 0;
 m->is_moved_to_asylum= 0;
 m->is_second_try= 0;
 m->node= NULL;
 m->dirnode= NULL;
 m->cstate= 0;
 m->is_dir= 0;
 m->file_type= 0;
 m->times_valid= 0;
 m->owner_valid= 0;
 m->perms_valid= 0;
 m->perms= 0;
 m->atime= 0;
 m->mtime= 0;
 m->owner_id= 0;
 m->group_id= 0;
 m->source[0]= 0;
 m->target[0]= 0;
 m->old_type= 0;
 m->truncate_old= 0;
 m->is_surely_clean= 0;
 m->original_altered= 0;
 m->file_is_new= 0;
 m->dir_has_valid_file= 0;
 m->error_is_handled= 0;
 m->change_list= NULL;
 m->permission_list= NULL;
 m->plcount= 0;
 m->path_is_prepared= 0;
 m->write_fd= -1;
 m->restored_bytes= 0.0;
 m->prev= NULL;
 m->next= NULL;
 if(askme!=NULL) {
   m->next= askme->restore_ctl_stack;
   if(m->next!=NULL)
     m->next->prev= m;
   askme->restore_ctl_stack= m;
 }
 return(1); 
}
 
 
int Askme_rescp_destroy(rescp,flag)
struct Askme_rescP **rescp;
int flag;
/*
 bit0= destroy whole ->next chain
*/
{
 struct Askme_rescP *m,*next_m;

 m= *rescp;
 if(m==NULL)
   return(0);
destroy_it:;
 next_m= m->next;
 if(m->askme!=NULL)
   if(m->askme->restore_ctl_stack==m)
     m->askme->restore_ctl_stack= m->next;
 if(m->next!=NULL)
   m->next->prev= m->prev;
 if(m->prev!=NULL)
   m->prev->next= m->next;
 if(m->write_fd!=-1)
   close(m->write_fd);
 if(m->permission_list!=NULL)
   free((char *) m->permission_list);
 if(m->change_list!=NULL)
   free((char *) m->change_list);
 free((char *) m);
 if((flag&1) && next_m!=NULL) {
   m= next_m;
   goto destroy_it;
 }
 *rescp= NULL;
 return(1);
}


int Askme_rescp_setup(o,flag)
struct Askme_rescP *o;
int flag;
/*
 bit0-5: as in Askme_restore_copy, in effect here:
  bit0= recursion
  bit2= according to the model, this is a directory
  bit3= silently ignore if backup_adr is missing
  bit4= this is a move-to-asylum
  bit5= this is a second try
*/
/* return: <=0 error , 1=ok, go on , 2=caller shall return 2  */
{
 int ret,node_file_type,dummy_depth;
 struct AskmE *askme;
 struct FstraversaL *trav;
 char *backup_adr;
 char *restore_adr;
 struct stat backup_stbuf;

 askme= o->askme;
 trav= o->trav;
 backup_adr= o->backup_adr;
 restore_adr= o->restore_adr;

 o->is_surely_clean= 1;
 if(askme->request_to_abort || askme->request_to_skip_volume>0)
   {ret= 2; goto ex;}
 o->simulate_action= !!(askme->restore_mode&64);
 if(o->simulate_action)
   o->use_model_tree= 1;
 o->verbous_action= ((askme->restore_mode&384)==384);
 o->verbous_file= ((askme->restore_mode&384)>=128);
 if(flag&16) {
   o->move_to_asylum= 1;
   o->overwrite= 0;
 } else {
   o->overwrite= (askme->restore_mode>>11)&15;
 }
 if(flag&32)
   o->is_second_try= 1;
 if(o->verbous_file && askme->resume_at_source[0]==0 &&
    ((flag&1) || o->move_to_asylum) && !(flag&8)){
   ret= Askme_announce_copy(askme,backup_adr,restore_adr,
          (!!o->is_second_try)|(o->simulate_action<<1)|(o->move_to_asylum<<2));
   if(ret==2)
     goto ex;
 }
 if(askme->restore_mode&32) {
   o->is_dir= !!(flag&4);
   o->file_type= 1+o->is_dir;
   o->perms= 0700;
   o->use_model_tree= 1;
   o->simulate_action= 1;
   o->atime= o->mtime= time(0);
   o->owner_id= 0;
   o->group_id= 0;
 } else {
   if(stat(backup_adr,&backup_stbuf)==-1) {
     if(flag&8)
       {ret= 2; goto ex;}
     sprintf(askme->info_text,
             "--- Cannot stat(\"%s\") : %s\n",backup_adr,
             (errno>0?strerror(errno):"-unknown error-"));
     Askme_info(askme,0);
     {ret= 0; goto ex;}
   }
   o->perms= backup_stbuf.st_mode&Fsinfo_mode_masK;
   o->file_type= Sfile_type(backup_adr,1|8);
   o->is_dir= (o->file_type==2);
   o->atime= backup_stbuf.st_atime;
   o->mtime= backup_stbuf.st_mtime;
   o->owner_id= backup_stbuf.st_uid;
   o->group_id= backup_stbuf.st_gid;
   o->times_valid= 1;
   o->owner_valid= 1;
   o->perms_valid= 1;
 }
 if(o->verbous_file  && askme->resume_at_source[0]==0 && 
    ((flag&1) || o->move_to_asylum) && (flag&8)) {
                                    /* this was delayed until existence test */
   ret= Askme_announce_copy(askme,backup_adr,restore_adr,
          (!!o->is_second_try)|(o->simulate_action<<1)|(o->move_to_asylum<<2));
   if(ret==2)
     goto ex;
 }

 if(trav!=NULL) {
   Fstraversal_get_pos(trav,&o->node,&o->dirnode,&o->cstate,0);
   Fsnode_get_type(o->node,&node_file_type,0);
   if(askme->restore_mode&32) {
     o->file_type= node_file_type;
     o->is_dir= (o->file_type==2);
   } else {
     if((node_file_type==2 && o->file_type!=2) || 
        (node_file_type!=2 && o->file_type==2)) {
       Askme_de_eN(
         sprintf(askme->info_text,
   "--- Widerspruechlicher Filetyp auf Datentraeger und im Modell:\n    %s\n",
               backup_adr);
       ,
         sprintf(askme->info_text,
   "--- Contradicting filetypes on backup volume and in model tree:\n    %s\n",
               backup_adr);
       )
       Askme_info(askme,1);
       {ret= 0; goto ex;}
     }
   }
   if(askme->resume_at_source[0]!=0) { /* for fine granulated -resume */
     ret= Fsnode_get_adr(o->node,o->dirnode,o->source,o->target,
                         &dummy_depth,4);
     if(ret<=0)
       {ret= -1; goto ex;}
   }

   /* >>> try to obtain ownership from model */;
   /* >>> translate into  owner_id , group_id */;

   o->use_model_tree= 1;
 } else if(o->use_model_tree) {
   sprintf(askme->info_text,
   "--- program error in Askme_restore_copy : use_model_tree && trav==NULL\n");
   Askme_info(askme,0);
   {ret= -1; goto ex;}
 } else {
   if(askme->resume_at_source[0]!=0) { /* for fine granulated -resume */
     strcpy(o->source,o->backup_adr);
     strcpy(o->target,o->backup_adr);
   }
 }

 /* try to obtain recorded permissions from tree */
 if(o->node!=NULL) {
   ret= Fsnode_get_permissions(o->node,&(o->perms),1);
   if(ret<0)
     goto ex;
 }

 o->old_type= Sfile_type(restore_adr,1|8);
 o->truncate_old= 0;
 if(o->old_type!=-1) {
   ret= Sfile_identical(backup_adr,restore_adr,1);
   if(ret==1) {
     Askme_de_eN(
       sprintf(askme->info_text,
  "--- Backupadresse und Restoreadresse zeigen auf den selben File:\n    %s\n",
               restore_adr);
     ,
       sprintf(askme->info_text,
  "--- Backup address and restore address point to the same file:\n     %s\n",
               restore_adr);
     )
     Askme_info(askme,1);
     {ret= 0; goto ex;}
   }
   if(o->old_type==2) {
     if(!o->is_dir) {
       Askme_de_eN(
         sprintf(askme->info_text,
                "--- Typaenderung von Directory verweigert: %s\n",restore_adr);
       ,
         sprintf(askme->info_text,
                 "--- Type change of directory refused: %s\n",restore_adr);
       )
       Askme_info(askme,1);
       {ret= 0; goto ex;}
     }
   }
 }
 ret= 1;
ex:;
 return(ret);
}


int Askme_rescp_prepare_path(o,flag)
struct Askme_rescP *o;
int flag;
/*
 bit0= reset altered permissions and free lists
 bit1= with bit0: remove all newly created directories
 bit2= do not use pager with Askme_info
 bit3= do not issue messages about removed directories
*/
{
 int ret,count,relationship,perms,i,do_change,verbous_action,simulate_action;
 int backup_perms,backup_perms_valid= 0,backup_count,min_ret,is_new_dir,idx;
 int length_overhead= -1;
 mode_t st_mode,backup_st_mode;
 char path[4096],*cpt,*spt,*dpt,dpt_recorded;
 char backup_path[4096],*backup_pt,*backup_dpt;

 struct FsnodE *upnode;
 struct AskmE *askme;
 struct FstraversaL *trav;
 char *backup_adr;
 char *restore_adr;

 askme= o->askme;
 trav= o->trav;
 backup_adr= o->backup_adr;
 restore_adr= o->restore_adr;
 simulate_action= !!(askme->restore_mode&64);
 verbous_action= ((askme->restore_mode&384)==384);

 if(flag&1)
   goto revoke_alterations;

 o->plcount= 0;
 if(o->change_list!=NULL)
   free((char *) o->change_list);
 o->change_list= NULL;
 if(o->permission_list!=NULL)
   free((char *) o->permission_list);
 o->permission_list= NULL;
 o->path_is_prepared= 0;
 
 strcpy(path,restore_adr);
 count= Sfile_count_components(path,0);
 if(count<=1)
   return(2);

 o->plcount= count-1;
 o->permission_list= TSOB_FELD(int,o->plcount);
 o->change_list= TSOB_FELD(int,o->plcount);
 if((o->permission_list)==NULL || (o->change_list)==NULL) {
   o->plcount= 0;
   return(-1);
 }
 for(i=0;i<o->plcount;i++)
   (o->permission_list)[i]= (o->change_list)[i]= 0;

 spt= path;
 for(i=0;i<o->plcount;i++) {
   ret= Sfile_component_pointer(spt,&cpt,1,0);
   if(ret<=0 || cpt<=spt)
 break; /* >>> this would be actually quite a program error */
   if(cpt-1==path)
     dpt= cpt;
   else
     dpt= cpt-1;
   dpt_recorded= *dpt;
   (*dpt)= 0;
   ret= Sfile_lookup_permissions(path,&perms,&st_mode,0);
   if(ret<0)
     goto ex;
   relationship= ret; /* 1= owner  2=group member  3=other */

   /* make way for creating a new leaf */;
   is_new_dir= 0;
   if(ret==0) { /* path component does not exist yet */
     is_new_dir= 1;
     backup_perms_valid= 0;
     if(o->move_to_asylum) {
       /* try to obtain permissions from o->boss->permission_list */
       if(o->boss!=NULL) 
         if(o->boss->path_is_prepared>0) {
           idx= o->boss->plcount - (o->plcount-i);
           if(idx>=0) {
             if(o->boss->change_list[idx]&1) {
               backup_perms= o->boss->permission_list[idx];
               backup_perms_valid= 1;
             }
           }
         }
     } else if(trav!=NULL) {
       /* try to obtain directory permissions from tree model */
       ret= Fstraversal_get_upnode(trav,&upnode,o->plcount-i,0);
       if(ret<0)
         goto ex;
       if(ret>0 && upnode!=NULL) {
         ret= Fsnode_get_permissions(upnode,&backup_perms,0);
         if(ret<0)
           goto ex;
         if(ret>0)
           backup_perms_valid= 1;
       }
     }
     if(backup_perms_valid)
       st_mode= backup_perms;
     else {
       /* try to obtain directory permissions from backup media */
       /* determine adr by stepping up backup_adr :
          clip off o->plcount+1-i components */;
       strcpy(backup_path,backup_adr);
       backup_count= Sfile_count_components(backup_path,0);
       length_overhead= 0;
       if(!o->move_to_asylum)
         length_overhead= askme->restore_from_compcount;
       if(backup_count+length_overhead>=(o->plcount+1-i)) {

         ret= Sfile_component_pointer(backup_path,&backup_pt,
                                      backup_count-(o->plcount+1-i)+1,0);
         if(backup_pt-1==backup_path)
           backup_dpt= backup_pt;
         else
           backup_dpt= backup_pt-1;
         (*backup_dpt)= 0;
         if(verbous_action) {
           sprintf(askme->info_text,"+%s look up permissions of dir %s\n",
                            (simulate_action?" simulate:":""),backup_path);
           Askme_info(askme,3*!(flag&4));
         }
         if(!(askme->restore_mode&32)) {
           ret= Sfile_lookup_permissions(backup_path,&backup_perms,
                                         &backup_st_mode,0);
           if(ret<0)
             goto ex;
           if(ret>0) {
             st_mode= backup_st_mode&Fsinfo_mode_masK;
             backup_perms_valid= 2; 
           }
         }
       }
     }
     if(!backup_perms_valid)
       st_mode= askme->default_dir_perms;
     if(askme->restore_mode&(1<<10))
       st_mode|= S_IXUSR|S_IWUSR|S_IRUSR;
     if(verbous_action) {
       sprintf(askme->info_text,"+%s creation of missing directory %s\n",
                            (simulate_action?" simulate:":""),path);
       Askme_info(askme,3*!(flag&4));
     }
     if(! simulate_action) { /* create directory */
       ret= mkdir(path,st_mode&Fsinfo_mode_masK); /* influenced by umask */
       if(ret==-1) {
         Askme_de_eN(
           sprintf(askme->info_text,
                   "--- Kann Directory nicht erzeugen: %s\n   %s\n",
                   path,(errno>0?strerror(errno):"-unbekannter fehler-"));
         ,
           sprintf(askme->info_text,
                   "--- Cannot create directory: %s\n   %s\n",
                   path,(errno>0?strerror(errno):"-unknown error-"));
         )
         Askme_info(askme,!(flag&4));
         {ret= 0; goto ex;}
       }
       (o->change_list)[i]|= 2;
     }

     if(verbous_action) {
       sprintf(askme->info_text,"+%s original  chmod %4.4o\n",
               (simulate_action?" simulate:":""),st_mode&Fsinfo_mode_masK);
       Askme_info(askme,3*!(flag&4));
     }
     if(!simulate_action) { /* set directory permissions */
       ret= chmod(path,st_mode&Fsinfo_mode_masK);
       if(ret==-1) {
cannot_set_perms:;
         Askme_de_eN(
           sprintf(askme->info_text,
                  "--- Kann Directoryrechte nicht setzen: %4.4o  %s\n    %s\n",
                   st_mode&Fsinfo_mode_masK,
                   path,(errno>0?strerror(errno):"-unbekannter fehler-"));
         ,
           sprintf(askme->info_text,
                   "--- Cannot set directory permissions: %4.4o  %s\n    %s\n",
                   st_mode&Fsinfo_mode_masK,
                   path,(errno>0?strerror(errno):"-unknown error-"));
         )
         Askme_info(askme,!(flag&4));
         {ret= 0; goto ex;}
       }
     }

     if(simulate_action) {
       relationship= 1;
       if(!backup_perms_valid) {
                                             /* deliberately not good enough */
         st_mode= S_IWUSR|S_IRUSR|S_IXGRP|S_IRGRP; 
                                                /* unless open_dir_for_owner */
         if(askme->restore_mode&(1<<10))
           st_mode|= S_ISVTX|S_IWUSR|S_IRUSR;
       }
       perms= 0;
       if(st_mode&S_IXUSR)
         perms|= 1;
       if(st_mode&S_IWUSR)
         perms|= 2;
       if(st_mode&S_IRUSR)
         perms|= 4;
     } else {
       ret= Sfile_lookup_permissions(path,&perms,&st_mode,0);
       if(ret<0)
         goto ex;
       relationship= ret; /* 1= owner  2=group member  3=other */
     }
   }
   (o->permission_list)[i]= (st_mode&Fsinfo_mode_masK);
   do_change= 0;

   if(0 && i<o->plcount-1) { /* <<< preliminary defunct */

     /* >>> branch needs rx permissions (is x alone enough ?) 
       if next directory exists. But needs rwx (is wx enough ?) else.  */
     /* >>> needed is a preview wether component i+1 exists */

     if((perms&5)!=5) {
       if(relationship==1 || getuid()==0) {
         st_mode|= S_IXUSR|S_IRUSR;
         do_change= 1;
       }
     }
   } else {           /* leafdir needs rwx (is w alone enough ?) */
     if((perms&7)!=7) {
       if(relationship==1) {
         st_mode|= S_IXUSR|S_IWUSR|S_IRUSR;
         do_change= 1;
       }
     }
   }
   if(st_mode&S_ISVTX) {
     st_mode&= ~S_ISVTX;
     do_change= 1;
   }
   if((askme->restore_mode&(1<<23)) && !is_new_dir)
     do_change= 0;
   if(do_change) {
     if(verbous_action) {
       sprintf(askme->info_text,"+%s temporary  chmod %4.4o %s\n",
               (simulate_action?" simulate:":""),st_mode&Fsinfo_mode_masK,
               path);
       Askme_info(askme,3*!(flag&4));
     }
     if(!simulate_action) { /* try set temporary permissions */
       ret= chmod(path,st_mode&Fsinfo_mode_masK);
     } else
       ret= 0;
     if(ret==-1) {
       if(verbous_action) {
         sprintf(askme->info_text,
                 "+%s temporary chmod failed (hoping for the best)\n    %s\n",
                 (simulate_action?" simulate:":""),
                 (errno>0?strerror(errno):"-unbekannter fehler-"));
         Askme_info(askme,!(flag&4));
       }
     } else {
       (o->change_list)[i]|= 1;
     }
   }
   *dpt= dpt_recorded;
   spt= cpt;
 }

 /* >>> if directory and existing : prepare too */

 o->path_is_prepared= 1;
 ret= 1;
ex:;
 return(ret);
revoke_alterations:;
 o->path_is_prepared= -1;
 min_ret= 1;
 strcpy(path,restore_adr);
 spt= path;
 for(i= o->plcount-1;i>=0;i--) {
   if(!((o->change_list)[i]&(1|2)))
 continue;
   ret= Sfile_component_pointer(spt,&cpt,i+1,0);
   if(ret<=0 || cpt<=spt) {
     sprintf(askme->info_text,
             "--- program error: Askme_rescp_prepare_path: number of path components shrunk\n");
     Askme_info(askme,0);
     {ret= -1; goto ex;}
   }
   if(cpt-1==path)
     dpt= cpt;
   else
     dpt= cpt-1;
   dpt_recorded= *dpt;
   (*dpt)= 0;
   if((flag&2) && ((o->change_list)[i]&2)) {
     ret= Askme_rm_dirty_dir(askme,path,(flag&(4|8)));
     if(ret<0)
       goto ex;
     if(ret<min_ret)
       min_ret= ret;
     *dpt= dpt_recorded;
 continue;
   }
   if(!((o->change_list)[i]&1)) {
     *dpt= dpt_recorded;
 continue;
   }
   st_mode= (o->permission_list)[i];
   if(verbous_action) {
     sprintf(askme->info_text,"+%s restoring  chmod %4.4o %s\n",
             (simulate_action?" simulate:":""),st_mode&Fsinfo_mode_masK,path);
     Askme_info(askme,3*!(flag&4));
   }
   if(!simulate_action) { /* set final permissions */
     ret= chmod(path,st_mode&Fsinfo_mode_masK);
     if(ret==-1) 
       goto cannot_set_perms;
   }
   *dpt= dpt_recorded;
 }
 ret= min_ret;
 if(o->change_list!=NULL)
   free((char *) o->change_list);
 o->change_list= NULL;
 if(o->permission_list!=NULL)
   free((char *) o->permission_list);
 o->permission_list= NULL;
 o->plcount= 0;
 goto ex;
}


int Askme_rescp_protect(o,flag)
struct Askme_rescP *o;
int flag;
/*
 bit0-3: as in Askme_restore_copy, in effect here:
  none
 bit17= do not issue messages
*/
/* return:
 <0 error
 0=item is protected (not shadowed) 
 1=ok, go on
 2=caller shall return 2 
 3=caller shall skip the current item (is shadowed)
*/
{
 int ret,check_link_target= 0,is_shadow;
 struct AskmE *askme;
 struct FprotectoR *pr,*pr_next;
 char *restore_adr,*adr,*redirection,*prefix,*blankfix;
 
 askme= o->askme;
 restore_adr= o->restore_adr;
 
 if(((askme->restore_mode>>11)&15)==2)
   check_link_target= 1;
 ret= Fprotector_find(askme->protections,restore_adr,&pr,check_link_target);
 if(ret<0)
   goto ex;
 if(ret==0)
   {ret= 1; goto ex;}
 Fprotector_get_boss(askme->protections,pr,&pr,0);
 Fprotector_get_par(pr,&adr,&redirection,&is_shadow,&pr_next,0);
 if(redirection!=NULL) {

   /* >>> change o-> restore_adr */;
   /* >>> re-read old_type */

 } else {
   if(!(flag&(1<<17))) {
     prefix= "---";
     blankfix= "   ";
     if(is_shadow) {
       prefix= "+";
       blankfix= " ";
     }
     Askme_de_eN(
       sprintf(askme->info_text,
               "%s Rueckspeicheradresse ist geschuetzt: %s\n%s durch:  ",
               prefix,restore_adr,blankfix);
     ,
       sprintf(askme->info_text,
               "%s Restore address is protected: %s\n%s by:  ",
               prefix,restore_adr,blankfix);
     )
     sprintf(askme->info_text+strlen(askme->info_text),
             "-%s:%s\n",(is_shadow?"shadow":"protect"),adr);
     Askme_info(askme,1);
   }
   ret= 3*!!is_shadow;
   if(ret!=0)
     if(askme->request_to_abort || askme->request_to_skip_volume>0)
       ret= 2;
   goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Askme_rescp_prepare(o,flag)
struct Askme_rescP *o;
int flag;
/*
 bit0-6: as in Askme_restore_copy, in effect here:
  bit0= recursion
  bit5= this is a second try

 bit16= do initially skipped preparation
*/
/* return:
 <=0 error
 1=ok, go on
 2=caller shall return 2 
 3=caller shall handle sub items,
 4=caller shall skip the current item
 5=caller shall handle sub items but not itself
*/
{
 int ret,skip_prepare= 0;
 char *backup_adr,*restore_adr;
 struct AskmE *askme;
 
 askme= o->askme;
 backup_adr= o->backup_adr;
 restore_adr= o->restore_adr;
 if(flag&32)
   o->is_second_try= 1;
 if(flag&(1<<16))
   goto prepare_path_and_try;
   
 skip_prepare= flag&1;
 if(o->move_to_asylum)
   if(o->boss!=NULL)
     if(o->boss->plcount>0)
       skip_prepare= 0;

 if(askme->resume_at_source[0]!=0) { /* for fine granulated -resume */
   if(strcmp(askme->resume_at_source,o->source)!=0) {
     if(!o->is_dir)
       {ret= 2; goto ex;}
     ret= Askme_rescp_protect(o,(flag&15)|(1<<17));
     if(ret<0 || ret==2)
       goto ex;
     if(ret==3 || ret==0)
       {ret= 5; goto ex;}
     {ret= 3; goto ex;}
   }
   askme->resume_at_source[0]= 0;
   if(o->verbous_file) {
     ret= Askme_announce_copy(askme,backup_adr,restore_adr,
          (!!o->is_second_try)|(o->simulate_action<<1)|(o->move_to_asylum<<2));
     if(ret==2)
       goto ex;
   }
 }

 ret= Askme_rescp_protect(o,flag&15);
 if(ret<=0 || ret==2)
   goto ex;
 if(ret==3)
   {ret= 4; goto ex;}

 if(o->old_type!=-1 && o->old_type!=2 && o->overwrite==0) {
   Askme_de_eN(
     sprintf(askme->info_text,"--- Existiert schon: %s\n",restore_adr);
   ,
     sprintf(askme->info_text,"--- Already existing: %s\n",restore_adr);
   )
   if(o->move_to_asylum && o->boss!=NULL)
     if(o->boss->path_is_prepared==0)
       {ret= 0; goto ex;}
   Askme_info(askme,1);
   {ret= 0; goto ex;}
 }

 o->is_surely_clean= 0;
 if(askme->resume_at_source[0]==0 && !o->move_to_asylum)
   Askme_record_resume_dirty(askme,restore_adr,0);

 if(!skip_prepare) {

   /* <<< eventually skip this on first try and rely on after-error preparation
    */;
prepare_path_and_try:;
   o->is_surely_clean= 0;
   if(o->verbous_action) {
     sprintf(askme->info_text,
             "+%s creation of missing directories resp. chmod u+rwx\n",
             (o->simulate_action?" simulate:":""));
     Askme_info(askme,3);
     if(askme->request_to_abort || askme->request_to_skip_volume>0)
       {ret= 2; goto ex;}
   }
   /* prepare path up to restore_adr */
   ret= Askme_rescp_prepare_path(o,0);
   if(ret<=0) {
     Askme_de_eN(
       sprintf(askme->info_text,
               "--- Kann Directorypfad nicht bereiten fuer : %s\n",
               restore_adr);
     ,
       sprintf(askme->info_text,
               "--- Cannot prepare directory path for : %s\n",
               restore_adr);
     )
     Askme_info(askme,1);
     {ret= 0; goto ex;}
   }
 }
 ret= 1;
ex:;
 return(ret);
}


int Askme_rescp_failed_chmod(o,restore_adr,st_mode,flag)
struct Askme_rescP *o;
char *restore_adr;
mode_t st_mode;
int flag;
/*
 bit0= do not set bit0 of Askme_info()
*/
{
 if(o->askme->german_dialog)
   sprintf(o->askme->info_text,
         "--- Kann Zugriffsrechte nicht setzen: %4.4o  %s\n   %s\n",
         st_mode,restore_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
 else
   sprintf(o->askme->info_text,
         "--- Cannot set file permissions: %4.4o  %s\n   %s\n",
         st_mode,restore_adr,(errno>0?strerror(errno):"-unknown error-"));
 Askme_info(o->askme,!(flag&1));
 return(1);
}


int Askme_rescp_unlink(o,adr,flag)
struct Askme_rescP *o;
char *adr;
int flag;
/*
 bit0= adr is a directory
 bit16= assume path unprepared (despite o->path_is_prepared)
 bit17= do not issue error message
*/
/* return: <=0 error , 1=ok, go on , 2=caller shall return 2 
                       3=caller shall prepare path and retry */
{
 int ret;
 struct AskmE *askme;

 askme= o->askme;

 if(o->verbous_action) {
   sprintf(askme->info_text,"+%s %s of  %s\n",
         (o->simulate_action?" simulate:":""),((flag&1)?"rmdir":"unlink"),adr);
   Askme_info(askme,3);
   if(askme->request_to_abort || askme->request_to_skip_volume>0)
     {ret= 2; goto ex;}
 }
 if(o->simulate_action)
   {ret= 1; goto ex;}
 if(flag&1)
   ret= rmdir(adr);
 else
   ret= unlink(adr);
 if(ret==-1) {
   if(((flag&(1<<16)) || !o->path_is_prepared) && errno!=ENOTEMPTY)
     {ret= 3; goto ex;}
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- Kann %s nicht entfernen:  %s\n    %s\n",
             ((flag&1)?"Directory":"File"),adr,
             (errno>0?strerror(errno):"-unbekannter fehler-"));
   ,
     sprintf(askme->info_text,
             "--- Cannot unlink %s:  %s\n    %s\n",
             ((flag&1)?"directory":"file"),adr,
             (errno>0?strerror(errno):"-unknown error-"));
   )
   if(!(flag&(1<<17)))
     Askme_info(askme,1);
   o->is_surely_clean= 1;
   {ret= 0; goto ex;}
 }
 o->original_altered= 1;
 ret= 1;
ex:;
 return(ret);
}


int Askme_rescp_overwrite(o,flag)
struct Askme_rescP *o;
int flag;
/*
 bit0-3: as in Askme_restore_copy, in effect here:
  none
*/
/* return: <=0 error , 1=ok, go on , 2=caller shall return 2 
                       3=caller shall prepare path and retry */
{
 int ret;
 char *backup_adr,*restore_adr;
 char asylum_adr[4096];
 struct AskmE *askme;
 mode_t st_mode;
 
 askme= o->askme;
 backup_adr= o->backup_adr;
 restore_adr= o->restore_adr;

 if(o->old_type==-1 || o->old_type==2)
   {ret= 1; goto ex;}


 /* >>> handle existing non-directory file */;
 if(o->overwrite==1) {
   ret= Askme_rescp_unlink(o,restore_adr,0);
   if(ret!=1)
     goto ex;
 } else if(o->overwrite==2) {
   if(!(o->old_type==1 && o->file_type==1)) {
     Askme_de_eN(
       sprintf(askme->info_text,
"--- Existiert schon mit unpassendem Typ fuer overwrite_truncate:\n    %s\n",
               restore_adr);
     ,
       sprintf(askme->info_text,
"--- Already existing with non-matching type for overwrite_truncate:\n    %s\n",
               restore_adr);
     )
     Askme_info(askme,1);
     o->is_surely_clean= 1;
     {ret= 0; goto ex;}
   }
   if(o->verbous_action) {
     sprintf(askme->info_text,"+%s scheduling truncation of  %s\n",
                          (o->simulate_action?" simulate:":""),restore_adr);
     Askme_info(askme,3);
     if(askme->request_to_abort || askme->request_to_skip_volume>0)
       {ret= 2; goto ex;}
   }
   if(!((askme->restore_mode&(1<<23)) || o->simulate_action)) {
     st_mode= S_IRUSR|S_IWUSR;
     ret= chmod(restore_adr,st_mode&Fsinfo_mode_masK);
     if(ret==-1) {
       if(!o->path_is_prepared)
         {ret= 3; goto ex;}
       Askme_rescp_failed_chmod(o,restore_adr,st_mode&Fsinfo_mode_masK,0);
       o->is_surely_clean= 1;
       {ret= 0; goto ex;}
     }
     o->original_altered= 1;
   }
   o->truncate_old= 1;
 } else if(o->overwrite==3) {
   ret= Askme_asylum_adr(askme,restore_adr,asylum_adr,0);
   if(ret<=0) 
     goto ex;
   if(o->verbous_action) {
     sprintf(askme->info_text,"+%s move of  %s\n       to  %s",
                          (o->simulate_action?" simulate:":""),
                          restore_adr,asylum_adr);
     Askme_info(askme,3);
     if(askme->request_to_abort || askme->request_to_skip_volume>0)
       {ret= 2; goto ex;}
   }
   if(!o->simulate_action) {
     /* move to asylum directory tree */;
     ret= Askme_restore_copy(askme,NULL,-1,-1,restore_adr,asylum_adr,
                             16|(32*!!o->is_second_try));
     if(ret!=1) {
       if(ret>0)
         if(askme->request_to_abort || askme->request_to_skip_volume>0)
           {ret= 2; goto ex;}
       if(!o->path_is_prepared)
         {ret= 3; goto ex;}
       Askme_de_eN(
         sprintf(askme->info_text,
          "--- Kann existierenden File nicht in Sicherheit bringen:\n    %s\n",
                 restore_adr);
       ,
         sprintf(askme->info_text,
          "--- Cannot move existing file to restore asylum:\n    %s\n",
                 restore_adr);
       )
       Askme_info(askme,1);
       /* >>> under what circumstances ? : o->is_surely_clean= 1; */
       {ret= 0; goto ex;}
     }
     o->original_altered= 1;
   }
 }
 ret= 1;
ex:;
 return(ret);
}


int Askme_rescp_rename(o,flag)
struct Askme_rescP *o;
int flag;
/*
 bit16= use rename() in any case
 bit17= do not issue error message on failed rename or link 
        (but do complain about failed unlink)
*/
/* return: <=0 error , 1=ok, go on , 2=caller shall return 2 
                       3=caller shall prepare both paths and retry */
{
 int ret;
 char *backup_adr,*restore_adr;
 struct AskmE *askme;

 askme= o->askme;
 backup_adr= o->backup_adr;
 restore_adr= o->restore_adr;

 if(o->old_type==2||(flag&(1<<16))) {
   ret= rename(backup_adr,restore_adr);
   if(ret==-1) {
     if(!o->path_is_prepared)
       {ret= 3; goto ex;}
     Askme_de_eN(
       sprintf(askme->info_text,
               "--- Kann Fileobjekt nicht umnennen:\n    %s\n -> %s\n    %s\n",
               backup_adr,restore_adr,
               (errno>0?strerror(errno):"-unbekannter fehler-"));
     ,
       sprintf(askme->info_text,
               "--- Cannot rename file object:\n    %s\n -> %s\n    %s\n",
               backup_adr,restore_adr,
               (errno>0?strerror(errno):"-unknown error-"));
     )
     if(!(flag&(1<<17)))
       Askme_info(askme,1);
     {ret= 0; goto ex;}
   }
   {ret= 1; goto ex;}
 }
 ret= link(backup_adr,restore_adr);
 if(ret==-1) {
   if(!o->path_is_prepared)
     {ret= 3; goto ex;}
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- Kann kein Hardlink auf File setzen:  %s\n    %s\n",
             restore_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
   ,
     sprintf(askme->info_text,
             "--- Cannot set hardlink on file:  %s\n    %s\n",
             restore_adr,(errno>0?strerror(errno):"-unknown error-"));
   )
   if(!(flag&(1<<17)))
     Askme_info(askme,1);
   {ret= 0; goto ex;}
 } else {
   ret= Askme_rescp_unlink(o,backup_adr,0);
   if(ret!=1) {
     if(ret==3)
       ret= 0;
     Askme_rescp_unlink(o,restore_adr,0);
     goto ex;
   }
 }
 ret= 1;
ex:;
 return(ret);
}


int Askme_rescp_create(o,flag)
struct Askme_rescP *o;
int flag;
/*
 bit0-4: as in Askme_restore_copy, in effect here:
*/
/* return: <=0 error , 1=ok, go on , 2=caller shall return 2 
                       3=caller shall prepare path and retry */
{
 int ret;
 char *backup_adr,*restore_adr;
 struct AskmE *askme;
 mode_t creation_mode= 0;
 char backup_link_target[4096],source_link_target[4096];
 char restore_link_target[4096],dummy[4096];
 
 askme= o->askme;
 backup_adr= o->backup_adr;
 restore_adr= o->restore_adr;

 if(o->move_to_asylum) {
   /* try to rename rather than to copy and unlink */
   ret= Askme_rescp_rename(o,(1<<17));
   if(ret==1)
     o->is_moved_to_asylum= 1;
   if(ret!=0)
     goto ex;
 }

 if(o->old_type==2) 
   goto all_done;
 
 if(o->verbous_action) {
   sprintf(askme->info_text,"+%s creation of  %s\n",
                            (o->simulate_action?" simulate:":""),restore_adr);
   Askme_info(askme,3);
   if(askme->request_to_abort || askme->request_to_skip_volume>0)
     {ret= 2; goto ex;}
 }
 if(!o->simulate_action) {
   /* >>> truncate or create file object with narrow permissions */;

   creation_mode= S_IRUSR|S_IWUSR;
   if(o->file_type==1) {
     if(o->truncate_old) {
       /* ??? Can a failed open spoil the old file ? 
              May  original_altered=1  be delayed until success of open ? */
       o->original_altered= 1;
       o->write_fd= open(restore_adr,O_WRONLY|O_TRUNC);
       if(o->write_fd==-1) {
         if(!o->path_is_prepared)
           {ret= 3; goto ex;}
         Askme_de_eN(
           sprintf(askme->info_text,
                   "--- Kann File nicht leer machen: %s\n    %s\n",
               restore_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
         ,
           sprintf(askme->info_text,
                   "--- Cannot truncate file object: %s\n    %s\n",
                   restore_adr,(errno>0?strerror(errno):"-unknown error-"));
         )
         Askme_info(askme,1); 
         if(!o->original_altered)
           o->is_surely_clean= 1;
         {ret= 0; goto ex;}
       }
       o->file_is_new= 1;
     } else {
       o->write_fd= creat(restore_adr,creation_mode);
       if(o->write_fd==-1) {
         if(!o->path_is_prepared)
           {ret= 3; goto ex;}
         Askme_de_eN(
           sprintf(askme->info_text,
                   "--- Kann File nicht erzeugen: %s\n    %s\n",
                 restore_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
         ,
           sprintf(askme->info_text,
                   "--- Cannot create file object: %s\n    %s\n",
                   restore_adr,(errno>0?strerror(errno):"-unknown error-"));
         )
         Askme_info(askme,1); 
         if(!o->original_altered)
           o->is_surely_clean= 1;
         {ret= 0; goto ex;}
       }
       o->file_is_new= 1;
       o->path_is_prepared= -1; /*from here on path preparation will not help*/
       ret= chmod(restore_adr,S_IRUSR|S_IWUSR); /* do not rely on umask */
       if(ret==-1) {
         Askme_rescp_failed_chmod(o,restore_adr,S_IRUSR|S_IWUSR,0);
         {ret= 0; goto ex;}
       }
     }
   } else if(o->file_type==2) {
     creation_mode|= S_IXUSR;
     ret= mkdir(restore_adr,creation_mode);
     if(ret==-1) {
       if(!o->path_is_prepared)
         {ret= 3; goto ex;}
       Askme_de_eN(
         sprintf(askme->info_text,
                 "--- Kann Directory nicht erzeugen: %s\n   %s\n",
               restore_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
       ,
         sprintf(askme->info_text,
                 "--- Cannot create directory: %s\n   %s\n",
                 restore_adr,(errno>0?strerror(errno):"-unknown error-"));
       )
       Askme_info(askme,1);
       if(!o->original_altered)
         o->is_surely_clean= 1;
       {ret= 0; goto ex;}
     }
     o->file_is_new= 1;
     o->path_is_prepared= -1; /* from here on path preparation will not help */
     ret= chmod(restore_adr,creation_mode); /* so umask cannot spoil */
     if(ret==-1) {
       Askme_rescp_failed_chmod(o,restore_adr,creation_mode,0);
       {ret= 0; goto ex;}
     }

   } else if(o->file_type==3) {

     ret= readlink(backup_adr,backup_link_target,sizeof(backup_link_target));
     if(ret<0) {
       Askme_de_eN(
         sprintf(askme->info_text,
              "--- Kann Linkadresse nicht von Softlink ablesen: %s\n   %s\n",
                backup_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
       ,
         sprintf(askme->info_text,
            "--- Cannot read linked address from symbolic link: %s\n   %s\n",
                 backup_adr,(errno>0?strerror(errno):"-unknown error-"));
       )
       Askme_info(askme,1);
       if(!o->original_altered)
         o->is_surely_clean= 1;
       {ret= 0; goto ex;}
     }
     backup_link_target[ret]= 0;

     /* eventually translate link target */
     if(backup_link_target[0]=='/') {
       if(o->verbous_action) {
         sprintf(askme->info_text,
                 "+%s translating symbolic link from:  %s\n",
                 (o->simulate_action?" simulate:":""),backup_link_target);
         Askme_info(askme,3);
         if(askme->request_to_abort || askme->request_to_skip_volume>0)
           {ret= 2; goto ex;}
       }
       strcpy(source_link_target,backup_link_target);
       ret= Askme_make_restore_adr(askme,source_link_target,backup_link_target,
                                   dummy,restore_link_target,0);
       if(ret<=0)
         goto ex;
     } else
       strcpy(restore_link_target,backup_link_target);

     if(o->verbous_file) {
       sprintf(askme->info_text,"+%s symbolic link to:  %s\n",
         (o->simulate_action?" simulate:":""),restore_link_target);
       Askme_info(askme,3);
       if(askme->request_to_abort || askme->request_to_skip_volume>0)
         {ret= 2; goto ex;}
     }
     ret= symlink(restore_link_target,restore_adr);
     if(ret==-1) {
       if(!o->path_is_prepared)
         {ret= 3; goto ex;}
       Askme_de_eN(
         sprintf(askme->info_text,
                 "--- Kann Softlink nicht erzeugen: %s\n   %s\n",
               restore_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
       ,
         sprintf(askme->info_text,
                 "--- Cannot create symbolic link: %s\n   %s\n",
                 restore_adr,(errno>0?strerror(errno):"-unknown error-"));
       )
       Askme_info(askme,1);
       if(!o->original_altered)
         o->is_surely_clean= 1;
       {ret= 0; goto ex;}
     }
     o->file_is_new= 1;

   } else if(o->file_type==4) {
     creation_mode|= S_IFIFO;
     ret= mknod(restore_adr,creation_mode,(dev_t) 0);
     if(ret==-1) {
       if(!o->path_is_prepared)
         {ret= 3; goto ex;}
       Askme_de_eN(
         sprintf(askme->info_text,
                 "--- Kann Named Pipe nicht erzeugen: %s\n   %s\n",
               restore_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
       ,
         sprintf(askme->info_text,
                 "--- Cannot create named pipe: %s\n   %s\n",
                 restore_adr,(errno>0?strerror(errno):"-unknown error-"));
       )
       Askme_info(askme,1);
       if(!o->original_altered)
         o->is_surely_clean= 1;
       {ret= 0; goto ex;}
     }
     o->file_is_new= 1;
     o->path_is_prepared= -1; /*from here on path preparation will not help*/

   } else if(o->file_type==5) {
     /* cannot happen as long as bit1 of Sfile_type() stays unset */;
   } else if(o->file_type==6) {

#ifdef Not_yeT

     /* >>> how to read and create a block device ? */;

#else /* Not_yeT */
     sprintf(askme->info_text,
             "--- Not implemented yet : creation of block device %s\n",
             restore_adr);
     Askme_info(askme,1);
     if(!o->original_altered)
       o->is_surely_clean= 1;
     {ret= 0; goto ex;}
#endif /* ! Not_yeT */

   } else if(o->file_type==7) {

#ifdef Not_yeT

/* ??? Does it make sense to restore a socket file ? */

     { struct sockaddr_un addr; int sock;

       if(strlen(restore_adr>=UNIX_PATH_MAX) {

         /* >>> complain about name too long for unix socket */;

       }
       sock= socket(PF_UNIX,SOCK_STREAM,0);
       addr.sun_family= AF_UNIX;
       strcpy(addr.sun_path,restore_adr);
       bind(sock,(struct sockaddr *) &addr,sizeof(addr));
       close(sock)
     }

#else /* Not_yeT */
     sprintf(askme->info_text,
             "--- Not implemented yet : creation of socket file %s\n",
             restore_adr);
     Askme_info(askme,1);
     if(!o->original_altered)
       o->is_surely_clean= 1;
     {ret= 0; goto ex;}
#endif /* ! Not_yeT */

   } else if(o->file_type==8) {

#ifdef Not_yeT

     /* >>> how to read and create a character device ? */;

#else /* Not_yeT */
     sprintf(askme->info_text,
             "--- Not implemented yet : creation of character device %s\n",
             restore_adr);
     Askme_info(askme,1);
     if(!o->original_altered)
       o->is_surely_clean= 1;
     {ret= 0; goto ex;}
#endif /* ! Not_yeT */

   } else {
     sprintf(askme->info_text,
             "--- Cannot create unknown filetype (%d) : %s\n",
             o->file_type,backup_adr);
     Askme_info(askme,1);
     if(!o->original_altered)
       o->is_surely_clean= 1;
     {ret= 0; goto ex;}
   }
 }
 
all_done:;
 o->path_is_prepared= -1; /* from here on path preparation will not help */
 ret= 1;
ex:;
 return(ret);
}


int Askme_rescp_content(o,flag)
struct Askme_rescP *o;
int flag;
/*
 bit0-4: as in Askme_restore_copy, in effect here:
  none
*/ 
/* return: <=0 error , 1=ok, go on , 2=caller shall return 2  */
{
 int ret;
 int read_fd= -1;
 struct AskmE *askme;
 char *backup_adr,*restore_adr;
 char buffer[64*1024];

 askme= o->askme;
 backup_adr= o->backup_adr;
 restore_adr= o->restore_adr;

 if(o->is_moved_to_asylum)
   {ret=1 ; goto ex;}
 if(!(o->file_type==1 && (o->write_fd>=0 || o->simulate_action)))
   {ret=1 ; goto ex;}

 if(o->verbous_action) {
   sprintf(askme->info_text,"+%s content copy from  %s\n",
                            (o->simulate_action?" simulate:":""),backup_adr);
   Askme_info(askme,3);
   if(askme->request_to_abort || askme->request_to_skip_volume>0)
     {ret= 0; goto ex;}
 }
 if(o->simulate_action)
   {ret=1 ; goto ex;}
 
 /* copy file content */;
 read_fd= open(backup_adr,O_RDONLY);
 if(read_fd==-1) {
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- Kann Backupfile nicht lesen: %s\n    %s\n",
             backup_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
   ,
     sprintf(askme->info_text,
             "--- Cannot read backup file: %s\n    %s\n",
             backup_adr,(errno>0?strerror(errno):"-unknown error-"));
   )
   Askme_info(askme,1); 
   {ret= 0; goto ex;}
 }
 while(1) {
   ret= read(read_fd,buffer,sizeof(buffer));
   if(ret==0)
 break;
   if(ret<0) {
     Askme_de_eN(
       sprintf(askme->info_text,
               "--- Fehler bei Lesen des Backupfiles: %s\n    %s\n",
              backup_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
     ,
       sprintf(askme->info_text,
               "--- Error while reading backup file: %s\n    %s\n",
               backup_adr,(errno>0?strerror(errno):"-unknown error-"));
     )
     Askme_info(askme,1); 
     {ret= 0; goto ex;}
   }
   ret= write(o->write_fd,buffer,ret);
   if(ret==-1) {
     Askme_de_eN(
       sprintf(askme->info_text,
               "--- Fehler bei Schreiben von: %s\n    %s\n",
             restore_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
     ,
       sprintf(askme->info_text,
               "--- Error while writing: %s\n    %s\n",
               restore_adr,(errno>0?strerror(errno):"-unknown error-"));
     )
     Askme_info(askme,1); 
     {ret= 0; goto ex;}
   }
   o->restored_bytes+= ret;
 }
 ret= 1;
ex:;
 if(read_fd!=-1)
   close(read_fd);
 if(o->write_fd!=-1)
   {close(o->write_fd); o->write_fd= -1;}
 return(ret);
}


int Askme_rescp_recursion(o,flag)
struct Askme_rescP *o;
int flag;
/*
 bit0-3: as in Askme_restore_copy, in effect here:
  bit1= do not go into recursion
  bit4= this is a move-to-asylum
*/
/* return: <=0 error , 1=ok, go on , 2=caller shall return 2  */
{
 int ret,allow_missing= 0,recursion_resumed= 0;
 int sub_level,sub_volume,sub_file_type,sub_depth;
 struct AskmE *askme;
 struct FstraversaL *trav,*sub_trav= NULL;
 struct DirseQ *dirseq= NULL;
 struct FsnodE *sub_node= NULL,*sub_dirnode= NULL;
 char *backup_adr,*restore_adr,*sub_namept;
 char sub_backup_adr[4096],sub_restore_adr[4096],sub_name[4096];
 char sub_source[4096],sub_target[4096];
 askme= o->askme;
 trav= o->trav;
 backup_adr= o->backup_adr;
 restore_adr= o->restore_adr;

 if(o->is_moved_to_asylum)
   {ret=1 ; goto ex;}
 if((!o->is_dir) || (flag&2))
   {ret= 1; goto ex;}

 if(o->verbous_action) {
   sprintf(askme->info_text,"+%s Directory recursion from %s\n",
                            (o->simulate_action?" simulate:":""),backup_adr);
   Askme_info(askme,3);
   if(askme->request_to_abort || askme->request_to_skip_volume>0)
     {ret= 0; goto ex;}
 }
 if(trav!=NULL) {
   ret= Fstraversal_new(&sub_trav,NULL,0);
   if(ret<=0)
     goto ex;
   Fstraversal_copy(trav,sub_trav,0);
 }
 if(!o->use_model_tree) {
   ret= Dirseq_new(&dirseq,backup_adr,0);
   if(ret<=0)
     goto ex;
 }

 o->is_surely_clean= 0;
 while(1) { /* loop over directory content */
   if(askme->request_to_abort || askme->request_to_skip_volume>0)
     {ret= 2; goto ex;}

   /* get next sub_name */
   if(o->use_model_tree) {
     /* get next level,volume member of directory */;
     allow_missing= 0;
     while(!(askme->request_to_abort || askme->request_to_skip_volume>0)) {
       ret= Fstraversal_next(sub_trav,o->dirnode,&sub_node,&sub_dirnode,
                             recursion_resumed<<4);
       recursion_resumed= 1;
       if(ret<0)
         goto ex;
       if(sub_node==NULL)
         ret= 0;
       if(ret==0)
     break;
       ret= Fsnode_get_data(sub_node,&sub_namept,&sub_level,&sub_volume,0);
       if(ret<=0)
         {ret= -1; goto ex;}
       if(sub_level==o->level && sub_volume==o->volume)
     break;
       Fsnode_get_type(sub_node,&sub_file_type,0);
       if(sub_file_type==2) {


         /* <<< this is a dirty shortcut */
         allow_missing= 1;
         /* >>> more clean would be a tree traversal for subitems on
                the desired level and volume */;

     break;
       }
     }
     if(ret==0)
 break;
     strcpy(sub_name,sub_namept);
     if(askme->resume_at_source[0]==0) {
                            /* record position for fine granulated -resume */ 
       ret= Fsnode_get_adr(sub_node,sub_dirnode,sub_source,sub_target,
                           &sub_depth,4);
       if(ret<=0)
         {ret= -1; goto ex;}
       if(!o->move_to_asylum)
         Askme_record_resume_source(askme,sub_source,2);
     } else if(strcmp(askme->resume_last_source,askme->resume_at_source)!=0){
       if(!o->move_to_asylum)
         Askme_record_resume_source(askme,askme->resume_at_source,2);
     }
   } else {
     ret= Dirseq_next_adr(dirseq,sub_name,0);
     if(ret<0)
       goto ex; 
     if(ret==0)
 break;

     /* >>> how to obtain source for askme->resume_last_source ? */;
       /* ???  source = backup_adr ? */

   }

   /* make sub_backup_adr, sub_restore_adr */
   strcpy(sub_backup_adr,backup_adr);
   ret= Sfile_add_to_path(sub_backup_adr,sub_name,0);
   if(ret<=0)
     goto ex;
   strcpy(sub_restore_adr,restore_adr);
   ret= Sfile_add_to_path(sub_restore_adr,sub_name,0);
   if(ret<=0)
     {ret= -1; goto ex;}

   /* recursion */
retry_recursion:;
   ret= Askme_restore_copy(askme,sub_trav,o->level,o->volume,
                           sub_backup_adr,sub_restore_adr,
                           1|(allow_missing<<3)|(flag&16));
   if(ret<0)
     goto ex;
   if(ret==0) { /* react according to restore error mode */
     ret= Askme_error_on_restore(askme,o,0);
     if(ret<=0) {
       o->error_is_handled= 1;
       goto ex;
     }
     if(ret==2)
       goto retry_recursion;
   }
   if(askme->resume_at_source[0]==0) {
     if(ret==1)
       o->dir_has_valid_file= 1;
     if(!o->move_to_asylum)
       Askme_record_resume_dirty(askme,restore_adr,0);
   }
 }

 ret= 1;
ex:;
 Fstraversal_destroy(&sub_trav,0);
 Dirseq_destroy(&dirseq,0);
 return(ret);
}


int Askme_rescp_properties(o,flag)
struct Askme_rescP *o;
int flag;
/*
 bit0-3: as in Askme_restore_copy(), in effect here:
  none
 bit16= try to perform all operations even after one of them failed
 bit17= do not issue error messages
 bit18= do not return 0 on request_to_abort
*/
/* return: <=0 error , 1=ok, go on , 2=caller shall return 2  */
{
 int ret,min_ret= 1,silent_on_error,hflag;
 char atext[81],mtext[81];
 struct AskmE *askme;
 struct Askme_rescP *boss;
 char *backup_adr,*restore_adr;
 struct utimbuf utime_buffer;
 mode_t st_mode;

 if(o->is_moved_to_asylum) /* properties have been moved with inode */
   return(1);
 if(o->overwrite==0 && o->old_type==2) /* protect directory properties */
   return(1);
 askme= o->askme;
 backup_adr= o->backup_adr;
 restore_adr= o->restore_adr;
 silent_on_error= !!(flag&(1<<17));

 if(o->verbous_action) {
   sprintf(askme->info_text,"+%s setting properties of %s\n",
           (o->simulate_action?" simulate:":""),restore_adr);
   Askme_info(askme,3);
   if(askme->request_to_abort || askme->request_to_skip_volume>0)
     {ret= 0; goto ex;}
 }

 /* if superuser and ownership transfer is not disabled */
 if(o->owner_valid && getuid()==0 && !(askme->restore_mode&(1<<15))) {
   if(o->verbous_action) {
     sprintf(askme->info_text,"+%s  owner= %.lf    group= %.lf\n",
             (o->simulate_action?" simulate:":""),
             (double) o->owner_id,(double) o->group_id);
     Askme_info(askme,3);
     if(askme->request_to_abort || askme->request_to_skip_volume>0) 
       if(!(flag&(1<<16)))
         {ret= 0; goto ex;}
   }
   if(!o->simulate_action) {
     ret= chown(restore_adr,o->owner_id,o->group_id);
     if(ret==-1) {
       if(!silent_on_error) {
         Askme_de_eN(
           sprintf(askme->info_text,
      "--- Kann Besitzer %.lf und Gruppe %.lf nicht setzen fuer:  %s\n   %s\n",
                   (double) o->owner_id,(double) o->group_id,
                 restore_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
         ,
           sprintf(askme->info_text,
                  "--- Cannot set owner %.lf and group %.lf for: %s\n   %s\n",
                    (double) o->owner_id,(double) o->group_id,
                    restore_adr,(errno>0?strerror(errno):"-unknown error-"));
         )
         Askme_info(askme,
                !(askme->request_to_abort || askme->request_to_skip_volume>0));
       }
       if(min_ret>0)
         min_ret= 0;
       if(!(flag&(1<<16)))
         {ret= 0; goto ex;}
     }
   }
 }

 if(o->times_valid && !(askme->restore_mode&(1<<19))) {
   if(o->verbous_action) {
     strcpy(atext,Sfile_datestr(o->atime,1|2|256));
     strcpy(mtext,Sfile_datestr(o->mtime,1|2|256));
     sprintf(askme->info_text,"+%s  atime= %s    mtime= %s\n",
             (o->simulate_action?" simulate:":""),atext,mtext);
     Askme_info(askme,
              3*!(askme->request_to_abort || askme->request_to_skip_volume>0));
     if(askme->request_to_abort || askme->request_to_skip_volume>0)
       if(!(flag&(1<<16)))
         {ret= 0; goto ex;}
   }
   if(!o->simulate_action) {
     utime_buffer.actime= o->atime;
     utime_buffer.modtime= o->mtime;
     ret= utime(restore_adr,&utime_buffer);
     if(ret==-1) {
       if(!silent_on_error) {
         Askme_de_eN(
           sprintf(askme->info_text,
            "--- Kann Zeitmarken (a%.lf,m%.lf) nicht setzen von: %s\n    %s\n",
               (double) o->atime,(double) o->mtime,
               restore_adr,(errno>0?strerror(errno):"-unbekannter fehler-"));
         ,
           sprintf(askme->info_text,
                   "--- Cannot set timestamps (a%.lf,m%.lf) of: %s\n    %s\n",
                   (double) o->atime,(double) o->mtime,
                   restore_adr,(errno>0?strerror(errno):"-unknown error-"));
         )
         Askme_info(askme, 
                !(askme->request_to_abort || askme->request_to_skip_volume>0));
       }
       if(min_ret>0)
         min_ret= 0;
       if(!(flag&(1<<16)))
         {ret= 0; goto ex;}
     }
   }
 }

 if(o->perms_valid) {
   if(o->is_dir && (askme->restore_mode&(1<<10)))
     o->perms|= S_ISVTX|S_IWUSR|S_IRUSR;
   if(o->verbous_action) {
     sprintf(askme->info_text,"+%s  chmod  %4.4o\n",
             (o->simulate_action?" simulate:":""),o->perms);
     Askme_info(askme,
              3*!(askme->request_to_abort || askme->request_to_skip_volume>0));
     if(askme->request_to_abort || askme->request_to_skip_volume>0)
       if(!(flag&(1<<16)))
         {ret= 0; goto ex;}
   }
   if(!o->simulate_action) {
     st_mode= o->perms;
     ret= chmod(restore_adr,st_mode&Fsinfo_mode_masK);
     if(ret==-1) {
       if(!silent_on_error) 
         Askme_rescp_failed_chmod(o,restore_adr,st_mode&Fsinfo_mode_masK,
                !(askme->request_to_abort || askme->request_to_skip_volume>0));
       if(min_ret>0)
         min_ret= 0;
       if(!(flag&(1<<16)))
         {ret= 0; goto ex;}
     }
   }
 }
 if(!(flag&(1<<18)))
   if(askme->request_to_abort || askme->request_to_skip_volume>0)
     {ret= 0; goto ex;}
 ret= 1;
ex:;
 if(min_ret<=0 && min_ret<ret)
   ret= min_ret;
 if(ret==1 && o->move_to_asylum) {
   /* unlink resp. rmdir source */
   hflag= (o->old_type==2);
   boss= o->boss;
   if(boss!=NULL)
     if(boss->path_is_prepared==0)
       hflag|= (1<<16)|(1<<17); /* allow retry, be silent */
   if(o->path_is_prepared==0)
     o->path_is_prepared= -1;
   ret= Askme_rescp_unlink(o,backup_adr,hflag);
   if(ret==3) {
     return(0);
   }
 }
 return(ret);
}


int Askme_rescp_release_dirty(o,flag)
struct Askme_rescP *o;
int flag;
/*
 bit0= allow release even if restore_adr does not match or o->move_to_asylum
 bit1= be verbous
*/
{
 char adr[4096],*cpt;
 struct AskmE *askme;

 askme= o->askme;
 if((strcmp(askme->resume_last_dirty,o->restore_adr)!=0 || o->move_to_asylum)
     && !(flag&1))
   return(0);
 adr[0]= 0;
 if(o->next!=NULL) {/* <<< better have an official is_recursion flag */
   /* set resume_last_dirty to boss directory */;
   if(strcmp(askme->resume_last_dirty,o->restore_adr)==0) {
     strcpy(adr,askme->resume_last_dirty);
     for(cpt= adr+strlen(adr); *cpt=='/' && cpt>adr; cpt--);
     for(;*cpt!='/' && cpt>adr;cpt--);
     *cpt= 0;
   }
 }
 if(flag&2) {
   if(adr[0]!=0)
     sprintf(askme->info_text,"    resume_last_dirty set to: %s\n",adr);
   else
     sprintf(askme->info_text,"    resume_last_dirty cleared\n");
   Askme_info(askme,3*!(flag&4));
 }
 Askme_record_resume_dirty(askme,adr,0);
 return(1);
}
  

/* ------------------------------------------------------------------------ */


int Askme_new(askme,progname,filename,flag)
struct AskmE **askme;
char *progname;
char *filename;
int flag;
{
 int ret;
 struct AskmE *m;
 mode_t st_mode;
 
 *askme= m= TSOB_FELD(struct AskmE,1);
 if(m==NULL)
   return(-1);
 strcpy(m->progname,progname);
 strcpy(m->filename,filename);
 m->initial_wd[0]= 0;
 m->wd[0]= 0;
 m->did_something_useful= 0;
 m->backup_id[0]= 0;
 m->backup_id_level= -1;
 m->recent_backup_date[0]= 0;
 m->num_level_info= 0;
 m->count_level_info= 0;
 m->level_info= NULL;
 m->num_sublevels= 0;
 m->sublevels= NULL;
 m->num_volumes= 0;
 m->num_highest_level_volumes= 0;
 m->conf_dir[0]= 0;
 m->root_node= NULL;
 m->target_root= NULL;
 m->highest_level= -1;
 m->highest_volno= 0;
 m->lev_vol_map= NULL;
 m->files_in_volume= NULL;
 m->count_in_volume= NULL;
 m->volume_info= NULL;
 m->volumes= NULL;
 m->sub= 1;
 m->suball= 0;
 m->levelall= 0;
 m->nolevel= 0;
 m->dialog= 0;
 m->fully_compatible= 1;
 m->search_mode= 0;
 m->structured_search= 0;
 m->search_target= 0;
 m->result_format= 1|4|16|32;  /* "levhdr:volno:source:expr" */
 m->use_stdin= 0;
 m->result_page_length= 0;
 m->result_page_width= 80;
 m->german_dialog= 0;
 m->mark_text[0]= 0;
 m->packet_output= 0;
 m->read_pickiness= 2;
 m->show_outdated= 0;
 m->volumewise= 0;
 m->no_dir_list= 0;
 m->status_history_max= Askme_status_history_maX;
 m->abort_handler= Askme_abort_handler_defaulT;
#ifdef Askme_with_regeX
 m->re= NULL;
 /* >>> ??? how to initialize m->match[0] ? */
#endif /* Askme_with_regeX */
 m->re_constants= NULL;
 m->re_count= 0;
 m->re_fill= 0;
 m->reg_expr[0]= 0;
 m->search_lv_map= NULL;
 m->resume_at_level= -1;
 m->resume_at_volume= -1;
 m->resume_at_source[0]= 0;
 m->resume_last_level= -1;
 m->resume_last_volume= -1;
 m->resume_last_source[0]= 0;
 m->resume_last_dirty[0]= 0;
 m->resume_search_text[0]= 0;
 m->resume_was_restore= 0;
 m->resume_state_file[0]= 0;
 m->media_mode= 0; 
 m->restore_mode= Askme_restore_mode_startuP; 
 m->restore_from[0]= 0;
 m->restore_from_compcount= 0;
 m->restore_to[0]= 0;
 m->restore_skip[0]= 0;
 m->restore_asylum[0]= 0;
 m->restore_pre_prompt[0]= 0;
 m->restore_post_prompt[0]= 0;
 m->protections= NULL;
 m->default_dir_perms= 0700;
 strcpy(m->media_load_cmd,Askme_media_load_cmD);
 strcpy(m->media_mount_cmd,Askme_media_mount_cmD);
 strcpy(m->media_umount_cmd,Askme_media_umount_cmD);
 strcpy(m->media_eject_cmd,Askme_media_eject_cmD);
 m->prescan_state= NULL;
 m->is_dialog= 0;
 m->bar_is_fresh= 0;
 m->pending_option[0]= 0;
 m->request_to_abort= 0;
 m->request_to_skip_volume= 0;
 m->suppress_result= 0;
 m->look_for_dir= 0;
 m->idle_time= 0.0;
 m->re_failed_at= -1;
 m->prepended_wd= 0;
 m->mounted_level= -1;
 m->mounted_volume= -1;
 m->restore_count= 0;
 m->restore_bytes= 0;
 m->error_count= 0;
 m->restore_ctl_stack= NULL;
 m->found= 0;
 m->search_failed= 0;
 m->result_line[0]= 0;
 m->result_line_counter= 0;
 m->result_page_counter= 0;
 m->result_open_line_len= 0;
 m->found_trav= NULL;
 m->info_text[0]= 0;

 st_mode= umask((mode_t) 0022); /* read */
 umask(st_mode);                /* write back */
 m->default_dir_perms= (~st_mode)&0777;

 ret= Askme_prestate_new(&(m->prescan_state),0);
 if(ret<=0)
   goto failed;

 return(1);
failed:
 Askme_destroy(askme,0);
 return(-1);
}


int Askme_destroy_re(m,flag)
struct AskmE *m;
int flag;
{
 int i;

#ifdef Askme_with_regeX
 if(m->re!=NULL) {
   for(i=0;i<m->re_fill;i++) {
     if(m->re_constants!=NULL)
       if(m->re_constants[i]!=NULL)
   continue; /* ,->re[i] was never subject to regcomp() */
     regfree(&(m->re[i]));
   }
   free((char *) m->re);
   m->re= NULL;
 }
#endif /* Askme_with_regeX */

 if(m->re_constants!=NULL) {
   for(i=0;i<m->re_fill;i++)
     if(m->re_constants[i]!=NULL)
       free(m->re_constants[i]);
   free((char *) m->re_constants);
   m->re_constants= NULL;
 }
 m->re_count= 0;
 m->re_fill= 0;
 return(1);
}


int Askme_destroy(askme,flag)
struct AskmE **askme;
int flag;
{
 int i,j;
 struct AskmE *m;

 m= *askme;
 if(m==NULL)
   return(0);
 if(m->level_info!=NULL) {
   for(i=0;i<m->count_level_info;i++) {
     if(m->level_info[i]!=NULL)
       free(m->level_info[i]);
     m->level_info[i]= NULL;
   }
   free((char *) m->level_info);
 }
 if(m->sublevels!=NULL)
   for(i=0;i<m->num_sublevels;i++)
     Askme_destroy(&(m->sublevels[i]),0);
 if(m->volumes!=NULL) {
   for(i=0;i<m->num_volumes;i++) {
     if(m->volumes[i]==NULL)
   continue;
     for(j=0;j<m->files_in_volume[i];j++) 
       if(m->volumes[i][j]!=NULL) {
         free(m->volumes[i][j]);
         m->volumes[i][j]= NULL;
       }
     free((char *) m->volumes[i]);
     m->volumes[i]= NULL;
   }
   free((char *) m->volumes);
   m->volumes= NULL;
 }
 if(m->files_in_volume!=NULL)
   free((char *) m->files_in_volume);
 if(m->count_in_volume!=NULL)
   free((char *) m->count_in_volume);
 if(m->volume_info!=NULL) {
   for(i=0;i<m->num_volumes;i++) {
     if(m->volume_info[i]!=NULL)
       free(m->volume_info[i]);
     m->volume_info[i]= NULL;
   }
   free((char *) m->volume_info);
 }
 m->files_in_volume= NULL;
 Bitmap2d_destroy(&(m->search_lv_map),0);
 Fsnode_destroy(&(m->root_node),1);
 Bitmap2d_destroy(&(m->lev_vol_map),0);
 Askme_destroy_re(m,0);
 Fstraversal_destroy(&(m->found_trav),0);
 Askme_prestate_destroy(&(m->prescan_state),0);
 Askme_rescp_destroy(&(m->restore_ctl_stack),1);

 free((char *) m);
 *askme= NULL;
 return(1);
}


int Askme_clone_options(target,source,flag)
struct AskmE *target;
struct AskmE *source;
int flag;
{
 int ret;

 target->sub= source->sub;
 target->suball= source->suball;
 target->levelall= source->levelall;
 target->nolevel= source->nolevel;
 target->dialog= source->dialog;
 target->fully_compatible= source->fully_compatible;
 target->search_mode= source->search_mode;
 target->structured_search= source->structured_search;
 target->search_target= source->search_target;
 target->result_format= source->result_format;
 target->use_stdin= source->use_stdin;
 target->result_page_length= source->result_page_length;
 target->result_page_width= source->result_page_width;
 target->german_dialog= source->german_dialog;
 strcpy(target->mark_text,source->mark_text);
 target->packet_output= source->packet_output;
 target->read_pickiness= source->read_pickiness;
 target->show_outdated= source->show_outdated;
 target->volumewise= source->volumewise;
 target->no_dir_list= source->no_dir_list;
 target->status_history_max= source->status_history_max;
 target->abort_handler= source->abort_handler;
 ret= Askme_prepare_regex(target,source->reg_expr,0);
 if(ret<=0)
   return(ret);
 target->resume_at_level= source->resume_at_level;
 target->resume_at_volume= source->resume_at_volume;
 strcpy(target->resume_at_source,source->resume_at_source);
 target->resume_last_level= source->resume_last_level;
 target->resume_last_volume= source->resume_last_volume;
 strcpy(target->resume_last_source,source->resume_last_source);
 strcpy(target->resume_last_dirty,source->resume_last_dirty);
 strcpy(target->resume_search_text,source->resume_search_text);
 target->resume_was_restore= source->resume_was_restore;
 strcpy(target->resume_state_file,source->resume_state_file);
 target->media_mode= source->media_mode;
 target->restore_mode= source->restore_mode;
 strcpy(target->restore_from,source->restore_from);
 target->restore_from_compcount= source->restore_from_compcount;
 strcpy(target->restore_to,source->restore_to);
 strcpy(target->restore_skip,source->restore_skip);
 strcpy(target->restore_asylum,source->restore_asylum);
 strcpy(target->restore_pre_prompt,source->restore_pre_prompt);
 strcpy(target->restore_post_prompt,source->restore_post_prompt);
 ret= Fprotector_copy_chain(source->protections,&(target->protections),0);
 if(ret<=0)
   return(ret);
 target->default_dir_perms= source->default_dir_perms;
 strcpy(target->media_load_cmd,source->media_load_cmd);
 strcpy(target->media_mount_cmd,source->media_mount_cmd);
 strcpy(target->media_umount_cmd,source->media_umount_cmd);
 strcpy(target->media_eject_cmd,source->media_eject_cmd);
 target->is_dialog= source->is_dialog;
 target->bar_is_fresh= source->bar_is_fresh;
 target->request_to_abort= source->request_to_abort;
 target->request_to_skip_volume= source->request_to_skip_volume;
 target->error_count= source->error_count;
 target->restore_count= source->restore_count;
 target->restore_bytes= source->restore_bytes;
 target->suppress_result= source->suppress_result;
 target->look_for_dir= source->look_for_dir;
 target->prepended_wd= source->prepended_wd;
 return(1);
}


int Askme_dialog_input(askme,line,linesize,flag)
struct AskmE *askme;
char line[];
int linesize;
int flag;
/*
 bit0= do not write to history
 bit1= do not read input (but eventually write to history)
 bit2= do not write to history line which begin with "-history:" or "-history "
*/
{
 char *cpt= NULL;
 int ret;
#ifdef Askme_with_readlinE
 int l;
 static char last_input[4096]= {""};
#endif /* ! Askme_with_readlinE */
 double tdiff;
 struct timeval tv;
 struct timezone tz;

 gettimeofday(&tv,&tz);
 tdiff= tv.tv_sec+(1.e-6*(double) tv.tv_usec);

 fflush(stdout);

#ifdef Askme_with_readlinE

 if(askme->use_stdin) {
   if(flag&2)
     {ret= 1; goto ex;}
   if(Sfile_fgets(line,linesize-1,stdin)==NULL) {
     /* need a very dramatic end */
     kill(getpid(),SIGHUP);
     {ret= -1; goto ex;}
   }
   {ret= 1; goto ex;}
 }
 if(flag&2) {
   cpt= NULL;
 } else {
   cpt= readline("");
   if(cpt==NULL) {
     /* need a very dramatic end */
     kill(getpid(),SIGHUP);
     {ret= -1; goto ex;}
   }
   l= strlen(cpt);
   if(l>=linesize) {
     strncpy(line,cpt,linesize-1);
     line[sizeof(line)-1]= 0;
   } else 
     strcpy(line,cpt);
 }
 if(line[0]!=0 && strcmp(last_input,line)!=0 && !(flag&1))
   if(!((flag&4) && 
      (strncmp(line,"-history:",9)==0 || strncmp(line,"-history ",9)==0))) {
     add_history(line);
     strncpy(last_input,line,sizeof(last_input)-1);
     last_input[sizeof(last_input)-1]= 0;
   }

#else /*  Askme_with_readlinE */

 if(flag&2)
   {ret= 1; goto ex;}
 if(Sfile_fgets(line,linesize-1,stdin)==NULL) {
   /* need a very dramatic end */
   kill(getpid(),SIGHUP);
   {ret= -1; goto ex;}
 }

#endif /* ! Askme_with_readlinE */

 ret= 1;
ex:;
 if(cpt!=NULL)
   free(cpt);
 gettimeofday(&tv,&tz);
 askme->idle_time+= tv.tv_sec+(1.e-6*(double) tv.tv_usec)-tdiff;
 return(ret);
}


int Askme_request_confirmation(askme,flag)
struct AskmE *askme;
int flag;
/*
 bit0= important operation going on: 
       demand confirmation of abort, only abort on @@@
 bit1= mark '@' and '@@' by return 4
 bit2= accept: i=ignore , r=retry , s=skip up to next volume , x=abort
 bit3= @@@ = 'done reading' rather than 'abort'
*/
/* return: <=0 error , 1=go on , 2=abort , 3=redo request, 4=see flag bit1
                       5=skip volume , 6=retry failed operation
 */
{
 int ret;
 char line[4096],*cpt,previous_line[4096],*abort_req_text,*abort_really_text;

 if(flag&8) {
   Askme_de_eN(
     abort_req_text= "Ende-Eingabe";
     abort_really_text= "ausgelesen";
   ,
     abort_req_text= "request to end";
     abort_really_text= "done reading";
   )
 } else {
   Askme_de_eN(
     abort_req_text= "Abbruchaufforderung";
     abort_really_text= "abbrechen";
   ,
     abort_req_text= "request to abort";
     abort_really_text= "abort";
   )
 }
 if(askme->restore_mode&(1<<20)) /* update_state_sparse: update _src, _dirty */
   Askme_record_resume_state(askme,-1,-1,"","",-1,2|8|16);
 ret= Askme_dialog_input(askme,line,sizeof(line),1);
 askme->result_line_counter= 0;
 askme->result_page_counter++;
 if(ret<=0)
   if(askme->result_page_length>0)
     askme->result_page_length= -askme->result_page_length;

 cpt= line;
 if(*cpt=='@') {
   if(strcmp(cpt,"@@@")==0) {
request_to_abort:;
     if(flag&1) {
       strcpy(previous_line,cpt);
       Askme_de_eN(
         sprintf(askme->info_text,
                 ".... [%s = %s registriert. Wirklich %s ? (j/n) ] ....\n",
                 cpt,abort_req_text,abort_really_text);
       ,
         sprintf(askme->info_text,
                 "....... [%s = %s registered. Really %s ? (y/n) ] .......\n",
                 cpt,abort_req_text,abort_really_text);
       )
       Askme_info(askme,0);
       ret= Askme_dialog_input(askme,line,sizeof(line),1);
       if(ret<=0)
         return(ret);
       cpt= line;
       if(strcmp(cpt,previous_line)==0 || 
          ((*cpt=='Y' || *cpt=='y' || *cpt=='j' || *cpt=='J' || *cpt=='1') &&
             *(cpt+1)==0)) {
         askme->request_to_abort= 1;
         Askme_de_eN(
           sprintf(askme->info_text,
                   "--- ( %s bestaetigt )\n",abort_req_text);
         ,
           sprintf(askme->info_text,
                   "------- ( %s confirmed )\n",abort_req_text);
         )
         Askme_info(askme,0);
         return(2);
       }
       Askme_de_eN(
         sprintf(askme->info_text,
                 "... ( %s zurueckgenommen )\n",abort_req_text);
       ,
         sprintf(askme->info_text,
                 "....... ( %s revoked )\n",abort_req_text);
       )
       Askme_info(askme,0);
       return(3);
     }
     askme->request_to_abort= 1;
     Askme_de_eN(
       sprintf(askme->info_text,
"------ [%s = Abbruchaufforderung registriert. Vorgang wird beendet ] -------\n",
               cpt);
     ,
       sprintf(askme->info_text,
"----------- [%s = request to abort registered. Operation ends ] ------------\n",
               cpt);
     )
     Askme_info(askme,0);
     return(2);
   } else if(strcmp(cpt,"@@")==0) {
     askme->suppress_result= 1;
     Askme_de_eN(
       sprintf(askme->info_text,
".... [@@ = Schweigeaufforderung registriert. Ausgabe wird unterdrueckt ] ....\n");
     ,
       sprintf(askme->info_text,
"..... [@@ = request for silence registered. Output will be suppressed ] .....\n");
     )
     Askme_info(askme,0);
   } else if(strcmp(cpt,"@")==0) {
     if(askme->result_page_length>0)
       askme->result_page_length= -askme->result_page_length;
     if(flag&1) {
       Askme_de_eN(
         sprintf(askme->info_text,
"... [@ = Frageunterdrueckung registriert. -page zeitweise abgeschaltet ] ....\n");
       ,
         sprintf(askme->info_text,
"... [@ = page prompt suppression registered. -page disabled temporarily ] ...\n");
       )
       Askme_info(askme,0);
     }
   } else if(strcmp(cpt,"@skip")==0 && (askme->volumewise&1)) {
skip_volume:;
     Askme_de_eN(
       sprintf(askme->info_text,
"[%s = Aufforderung zum Ueberspringen registriert. Wirklich machen ? (j/n) ]\n",
              cpt);
     ,
       sprintf(askme->info_text,
".. [%s = request to skip volume registered. Really leave out ? (y/n) ] ..\n",
               cpt);
     )
     Askme_info(askme,0);
     strcpy(previous_line,cpt);
     ret= Askme_dialog_input(askme,line,sizeof(line),1);
     if(ret<=0)
       return(ret);
     cpt= line;
     if(strcmp(cpt,previous_line)==0 || 
        ((*cpt=='Y' || *cpt=='y' || *cpt=='j' || *cpt=='J' || *cpt=='1') &&
           *(cpt+1)==0)) {
       askme->request_to_skip_volume= 1;
                                                         /* write to history */
       Askme_dialog_input(askme,"@skip",6,2);
       Askme_de_eN(
         sprintf(askme->info_text,
                 "--- ( Aufforderung zum Ueberspringen bestaetigt )\n");
       ,
         sprintf(askme->info_text,
                 "------- ( request to skip volume confirmed )\n");
       )
       Askme_info(askme,0);
       return(5);
     }
     Askme_de_eN(
       sprintf(askme->info_text,
               "... ( Aufforderung zum Ueberspringen zurueckgenommen )\n");
     ,
       sprintf(askme->info_text,
               "....... ( request to skip volume revoked )\n");
     )
     Askme_info(askme,0);
     return(3);
   } else {
     Askme_dialog_input(askme,cpt,strlen(line)+1,2); /* write to history */
     Askme_de_eN(
       sprintf(askme->info_text,
 "--- Unbekannte Eingabe, die mit @ beginnt. Bitte etwas anderes eingeben.\n");
     ,
       sprintf(askme->info_text,
 "--- Unrecognized input beginning with @. Please enter someting else.\n");
    )
     Askme_info(askme,0);
     return(3);
   }
   if(flag&2)
     return(4);
   if(flag&1)
     return(3);
   return(1);
 } else if(flag&4) {

   if(strcmp(cpt,"i")==0 || strcmp(cpt,"I")==0 || *cpt==0) { 
     return(1);
   } else if(strcmp(cpt,"r")==0 || strcmp(cpt,"R")==0) {
     return(6);
   } else if(strcmp(cpt,"s")==0 || strcmp(cpt,"S")==0) {
     if(flag&1)
       goto skip_volume;
     askme->request_to_skip_volume= 1;
     return(5);
   } else if(strcmp(cpt,"x")==0 || strcmp(cpt,"X")==0) {
     goto request_to_abort;
   } else {
     /* >>> unknown input */
     Askme_de_eN(
       sprintf(askme->info_text,
      "--- Als Eingaben sind erlaubt : Leere Zeile, i, r, s, x, @, @@, @@@\n");
     ,
       sprintf(askme->info_text,
      "--- Please enter one of : empty line, i, r, s, x, @, @@, @@@\n");
     )
     Askme_info(askme,0);
     return(3);
   }

 } else if(*cpt=='-' && !(flag&1)) {
   Askme_dialog_input(askme,cpt,strlen(line)+1,2); /* write to history */
   strcpy(askme->pending_option,cpt);
   askme->request_to_abort= 1;
   Askme_de_eN(
     sprintf(askme->info_text,
"--------- [-* = Optionseingabe registriert. Vorgang wird beendet ] ----------\n");
   ,
     sprintf(askme->info_text,
"------------ [-* = input of option registered. Operation ends ] -------------\n");
   )
   Askme_info(askme,0);
   return(2);

 } else if(*cpt!=0 && !(flag&1)) {
   Askme_dialog_input(askme,cpt,strlen(line)+1,2); /* write to history */
   strcpy(askme->pending_option,cpt);
   askme->request_to_abort= 1;
   Askme_de_eN(
     sprintf(askme->info_text,
"----------- [ Suchtexteingabe registriert. Vorgang wird beendet ] -----------\n");
   ,
     sprintf(askme->info_text,
"------------ [ input of searchtext registered. Operation ends ] -------------\n");
   )
   Askme_info(askme,0);
   return(2);
 } else if(*cpt!=0) {
   Askme_dialog_input(askme,cpt,strlen(line)+1,2); /* write to history */
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- Als Eingaben sind erlaubt : Leere Zeile, @, @@, @@@\n");
   ,
     sprintf(askme->info_text,
             "--- Please enter one of : empty line, @, @@, @@@\n");
   )
   Askme_info(askme,0);
   return(3);
 }
 return(1);
}


int Askme_predict_linecount(askme,line,linecount,flag)
struct AskmE *askme;
char *line;
int *linecount;
int flag;
{
 int width,l;
 char *spt,*ept;

 *linecount= 0;
 spt= line;
 width= askme->result_page_width;
 while(1) {
   ept= strchr(spt,'\n');
   if(ept==NULL)
     l= strlen(spt);
   else
     l= ept-spt;
   l+= askme->result_open_line_len;
   if(ept!=NULL && l==0)
     (*linecount)++;
   else {
     (*linecount)+= l/width;
     if(ept==NULL) {
       askme->result_open_line_len= l%width;
 break;
     }
     (*linecount)+= !!(l%width);
   }
   askme->result_open_line_len= 0;
   spt= ept+1;
 }
 return(1);
}


int Askme_pager(askme,line,flag)
struct AskmE *askme;
char *line;
int flag;
/*
 bit1= mark '@' by return 4
*/
/* return: <=0 error , 1=go on , 2=abort , 4=see flag bit1*/
{
 int ret,linecount;
 char info_text[10*4096];

 if(askme->result_page_length<=0)
   return(1);
 Askme_predict_linecount(askme,line,&linecount,0);
 if(askme->result_line_counter+linecount>askme->result_page_length) {
ask_for_page:;
   strcpy(info_text,askme->info_text);
   sprintf(askme->info_text,"\n");
   Askme_info(askme,0);
   Askme_de_eN(
     sprintf(askme->info_text,
"... [Bitte Enter-Taste druecken. @,Enter verhindert Frage. @@@ bricht ab] ...\n");
   ,
     sprintf(askme->info_text,
".... [Press Enter to continue. @,Enter avoids further stops. @@@ aborts] ....\n");
   )
   Askme_info(askme,0);
   ret= Askme_request_confirmation(askme,(flag&2) | !!(askme->restore_mode&1));
   strcpy(askme->info_text,info_text);
   if(ret<=0)
     return(ret);
   if(ret==2 || askme->suppress_result)
     return(2);
   if(ret==3)
     goto ask_for_page;
 }
 askme->result_line_counter+= linecount;
 return(1);
}


int Askme_result(askme,flag)
struct AskmE *askme;
int flag;
/*
 bit0= no considerations or computations or dialog. Just put out.
*/
{
 int ret;

 if(flag&1)
   goto put_it_out;
 if(askme->request_to_abort || askme->suppress_result)
   return(1);
 if(askme->result_page_length>0) {
   ret= Askme_pager(askme,askme->result_line,2);
   if(ret<=0)
     return(ret);
   if(ret==2)
     return(1);
   if(askme->request_to_abort || askme->suppress_result ||
      askme->request_to_skip_volume)
     return(1);
 }
put_it_out:;
 askme->bar_is_fresh= 0;
 ret= Write_to_channel(askme->result_line,1,!askme->packet_output);
 return(ret);
}


int Askme_info(askme,flag)
struct AskmE *askme;
int flag;
/*
 bit0= use pager (as with result)
 bit1= permission to suppress output
*/
{
 int ret;

 if(flag&2)
   if(askme->request_to_abort || askme->suppress_result)
     return(1);
 if(flag&1) {
   ret= Askme_pager(askme,askme->info_text,2);
   if(ret<=0)
     return(ret);
   if(ret==2)
     return(1);
   if(flag&2)
     if(askme->request_to_abort || askme->suppress_result ||
        askme->request_to_skip_volume)
       return(1);
 }
 askme->bar_is_fresh= 0;
 ret= Write_to_channel(askme->info_text,2,!askme->packet_output);
 return(ret);
}


int Askme_mark(askme,flag)
struct AskmE *askme;
int flag;
{
 int ret= 1,r_ret,i_ret;

 if(askme->mark_text[0]==0)
   return(1);
 if(askme->packet_output) 
   ret= Write_to_channel(askme->mark_text,3,0);
 else {
   sprintf(askme->result_line,"%s\n",askme->mark_text);
   r_ret= Askme_result(askme,1);
   strcpy(askme->info_text,askme->result_line);
   i_ret= Askme_info(askme,0);
   if(r_ret==0 || i_ret==0)
     ret= 0;
 }
 return(ret);
}


int Askme_restxt(askme,text)
struct AskmE *askme;
char *text;
{
 int ret;

 strcpy(askme->result_line,text);
 ret= Askme_result(askme,0);
 return(ret);
}


int Askme_rm_dirty_dir(askme,restore_adr,flag)
struct AskmE *askme;
char *restore_adr;
int flag;
/*
 bit2= do not use pager with Askme_info
 bit3= do not issue messages about removed directories
*/
{
 int ret;
 
 ret= rmdir(restore_adr);
 if(ret==-1)
   return(0);
 Askme_de_eN(
   sprintf(askme->info_text,
   "--- Nach Fehler beim Rueckspeichern leeres Directory wieder geloescht:\n");
 ,
   sprintf(askme->info_text,
           "--- Empty directory deleted after error while restoring:\n");
 )
 sprintf(askme->info_text+strlen(askme->info_text),"    %s\n",restore_adr);
 if(!(flag&8))
   Askme_info(askme,!(flag&4));
 return(1);
}


int Askme_error_on_restore(askme,ctl,flag)
struct AskmE *askme;
struct Askme_rescP *ctl;
int flag;
/*
 bit0= no error encountered, but final query (e.g. for after_error_exit) 
 bit1= be verbous about cleanup
 bit2= emergency mode, Askme_info without pager, always assume ret=0
*/
/*
 return:
  -1= exit program
   0= end operation 
   1= go on as good as possible 
   2= retry failed operation
*/
{
 int error_mode,ret,ex_ret,do_clear_dirty= 0,silent= 0;

 if(askme->request_to_abort || (flag&4))
   {ret= 0; goto ex;}
 if(ctl!=NULL)
   if(ctl->move_to_asylum)
     {ret= 0; goto ex;}
 error_mode= (askme->restore_mode>>16)&7;
 if(!(flag&1))
   askme->error_count++;
 if(error_mode==1) {
   {ret= 1; goto ex;}
 } else if(error_mode==2) {
   if((flag&1) && askme->error_count<=0)
     {ret= 1; goto ex;}
   {ret= -1; goto ex;}
 } else if(error_mode==3) {
   if(flag&1)
     if(askme->error_count>0)
       {ret= -1; goto ex;}
 } else if(error_mode==4) {
   if(flag&1)
     {ret= 1; goto ex;}
ask_user:;
   Askme_de_eN(
     sprintf(askme->info_text,
 "Wie weiter ? i=ignorieren , r=wiederholen , s=naechstes Stueck , x=Abbruch\n"
            );
   ,
     sprintf(askme->info_text,
 "What to do ? i=ignore , r=retry , s=skip up to next volume , x=abort\n");
   )
   Askme_info(askme,0);
   ret= Askme_request_confirmation(askme,1|4);
   if(ret<0)
     {ret= -1; goto ex;}
   if(ret<=0 || ret==2)
     {ret= 0; goto ex;}
   if(ret==3)
     goto ask_user;
   if(ret==6)
     {ret= 2; goto ex;}
   {ret= 1; goto ex;}
 }
 if(flag&1)
   {ret= 1; goto ex;}
 askme->request_to_abort= 1;
 ret=0;
ex:;
 if(ctl==NULL || (flag&1)) 
   return(ret);
 if(flag&2) {
   sprintf(askme->info_text,
           "--- handling possibly dirty file:\n    %s\n",ctl->restore_adr);
   Askme_info(askme,3*!(flag&4));
 }
 if(ctl->error_is_handled) {
   if(flag&2) {
     sprintf(askme->info_text,"    error already handled\n");
     Askme_info(askme,3*!(flag&4));
   }
   return(ret);
 }
 if(!ctl->file_is_new) {
   if(flag&2) {
     sprintf(askme->info_text,
             "    file object was not created by this restore operation\n");
     Askme_info(askme,3*!(flag&4));
   }
   do_clear_dirty= 1;
   if(ctl->original_altered)
      goto try_to_set_properties;
   if(flag&2) {
     sprintf(askme->info_text,
       "    original file object was not altered by this restore operation\n");
     Askme_info(askme,3*!(flag&4));
   }
   Askme_rescp_release_dirty(ctl,flag&2);
   return(ret);
 }

 /* file restored incompletely */
 if((askme->restore_mode&(1<<22))) {
   if(flag&2) {
     sprintf(askme->info_text,
             "    -restore_mode:keep_on_error protects dirty files\n\n");
     Askme_info(askme,3*!(flag&4));
   }
 } else if(ctl->is_dir) {
   if(ret<=0 && !ctl->dir_has_valid_file) {
     ex_ret= Askme_rm_dirty_dir(askme,ctl->restore_adr,flag&4);
     if(ex_ret==1) {
       /* delete freshly created part of prepare path */
       ex_ret= Askme_rescp_prepare_path(ctl,1|2|(flag&4));
       if(ex_ret==1) {
         Askme_rescp_release_dirty(ctl,flag&2);
       }
     }
     return(ret);
   } else {
     if(flag&2) {
       sprintf(askme->info_text,"    directory restored some valid files\n");
       Askme_info(askme,3*!(flag&4));
     }
   }
 } else {
   ex_ret= unlink(ctl->restore_adr);
   if(ex_ret!=-1) {
     Askme_de_eN(
       sprintf(askme->info_text,
               "--- Nach Fehler beim Rueckspeichern wieder geloescht:\n");
     ,
       sprintf(askme->info_text,
               "--- Deleted after error while restoring:\n");
     )
     sprintf(askme->info_text+strlen(askme->info_text),
             "    %s\n",ctl->restore_adr);
     silent= 0;
     if(ctl->move_to_asylum)
       if(ctl->boss!=NULL)
         if(!ctl->boss->path_is_prepared)
           silent= 1;
     if(!silent)
       Askme_info(askme,!(flag&4));
     ex_ret= Askme_rescp_prepare_path(ctl,1|2|(flag&4)|(silent<<3));
     if(ex_ret==1) {
       Askme_rescp_release_dirty(ctl,flag&2);
     }
   }
   return(ret);
 }

try_to_set_properties:;
 /* try to set properties of surviving file object */
 if(flag&2) {
   sprintf(askme->info_text,
           "    trying to restore original properties of file object\n");
   Askme_info(askme,3*!(flag&4));
 }

 if((ctl->original_altered) && !ctl->file_is_new) {
   /* >>> restore properties of original file object rather than those of
          failed restore file object */;
   do_clear_dirty= 0; /* <<< */

 }

 ctl->times_valid= 0; /* but do not set atime,mtime to old values */
 ex_ret= Askme_rescp_properties(ctl,(1<<16)|((2^(flag&2))<<16));
 if(ex_ret>0) {
   if((ctl->is_dir || do_clear_dirty) && 
      strcmp(askme->resume_last_dirty,ctl->restore_adr)==0) {
     Askme_rescp_release_dirty(ctl,flag&2);
     askme->info_text[0]= 0;
   } else if((flag&2) && (ctl->is_dir || do_clear_dirty)) {
     sprintf(askme->info_text,
     "    address does not match resume_last_dirty (may not clear):\n    %s\n",
             askme->resume_last_dirty);
   } else {
     sprintf(askme->info_text,
     "    fileobject neither unaltered original nor completely restored\n%s\n",
     "    (may not clear resume_last_dirty)");
   }
 } else {
   if((flag&2))
     sprintf(askme->info_text,
       "    reset of properties failed (may not clear resume_last_dirty)\n\n");
 }
 if((flag&2) && askme->info_text[0]!=0)
   Askme_info(askme,3*!(flag&4));
 return(ret); 
}


int Askme_abort_handler(askme,signum,flag)
struct AskmE *askme;
int signum;
int flag;
{
 struct Askme_rescP *ctl;
 int verbous_action;

 if(askme==NULL)
   return(0);
 verbous_action= (((askme->restore_mode>>7)&3)==3);
 sprintf(askme->info_text,"ABORT");
 if(((askme->abort_handler>>4)&15)==2) {
   Cleanup_set_handlers(NULL,NULL,2); /* ignore all signals */
   strcat(askme->info_text,": trying to ignore further signals");
 } else if(((askme->abort_handler>>4)&15)==0) {
   Cleanup_set_handlers(NULL,NULL,1); /* allow abort */
   strcat(askme->info_text,": enabled default action on further signals");
 }
 strcat(askme->info_text,"\n");
 if(!verbous_action) {
   Askme_de_eN(
     sprintf(askme->info_text,
      "Rueckspeicherung in heiklem Moment unterbrochen. Aufraeumen beginnt.\n"
            );
   ,
     sprintf(askme->info_text,
      "Restore operation interrupted in precarious state. Cleanup started.\n");
   )
 }
 Askme_info(askme,0);

 while(askme->restore_ctl_stack!=NULL) {
   ctl= askme->restore_ctl_stack;
   sprintf(askme->info_text,
           "ABORT: handling file : %s\n",ctl->restore_adr);
   if(verbous_action)
     Askme_info(askme,0);
   askme->request_to_abort= 0;
   Askme_error_on_restore(askme,ctl,((askme->abort_handler&256)>>7)|4);
   Askme_rescp_destroy(&ctl,0);   
 }
 if(askme->resume_last_dirty[0]!=0) {
   Askme_de_eN(
     sprintf(askme->info_text,
             "ABORT: File verbleibt in fragwuerdigem Zustand:\n");
   ,
     sprintf(askme->info_text,
             "ABORT: file recorded as potentially dirty:\n");
   )
   sprintf(askme->info_text+strlen(askme->info_text),
             "       %s\n",askme->resume_last_dirty);
   Askme_de_eN(
     sprintf(askme->info_text+strlen(askme->info_text),
        "       Der File selbst ist fragwuerdig und ebenso\n");
     sprintf(askme->info_text+strlen(askme->info_text),
        "       die Besitzer-Zugriffsrechte der Direcories in seinem Pfad.\n");
   ,
     sprintf(askme->info_text+strlen(askme->info_text),
        "       The file itself is questionable as well as\n");
     sprintf(askme->info_text+strlen(askme->info_text),
        "       the owner access permissions of directories in its path.\n");
   )
   Askme_info(askme,0);
 } else {
   Askme_de_eN(
     sprintf(askme->info_text,
             "\nDie nun existierenden Files sind vermutlich in Ordnung.\n\n");
   ,
     sprintf(askme->info_text,
             "\nAll existing files are supposed to be clean.\n\n");
   )
   Askme_info(askme,0);
 }
 if((askme->media_mode&(4|16)) && (askme->media_mode&1)) {
   Cleanup_set_handlers(NULL,NULL,1); /* allow abort */
   strcpy(askme->info_text,
          "ABORT: enabled default action on further signals\n");
   if(verbous_action)
     Askme_info(askme,0);
   Askme_de_eN(
     sprintf(askme->info_text,"Abschlussbehandlung von Datentrager an: %s\n",
             askme->restore_from);
   ,
     sprintf(askme->info_text,"Stand by for final treatment of media at: %s\n",
             askme->restore_from);
   )
   Askme_info(askme,0);
   sleep(1); /* make "device is busy" errors more unlikely */
   Askme_umount_eject(askme,askme->restore_from,askme->restore_from,0);
 }
 return(1);
}


int Askme_reset_counters(askme,flag)
struct AskmE *askme;
int flag;
/*
 bit0= reset askme->found to -1
*/
{
 if(flag&1)
   askme->found= -1;
 else
   askme->found= 0;
 askme->error_count= 0;
 askme->restore_count= 0;
 askme->restore_bytes= 0;
 return(1);
}


int Askme_count_volumes_and_subs(askme,flag)
struct AskmE *askme;
int flag;
{
 int ret,maxlev= -1,maxvol= -1;
 FILE *fp= NULL;
 char line[2*4096],*cpt;

 fp= fopen(askme->filename,"r");
 if(fp==NULL)
   {ret= 0; goto ex;}
 while(Sfile_fgets(line,sizeof(line)-1,fp)!=NULL) {
   if(strncmp(line,"for lev in ",11)==0) {
     maxlev= -1;
     sscanf(line+11,"%d",&maxlev);
     if(maxlev>=0)
       askme->num_sublevels= maxlev+1;
   }
   if(strncmp(line,"  for cd in ",12)==0) {
     cpt= strrchr(line,' ');
     maxvol= -1;
     sscanf(cpt+1,"%d",&maxvol);
     if(maxvol>0)
       askme->num_volumes= askme->highest_volno= maxvol;
   }
 }
 ret= 1;
 if(maxvol<=0) {
   Askme_de_eN(
     sprintf(askme->info_text,
            "Kein verwendbarer Volumenzaehler gefunden. Ist das ueberhaupt ein ASKME File ?\n");
   ,
     sprintf(askme->info_text,
            "No usable volume count found. Is this an ASKME file at all ?\n");
   )
   Askme_info(askme,0);
   sprintf(askme->info_text,"  %s\n",askme->filename);
   Askme_info(askme,0);
   ret= -1;
 }
ex:;
 if(fp!=NULL)
   fclose(fp);
 return(ret);
}


int Askme__bourne_to_reg(bourne_expr,reg_expr,flag)
char bourne_expr[],reg_expr[];
int flag;
/* reg_expr should be twice as large as bourne_expr ( + 2 to be exact) */
/* return: 2= bourne_expr is surely a constant */
{
 char *wpt,*lpt;
 int backslash= 0,is_constant= 1,in_square_brackets= 0;
 int first_in_square_brackets=0;

 wpt= reg_expr;
 lpt= bourne_expr;

 *(wpt++)= '^';

 while(*lpt!=0){
  if(first_in_square_brackets>0)
    first_in_square_brackets--;
  if(!backslash){
    switch(*lpt){
    case '?':
      *(wpt++)= '.';
      is_constant= 0;
    break;case '*':
      *(wpt++)= '.';
      *(wpt++)= '*';
      is_constant= 0;
    break;case '.':
      *(wpt++)= '\\';
      *(wpt++)= '.';
    break;case '+':
      *(wpt++)= '\\';
      *(wpt++)= '+';
    break;case '[':
      *(wpt++)= *lpt;
      first_in_square_brackets= 2;
      in_square_brackets= 1;
      is_constant= 0;
    break;case ']':
      *(wpt++)= *lpt;
      in_square_brackets= 0;
    break;case '!':
      if(first_in_square_brackets)
        *(wpt++)= '^';
      else if(in_square_brackets)
        *(wpt++)= '!';
      else {
        *(wpt++)= '\\';
        *(wpt++)= '!';
      }
    break;case '^':
      if(in_square_brackets)
        *(wpt++)= '^';
      else
        *(wpt++)= '\\';
        *(wpt++)= '^';
    break;case '$':
      *(wpt++)= '\\';
      *(wpt++)= '$';
    break;case '\\':
      backslash= 1;
      *(wpt++)= '\\';
      is_constant= 0;
    break;default:
      *(wpt++)= *lpt;
    }
  } else {
    backslash= 0;
    *(wpt++)= *lpt;
  }
  lpt++;
 }
 *(wpt++)= '$';
 *wpt= 0;
 return(1+(is_constant>0));
}


int Askme_fsm_add_level_info(askme,reader,idx,flag)
struct AskmE *askme;
struct LreadeR *reader;
int *idx;
int flag;
/*
 bit0= write headline
*/
{
 int new_linecount,num_volumes= 0,level,i,ret;
 char **new_lines= NULL;
 char *datept,*younger_than_pt,line[4096];

 Lreader_get_id_info(reader,&datept,&level,&younger_than_pt,&num_volumes,0);
 if(num_volumes>0)
   askme->num_volumes+= num_volumes;
 if(num_volumes>askme->highest_volno)
   askme->highest_volno= num_volumes;
 if(askme->highest_level<level) {
   askme->highest_level= level;
   askme->num_highest_level_volumes= num_volumes;
 }

 if(*idx+4>=askme->num_level_info) {
   /* enlarge */;
   new_linecount= askme->num_level_info*4+4;
   new_lines= TSOB_FELD(char *,new_linecount);
   if(new_lines==NULL)
     return(-1);
   for(i=0;i<askme->num_level_info;i++) {
     new_lines[i]= askme->level_info[i];
     askme->level_info[i]= NULL;
   }
   for(;i<new_linecount;i++)
     new_lines[i]= NULL;
   if(askme->level_info!=NULL)
     free((char *) askme->level_info);
   askme->level_info= new_lines;
   askme->num_level_info= new_linecount;
 }
 i= *idx;

 if(flag&1) {
   if(Sregex_string(&(askme->level_info[i]),"scdbackup backup tree model",0)<=0)
     {ret= -1; goto ex;}
   i++;
 }
 sprintf(line,"backup date %s",datept);
 if(Sregex_string(&(askme->level_info[i]),line,0)<=0)
   {ret= -1; goto ex;}
 i++;
 if(level>=0 && younger_than_pt!=NULL) {
   sprintf(line,"-level %d   -younger_than %s",level,younger_than_pt);
   if(Sregex_string(&(askme->level_info[i]),line,0)<=0)
     {ret= -1; goto ex;}
   i++;
 }
 sprintf(line,"this backup consists of %d parts",num_volumes);
 if(Sregex_string(&(askme->level_info[i]),line,0)<=0)
   {ret= -1; goto ex;}
 i++;
 ret= 1;
ex:;
 *idx= i;
 return(ret);
}


int Askme_mount_for_lists(askme,level,volume,adr,flag)
struct AskmE *askme;
int level,volume;
char *adr;
int flag;
/*
 bit0= force umount attempt
 bit1= disallow bold first try wether the mounted volume is the right one
 bit15= umount
*/
/* return: <=0 error , 1=ok , 2=not mounted, abort , 
            5=not mounted, skip volume
*/
{
 int ret,l,i,addon,media_level,media_volume,media_vtotal,is_mounted= 0;
 int bold_first_try= 1,report_err_dummy;
 char lev_vol_line[161],*ipt;
 char media_date[81],media_yt[81],volno_text[81];
 char media_id[81],media_id_date[81];

 is_mounted= (askme->mounted_level>=0 && askme->mounted_volume>0) || (flag&1);
 if(!flag&2)
   if(!(is_mounted && (askme->mounted_level!=level 
                       || (volume>0 && askme->mounted_volume!=volume))))
     if(!(flag&(1<<15)))
       goto check_level_volume;

umount_eject:;
 bold_first_try= 0;
 ret= Askme_umount_eject(askme,adr,adr,!!is_mounted);
 if(ret<0)
   goto ex;
 is_mounted= 0;
 if(flag&(1<<15))
   {ret= 1; goto ex;}

ask_for_volume:;
 /* ask for level,volume */
 strcpy(askme->info_text,"\n");
 if(volume>0)
   sprintf(volno_text,"%d",volume);
 else
   volno_text[0]= 0;
 Askme_de_eN(
   sprintf(lev_vol_line,
           " Bitte einlegen%s:   LEVEL  %d   %sBACKUPSTUECK  %s   ",
           (askme->media_mode&4)?"":" und mounten",level,
           (volume>0?"":"letztes "),volno_text);
 ,
   sprintf(lev_vol_line,
           " Please insert%s:   LEVEL  %d   %sVOLUME  %s   ",
           (askme->media_mode&4)?"":" and mount",level,
           (volume>0?"":"last "),volno_text);
 )
 /* add flexible length ### and \n */
 l= strlen(lev_vol_line);
 addon= (79-1-l)/2;
 if(addon>0) {
   for(i=l;i>=0;i--)
     lev_vol_line[i+addon]= lev_vol_line[i];
   for(i=0;i<addon;i++)
     lev_vol_line[i]= lev_vol_line[l+addon+i]= '#';
 }
 lev_vol_line[l+2*addon]= '\n';
 lev_vol_line[l+2*addon+1]= 0;
 l+= 2*addon+1;
 ipt= askme->info_text;
 strcpy(ipt,lev_vol_line);
 for(i=0;i<l-1;i++)
   ipt[i]= '#';
 ipt= ipt+strlen(ipt);
 strcpy(ipt,lev_vol_line);
 ipt= ipt+strlen(ipt);
 strcpy(ipt,lev_vol_line);
 for(i=0;i<l-1;i++)
   ipt[i]= '#';
 ipt= ipt+strlen(ipt);
 Askme_de_eN(
   strcpy(ipt,
".... [ Wenn fertig, Enter-Taste druecken. @@@ = Alle Levels eingelesen ] ....\n");
 ,
   strcpy(ipt,
"............. [ Press enter when done. @@@ = All levels done ] .............\n");
 )
 Askme_info(askme,0);

 ret= Askme_request_confirmation(askme,1|8);
 if(ret<=0 || ret==2)
   goto ex;
 if(ret==3) 
   goto ask_for_volume;
 if(askme->request_to_skip_volume>0)
   {ret= 5; goto ex;}

 Askme_de_eN(
   sprintf(askme->info_text,"Verarbeitung des Backupstuecks beginnt ...\n");
 ,
   sprintf(askme->info_text,"Processing of volume started ...\n");
 )
 Askme_info(askme,1);
 ret= Askme_load_mount(askme,adr,adr,0);
 if(ret<0)
   goto ex;
 if(ret==0)
   goto umount_eject;
 if(ret==2)
   is_mounted= 1;

check_level_volume:;
 ret= Askme_identify_volume(askme,adr,media_id,media_id_date,
                            &media_level,&media_volume,&media_vtotal,
                            media_date,media_yt,0);
 if(ret<0)
   goto ex;
 if(ret==0) {
   if(bold_first_try)
     goto umount_eject;
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- KEIN IDENTIFIZIERBARES BACKUPSTUECK IN: %s\n",adr);
   ,
     sprintf(askme->info_text,
             "--- CANNOT IDENTIFY A VALID BACKUP VOLUME IN: %s\n",adr);
   )
   Askme_info(askme,0);
   goto umount_eject;
 }
 if(askme->backup_id[0]!=0 && media_level>=askme->backup_id_level &&
    strcmp(askme->backup_id,media_id)!=0) {
   if(bold_first_try)
     goto umount_eject;
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- FALSCHES BACKUP ID !\n    ist:  '%s'\n    soll: '%s'\n\n",
             media_id,askme->backup_id);
   ,
     sprintf(askme->info_text,
           "--- WRONG BACKUP ID !\n    read:     '%s'\n    expected: '%s'\n\n",
             media_id,askme->backup_id);
   )
   ret= Askme_handle_picky0(askme,askme->info_text,&report_err_dummy,0);
   if(ret<0)
     goto ex;
   if(ret==0)
     goto umount_eject;
 }
 if((media_volume!=volume && volume>0) || media_level!=level) {
   if(bold_first_try)
     goto umount_eject;
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- FALSCHES BACKUPSTUECK !\n    (Level= %d, Stueck= %d, %s)\n",
             media_level,media_volume,media_date);
   ,
     sprintf(askme->info_text,
             "--- WRONG VOLUME !\n    (level= %d, volume= %d, %s)\n",
             media_level,media_volume,media_date);
   )
   Askme_info(askme,0);
   goto umount_eject;
 }
 askme->mounted_level= level;
 askme->mounted_volume= volume;
 ret= 1;
ex:;
 return(ret);
}


int Askme_handle_picky0(askme,errmsg,report_error,flag)
struct AskmE *askme;
char *errmsg;
int *report_error;
int flag;
{
 Write_to_channel(errmsg,2,!askme->packet_output);
 if(((askme->read_pickiness)&3)>0) {
   Askme_de_eN(
     sprintf(errmsg,
             "    ( Problem waere tolerabel mit  -read_pickiness:zero )\n\n");
   ,
     sprintf(errmsg,
             "    ( this would be tolerable with  -read_pickiness:zero )\n\n");
   )
   Write_to_channel(errmsg,2,!askme->packet_output);
   *report_error= 0;
   return(0);
 }
 return(1);
}


int Askme_read_tree(askme,adr,level,returned_reader,flag)
struct AskmE *askme;
char *adr;
int level;
struct LreadeR **returned_reader;
int flag;
/*
 bit0= create all inodes with struct FsinfO 
 bit1= only accept content list or complete list 
       resp. with bit15 vanish list, resp. with bit13 directory list
 bit2= do not issue error message about wrong list type
 bit3= do not destroy LreadeR
 bit4= do not destroy old LreadeR before creating new one
 bit5= do only accept list file of given level

 bit12= search for youngest file which matches adr_body*.adr_suffix
 bit13= do only update permissions of existing directories 
 >>> bit14= do only update (resp. create) FsinfO of existing nodes 
 bit15= delete files from tree rather than inserting them
*/
/*
 return:
 >0   highest volume number (at least 1)
 <= 0 error, especially:
 -4   wrong list type with bit1
*/
{
 int popened= 0,ret,is_dir= 0,node_level=- 1,dummy,report_error= 1,pickiness;
 int packeted,volume,file_type,partno,partcount,part_status;
 FILE *fpin= NULL;
 struct LreadeR *reader= NULL;
 struct FsnodE *tcr,*newtn= NULL,*wd_node= NULL;
 char *linept,*sourcept,*targetpt,*typept,*idpt,errmsg[2*4096];
 char *eff_adr,youngest_adr[4096],*date_pt,*younger_than_pt;
 off_t count,part_start,part_size,part_full_size;
 struct FsnodE **root_node,**target_root,*existing_node;
 struct FspartreC *partrec= NULL;

 /* these where once parameters of Fsnode_read_tree() */
 root_node= &(askme->root_node);
 target_root= &(askme->target_root);
 pickiness= ((askme->read_pickiness)&3)<<7;
 packeted= (!!askme->packet_output)<<6;

 eff_adr= adr;
 if(flag&(1<<12)) {
   ret= A_youngest_listfile(adr,youngest_adr,0);
   if(ret>0)
     eff_adr= youngest_adr;
 }
 popened= A_open_listfile(&fpin,eff_adr,"r",1|packeted);
 if(popened!=2)
   popened= 0;
 if(fpin==NULL) 
   {ret= 0; goto ex;}
 if(!(flag&16))
   Lreader_destroy(returned_reader,0);
 ret= Lreader_new(&reader,eff_adr,fpin,0);
 if(ret<=0) {
   sprintf(errmsg,"Failed to create reader object (out of memory ?)\n");
   Write_to_channel(errmsg,2,!askme->packet_output);
   goto ex;
 }
 Lreader_set_root_nodes(reader,*root_node,*target_root,0);

 tcr= *root_node;
 while(1) {
   ret= Lreader_read_line(reader,0);
   if(ret<=0) {
     if(ret!= -4)
       ret= 0; 
     goto ex;
   }
   if(ret==1 || ret==2) {
     is_dir= (ret==1);
     Lreader_get_src_target(reader,&sourcept,&targetpt,0);
     ret= Fsnode_graft_branch(root_node,&newtn,sourcept,reader,NULL,
                              (flag&(1|(7<<13)))|2|packeted|pickiness);
     if(ret<=0)
       goto ex;
     if(is_dir)
       tcr= newtn;

   } else if (ret==3 || ret==4) {
     is_dir= (ret==3);
     Lreader_get_src_target(reader,&sourcept,&targetpt,1);
     ret= Fsnode_graft_branch(&tcr,&newtn,sourcept,reader,NULL,
                              (flag&(1|(7<<13)))|packeted|pickiness);
     if(ret<=0)
       goto ex;
     if(is_dir)
       tcr= newtn;

   } else if (ret==5) {
     /* move tree one directory up */;
     if(tcr==NULL) {
no_dir_up:;
       sprintf(askme->info_text,
               "--- list format error : no parent directory on ..\n");
       Askme_info(askme,0);
       {ret= -1; goto ex;}
     }
     if(tcr->extra==NULL)
       goto no_dir_up;
     tcr= tcr->extra->up;

   } else if (ret==6) {
     /* move to list end of initial wd node */;
     tcr= wd_node; 

   } else if (ret==9) { /* list type determined */
     if(flag&2) {
       Lreader_get_list_type(reader,&typept,0);
       if(flag&(1<<15)) {
         if(strcmp(typept,"vanished file objects list")!=0)
           {ret= -4; goto ex;}
       } else if(flag&(1<<13)) {
         if(strcmp(typept,"complete backup area directory list")!=0)
           {ret= -4; goto ex;}
       } else {
         if(strcmp(typept,"complete backup area file list")!=0 &&
            strcmp(typept,"backup content list")!=0)
           {ret= -4; goto ex;}
       }
     }
     sprintf(errmsg,"reading  %s\n",eff_adr);
     Write_to_channel(errmsg,2,!askme->packet_output);
 
   } else if(ret==11) { /* level */
     /*
     Lreader_get_nodedata(reader,&sourcept,&node_level,&dummy,&dummy,0);
     */
     Lreader_get_id_info(reader,&date_pt,&node_level,&younger_than_pt,&dummy);
     if(flag&32) {
       if(node_level!=level) {
         sprintf(errmsg,
                 "\n--- unexpected level number  %d . Expected is %d\n\n",
                 node_level,level);
         ret= Askme_handle_picky0(askme,errmsg,&report_error,0);
         if(ret<=0)
           goto ex;
       }
     }
     if(younger_than_pt[0]!=0 && node_level>0 && !(flag&(1<<13))) {
       if(strcmp(younger_than_pt,askme->recent_backup_date)!=0) {
         Askme_de_eN(
           sprintf(askme->info_text,
                   "--- FALSCHES BASISDATUM !\n    ist : -younger_than '%s'\n    soll: -younger_than '%s'\n\n",
                   younger_than_pt,askme->recent_backup_date);
         ,
           sprintf(askme->info_text,
                   "--- WRONG BASE DATE !\n    read    : -younger_than '%s'\n    expected: -younger_than '%s'\n\n",
                   younger_than_pt,askme->recent_backup_date);
         )
         ret= Askme_handle_picky0(askme,askme->info_text,&dummy,0);
         if(ret<=0)
           goto ex;
       }
     }

   } else if(ret==12) { /* date */
     Lreader_get_id_info(reader,&date_pt,&node_level,&younger_than_pt,&dummy);
     if(!(flag&((1<<13)|(1<<15)))) {
       strncpy(askme->recent_backup_date,date_pt,
               sizeof(askme->recent_backup_date)-1);
       askme->recent_backup_date[sizeof(askme->recent_backup_date)-1]= 0;
     }
     if(date_pt[0]!=0 && (flag&(1<<13))) {
        if(strcmp(date_pt,askme->recent_backup_date)!=0) {
          Askme_de_eN(
            sprintf(askme->info_text,
                    "--- FALSCHES BACKUPDATUM !\n    ist : '%s'\n    soll: '%s'\n\n",
                    date_pt,askme->recent_backup_date);
          ,
            sprintf(askme->info_text,
                    "--- WRONG BACKUP DATE !\n    read    : '%s'\n    expected: '%s'\n\n",
                    date_pt,askme->recent_backup_date);
          )
          ret= Askme_handle_picky0(askme,askme->info_text,&dummy,0);
          if(ret<=0)
            goto ex;
        }
     }

   } else if(ret==15) { /* working directory */
     Lreader_get_line(reader,&linept,&count,0);
     ret= Fsnode_graft_branch(root_node,&newtn,linept,reader,NULL,
                              (flag&1)|8|packeted|pickiness);
     if(ret<=0)
       goto ex;
     wd_node= newtn;
     tcr= newtn;

   } else if(ret==17) { /* #=# end */
 break;

   } else if(ret==18) { /* id */
     Lreader_get_id(reader,&idpt,0);
     if(askme->backup_id[0]==0) {
       if(strlen(idpt)>=sizeof(askme->backup_id)) {
         sprintf(errmsg,
           "\n--- oversized backup id string (chars: %d , admissible: %d)\n\n",
           strlen(idpt),sizeof(askme->backup_id)-1);
         ret= Askme_handle_picky0(askme,errmsg,&report_error,0);
         if(ret<=0)
           goto ex;
       } else
         strcpy(askme->backup_id,idpt);
     } else {
       if(strcmp(askme->backup_id,idpt)!=0) {
         sprintf(errmsg,
                 "\n--- backup id of list does not match id of previous lists\n    read: '%s'\n    prev: '%s'\n\n",
                 idpt,askme->backup_id);
         ret= Askme_handle_picky0(askme,errmsg,&report_error,0);
         if(ret<=0)
           goto ex;
       }
     }

   } else if(ret==19) { /* part record "p" */
     /* Might become interesting for unmapped archives (afio, star) */
     /* With a target mapped filesystem use "P" record */;

   } else if(ret==20) { /* part record "P" */

  continue; /* <<< */

     /* >>> obtain partno and partcount from target name */;

     Lreader_get_nodedata(reader,&sourcept,&node_level,&volume,&file_type,0);
     Lreader_get_part_info(reader,&part_status,&part_start,
                           &part_size,&part_full_size,0);
     ret= Fspartrec_new(&partrec,node_level,volume,partno,partcount,
                                 part_start,part_size,0);
     if(ret<=0)
       goto ex;

     existing_node= NULL;

     /* cause partrec to be attached to tlink */
     Lreader_get_src_target(reader,&sourcept,&targetpt,0);
     ret= Fsnode_graft_branch(root_node,&newtn,sourcept,reader,partrec,
                             (flag&(1|(7<<13)))|2|(1<<12)|packeted|pickiness);
     if(ret<=0)
       goto ex;

   } else {
     /* ignored for now */;
   }
 }
 ret= 1;
ex:
 if(ret>0 && reader!=NULL) {
   Lreader_get_line(reader,&linept,&count,0);
   sprintf(errmsg,"    %12.lf lines read\n",(double) count);
   Write_to_channel(errmsg,2,!askme->packet_output);
 } else if(ret<=0 && reader!=NULL && report_error) {
   if(ret==-4)
     Lreader_set_errcode(reader,4,0);
   if(!((flag&4) && Lreader_get_errcode(reader,0)==4)) {
     sprintf(errmsg,"\n%s\n",Lreader_strerror(reader,0));
     Write_to_channel(errmsg,2,!askme->packet_output);
     Lreader_get_line(reader,&linept,&count,0);
     sprintf(errmsg,"Last input: %.lf :  %s\n", (double) count,linept);
     Write_to_channel(errmsg,2,!askme->packet_output);
     Lreader_get_src_target(reader,&sourcept,&targetpt,0);
     sprintf(errmsg,"Last  addr: %s = %s\n",targetpt,sourcept);
     Write_to_channel(errmsg,2,!askme->packet_output);
   }
 }
 if(fpin!=NULL){
   if(popened)
     pclose(fpin);
   else
     Sfile_fclose(fpin);
 }
 if(ret>0) {
   Lreader_get_highest_volume(reader,&ret,0);
   if(ret<=0)
     ret= 1;
 }
 if(reader!=NULL)
   Lreader_get_root_nodes(reader,root_node,target_root,0);
 if(!(flag&8)) {
   Lreader_destroy(&reader,0);
   *returned_reader= NULL;
 } else
   *returned_reader= reader;
 if(askme->backup_id_level<0 && askme->backup_id[0]!=0) {
   if(node_level>=0)
     askme->backup_id_level= node_level;
   else
     askme->backup_id[0]= 0; /* no id without level */
 }
 return(ret);
}



int Askme_load_old_tree(askme,content_list_adr,vanish_list_adr,flag)
struct AskmE *askme;
char *content_list_adr;
char *vanish_list_adr;
int flag;
/*
 bit0..3= mode
     0= read content_list_adr only
     1= read under dialog control from fixed addresses
     2= read  $conf_dir/level_$i/$*_list.gz, i<=$conf_dir/highest_level
 bit4= do not issue error messages about non-matching file types
 bit5= do not sort finally
*/
/*
 return:
   1 = success
 <=0 = error, especially:
  -4 = neither content list nor complete list
*/
{
 int ret,highest_level= -1,i,mode,info_idx= 0,hret,l,picky;
 int was_mounted= 0,have_vanished_list,have_content_list,have_dir_list;
 struct LreadeR *reader= NULL;
 FILE *fp= NULL;
 char adr[4096],*cpt;

 picky= askme->read_pickiness<<7;

 mode= flag&15;
 if(mode==1) {
   for(i=0;1;i++) {
     have_vanished_list= 0;
     have_content_list= 0;
     have_dir_list= 0;
prompt_for_level:;
     strcpy(adr,askme->filename);
     cpt= strstr(adr,"/added_by_scdbackup");
     if(cpt!=NULL && cpt!=adr)
       *cpt= 0;
     ret= Askme_mount_for_lists(askme,i,-1,adr,((i>0) || was_mounted)*3);
     if(ret<0)
       goto ex;
     if(ret==2) {
   break;
     }
     if(ret==5)
   continue;
     was_mounted= 1;
     if(i>0 && have_vanished_list==0) {
       /* read vanish list */
       sprintf(adr,"%s/%s",askme->filename,"vanished_file_list.gz");

#ifdef Use_fsnode_read_treE
       ret= Fsnode_read_tree(&(askme->root_node),&(askme->target_root),
                           adr,askme->highest_level+1,&reader,
                           2|8|32|((!!askme->packet_output)<<6)|(1<<15)|picky);
#else /* Use_fsnode_read_treE */
       ret= Askme_read_tree(askme,adr,askme->highest_level+1,&reader,
                            2|8|32|(1<<15));
#endif /* ! Use_fsnode_read_treE */

       if(ret<0)
         goto ex;
       if(ret==0) 
         goto prompt_for_level;
       have_vanished_list= 1;
     }
     /* read content list */;
     if(have_content_list==0) {
       sprintf(adr,"%s/%s",askme->filename,"content_file_list.gz");

#ifdef Use_fsnode_read_treE
       ret= Fsnode_read_tree(&(askme->root_node),&(askme->target_root),
              adr,askme->highest_level+1,&reader,
            2|((flag&16)>>2)|8|((i>0)<<5)|((!!askme->packet_output)<<6)|picky);
#else /* Use_fsnode_read_treE */
       ret= Askme_read_tree(askme,adr,askme->highest_level+1,&reader,
                            2|((flag&16)>>2)|8|((i>0)<<5));
#endif /* ! Use_fsnode_read_treE */

       if(ret<0)
         goto ex;
       if(ret==0)
         goto prompt_for_level;
       have_content_list= 1;
       /* emulate ASKME level_info */
       ret= Askme_fsm_add_level_info(askme,reader,&info_idx,(i==0));
       if(ret<=0)
         goto ex;
       askme->count_level_info= info_idx;
       if(i==0)
         Lreader_get_initial_wd(reader,askme->initial_wd,0);
     }

     /* read directory list */;
     if(have_dir_list==0 && !askme->no_dir_list) {
       sprintf(adr,"%s/%s",askme->filename,"complete_dir_list.gz");

#ifdef Use_fsnode_read_treE
       ret= Fsnode_read_tree(&(askme->root_node),&(askme->target_root),
                             adr,askme->highest_level,&reader,
                   2|((flag&16)>>2)|8|((i>0)<<5)|((!!askme->packet_output)<<6)|
                   picky|(1<<13));
#else /* Use_fsnode_read_treE */
       ret= Askme_read_tree(askme,adr,askme->highest_level,&reader,
                            2|((flag&16)>>2)|8|((i>0)<<5)|(1<<13));
#endif /* ! Use_fsnode_read_treE */

       if(ret<0)
         goto ex;
       if(ret==0) 
         goto prompt_for_level;
       have_dir_list= 1;
     } else if(askme->no_dir_list)
       have_dir_list= 2;
   }
   if(askme->root_node==NULL) {
     sprintf(askme->info_text,"--- no tree model has been loaded.\n");
     Askme_info(askme,0);
     ret= 0; goto ex;
   }

 } else if(mode==2) {

   sprintf(adr,"%s/highest_level",askme->conf_dir);
   fp= fopen(adr,"r");
   if(fp==NULL) {
     sprintf(askme->info_text,"%s : %s",askme->progname,strerror(errno));
     Askme_info(askme,0);
     sprintf(askme->info_text,"%s : cannot open file for reading : %s\n",
             askme->progname,adr);
     Askme_info(askme,0);
     ret= 0; goto ex;
   }
   ret= fscanf(fp,"%d",&highest_level);
   if(ret!=1 || highest_level<0) {
     sprintf(askme->info_text,"%s : read unusable number %d from file : %s\n",
             askme->progname,highest_level,adr);
     Askme_info(askme,0);
     ret= 0; goto ex;
   }
   fclose(fp); fp= NULL;
   for(i=0;i<=highest_level;i++) {
     if(i>0) {
       /* read vanish list */
       sprintf(adr,"%s/level_%d/vanished_file_list.gz",askme->conf_dir,i);

#ifdef Use_fsnode_read_treE
       ret= Fsnode_read_tree(&(askme->root_node),&(askme->target_root),
                             adr,0,&reader,
                             2|8|((!!askme->packet_output)<<6)|picky|(1<<15));
#else /* Use_fsnode_read_treE */
       ret= Askme_read_tree(askme,adr,0,&reader,2|8|(1<<15));
#endif /* ! Use_fsnode_read_treE */

       if(ret<=0)
         goto ex;
     }
     /* read content list */;
     sprintf(adr,"%s/level_%d/content_file_list.gz",askme->conf_dir,i);

#ifdef Use_fsnode_read_treE
     ret= Fsnode_read_tree(&(askme->root_node),&(askme->target_root),
                       adr,0,&reader,
                       2|((flag&16)>>2)|8|((!!askme->packet_output)<<6)|picky);
#else /* Use_fsnode_read_treE */
     ret= Askme_read_tree(askme,adr,0,&reader,2|((flag&16)>>2)|8);
#endif /* ! Use_fsnode_read_treE */

     if(ret<=0)
       goto ex;
     /* emulate ASKME level_info */
     ret= Askme_fsm_add_level_info(askme,reader,&info_idx,(i==0));
     if(ret<=0)
       goto ex;
     if(askme->highest_level>i)
       askme->highest_level= i;
     askme->count_level_info= info_idx;
     if(i==0)
       Lreader_get_initial_wd(reader,askme->initial_wd,0);

     /* read directory list */;
     if(!askme->no_dir_list) {
       sprintf(adr,"%s/level_%d/complete_dir_list.gz",askme->conf_dir,i);

#ifdef Use_fsnode_read_treE
       ret= Fsnode_read_tree(&(askme->root_node),&(askme->target_root),
                             adr,0,&reader,
                             2|((flag&16)>>2)|8|((!!askme->packet_output)<<6)|
                             picky|(1<<12)|(1<<13));
#else /* Use_fsnode_read_treE */
       ret= Askme_read_tree(askme,adr,0,&reader,
                            2|((flag&16)>>2)|8|(1<<12)|(1<<13));
#endif /* ! Use_fsnode_read_treE */

       if(ret<=0)
         goto ex;
     }
   }
 } else {
   /* read single content list or complete list */

   strcpy(adr,askme->filename);
#ifdef Use_fsnode_read_treE
   ret= Fsnode_read_tree(&(askme->root_node),&(askme->target_root),
                     adr,0,&reader,
                     2|((flag&16)>>2)|8|((!!askme->packet_output)<<6)|picky);
#else /* Use_fsnode_read_treE */
   ret= Askme_read_tree(askme,adr,0,&reader,2|((flag&16)>>2)|8);
#endif /* ! Use_fsnode_read_treE */

   if(ret<=0)
     goto ex;
   /* emulate ASKME level_info */
   ret= Askme_fsm_add_level_info(askme,reader,&info_idx,1);
   if(ret<=0)
     goto ex;
   askme->count_level_info= info_idx;
   Lreader_get_initial_wd(reader,askme->initial_wd,0);

   if(!askme->no_dir_list) {
     /* give a chance to read dir_list */
     strcpy(adr,askme->filename);
     ret=  Sfile_count_components(adr,0);
     if(ret>0) {
       Sfile_component_pointer(adr,&cpt,ret-1,0);
       strcpy(cpt,"complete_dir_list.gz");

#ifdef Use_fsnode_read_treE
       ret= Fsnode_read_tree(&(askme->root_node),&(askme->target_root),
                            adr,0,&reader,
                            4|2|((flag&16)>>2)|8|((!!askme->packet_output)<<6)|
                            picky|(1<<12)|(1<<13));
#else /* Use_fsnode_read_treE */
       ret= Askme_read_tree(askme,adr,0,&reader,
                            4|2|((flag&16)>>2)|8|(1<<12)|(1<<13));
#endif /* ! Use_fsnode_read_treE */

       if(ret<0)
         goto ex;
     }
   }

 }
 ret= 1;
ex:;
 if(fp!=NULL)
   fclose(fp);
 if(ret>0 && !(flag&32) && askme->root_node!=NULL) {
   sprintf(askme->info_text,"sorting  directories of file tree model\n");
   Askme_info(askme,0);
   hret= Fsnode_sort_tree(askme->root_node,0);
   if(hret<=0) {
     ret= hret;
     sprintf(askme->info_text,"         sorting failed\n");
   } else
     sprintf(askme->info_text,"         sorting done\n");
   Askme_info(askme,0);
 }
 if(ret>0 && !(flag&32) && askme->target_root!=NULL) {
   sprintf(askme->info_text,"sorting  directories of file target model\n");
   Askme_info(askme,0);
   hret= Fsnode_sort_tree(askme->target_root,0);
   if(hret<=0) {
     ret= hret;
     sprintf(askme->info_text,"         sorting failed\n");
   } else
     sprintf(askme->info_text,"         sorting done\n");
   Askme_info(askme,0);
 }
 Lreader_destroy(&reader,0);
 l= strlen(askme->initial_wd);
 while(l>0)
   if(askme->initial_wd[l-1]=='/')
     askme->initial_wd[--l]= 0;
   else
 break;
 strcpy(askme->wd,askme->initial_wd);
 return(ret);
}


int Askme_try_old_tree(askme,flag)
struct AskmE *askme;
int flag;
/* return:
 <=0  error
 1    successfully read old tree
 2    must be some other file format
*/
{
 int ret,mode= 0;
 char adr[4096],*cpt;
 FILE *fp;

 sprintf(adr,"%s/highest_level",askme->filename);
 fp= fopen(adr,"r");
 if(fp!=NULL) {
   fclose(fp);
   mode= 2;
   strcpy(askme->conf_dir,askme->filename);
 } else {
   cpt= strrchr(askme->filename,'/');
   if(cpt!=NULL)
     cpt++;
   else
     cpt= askme->filename;
   if(strcmp(cpt,"added_by_scdbackup")==0)
     mode= 1;
 }

 ret= Askme_load_old_tree(askme,askme->filename,"",mode|16);
 if(ret==-4) {
   /* this is neither a content list nor a complete list */
   return(2);
 }
 if(ret<=0)
   return(ret);
 askme->search_mode= 3;
 askme->structured_search= 1;
 askme->levelall= 1;
 askme->result_format= 1|4|8|16;
 return(1); 
}


/* >>> might get extended to an on-the-fly searcher */
int Askme_test_lreader(askme,adr,flag)
struct AskmE *askme;
char *adr;
int flag;
{
 int popened= 0,ret= 0;
 FILE *fpin= NULL;
 struct LreadeR *reader= NULL;
 char *linept,*sourcept,*targetpt;
 off_t linecount= 0;

 popened= A_open_listfile(&fpin,adr,"r",1|((!!askme->packet_output)<<6));
 if(popened!=2)
   popened= 0;
 if(fpin==NULL) {

   /* >>> complain */;

   ret= 0; goto ex;
 }
 ret= Lreader_new(&reader,adr,fpin,0);
 if(ret<=0)
   goto ex;
 while(1) {
   ret= Lreader_read_line(reader,0);
   if(ret<=0) {
     ret= 0; goto ex;
   }
   if(ret==1 || ret==2 || ret==3 || ret==4) {

     /* >>> searchworthy address */;

     Lreader_get_line(reader,&linept,&linecount,0);
     Lreader_get_src_target(reader,&sourcept,&targetpt,0);

     sprintf(askme->result_line,
             "%7.lf   %s = %s\n",(double) linecount,targetpt,sourcept);
     Askme_result(askme,0);

   } else if(ret==17) { /* #=# end */
 break;
   } else if(ret>17) { 
     sprintf(askme->info_text,
             "--- Program error: unknown Lreader return code %d\n",ret);
     Askme_info(askme,0);
   } 
 }

 ret= 1;
ex:;
 if(ret<=0 && reader!=NULL) {
   sprintf(askme->info_text,"\n%s\n",Lreader_strerror(reader,0));
   Askme_info(askme,0);
   Lreader_get_line(reader,&linept,&linecount,0);
   sprintf(askme->info_text,
           "Last input: %.lf :  %s\n", (double) linecount,linept);
   Askme_info(askme,0);
   Lreader_get_src_target(reader,&sourcept,&targetpt,0);
   sprintf(askme->info_text,"Last addr : %s = %s\n",targetpt,sourcept);
   Askme_info(askme,0);
 }
 if(fpin!=NULL){
   if(popened)
     pclose(fpin);
   else
     Sfile_fclose(fpin);
 }
 Lreader_destroy(&reader,0); 
 return(ret);

 return(ret);
}


int Askme_check_volno(askme,line,volno,flag)
struct AskmE *askme;
char *line;
int *volno;
int flag;
/* return: <=0 = error , 1=is list_start , 2= is something else */
{
 char *cpt;

 if(strncmp(line,"cd_liste_",9)!=0) 
   return(2);
 for(cpt=line+9;(*cpt)!='(' && (*cpt)!='_' && (*cpt)!=0;cpt++);
 *cpt= 0;
 *volno= -1;
 sscanf(line+9,"%d",volno);
 if(*volno<=0) {
   sprintf(askme->info_text,
           "Does not look like an ASKME file : %s\n",askme->filename);
   Askme_info(askme,0);
   return(0);
 }
 return(1);
}


int Askme_read_items(askme,flag)
struct AskmE *askme;
int flag;
/*
 bit0= count items , do not store them in **askme->volumes
*/
{
 int ret,list_open= 0,volno,l,info_open= 0;
 FILE *fp= NULL;
 char line[2*4096],*cpt,*apt;

 fp= fopen(askme->filename,"r");
 if(fp==NULL)
   {ret= 0; goto ex;}
 while(Sfile_fgets(line,sizeof(line)-1,fp)!=NULL) {
   if(list_open==0 && info_open==0) {
     ret= Askme_check_volno(askme,line,&volno,0);
     if(ret<=0)
       goto ex;
     if(ret==1) {
       if(volno>askme->num_volumes)
         goto false_format;
       list_open= 1;
     } else if(!(flag&1) && strncmp(line,"  echo '########## part ",24)==0) {
       volno= -1;
       sscanf(line+24,"%d",&volno);
       if(volno<1||volno>askme->num_volumes)
         goto false_format;
       cpt= strchr(line+24,'(');
       if(cpt==NULL)
         goto false_format;
       l= strlen(cpt);
       if(l<2||cpt[l-1]!='\''||cpt[l-2]!=')')
         goto false_format;
       cpt[l-2]= 0;
       askme->volume_info[volno-1]= TSOB_FELD(char,strlen(cpt+1)+1);
       if(askme->volume_info[volno-1]==NULL)
         {ret= -1; goto ex;}
       strcpy(askme->volume_info[volno-1],cpt+1);
     } else if(strncmp(line,"  echo 'cd_backup_planer",24)==0 &&
               strcmp(line+strlen(line)-19,"information script'")==0) {
       info_open= 1;
       goto handle_level_info;
     }
   } else if (list_open==1) {
     if(strcmp(line,"cat  <<+")!=0 && strcmp(line,"cat  <<'+'")!=0) {
false_format:;
       sprintf(askme->info_text,"Does not look like an ASKME file : %s\n",
               askme->filename); 
       Askme_info(askme,0);
       ret= 0; goto ex;
     }
     list_open= 2;
   } else if(list_open==2) {
     if(strcmp(line,"+")==0)
       list_open= 0;
     else {
       if(flag&1) {
         askme->files_in_volume[volno-1]++;
       } else {
         cpt= TSOB_FELD(char,strlen(line)+1);
         if(cpt==NULL)
           {ret= -1; goto ex;}
         strcpy(cpt,line);
         if(askme->count_in_volume[volno-1]>=askme->files_in_volume[volno-1]) 
           goto false_format;
         askme->volumes[volno-1][askme->count_in_volume[volno-1]]= cpt;
         askme->count_in_volume[volno-1]++;
       }
     }
   } else if(info_open) {
     if(line[0]==0)
 continue;
     if(strcmp(line,"fi")==0) {
       info_open= 0;
 continue;
     }
handle_level_info:;
     if(flag&1) {
       askme->num_level_info++;
     } else {
       if(askme->count_level_info>=askme->num_level_info)
         goto false_format;
       cpt= TSOB_FELD(char,strlen(line)+1);
       if(cpt==NULL)
         {ret= -1; goto ex;}
       apt= strchr(line,'\'');
       if(apt==NULL)
 continue;
       apt++;
       strcpy(cpt,apt);
       apt= strrchr(cpt,'\'');
       if(apt!=NULL)
         (*apt)= 0;
       askme->level_info[askme->count_level_info]= cpt;
       askme->count_level_info++;
     }
   }
 }

 ret= 1;
ex:;
 if(fp!=NULL)
   fclose(fp);
 return(ret);
}


int Askme_count_items(askme,flag)
struct AskmE *askme;
int flag;
{
 int ret;

 ret= Askme_read_items(askme,1);
 return(ret);
}


int Askme_read_level(askme,flag)
struct AskmE *askme;
int flag;
{
 int ret,i,j;

 if(askme->volumes!=NULL)
   return(-2);                                   /* programming error caught */
 ret= Askme_count_volumes_and_subs(askme,0);
 if(ret<=0)
   return(ret);
 if(askme->num_volumes<=0)
   return(2);
 askme->files_in_volume= TSOB_FELD(int,askme->num_volumes);
 askme->count_in_volume= TSOB_FELD(int,askme->num_volumes);
 askme->volume_info= TSOB_FELD(char *,askme->num_volumes);
 askme->volumes= TSOB_FELD(char **,askme->num_volumes);
 if(askme->count_in_volume==NULL || askme->files_in_volume==NULL ||
    askme->volume_info==NULL || askme->volumes==NULL)
   return(-1);
 for(i=0;i<askme->num_volumes;i++) {
   askme->files_in_volume[i]= 0;
   askme->count_in_volume[i]= 0;
   askme->volume_info[i]= NULL;
   askme->volumes[i]= NULL;
 }
 ret= Askme_count_items(askme,0);
 if(ret<=0)
   return(ret);
 if(askme->num_level_info>=0) {
   askme->level_info= TSOB_FELD(char *,askme->num_level_info);
   if(askme->level_info==NULL)
     return(-1);
   for(i=0;i<askme->num_level_info;i++) 
     askme->level_info[i]= NULL;
 }
 for(i=0;i<askme->num_volumes;i++) {
   if(askme->files_in_volume[i]<=0)
 continue;
   askme->volumes[i]= TSOB_FELD(char *,askme->files_in_volume[i]);
   if(askme->volumes[i]==NULL)
     return(-1);
   for(j=0;j<askme->files_in_volume[i];j++)
     askme->volumes[i][j]= NULL;
 }  
 ret= Askme_read_items(askme,0);
 if(ret<=0)
   return(ret);
 return(1);
}


int Askme_read_all(askme,flag)
struct AskmE *askme;
int flag;
{
 int ret,i;
 char levelname[4096],*cpt;

 ret= Askme_try_old_tree(askme,0);
 if(ret<=0)
   return(ret);
 if(ret==1)
   return(1);
 ret= Askme_read_level(askme,0);
 if(ret<=0)
   return(ret);
 if(askme->num_sublevels>0) {
   askme->sublevels= TSOB_FELD(struct AskmE *,askme->num_sublevels);
   if(askme->sublevels==NULL)
     return(-1);
   for(i=0;i<askme->num_sublevels;i++)
     askme->sublevels[i]= NULL;
   for(i=0;i<askme->num_sublevels;i++) {
     strcpy(levelname,askme->filename);
     cpt= strrchr(levelname,'/');
     if(cpt!=NULL)
       *(cpt+1)= 0;
     else
       levelname[0]= 0;
     sprintf(levelname+strlen(levelname),"ASKME_level_%d",i);
     ret= Askme_new(&(askme->sublevels[i]),askme->progname,levelname,0);
     if(ret<=0)
       return(ret);
     ret= Askme_read_level(askme->sublevels[i],0);
     if(ret<=0)
       return(ret); 
   }
 }
 return(1);
}


int Askme_is_already_reported_sub(askme,level,volume,flag)
struct AskmE *askme;
int level;
int volume;
int flag;
/*
 bit0= initialize and return
 bit1= set bit
*/
{
 int ret;
 static int prog_error_reported= 0;

 if(askme->search_lv_map==NULL)
   return(0);
 if(flag&1) {
   Bitmap2d_init(askme->search_lv_map,0);
   return(1);
 }
 ret= Bitmap2d_is_set(askme->search_lv_map,volume,level,
                      (prog_error_reported<<1)|((flag&2)>>1));
 if(ret<0)
   prog_error_reported= 1;
 return(ret);
}


int Askme_make_lev_vol_map(askme,flag)
struct AskmE *askme;
int flag;
{
 int ret,node_level,node_volume;
 char *namept;
 struct FsnodE *dirnode,*node;
 struct FstraversaL *trav= NULL;

 Bitmap2d_destroy(&(askme->lev_vol_map),0);
 ret= Bitmap2d_new(&(askme->lev_vol_map),
                   askme->highest_volno+1,askme->highest_level+1,0);
 if(ret<=0)
   goto ex;
 ret= Fstraversal_new(&trav,askme->root_node,(!!askme->show_outdated)<<6);
 if(ret<=0)
   goto ex;
 while(1) {
   ret= Fstraversal_next(trav,NULL,&node,&dirnode,0);
   if(ret<=0)
     goto ex;
   if(node==NULL)
 break;
   if(askme->show_outdated==2 && (!Fstraversal_is_outdated(trav,0)))
 continue;
   ret= Fsnode_get_data(node,&namept,&node_level,&node_volume,0);
   if(ret<=0)
     goto ex;
   if(node_level>=0 && node_level<=askme->highest_level &&
      node_volume>0 && node_volume<=askme->highest_volno)
     Bitmap2d_set(askme->lev_vol_map,node_volume,node_level,0);
 }
 ret= 1;
ex:;
 Fstraversal_destroy(&trav,0);
 return(ret);
}


int Askme_identify_volume(askme,mounted_adr,
                          id,id_date,level,volume,volume_total,
                          backup_date,younger_than,flag)
struct AskmE *askme;
char *mounted_adr;
char id[81],id_date[81];
int *level,*volume,*volume_total;
char backup_date[81],younger_than[81];
int flag;
{
 int ret,got_date= 0,got_yt= 0,got_level= 0,got_volume= 0,got_vtotal= 0;
 int got_id= 0,line_count= 0,report_error_dummy= 0;
 char adr[4096],line[4096],*spt,*ept,*qpt,errmsg[2*4096];
 FILE *fp= NULL;

 id[0]= id_date[0]= 0;
 *level= 0;
 *volume= *volume_total= -1;
 backup_date[0]= younger_than[0]= 0;
 strcpy(adr,mounted_adr);
 ret= Sfile_add_to_path(adr,"ASKME",0);
 if(ret<=0)
   goto ex;
 fp= fopen(adr,"r");
 if(fp==NULL)
   {ret= 0; goto ex;}
 while(Sfile_fgets(line,sizeof(line)-1,fp)!=NULL) {
   line_count++;
   if(!got_date) {
     spt= strstr(line,"echo 'backup date ");
     if(spt!=NULL) {
       ept= strchr(spt+18,'\'');
       if(ept==NULL) {
picky:;
         sprintf(errmsg,"--- improper line format in line %d of file %s\n\n",
                 line_count,adr);
         ret= Askme_handle_picky0(askme,errmsg,&report_error_dummy,0);
         if(ret<=0)
           goto ex;
 continue;
       }
       *ept= 0;
       if(strlen(spt)>80)
         goto picky;
       strcpy(backup_date,spt+18);
       got_date= 1;
 continue;
     }
   }
   if(!got_level) {
     spt= strstr(line,"echo '-level ");
     if(spt!=NULL) {
       if(spt[13]<'0' || spt[13]>'9')
         {ret= 0; goto ex;}
       sscanf(spt+13,"%d",level);
       got_level= 1;
       spt= strstr(line,"-younger_than ");
       if(spt!=NULL && !got_yt) {
         ept= strchr(spt+14,'\'');
         if(ept==NULL)
           {ret= 0; goto ex;}
         *ept= 0;
         if(strlen(spt)>80)
           {ret= 0; goto ex;}
         strcpy(younger_than,spt+14);
         got_yt= 1;
       }
 continue;
     }
   }
   if(!got_volume) {
     spt= strstr(line,"echo 'this is backup part no ");
     if(spt!=NULL) {
       if(spt[29]<'0' || spt[29]>'9')
         {ret= 0; goto ex;}
       sscanf(spt+29,"%d",volume);
       got_volume= 1;
       spt= strstr(spt," of ");
       if(spt!=NULL && !got_vtotal) {
         if(spt[4]<'0' || spt[4]>'9')
           {ret= 0; goto ex;}
         sscanf(spt+4,"%d",volume_total);
         got_vtotal= 1;
       }
 continue;
     }
   }
   if(!got_id) {
     spt= strstr(line,"echo 'id ");
     if(spt!=NULL) {
       if(strlen(line)-9>80)
         goto picky;
       sscanf(spt+9,"%s %s",id,id_date);
       qpt= strchr(id_date,'\'');
       if(qpt!=NULL)
         *qpt= 0;
       if(id[0]==0 || id_date[0]==0)
         goto picky;
       got_id= 0;
     }
   }
   if(got_volume && strcmp(line,"fi")==0)
 break;
 }
 ret= 1;
ex:;
 if(fp!=NULL)
   fclose(fp);
 return(ret);
}


int Askme_execute_prompt_script(askme,prompt_cmd,level,volume,
                                backup_adr,restore_adr,flag)
struct AskmE *askme;
char *prompt_cmd;
int level,volume;
char *backup_adr,*restore_adr;
int flag;
{
 int ret,i,verbous_volume= 0,simulate_action= 0;
 char *exec_argv[4],argv_data[4][4096],errmsg[2*4096],*cmd;

 cmd= prompt_cmd;
 if(cmd[0]=='@')
   cmd++;
 else
   simulate_action= !!(askme->restore_mode&512);
 verbous_volume= ((askme->restore_mode&384)>=256);
 if(verbous_volume) {
   sprintf(askme->info_text,"+%s %s %d %d %s %s\n",
          (simulate_action?" simulate:":""),
          cmd,level,volume,backup_adr,restore_adr);
   Askme_info(askme,!askme->request_to_abort);
 }
 if(simulate_action)
   return(1);

 for(i=0;i<4;i++)
   exec_argv[i]= argv_data[i];
 sprintf(exec_argv[0],"%d",level);
 sprintf(exec_argv[1],"%d",volume);
 strcpy(exec_argv[2],backup_adr);
 strcpy(exec_argv[3],restore_adr);
 ret= Exec_cmd_line(askme->progname,cmd,4,exec_argv,errmsg,
                    !!(askme->packet_output));
 if(ret<0)
   goto ex;
 if(ret>0) {
   sprintf(askme->info_text,"--- failed :  %s %d %d %s %s\n",
           askme->restore_pre_prompt,level,volume,backup_adr,restore_adr);
   Askme_info(askme,0);
   if(errmsg[0]!=0) {
     sprintf(askme->info_text,"--- message:  %s\n",errmsg);
     Askme_info(askme,0);
   }
   {ret= 0; goto ex;}
 }
 ret= 1;
ex:;
 return(ret);
}


int Askme_exec_mount_cmd(askme,cmd,adr,flag)
struct AskmE *askme;
char *cmd;
char *adr;
int flag;
{
 int ret,verbous_volume,simulate_action;
 char *extra_argv[1],errmsg[2*4096];

 simulate_action= !!(askme->restore_mode&512);
 verbous_volume= ((askme->restore_mode&384)>=256);
 if(verbous_volume) {
   sprintf(askme->info_text,"+%s %s %s\n",
          (simulate_action?" simulate:":""),cmd,adr);
   Askme_info(askme,!askme->request_to_abort);
 }
 if(simulate_action)
   return(1);
 extra_argv[0]= adr;
 ret= Exec_cmd_line(askme->progname,cmd,
                    1,extra_argv,errmsg,!!(askme->packet_output));
 if(ret<0)
   return(ret);
 if(ret>0) {
   sprintf(askme->info_text,"--- failed :  %s %s\n",cmd,extra_argv[0]);
   Askme_info(askme,0);
   if(errmsg[0]!=0) {
     sprintf(askme->info_text,"--- message:  %s\n",errmsg);
     Askme_info(askme,0);
   }
   return(0);
 }
 return(1);
}


int Askme_umount_eject(askme,umount_adr,eject_adr,flag)
struct AskmE *askme;
char *umount_adr,*eject_adr;
int flag;
/*
 bit0= force is_mounted
*/
{
 int ret,is_mounted= 0;

 is_mounted= (flag&1) || (askme->mounted_level>=0 && askme->mounted_volume>0);
 if(is_mounted) {
   if((askme->media_mode&4) && umount_adr[0]!=0) {
     /* umount previous volume */
     ret= Askme_exec_mount_cmd(askme,askme->media_umount_cmd,umount_adr,0);
     if(ret<0)
       goto ex;
     askme->mounted_level= -1;
     askme->mounted_volume= -1;
     is_mounted= 0;
   }
 }
 if((askme->media_mode&16) && eject_adr[0]!=0 ){
   ret= Askme_exec_mount_cmd(askme,askme->media_eject_cmd,eject_adr,0);
   if(ret<0)
     goto ex;
 }
 askme->mounted_level= -1;
 askme->mounted_volume= -1;
 ret= 1;
ex:;
 return(ret);
}


int Askme_load_mount(askme,load_adr,mount_adr,flag)
struct AskmE *askme;
char *load_adr,*mount_adr;
int flag;
/* return: <=0 error , 1=ok, nothing was mounted , 2=ok, successfully mounted
*/
{
 int ret;

 if((askme->media_mode&16) && load_adr[0]!=0) {
   ret= Askme_exec_mount_cmd(askme,askme->media_load_cmd,load_adr,0);
   if(ret<0)
     goto ex;
 }
 if((askme->media_mode&4) && mount_adr[0]!=0) {
   ret= Askme_exec_mount_cmd(askme,askme->media_mount_cmd,mount_adr,0);
   if(ret>0)
     ret= 2;
   goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Askme_mount_for_restore(askme,level,volume,backup_adr,restore_adr,flag)
struct AskmE *askme;
int level,volume;
char *backup_adr,*restore_adr;
int flag;
/*
 bit15= umount
*/
/* return: <=0 error , 1=ok , 2=not mounted, abort , 
            5=not mounted, skip volume
*/
{
 int ret,l,i,addon,media_level,media_volume,simulate_action,media_vtotal;
 int verbous_action,simulate_existence,is_mounted= 0,verbous_volume;
 int bold_first_try= 1,report_err_dummy,num_parts,has_level_info= 0;
 char lev_vol_line[161],*ipt,*dpt;
 char media_date[81],media_yt[81],media_id[81],media_id_date[81];
 char level_date[81],level_younger_than[81];

 if(askme->request_to_abort && !(flag&(1<<15)))
   return(2);
 simulate_action= !!(askme->restore_mode&512);
 simulate_existence= !!(askme->restore_mode&32);
 verbous_action= ((askme->restore_mode&384)==384);
 verbous_volume= ((askme->restore_mode&384)>=256);
 is_mounted= (askme->mounted_level>=0 && askme->mounted_volume>0);
 if(!(is_mounted && 
    (askme->mounted_level!=level || askme->mounted_volume!=volume)))
   if(!(flag&(1<<15))) if(!simulate_existence)
     goto check_level_volume;

umount_eject:;
 bold_first_try= 0;
 ret= Askme_umount_eject(askme,askme->restore_from,askme->restore_from,
                         !!is_mounted);
 if(ret<0)
   goto ex;
 is_mounted= 0;
 if(flag&(1<<15))
   {ret= 1; goto ex;}

 if(askme->restore_pre_prompt[0]!=0) {
   ret= Askme_execute_prompt_script(askme,askme->restore_pre_prompt,
                                    level,volume,backup_adr,restore_adr,0);
   if(ret<=0)
     goto ex;
 }
 /* ??? eventually disable local prompting */;

ask_for_volume:;
 /* ask for level,volume */
 strcpy(askme->info_text,"\n");
 Askme_de_eN(
   sprintf(lev_vol_line,
           " Bitte einlegen%s:   LEVEL  %d   BACKUPSTUECK  %d   ",
           (askme->media_mode&4)?"":" und mounten",level,volume);
 ,  
   sprintf(lev_vol_line,
           " Please insert%s:   LEVEL  %d   VOLUME  %d   ",
           (askme->media_mode&4)?"":" and mount",level,volume);
 )
 /* add flexible length ### and \n */
 l= strlen(lev_vol_line);
 addon= (79-1-l)/2;
 if(addon>0) {
   for(i=l;i>=0;i--)
     lev_vol_line[i+addon]= lev_vol_line[i];
   for(i=0;i<addon;i++)
     lev_vol_line[i]= lev_vol_line[l+addon+i]= '#';
 }
 lev_vol_line[l+2*addon]= '\n';
 lev_vol_line[l+2*addon+1]= 0;
 l+= 2*addon+1;
 ipt= askme->info_text;
 strcpy(ipt,lev_vol_line);
 for(i=0;i<l-1;i++)
   ipt[i]= '#';
 ipt= ipt+strlen(ipt);
 strcpy(ipt,lev_vol_line);
 ipt= ipt+strlen(ipt);
 strcpy(ipt,lev_vol_line);
 for(i=0;i<l-1;i++)
   ipt[i]= '#';
 ipt= ipt+strlen(ipt);
 Askme_de_eN(
   strcpy(ipt,
"........... [ Wenn fertig, Enter-Taste druecken. @@@ = Abbruch ] ............\n");
 ,
   strcpy(ipt,
"................... [ Press enter when done. @@@ = Abort ] ..................\n");
 )
 Askme_info(askme,0);

 ret= Askme_request_confirmation(askme,1);
 if(ret<=0 || ret==2)
   goto ex;
 if(ret==3) 
   goto ask_for_volume;
 if(askme->request_to_skip_volume>0)
   {ret= 5; goto ex;}

 Askme_de_eN(
   sprintf(askme->info_text,"Verarbeitung des Backupstuecks beginnt ...\n");
 ,
   sprintf(askme->info_text,"Processing of volume started ...\n");
 )
 Askme_info(askme,1);
 if(askme->restore_post_prompt[0]!=0) {
   ret= Askme_execute_prompt_script(askme,askme->restore_post_prompt,
                                    level,volume,backup_adr,restore_adr,0);
   if(ret<=0)
     goto ex;
 }
 ret= Askme_load_mount(askme,askme->restore_from,askme->restore_from,0);
 if(ret<0)
   goto ex;
 if(ret==0)
   goto umount_eject;
 if(ret==2)
   is_mounted= 1;

check_level_volume:;
 if(verbous_action) {
   sprintf(askme->info_text,"+%s check for level and volume\n",
           (simulate_existence?" simulate:":""));
   Askme_info(askme,3);
 }
 has_level_info= Askme_get_level_info(askme,level,
                                   level_date,level_younger_than,&num_parts,0);
 if(simulate_existence) {
   ret= 1;
   media_level= level;
   media_volume= volume;
 } else 
   ret= Askme_identify_volume(askme,askme->restore_from,
                              media_id,media_id_date,
                              &media_level,&media_volume,&media_vtotal,
                              media_date,media_yt,0);
 if(ret<0)
   goto ex;
 if(ret==0) {
   if(bold_first_try)
     goto umount_eject;
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- KEIN IDENTIFIZIERBARES BACKUPSTUECK IN: %s\n",
             askme->restore_from);
   ,
     sprintf(askme->info_text,
             "--- CANNOT IDENTIFY A VALID BACKUP VOLUME IN: %s\n",
             askme->restore_from);
   )
   Askme_info(askme,0);
   goto umount_eject;
 }
 if(has_level_info && media_id_date[0]!=0 && 
    strcmp(media_id_date,level_date)!=0) {
   if(bold_first_try)
     goto umount_eject;
   Askme_de_eN(
     sprintf(askme->info_text,
     "--- FALSCHES BACKUPDATUM !\n    (Level= %d, Stueck= %d, Datum= %s)\n\n",
        media_level,media_volume,media_id_date);
   ,
     sprintf(askme->info_text,
     "--- WRONG BACKUP DATE !\n    (level= %d, volume= %d, date= %s)\n\n",
             media_level,media_volume,media_id_date);
   )
   goto wrong_backup;
 }
 if(askme->backup_id[0]!=0 && media_level>=askme->backup_id_level &&
    strcmp(askme->backup_id,media_id)!=0) {
   if(bold_first_try)
     goto umount_eject;
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- FALSCHES BACKUP ID !\n    ist : '%s'\n    soll: '%s'\n",
             media_id,askme->backup_id);
   ,
     sprintf(askme->info_text,
           "--- WRONG BACKUP ID !\n    read    : '%s'\n    expected: '%s'\n",
             media_id,askme->backup_id);
   )
wrong_backup:;
   if(has_level_info>0) {
     Askme_de_eN(
       sprintf(askme->info_text+strlen(askme->info_text),
               "    soll: Datum= %s   ,   -younger_than %s\n\n",
               level_date,level_younger_than);
     ,
       sprintf(askme->info_text+strlen(askme->info_text),
               "    expected: date= %s   ,   -younger_than %s\n\n",
               level_date,level_younger_than);
     )
   } else
     strcat(askme->info_text,"\n");
   ret= Askme_handle_picky0(askme,askme->info_text,&report_err_dummy,0);
   if(ret<0)
     goto ex;
   if(ret==0)
     goto umount_eject;
 }
 if(media_volume!=volume || media_level!=level) {
   if(bold_first_try)
     goto umount_eject;
   if(media_id_date[0]==0)
     dpt= media_date;
   else
     dpt= media_id_date;
   Askme_de_eN(
     sprintf(askme->info_text,
     "--- FALSCHES BACKUPSTUECK !\n    (Level= %d, Stueck= %d, Datum= %s)\n\n",
        media_level,media_volume,dpt);
   ,
     sprintf(askme->info_text,
             "--- WRONG VOLUME !\n    (level= %d, volume= %d, date= %s)\n\n",
             media_level,media_volume,dpt);
   )
   Askme_info(askme,0);
   goto umount_eject;
 }
 askme->mounted_level= level;
 askme->mounted_volume= volume;
 ret= 1;
ex:;
 return(ret);
}


int Askme_make_restore_adr(askme,source,target,backup_adr,restore_adr,flag)
struct AskmE *askme;
char *source;
char *target;
char backup_adr[4096];
char restore_adr[4096];
int flag;
/*
 bit0= do not allow restore source which does not match askme->restore_skip
*/
{
 int l,ret,trailing_slash= 0;
 char *rtgt;

 backup_adr[0]= restore_adr[0]= 0;
 if(askme->restore_mode&2)
   rtgt= target;
 else
   rtgt= source;
 if(askme->restore_skip[0]!=0) {
   l= strlen(askme->restore_skip);
   if(askme->restore_skip[l-1]=='/')
     trailing_slash= 1;
   if(strncmp(rtgt,askme->restore_skip,l)==0
      && (trailing_slash || rtgt[l]=='/' || rtgt[l]==0))
     rtgt+= strlen(askme->restore_skip);
   else if(flag&1) {
     strcpy(restore_adr,rtgt);
     Askme_de_eN(
       sprintf(askme->info_text,"--- passt nicht zu -restore_skip: %s\n",
               restore_adr);
     ,
       sprintf(askme->info_text,"--- does not match -restore_skip: %s\n",
               restore_adr);
     )
     Askme_info(askme,1);
     return(0);
   }
 }
 strcpy(restore_adr,askme->restore_to);
 ret= Sfile_add_to_path(restore_adr,rtgt,0);
 if(ret<=0)
   return(ret);
 strcpy(backup_adr,askme->restore_from);
 ret= Sfile_add_to_path(backup_adr,target,0);
 if(ret<=0)
   return(ret);
 return(1);
}


int Askme_prepare_path(askme,trav,backup_adr,restore_adr,
                       permission_list,change_list,plcount,flag)
struct AskmE *askme;
struct FstraversaL *trav;
char *backup_adr;
char *restore_adr;
int **permission_list;
int **change_list;  /* bit0=finally restore perms , bit1=newly created */
int *plcount;
int flag;
/*
 bit0= reset altered permissions and free lists
 bit1= with bit0: remove all newly created directories
 bit2= do not use pager with Askme_info
 bit3= do not issue messages about removed directories
*/
{
 int ret,count,relationship,perms,i,do_change,verbous_action,simulate_action;
 int backup_perms,backup_perms_valid= 0,backup_count,min_ret,is_new_dir;
 mode_t st_mode,backup_st_mode;
 char path[4096],*cpt,*spt,*dpt,dpt_recorded;
 char backup_path[4096],*backup_pt,*backup_dpt;

 struct FsnodE *upnode;

 simulate_action= !!(askme->restore_mode&64);
 verbous_action= ((askme->restore_mode&384)==384);

 if(flag&1)
   goto revoke_alterations;

 *plcount= 0;
 
 strcpy(path,restore_adr);
 count= Sfile_count_components(path,0);
 if(count<=1)
   return(2);

 *plcount= count-1;
 *permission_list= TSOB_FELD(int,*plcount);
 *change_list= TSOB_FELD(int,*plcount);
 if((*permission_list)==NULL || (*change_list)==NULL) {
   *plcount= 0;
   return(-1);
 }
 for(i=0;i<*plcount;i++)
   (*permission_list)[i]= (*change_list)[i]=0;

 spt= path;
 for(i=0;i<*plcount;i++) {
   ret= Sfile_component_pointer(spt,&cpt,1,0);
   if(ret<=0 || cpt<=spt)
 break; /* >>> this would be actually quite a program error */
   if(cpt-1==path)
     dpt= cpt;
   else
     dpt= cpt-1;
   dpt_recorded= *dpt;
   (*dpt)= 0;
   ret= Sfile_lookup_permissions(path,&perms,&st_mode,0);
   if(ret<0)
     goto ex;
   relationship= ret; /* 1= owner  2=group member  3=other */

   /* make way for creating a new leaf */;
   is_new_dir= 0;
   if(ret==0) { /* path component does not exist yet */
     is_new_dir= 1;
     backup_perms_valid= 0;
     if(trav!=NULL) {
       /* try to obtain directory permissions from tree model */
       ret= Fstraversal_get_upnode(trav,&upnode,*plcount-i,0);
       if(ret<0)
         goto ex;
       if(ret>0 && upnode!=NULL) {
         ret= Fsnode_get_permissions(upnode,&backup_perms,0);
         if(ret<0)
           goto ex;
         if(ret>0)
           backup_perms_valid= 1;
       }
     }
     if(backup_perms_valid)
       st_mode= backup_perms;
     else {
       /* try to obtain directory permissions from backup media */
       /* determine adr by stepping up backup_adr :
          clip off *plcount-i components */;
       strcpy(backup_path,backup_adr);
       backup_count= Sfile_count_components(backup_path,0);
       if(backup_count>(*plcount-i)) {
         ret= Sfile_component_pointer(backup_path,&backup_pt,
                                      backup_count-(*plcount-i)+1,0);
         if(backup_pt-1==backup_path)
           backup_dpt= backup_pt;
         else
           backup_dpt= backup_pt-1;
         (*backup_dpt)= 0;
         if(verbous_action) {
           sprintf(askme->info_text,"+%s look up permissions of dir %s\n",
                            (simulate_action?" simulate:":""),backup_path);
           Askme_info(askme,3*!(flag&4));
         }
         if(!(askme->restore_mode&32)) {
           ret= Sfile_lookup_permissions(backup_path,&backup_perms,
                                          &backup_st_mode,0);
           if(ret<0)
             goto ex;
           if(ret>0)
             st_mode= backup_st_mode&Fsinfo_mode_masK;
           backup_perms_valid= 2; 
         }
       }
     }

     if(askme->restore_mode&(1<<10))
       st_mode|= S_IXUSR|S_IWUSR|S_IRUSR;
     if(verbous_action) {
       sprintf(askme->info_text,"+%s creation of missing directory %s\n",
                            (simulate_action?" simulate:":""),path);
       Askme_info(askme,3*!(flag&4));
     }
     if(! simulate_action) { /* create directory */
       ret= mkdir(path,S_IXUSR|S_IWUSR|S_IRUSR);
       if(ret==-1) {
         Askme_de_eN(
           sprintf(askme->info_text,
                   "--- Kann Directory nicht erzeugen: %s\n   %s\n",
                   path,(errno>0?strerror(errno):"-unbekannter fehler-"));
         ,
           sprintf(askme->info_text,
                   "--- Cannot create directory: %s\n   %s\n",
                   path,(errno>0?strerror(errno):"-unknown error-"));
         )
         Askme_info(askme,!(flag&4));
         {ret= 0; goto ex;}
       }
       (*change_list)[i]|= 2;
       ret= chmod(path,S_IXUSR|S_IWUSR|S_IRUSR); /* do not rely on umask */
       if(ret==-1)
         goto cannot_set_perms;
     }

     if(backup_perms_valid) {
       if(verbous_action) {
         sprintf(askme->info_text,"+%s original  chmod %4.4o\n",
                 (simulate_action?" simulate:":""),st_mode&Fsinfo_mode_masK);
         Askme_info(askme,3*!(flag&4));
       }
       if(!simulate_action) { /* set directory permissions */
         ret= chmod(path,st_mode&Fsinfo_mode_masK);
         if(ret==-1) {
cannot_set_perms:;
           Askme_de_eN(
             sprintf(askme->info_text,
                  "--- Kann Directoryrechte nicht setzen: %4.4o  %s\n    %s\n",
                   st_mode&Fsinfo_mode_masK,
                   path,(errno>0?strerror(errno):"-unbekannter fehler-"));
           ,
             sprintf(askme->info_text,
                   "--- Cannot set directory permissions: %4.4o  %s\n    %s\n",
                   st_mode&Fsinfo_mode_masK,
                   path,(errno>0?strerror(errno):"-unknown error-"));
           )
           Askme_info(askme,!(flag&4));
           {ret= 0; goto ex;}
         }
       }
     }

     if(simulate_action) {
       relationship= 1;
       if(!backup_perms_valid) {
                                             /* deliberately not good enough */
         st_mode= S_IWUSR|S_IRUSR|S_IXGRP|S_IRGRP; 
                                                /* unless open_dir_for_owner */
         if(askme->restore_mode&(1<<10))
           st_mode|= S_ISVTX|S_IWUSR|S_IRUSR;
       }
       perms= 0;
       if(st_mode&S_IXUSR)
         perms|= 1;
       if(st_mode&S_IWUSR)
         perms|= 2;
       if(st_mode&S_IRUSR)
         perms|= 4;
     } else {
       ret= Sfile_lookup_permissions(path,&perms,&st_mode,0);
       if(ret<0)
         goto ex;
       relationship= ret; /* 1= owner  2=group member  3=other */
     }
   }
   (*permission_list)[i]= (st_mode&Fsinfo_mode_masK);
   do_change= 0;

   if(0 && i<*plcount-1) { /* <<< preliminary defunct */

     /* >>> branch needs rx permissions (is x alone enough ?) 
       if next directory exists. But needs rwx (is wx enough ?) else.  */
     /* >>> needed is a preview wether component i+1 exists */

     if((perms&5)!=5) {
       if(relationship==1 || getuid()==0) {
         st_mode|= S_IXUSR|S_IRUSR;
         do_change= 1;
       }
     }
   } else {           /* leafdir needs rwx (is w alone enough ?) */
     if((perms&7)!=7) {
       if(relationship==1 || getuid()==0) {
         st_mode|= S_IXUSR|S_IWUSR|S_IRUSR;
         do_change= 1;
       }
     }
   }
   if(st_mode&S_ISVTX) {
     st_mode&= ~S_ISVTX;
     do_change= 1;
   }
   if((askme->restore_mode&(1<<23)) && !is_new_dir)
     do_change= 0;
   if(do_change) {
     if(verbous_action) {
       sprintf(askme->info_text,"+%s temporary  chmod %4.4o %s\n",
               (simulate_action?" simulate:":""),st_mode&Fsinfo_mode_masK,
               path);
       Askme_info(askme,3*!(flag&4));
     }
     if(!simulate_action) { /* try set temporary permissions */
       ret= chmod(path,st_mode&Fsinfo_mode_masK);
     } else
       ret= 0;
     if(ret==-1) {
       if(verbous_action) {
         sprintf(askme->info_text,
                 "+%s temporary chmod failed (hoping for the best)\n    %s\n",
                 (simulate_action?" simulate:":""),
                 (errno>0?strerror(errno):"-unbekannter fehler-"));
         Askme_info(askme,!(flag&4));
       }
     } else {
       (*change_list)[i]|= 1;
     }
   }
   *dpt= dpt_recorded;
   spt= cpt;
 }
 ret= 1;
ex:;
 return(ret);
revoke_alterations:;
 min_ret= 1;
 strcpy(path,restore_adr);
 spt= path;
 for(i= *plcount-1;i>=0;i--) {
   if(!((*change_list)[i]&(1|2)))
 continue;
   ret= Sfile_component_pointer(spt,&cpt,i+1,0);
   if(ret<=0 || cpt<=spt) {
     sprintf(askme->info_text,
  "--- program error: Askme_prepare_path: number of path components shrunk\n");
     Askme_info(askme,0);
     {ret= -1; goto ex;}
   }
   if(cpt-1==path)
     dpt= cpt;
   else
     dpt= cpt-1;
   dpt_recorded= *dpt;
   (*dpt)= 0;
   if((flag&2) && ((*change_list)[i]&2)) {
     ret= Askme_rm_dirty_dir(askme,path,(flag&(4|8)));
     if(ret<0)
       goto ex;
     if(ret<min_ret)
       min_ret= ret;
     *dpt= dpt_recorded;
 continue;
   }
   if(!((*change_list)[i]&1)) {
     *dpt= dpt_recorded;
 continue;
   }
   st_mode= (*permission_list)[i];
   if(verbous_action) {
     sprintf(askme->info_text,"+%s restoring  chmod %4.4o %s\n",
             (simulate_action?" simulate:":""),st_mode&Fsinfo_mode_masK,path);
     Askme_info(askme,3*!(flag&4));
   }
   if(!simulate_action) { /* set final permissions */
     ret= chmod(path,st_mode&Fsinfo_mode_masK);
     if(ret==-1) 
       goto cannot_set_perms;
   }
   *dpt= dpt_recorded;
 }
 ret= min_ret;
 if(*change_list!=NULL)
   free((char *) *change_list);
 *change_list= NULL;
 if(*permission_list!=NULL)
   free((char *) *permission_list);
 *permission_list= NULL;
 *plcount= 0;
 goto ex;
}


int Askme_asylum_adr(askme,restore_adr,asylum_adr,flag)
struct AskmE *askme;
char *restore_adr;
char asylum_adr[4096];
int flag;
{
 int ret;

 if(askme->restore_asylum[0]==0) {
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- Kein -restore_asylum voreingestellt im -restore_mode:+:overwrite_by_asylum\n");
   ,
     sprintf(askme->info_text,
             "--- No -restore_asylum preset while in -restore_mode:+:overwrite_by_asylum\n");
   )
   Askme_info(askme,1);
   {ret= 0; goto ex;}
 }
 if(strlen(restore_adr)+strlen(askme->restore_asylum)+1>=4095) {
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- Asyladresse wird laenger als 4095 Zeichen.\n");
   ,
     sprintf(askme->info_text,
             "--- Asylum address becomes longer than 4095 characters.\n");
   )
   Askme_info(askme,1);
   {ret= 0; goto ex;}
 }
 strcpy(asylum_adr,askme->restore_asylum);
 ret= Sfile_add_to_path(asylum_adr,restore_adr,0);
 if(ret<=0) {
   Askme_de_eN(
     sprintf(askme->info_text,"--- Kann Asyladresse nicht bilden.\n");
   ,
     sprintf(askme->info_text,"--- Asylum address cannot be determined.\n");
   )
   Askme_info(askme,1);
   {ret= 0; goto ex;}
 }
 ret= 1;
ex:;
 return(ret);
}


int Askme_announce_copy(askme,backup_adr,restore_adr,flag)
struct AskmE *askme;
char *backup_adr;
char *restore_adr;
int flag;
/*
 bit0= do not announce (eventually to save an if() at the caller)
 bit1= simulation mode
 bit2= subordinate copy or move
*/
/* return:  1=did announce  2=abort copy operation  3=nothing done */
{
 if(!(flag&1)) {
   sprintf(askme->result_line,"%s%s %s  ->  %s\n",
           ((flag&4)?"++":"+"),
           ((flag&2)?" simulate:":""),backup_adr,restore_adr);
   Askme_result(askme,0);
   if(askme->request_to_abort || askme->request_to_skip_volume>0)
     return(2); 
   return(1);
 }
 return(3);
}


int Askme_restore_copy(askme,trav,level,volume,backup_adr,restore_adr,flag)
struct AskmE *askme;
struct FstraversaL *trav;
int level,volume;
char *backup_adr;
char *restore_adr;
int flag;
/*
 bit0= recursion
 bit1= do not go into recursion
 bit2= according to the model, this is a directory
 bit3= silently ignore if backup_adr is missing
 bit4= this is a move-to-asylum
 bit5= this is a second try
*/
/* return: <=0 error , 1=ok, did restore fine, 2=restored nothing valuable */
{
 int ret,do_prepare= 0,skip_properties= 0,second_try= 0;
 int rec_ret= 0,return_rec_ret= 0;
 struct Askme_rescP *ctl= NULL,*ctl_boss;

 ctl_boss= NULL;
 if(flag&16)
   ctl_boss= askme->restore_ctl_stack;
 ret= Askme_rescp_new(&ctl,ctl_boss,askme,trav,level,volume,
                      backup_adr,restore_adr,0);
 if(ret<=0)
   return(-1); /* must not jump to ex with ctl==NULL */

retry_restore:;
 ret= Askme_rescp_setup(ctl,(flag&63)|second_try);
 if(ret<=0 || ret==2)
   goto ex;

prepare_path_and_try:;
 ret= Askme_rescp_prepare(ctl,(flag&15)|second_try|do_prepare);
 if(ret<=0 || ret==2)        /* error or request to abort */
   goto ex;
 if(ret==3)                  /* still looking for resume start */
   goto handle_subitems;
 if(ret==4)                  /* shadowed */
   {ret= 2; goto ex;}
 if(ret==5) {                /* protected directory during resume search */
   skip_properties= 1;
   goto handle_subitems;
 }
 ret= Askme_rescp_overwrite(ctl,flag&15);
 if(ret<=0 || ret==2)
   goto ex;
 if(ret==3) {
   do_prepare= 1<<16;
   second_try= 32;
   goto prepare_path_and_try;
 }

 ret= Askme_rescp_create(ctl,flag&15);
 if(ret<=0 || ret==2)
   goto ex;
 if(ret==3) {
   do_prepare= 1<<16;
   goto prepare_path_and_try;
 }

 ret= Askme_rescp_content(ctl,flag&15);
 if(ret<=0 || ret==2)
   goto ex;

handle_subitems:;
 ret= Askme_rescp_recursion(ctl,flag&15);
 if(ret<=0 || ret==2) {
   if(ret<=0 && !(ctl->dir_has_valid_file || (askme->restore_mode&(1<<22))))
     goto ex;
   /* delay exit until Askme_rescp_properties() is done */
   rec_ret= ret;
   return_rec_ret= 1;
 }

 if(askme->resume_at_source[0]==0 && !skip_properties) {
   ret= Askme_rescp_properties(ctl,(flag&15)|((!!return_rec_ret)<<18));
   if(ret<=0 || ret==2)
     goto ex;
   if(!(ctl->move_to_asylum || return_rec_ret)) {
     askme->restore_count++;
     askme->restore_bytes+= ctl->restored_bytes;
   }
 }
 ctl->is_surely_clean= 1; /* sub file might be dirty, though */
 ret= 1;
ex:;
 if(return_rec_ret && (ret>0 || rec_ret<ret))
   ret= rec_ret;
 if(ctl->write_fd>=0)
   close(ctl->write_fd);
 if(ret==0 && !ctl->error_is_handled) {
                                   /* react according to restore error mode */
   ret= Askme_error_on_restore(askme,ctl,0);
   if(ret==2)
     goto retry_restore;
   if(ret==1)
     ret= 2;
 }
 if(ctl->plcount>0) {
   /* revoke eventual temporary permission changes */
   Askme_rescp_prepare_path(ctl,1);
 }
 if(!((flag&1)||(flag&16)))
   if((ret>0 && !(ret!=1 && ctl->original_altered)) || ctl->is_surely_clean)
     if(askme->resume_at_source[0]==0)
       if(askme->resume_last_dirty[0]==0 ||
          strcmp(askme->resume_last_dirty,restore_adr)==0)
         Askme_record_resume_dirty(askme,"",0);
 Askme_rescp_destroy(&ctl,0);
 return(ret);
}


int Askme_restore(askme,trav,level,volume,source,target,flag)
struct AskmE *askme;
struct FstraversaL *trav;
int level,volume;
char *source;
char *target;
int flag;
{
 int ret,cstate,file_type= 0,eventual_error_is_handled= 0;
 char backup_adr[4096],restore_adr[4096];
 struct FsnodE *node= NULL,*dirnode= NULL;

 if((askme->abort_handler&15)==1)
   Cleanup_set_handlers(askme,(Cleanup_app_handler_T) Askme_abort_handler,0);
 if((askme->abort_handler&15)==2)
   Cleanup_set_handlers(NULL,NULL,2); /* ignore all signals */
 
 if(askme->resume_at_source[0]==0)
   Askme_record_resume_source(askme,source,0);

retry_restore:;
 if(askme->request_to_abort || askme->request_to_skip_volume>0)
   return(2);

 if(trav!=NULL) {
   Fstraversal_get_pos(trav,&node,&dirnode,&cstate,0);
   if(node!=NULL)
     Fsnode_get_type(node,&file_type,0);
 }
 if(level<0 || volume<=0) {
   sprintf(askme->info_text,
        "--- program error: level or volume undefined : level=%d, volume=%d\n",
           level,volume);
   Askme_info(askme,1);
   {ret= -1; goto ex;}
 }

 ret= Askme_make_restore_adr(askme,source,target,backup_adr,restore_adr,1);
 if(ret<=0)
   goto ex;
 sprintf(askme->info_text,"restore:  %s  ->  %s\n",backup_adr,restore_adr);
 Askme_info(askme,3);
 if(askme->request_to_abort || askme->request_to_skip_volume>0)
   {ret= 2; goto ex;}

 if(askme->mounted_level!=level || askme->mounted_volume!=volume) {
   ret= Askme_mount_for_restore(askme,level,volume,backup_adr,restore_adr,0);
   if(ret<0) 
     goto ex;
   if(ret==0) {
     askme->request_to_abort= 1;
     goto ex;
   }
   if(ret==2) {
     askme->request_to_abort= 1;
     {ret= 2; goto ex;}
   }
   if(askme->request_to_skip_volume>0)
     {ret= 2; goto ex;}
 }

 eventual_error_is_handled= 1;
 ret= Askme_restore_copy(askme,trav,level,volume,backup_adr,restore_adr,
                         ((!!askme->suball)<<1)|((file_type==2)<<2));
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 if(ret==0 && !eventual_error_is_handled) { /* react according error mode */
   ret= Askme_error_on_restore(askme,NULL,0);
   if(ret==2)
     goto retry_restore;
 }
 if((askme->abort_handler&15)!=0)
   Cleanup_set_handlers(NULL,NULL,1); /* back to default signal handling */
 Askme_rescp_destroy(&(askme->restore_ctl_stack),1);   
 return(ret);
}


int Askme_found(askme,trav,level,volume,source,target,expr,mtime,ctime,flag)
struct AskmE *askme;
struct FstraversaL *trav;
int level,volume;
char *source;
char *target;
char *expr;
time_t mtime,ctime;
int flag;
/*
 bit0= do not report level headline
 bit1= append / to target
 bit2= if enabled by askme->result_format : append / to target and source
*/
{
 int show_target,add_slash,add_target_slash,ret;
 char *cpt;

 if(askme->request_to_skip_volume>0)
   return(2);

 if((!(askme->restore_mode&1)) && (askme->volumewise&3)==3)
   if(askme->resume_at_source[0]==0)
     Askme_record_resume_source(askme,source,0);

 add_slash= (askme->result_format&(1<<20)) && (flag&4);
 add_target_slash= (flag&2) || add_slash;
 if(target[0]=='/' && target[1]==0)
   add_target_slash= 0;
 if(source[0]=='/' && source[1]==0)
   add_slash= 0;
 cpt= askme->result_line;
 show_target= (askme->result_format&8);
 if(askme->result_format&1) {
   if(level>=0 && !(flag&1)) {
     sprintf(cpt,"+++ on backup level %d\n",level);
     Askme_result(askme,0);
     cpt= askme->result_line;
   }
 } 
 if(askme->result_format&2) {
   sprintf(cpt,"%d",level);
   cpt+= strlen(cpt);
 }
 if(askme->result_format&4) {
   if(askme->result_format&2) {
     sprintf(cpt,",");
     cpt++;
   }
   sprintf(cpt,"%d",volume+1);
   cpt+= strlen(cpt);
 }

#ifdef Askme_fsnode_with_timeS
 if(askme->result_format&64) {
   if(cpt!=askme->result_line) {
     sprintf(cpt," ");
     cpt++; 
   }
   sprintf(cpt,"%.f",(double) mtime);
   cpt+= strlen(cpt);
 }
 if(askme->result_format&128) {
   if(cpt!=askme->result_line) {
     sprintf(cpt," ");
     cpt++; 
   }
   sprintf(cpt,"%.f",(double) ctime);
   cpt+= strlen(cpt);
 }
#endif /* Askme_fsnode_with_timeS */

 if(show_target) {
   if(cpt!=askme->result_line) {
     sprintf(cpt,"   ");
     cpt+= 3;
   }
   if(askme->result_format&(1<<19)) {
     char esc_target[2*4096];
     strcpy(esc_target,target);
     Fileliste__escape_target(esc_target,0);
     if(add_target_slash)
       sprintf(cpt,"%s/",esc_target);  
     else
       sprintf(cpt,"%s",esc_target);  
   } else {
     if(add_target_slash)
       sprintf(cpt,"%s/",target);  
     else
       sprintf(cpt,"%s",target);  
   }
   cpt+= strlen(cpt);
 }
 if(askme->result_format&16) {
   if(show_target) {
     sprintf(cpt," = ");
     cpt+= 3;
   } else if(cpt!=askme->result_line) {
     sprintf(cpt,"   ");
     cpt+= 3;
   }
   if(add_slash)
     sprintf(cpt,"%s/",source);  
   else
     sprintf(cpt,"%s",source);  
   cpt+= strlen(cpt);
 }
 if(askme->result_format&32) {
   if(cpt!=askme->result_line) {
     sprintf(cpt,"   ");
     cpt+= 3;
   }
   sprintf(cpt,"%s",expr);  
   cpt+= strlen(cpt);
 }
 sprintf(cpt,"\n");
 Askme_result(askme,0);
 askme->found++;
 if(askme->restore_mode&1) {
   ret= Askme_restore(askme,trav,level,volume+1,source,target,0);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Askme_prepare_regex(askme,adr,flag)
struct AskmE *askme;
char *adr;
int flag;
/*
 bit0= do not augment relative structured search by askme->wd
*/
{
 int l,ret,i,count,bonked= 0,is_constant;
 char *cpt,*npt,adr_part[2*4096],absolute_adr[2*4096],*adr_start;

#ifndef Askme_with_regeX

 if(askme->search_mode==2 || 
    (askme->search_mode==3 && askme->structured_search==0)) {
no_regex_available:;
   Askme_de_eN(
     sprintf(askme->info_text,
             "%s : Regulaere Ausdruecke sind nicht verfuegbar\n",
             askme->progname);
   ,
     sprintf(askme->info_text,"%s : regular expressions not implemented\n",
                    askme->progname);
   )
   Askme_info(askme,0);
   return(0);
 }
   
#endif /* ! Askme_with_regeX */

 if(askme->search_mode>=2 && askme->search_mode<=4) {
   if(askme->search_mode==3 || askme->search_mode==4) {
     l= strlen(adr)+strlen(askme->wd)+1;
     if(l*2+2>sizeof(askme->reg_expr) || l*2+2>sizeof(adr_part)) {
       sprintf(askme->info_text,"%s : search text too long\n",askme->progname);
       Askme_info(askme,0);
       return(0);
     }
   }
   Askme_destroy_re(askme,0);
   if(askme->structured_search && askme->search_mode==3) {
     if(adr[0]!='/' && !(flag&1)) {
       /* relative expression : prepend working directory */
       sprintf(absolute_adr,"%s/%s",askme->wd,adr);
       adr_start= absolute_adr;
       askme->prepended_wd= 1;
     } else
       adr_start= adr;
     /* count slashes */;
     cpt= adr_start;
     while(*cpt=='/')
       cpt++;
     for(i= 0;1;i++) {
       cpt= strchr(cpt,'/');
       if(cpt==NULL)
     break;
       while(*cpt=='/')
         cpt++;
     }
     count= i+1;

#ifdef Askme_with_regeX
     askme->re= TSOB_FELD(regex_t,count);
     if(askme->re==NULL)
       return(-1);
#endif /* Askme_with_regeX */

     askme->re_constants= TSOB_FELD(char *,count);
     if(askme->re_constants==NULL)
       return(-1);
     for(i= 0;i<count;i++)
       askme->re_constants[i]= NULL;
     askme->re_count= count;
     askme->re_fill= 0;
       
     /* loop over slash chunks*/;
     cpt= adr_start;
     askme->re_fill= 0;
     while(*cpt=='/')
       cpt++;
     for(i= 0;i<count;i++) {
       npt= strchr(cpt,'/');
       if(npt==NULL) {
         strcpy(adr_part,cpt);
       } else {
         strncpy(adr_part,cpt,npt-cpt);
         adr_part[npt-cpt]= 0;
       }

       if(adr_part[0]==0)
         goto next_adr_part;
       if(adr_part[0]=='.' && adr_part[1]==0)
         goto next_adr_part;
       if(adr_part[0]=='.' && adr_part[1]=='.' && adr_part[2]==0) {
         /* delete previous part */
         if(askme->re_fill<=0) {
           bonked= 1;
           goto next_adr_part;
         } 
         if(askme->re_constants[askme->re_fill-1]!=NULL) {
           free(askme->re_constants[askme->re_fill-1]);
           askme->re_constants[askme->re_fill-1]= NULL;
         } else
#ifdef Askme_with_regeX
           regfree(&(askme->re[askme->re_fill-1]));
#else /* Askme_with_regeX */
           ;
#endif /* ! Askme_with_regeX */
         (askme->re_fill)--;
         goto next_adr_part;
       }
       if(strcmp(adr_part,"*")==0) {
         adr_part[0]= 0;
         ret= 2;
       } else
         ret= Askme__bourne_to_reg(adr_part,askme->reg_expr,0);
       if(ret==2) {
         if(Sregex_string(&(askme->re_constants[askme->re_fill]),adr_part,0)
            <=0)
           return(-1);
       } else {

#ifdef Askme_with_regeX
         if(regcomp(&(askme->re[askme->re_fill]),askme->reg_expr,0)!=0)
           goto cannot_compile;
#else /* Askme_with_regeX */
         if(!warned_of_regex) {
           Askme_de_eN(
             sprintf(askme->info_text,
      "=== Hinweis: Es werden keine Wildcards ausser einzelnem '*' erkannt\n");
           ,
             sprintf(askme->info_text,
      "=== Note: no wildcards get recognized besides single '*'\n");
           )
           Askme_info(askme,0);
           warned_of_regex= 1;
         }
         if(Sregex_string(&(askme->re_constants[askme->re_fill]),adr_part,0)
            <=0)
           return(-1);
#endif /* ! Askme_with_regeX */

       }
       askme->re_fill++;
next_adr_part:;
       if(i==count-1)
     break;
       cpt= npt+1;
       while(*cpt=='/')
         cpt++;
     }
     if(bonked) {
       Askme_de_eN(
         sprintf(askme->info_text,
                 "=== Hinweis: '..' an root-Directory angestossen.\n");
       ,
         sprintf(askme->info_text,
                 "=== Note: Your '..' bonked at the root directory.\n");
       )
       Askme_info(askme,0);
     }

     Askme__bourne_to_reg(adr_start,askme->reg_expr,0); /* just for show */

   } else { 
     is_constant= 0;
     if(strcmp(adr,"*")==0) {
       is_constant= 1;
     } else if(askme->search_mode==3 || askme->search_mode==4) {
       ret= Askme__bourne_to_reg(adr,askme->reg_expr,0);
       is_constant= (ret==2);
     } else
       strcpy(askme->reg_expr,adr);
     askme->re_count= 0; /* tells matcher that this is not structured */
     askme->re_constants= TSOB_FELD(char *,1);
     if(askme->re_constants==NULL)
       return(-1);
     askme->re_constants[0]= NULL;
     if(is_constant) {
       if(strcmp(adr,"*")==0) {
         if(Sregex_string(&(askme->re_constants[0]),"",0)<=0)
           return(-1);
       } else {
         if(Sregex_string(&(askme->re_constants[0]),adr,0)<=0)
           return(-1);
       }
       askme->re_fill= 1;
     } else { 
#ifdef Askme_with_regeX
       askme->re= TSOB_FELD(regex_t,1);
       if(askme->re==NULL)
         return(-1);
       if(regcomp(&(askme->re[0]),askme->reg_expr,0)!=0) {
cannot_compile:;
         sprintf(askme->info_text,
                 "%s : cannot compile regular expression : %s\n",
                 askme->progname,askme->reg_expr);
         Askme_info(askme,0);
         return(0);
       }
#else /* Askme_with_regeX */
       Askme_de_eN(
         sprintf(askme->info_text,
      "=== Hinweis: Es werden keine Wildcards ausser einzelnem '*' erkannt\n");
       ,
        sprintf(askme->info_text,
      "=== Note: no wildcards get recognized besides single '*'\n");
       )
       Askme_info(askme,0);
       if(Sregex_string(&(askme->re_constants[0]),adr,0)<=0)
         return(-1);
       askme->re_fill= 1;
#endif /* ! Askme_with_regeX */
     }

   }
 }
 return(1);
}


int Askme_regexec(askme,to_match,failed_at,flag)
struct AskmE *askme;
char *to_match;
int *failed_at;
int flag;
/*
 bit0= do not shortcut last component of to_match
 bit2= retry beginning at failed last component 
*/
{
 int ret,i,re_start= 0,reg_nomatch= -1;
 char *cpt,*npt,adr_part[4096],*mpt;

#ifdef Askme_with_regeX
 reg_nomatch= REG_NOMATCH;
#endif /* Askme_with_regeX */

 *failed_at= 0;
 if(!(askme->structured_search && askme->re_count>0)) {
   if(askme->re_constants!=NULL)
     if(askme->re_constants[0]!=NULL) {
       if(askme->re_constants[0][0]==0)
         return(0);
       if(strcmp(askme->re_constants[0],to_match)!=0)
         return(reg_nomatch);
       return(0);
     }  
#ifdef Askme_with_regeX
   ret= regexec(&(askme->re[0]),to_match,1,askme->match,0);
#else
   ret= reg_nomatch;
#endif /* ! Askme_with_regeX */

   return(ret);
 }

 cpt= to_match;
 while(*cpt=='/')
  cpt++;
 if(flag&4)
   re_start= askme->re_failed_at;
 if(re_start<0)
   re_start= 0;
 for(i= re_start;i<askme->re_fill;i++) {
   *failed_at= i;
   npt= strchr(cpt,'/');
   if(npt==NULL) {
     if(i<askme->re_fill-1 && !(flag&1)) 
       return(reg_nomatch); /* this must be the last expression part */
     mpt= cpt;
   } else {
     strncpy(adr_part,cpt,npt-cpt);
     adr_part[npt-cpt]= 0;
     mpt= adr_part;
   }
   if(askme->re_constants[i]!=NULL) {
     if(askme->re_constants[i][0]!=0) /* empty constant matches anything */
       if(strcmp(askme->re_constants[i],mpt)!=0)
         return(reg_nomatch);
   } else {

#ifdef Askme_with_regeX
     ret= regexec(&(askme->re[i]),mpt,1,askme->match,0);
     if(ret!=0)
       return(ret);
#else
     return(reg_nomatch);
#endif /* ! Askme_with_regeX */

   }
   if(npt==NULL) {
     if(i>=askme->re_fill-1)
       return(0); /* MATCH */
     *failed_at= i+1;
     return(reg_nomatch);
   }
   cpt= npt+1;
   while(*cpt=='/')
     cpt++; 
 }
 *failed_at= askme->re_fill;
 return(reg_nomatch);
}


int Askme_match_file(askme,to_match,adr,flag)
struct AskmE *askme;
char *to_match;
char *adr;
int flag;
/*
 bit0= match under fsm rules: with search_mode 0 : do not accept partial match
 bit1= with askme->search_mode==4 : do not check for '/'
 bit2= single level bit for Askme_regexec()
*/
/* return:
  <0 error
   0 no match
   1 match
   2 match with request to break volume loop
*/
{
 int ret,l,failed_at;
 char *cpt;

 l= strlen(to_match);
 if(askme->search_mode==0) {
   if(flag&1) {
     if(strcmp(adr,to_match)==0)
       return(2);
   } else {
     if(strncmp(adr,to_match,l)==0)
       if(adr[l]=='/'||adr[l]==0)
         return(2);
   }
   if(askme->sub==0)
     return(0);
   l= strlen(adr);
   if(strncmp(to_match,adr,l)==0)
     if(to_match[l]=='/'||to_match[l]==0) {
       if(askme->suball==0)
         return(2);
       return(1);
     }
 } else if(askme->search_mode==1) {
   if(strstr(to_match,adr)!=NULL)
     return(1);
 } else if(askme->search_mode>=2 && askme->search_mode<=4) {
   if(askme->search_mode==4 && !(flag&2)) {
     cpt= strrchr(to_match,'/');
     if(cpt==NULL)
       cpt= to_match;
     else
       cpt++;
   } else
     cpt= to_match;
   ret= Askme_regexec(askme,cpt,&failed_at,1|(flag&4));
   if(ret==0)
     return(1);
   askme->re_failed_at= failed_at;
 }
 return(0);
}


int Askme_fsm_report_sub(askme,adr,boss_node,boss_dirnode,boss_trav,
                         level,volume,flag)
struct AskmE *askme;
char *adr;
struct FsnodE *boss_node;
struct FsnodE *boss_dirnode;
struct FstraversaL *boss_trav;
int level;
int volume;
int flag;
/*
 bit0= process levels one by one, beginning with parameter level
 bit1= level headline has already been reported
 bit2= report only the first occurence and then return
 bit3= -ls mode: report all on first directory level, 
                 eventually transform absolute addresses to wd-relative ones
 bit4= do not report but only mark askme->search_lv_map and return match
 bit5= eventually transform absolute addresses to wd-relative ones
*/
/* return:
  <=0 error
    1 some sub objects have been reported
    2 no sub objects have been reported
*/
{
 char *namept,source[4096],target[4096+100];
 char boss_source[4096+100],boss_target[4096+100];
 char *sourcept,*targetpt,*subpt,*origpt;
 int node_level,node_volume,ret,subs_reported= 0,found_on_level,file_type;
 int subl,tgtl,depth,nodive= 0,comps,boss_comps,source_comps,l;
 int delayed_repetition_test= 0,source_not_matched;
 time_t mtime= 0, ctime= 0;
 struct FsnodE *dirnode,*node;
 struct FstraversaL *trav= NULL;
 struct Bitmap2D *virt_lev_vol;
 struct FsrptmaP *virtual_sources= NULL,*rpt;

 source_not_matched= (askme->resume_at_source[0]!=0);
 found_on_level= (flag&2)>>1;
 Askme_is_already_reported_sub(askme,0,0,1); /* initialize */
 ret= Fsnode_get_adr(boss_node,boss_dirnode,boss_source,target,&depth,0);
 if(ret<=0)
   {ret= -1; goto ex;}
 Fsnode_get_mctimes(boss_node,&mtime,&ctime,0);
 dirnode= boss_dirnode;
 node= boss_node;
 boss_target[0]= 0;
 if(boss_trav!=NULL) { /* will not change boss_trav state */
   ret= Fstraversal_new(&trav,NULL,0);
   if(ret<=0)
     {ret= -1; goto ex;}
   Fstraversal_copy(boss_trav,trav,0);
   if((trav->cursor_state&33)==33) {
     if(boss_source[0]=='/' && boss_source[1]==0)
       strcat(boss_target,boss_source);
     else
       sprintf(boss_target,"%s/",boss_source);
     sprintf(boss_source,"[target: %s]",boss_target);
   }
 }
 while(node!=NULL) {
   ret= Fstraversal_next(trav,boss_dirnode,&node,&dirnode,nodive);
   if(ret<=0)
     {ret= -1; goto ex;}
   if(node==NULL)
 break;
   if(flag&8)
     nodive= 3;
   ret= Fsnode_get_data(node,&namept,&node_level,&node_volume,0);
   if(ret<=0)
     {ret= -1; goto ex;}
   if(level>=0 && node_level!=level && (askme->nolevel || (flag&1)))
 continue;
   if(volume>0 && node_volume!=volume)
 continue;
   Fsnode_get_type(node,&file_type,0);
   if(file_type==2 && !(flag&8))
 continue;
   if(askme->show_outdated==2 && !Fstraversal_is_outdated(trav,0)) {
     if(file_type==2 && (flag&8)) {

       /* find out wether there are outdated files below */
       ret= Askme_fsm_report_sub(askme,adr,node,dirnode,trav,level,volume,
                                 (flag&(1|2))|4|16); /* do not report */
       if(ret<=0)
         goto ex;
       if(ret==2) /* no files found */
 continue;

     } else
 continue;
   }
   if(node_volume<0 || (node_level<0 && !(flag&8)))
 continue;
   if((askme->suball || (flag&8)) && !(flag&16)) {
     /* report all */;
   } else if((!(flag&8)) && boss_target[0]!=0) {
     /* repetition test will be done on virtual source node as soon as
        its address is determined */;
     delayed_repetition_test= 1;     
   } else {
     /* report one representative per level,volume */
     ret= Askme_is_already_reported_sub(askme,node_level,node_volume,0);
     if(flag&16) {
       Askme_is_already_reported_sub(askme,node_level,node_volume,2);
       subs_reported= 1;
 continue;
     }
     if(ret==1)
 continue;
   } 
   ret= Fsnode_get_adr(node,dirnode,source,target,&depth,0);
   if(ret<=0)
     {ret= -1; goto ex;}

   if(askme->resume_at_source[0]!=0 && source_not_matched) {
                                             /* for fine granulated -resume */
     if(strcmp(askme->resume_at_source,source)!=0)
 continue;
     if(!(askme->restore_mode&1))
       askme->resume_at_source[0]= 0;
     source_not_matched= 0;
   }

   sourcept= source;
   virt_lev_vol= NULL;
   if((flag&8) || askme->suball) {
     Fsnode_get_mctimes(node,&mtime,&ctime,0);
   } else {
     if(boss_target[0]!=0) {
       sourcept= boss_source;
       /* try to obtain boss source from found source */
       boss_comps= Sfile_count_components(boss_target,0);
       comps= Sfile_count_components(target,0);
       source_comps= Sfile_count_components(source,0);
       if(source_comps-(comps-boss_comps)>0) {
         /* hop up the difference in source and truncate it there */
         ret= Sfile_component_pointer(source,&sourcept,
                                      source_comps-(comps-boss_comps),0);
         if(ret>0) {
           if(sourcept-source>1)
             sourcept--;
           *sourcept= 0;
           sourcept= source;
           if(delayed_repetition_test) {
             for(rpt= virtual_sources; rpt!=NULL; rpt= rpt->next)
               if(strcmp(rpt->name,sourcept)==0)
             break;
             if(rpt!=NULL) {
               virt_lev_vol= rpt->bitmap;
             } else {
               ret= Fsrptmap_new(&virtual_sources,sourcept,
                                 askme->highest_volno+1,askme->highest_level+1,
                                 virtual_sources,0);
               if(ret<=0)
                 goto ex;
               virt_lev_vol= virtual_sources->bitmap;
             }
             /* do delayed repetition test on virtual source */
             ret= Bitmap2d_is_set(virt_lev_vol,node_volume,node_level,0);
             if(ret>0)
 continue;
             if(flag&16) {
               Bitmap2d_set(virt_lev_vol,node_volume,node_level,0);
               subs_reported= 1;
 continue;
             }
           }
         }  
       }
       if(sourcept==boss_source && delayed_repetition_test) {
         ret= Askme_is_already_reported_sub(askme,node_level,node_volume,0);
         if(flag&16) {
           Askme_is_already_reported_sub(askme,node_level,node_volume,2);
           subs_reported= 1;
 continue;
         }
         if(ret==1)
 continue;
       }
       strcpy(target,boss_target);
     } else {

   /* >>> ??? is it ok to compare the end pieces without leading slash ? */
   /* >>> ??? i.e.  subpt= source+strlen(boss_source)+1;
             shouldn't that be  subpt= source+strlen();
    */

       /* try to obtain boss target from found target */
       subpt= source+strlen(boss_source)+1;
       subl= strlen(subpt);
       tgtl= strlen(target);
       if(subl<tgtl)
         if(strcmp(subpt,target+(tgtl-subl))==0) {
           target[tgtl-subl]= 0;
           sourcept= boss_source;
         }
       if(sourcept!=boss_source) { /* there was no target found */
         sprintf(target,"[source: %s]",boss_source);
         sourcept= boss_source;
       }
     }
   }
   targetpt= target;
   if(((flag&8) && askme->prepended_wd) || (flag&32)) {
     l= strlen(askme->wd);
     if(askme->search_target) {
       origpt= targetpt;
       if(l>0) if(strncmp(targetpt,askme->wd,l)==0 && targetpt[l]=='/') {
         targetpt= targetpt+strlen(askme->wd);
         for(;*targetpt=='/';targetpt++);
       }
       if(targetpt[0]==0)
         targetpt= origpt;
     } else {
       origpt= sourcept;
       if(l>0) if(strncmp(sourcept,askme->wd,l)==0 && sourcept[l]=='/') {
         sourcept= sourcept+strlen(askme->wd);
         for(;*sourcept=='/';sourcept++);
       }
       if(sourcept[0]==0)
         sourcept= origpt;
     }
   }
   if(!source_not_matched) {
     ret= Askme_found(askme,boss_trav,node_level,node_volume-1,
                      sourcept,targetpt,adr,mtime,ctime,
                      ((file_type==2)<<2) | (found_on_level>0 || !(flag&1)));
     if(ret<=0)
       goto ex;
     if(node_level>=0 && node_volume>0) {
       if(virt_lev_vol!=NULL)
         Bitmap2d_set(virt_lev_vol,node_volume,node_level,0);
       else
         Askme_is_already_reported_sub(askme,node_level,node_volume,2);
     }
     subs_reported= 1;
     found_on_level= 1;
     if((flag&4)||askme->request_to_abort)
 break;
   }
 }
 ret= 1+(subs_reported==0);
ex:;
 Fsrptmap_destroy(&virtual_sources,1);
 Fstraversal_destroy(&trav,0);
 return(ret);
}


int Askme_state_to_file(askme,adr,label,state_line,fp,flag)
struct AskmE *askme;
char *adr;
char *label;
char *state_line;
FILE **fp;
int flag;
/*
 bit0= do not close after write
 bit1= do not write state_line
 bit2= use truncate and fsync rather than unlink
*/
{
 int ret;
 char line[81];

 *fp= NULL;
 ret= Sfile_type(adr,1);
 if(ret==-1)
   /* ok. no file existing yet */;
 else if(ret==1 || ret==3) {
                 /* prevent link traps and overwriting running -resume file */
   if((flag&4) && ret!=3) {
     /* will truncate file by fopen() */;
   } else {
     ret= unlink(adr);
     if(ret==-1)
       return(0);
   }
 } else
   return(0);
 *fp= Afile_fopen(adr,"wb",1);
 if(*fp==NULL)
   return(0);
 sprintf(line,"# %s , %s\n",label,Sfile_datestr(time(0),3|256));
 ret= fwrite(line,strlen(line),1,*fp);
 if(ret<=0) {
   fclose(*fp);
   *fp= NULL;
   return(0);
 }
 if(!(flag&2)) {
   ret= fwrite(state_line,strlen(state_line),1,*fp);
   if(ret<=0) {
     fclose(*fp);
     *fp= NULL;
     return(0);
   }
 }
 if(flag&4) {
   fflush(*fp);
   fsync(fileno(*fp));
 }
 if(!(flag&1)) {
   fclose(*fp);
   *fp= NULL;
 }
 return(1);
}


int Askme_record_resume_state(askme,level,volume,source,search_text,restore,
                              flag)
struct AskmE *askme;
int level,volume;
char *source;
char *search_text;
int restore;       /* 1= -restore , 0= -volumwise -search */
int flag;
/*
 bit0= do not write to eventual state files
 bit1= do not set current resume state
 bit3= do not write big file and _opt
 bit4= do not write _lv
 bit5= do not write _src
 bit6= do not write _dirty 
*/
{
 int ret,indirect_last_source= 16,fsync_flag;
 char line[6*4096],adr[4096+5];
 FILE *fp= NULL;

 /* <<< might get removed after thorough testing of program */;
 if(askme->restore_ctl_stack!=NULL)
   if(askme->restore_ctl_stack->move_to_asylum) {
     sprintf(askme->info_text,
             "--- program error: attempt to record resume state of move_to_asylum operation\n");
     Askme_info(askme,0);
     {ret= -1; goto ex;}
   }

 if(flag&2)
  goto write_to_file;
 askme->resume_last_level= level;
 askme->resume_last_volume= volume;
 if(source!=NULL && source!=askme->resume_last_source) {
   if(askme->resume_at_source[0]==0)
     strcpy(askme->resume_last_source,source);
   else
     strcpy(askme->resume_last_source,askme->resume_at_source);
 }
 if(search_text!=askme->resume_search_text && search_text!=NULL)
   strcpy(askme->resume_search_text,search_text);
 askme->resume_was_restore= !!restore;

write_to_file:;
 if(flag&1)
   {ret= 1; goto ex;} 
 if(askme->resume_state_file[0]!=0) {
   if(!(flag&64)) {
     sprintf(adr,"%s_dirty",askme->resume_state_file);
     sprintf(line,"-resume_last_dirty:%s\n",askme->resume_last_dirty);
     fsync_flag= 0;
     if(askme->resume_last_dirty[0]!=0 && (askme->restore_mode&(1<<21)))
       fsync_flag= 4;
     ret= Askme_state_to_file(askme,adr,"resume_state_file_dirty",line,&fp,
                              fsync_flag);
     if(ret<=0) {
       indirect_last_source= 0;
       goto write_status_file;
     }
   }
   if(!(flag&32)) {
     sprintf(adr,"%s_src",askme->resume_state_file);
     sprintf(line,"-resume_last_source:%s\n",askme->resume_last_source);
     ret= Askme_state_to_file(askme,adr,"resume_state_file_src",line,&fp,0);
     if(ret<=0) {
       indirect_last_source= 0;
       goto write_status_file;
     }
   }
   if(!(flag&8)) {
     sprintf(adr,"%s_opt",askme->resume_state_file);
     sprintf(line,"-resume_opt_text:%s:%s\n",
                  (askme->resume_was_restore?"-restore":"-search"),
                  askme->resume_search_text);
     ret= Askme_state_to_file(askme,adr,"resume_state_file_opt",line,&fp,0);
     if(ret<=0) {
       indirect_last_source= 0;
       goto write_status_file;
     }
   }
   if(!(flag&16)) {
     sprintf(adr,"%s_lv",askme->resume_state_file);
     sprintf(line,"-resume:%d:%d\n",
             askme->resume_last_level,askme->resume_last_volume);
     ret= Askme_state_to_file(askme,adr,"resume_state_file_lv",line,&fp,0);
     if(ret<=0) {
       indirect_last_source= 0;
       goto write_status_file;
     }
   }
write_status_file:;
   if(!(flag&8)) {
     ret= Askme_state_to_file(askme,askme->resume_state_file,
                              "resume_state_file","",&fp,1|2);
     if(ret<=0)
       {ret= 0; goto ex;}
     ret= Askme_status(askme,NULL,fp,2|8|indirect_last_source);
     if(ret<=0)
       goto ex;
   }
 }
 ret= 1;
ex:;
 if(fp!=NULL)
   fclose(fp);
 return(ret);
}


int Askme_record_resume_source(askme,source,flag)
struct AskmE *askme;
char *source;
int flag;
/*
 bit0= force write to file
 bit1= do not invalidate  askme->resume_last_dirty
 bit2= disable write to file
*/
{
 int no_dirty= 0;

 strcpy(askme->resume_last_source,source);
 if(flag&2)
   no_dirty= 64;
 else
   askme->resume_last_dirty[0]= 0;
 if(!(flag&4))
   if((flag&1) || !(askme->restore_mode&(1<<20)))
     Askme_record_resume_state(askme,-1,-1,"","",-1,2|8|16|no_dirty);
 return(1);
}


int Askme_record_resume_dirty(askme,restore_adr,flag)
struct AskmE *askme;
char *restore_adr;
int flag;
/*
 bit0= force write to file
 bit2= disable write to file
*/
{
 strcpy(askme->resume_last_dirty,restore_adr);
 if(!(flag&4))
   if((flag&1) || !(askme->restore_mode&(1<<20))) /* write to file _dirty */
     Askme_record_resume_state(askme,-1,-1,"","",-1,2|8|16|32);
 return(1);
}


int Askme_fsm_search_file(askme,adr,level,volume,flag)
struct AskmE *askme;
char *adr;
int level;
int volume;
int flag;
/*
 bit0= process levels in descending order, beginning with parameter level
 bit1= only search below current working directory
 bit2= do not initialize level_vol_map
 bit3= do not destroy trav but transfer into askme.found_trav
 bit4= process volumes one by one
 bit5= process levels in ascending order up to askme->highest_level
*/
{
 int ret,l,found= 0,match_ret,found_on_level,file_type,no_dive;
 int depth= -1,structured_search,skip_next_and_match= 0,search_target;
 int node_level,node_volume,accept_only_level;
 int source_valid= 0,target_valid= 0,resume_dir_search=0;
 int structured_shortcut,resume_dir_depth= 0,adr_flag,in_target_tree= 0,cstate;
 int unique_find= 0,accept_only_volume,dummy_depth;
 char source[4096],target[4096],*namept,*to_match,*sourcept,*targetpt,*origpt;
 time_t mtime= 0,ctime= 0;
 struct FsnodE *dirnode,*node,*resume_dirnode= NULL;
 struct FsnodE *endnode= NULL,*masternode= NULL;
 struct FsnodE *prev_dirnode,*prev_node;
 struct FstraversaL *trav= NULL,*master_trav= NULL;

 askme->did_something_useful= 1;
 if((flag&16) && (flag&32) && adr!=askme->resume_search_text) { 
                              /* full volumewise search : record for -resume */
   if(strlen(adr)<sizeof(askme->resume_search_text))
     strcpy(askme->resume_search_text,adr);
   else
     askme->resume_search_text[0]= 0;
 }
 search_target= askme->search_target;
 ret= Askme_prepare_regex(askme,adr,0);
 if(ret<=0)
   goto ex;
 structured_search= askme->structured_search * (askme->re_count>0);
 structured_shortcut= structured_search;
 unique_find= (askme->search_mode==4)&&(askme->structured_search==4);

 if(askme->search_lv_map==NULL || !(flag&4)) {
   Bitmap2d_destroy(&(askme->search_lv_map),0);
   ret= Bitmap2d_new(&(askme->search_lv_map),
                       askme->highest_volno+1,askme->highest_level+1,0);
   if(ret<=0)
     {ret= -1; goto ex;}
 }
 if((askme->search_mode==3 && 
     (askme->structured_search==2 || askme->structured_search==3)) ||
     unique_find)
   level= -1;

 if(search_target) {
   node= askme->target_root;
   ret= Fstraversal_new(&master_trav,askme->target_root,
                        32|((!!askme->show_outdated)<<6));
 } else {
   node= askme->root_node;
   ret= Fstraversal_new(&master_trav,askme->root_node,
                        ((!!askme->show_outdated)<<6));
 }
 if(ret<=0)
   goto ex;
 if(flag&2) {
   /* position  master_trav  on askme->wd */
   ret= Fsnode_get_by_path(NULL,askme->wd ,&node,&dirnode,&master_trav,2);
   if(ret<=0)
     goto ex;
   masternode= node;
   endnode= dirnode;
 }
 if(flag&16) {
   ret= Askme_make_lev_vol_map(askme,0);
   if(ret<=0)
     goto ex;
   volume= 1;
   if(askme->resume_at_volume>0)
     volume= askme->resume_at_volume;
 }
 if(flag&32) {
   level= 0;
   if(askme->resume_at_level>0)
     level= askme->resume_at_level;
 }
 ret= Fstraversal_new(&trav,NULL,0); /* a dummy to be reprogrammed by master */
 if(ret<=0)
   goto ex;

once_again:;
 found_on_level= 0;
once_again_same_level:;
 if((flag&16) && (flag&32))
   Askme_record_resume_state(askme,level,volume,"",askme->resume_search_text,
                             askme->restore_mode&1,8);
 no_dive= 0;
 resume_dir_search= 0;
 dirnode= NULL;
 Fstraversal_copy(master_trav,trav,0);
 Fstraversal_get_pos(trav,&node,&dirnode,&cstate,0);

 if(structured_search && askme->re_fill==0)
   skip_next_and_match= 1;
 accept_only_level= ((level>=0) && 
                     ((askme->nolevel || (flag&1)) && !skip_next_and_match));
 accept_only_volume= (volume>0 && !skip_next_and_match) ;
 while(node!=NULL) {
   if(askme->request_to_abort)
 break;

   /* <<< debugging */
   prev_node= node;
   prev_dirnode= dirnode;

   if(!skip_next_and_match) {
     ret= Fstraversal_next(trav,endnode,&node,&dirnode,
                           no_dive|((!!(unique_find))<<2));
     if(ret<=0)
       {ret= -1; goto ex;}
   }
   no_dive= 0;
   if(node==NULL)
 break;
   if(search_target) {
     in_target_tree= Fstraversal_is_in_target_tree(trav,0);
     if(masternode!=NULL && !in_target_tree)
       if(node->extra!=NULL)
         if(node->extra->tlink!=NULL)
           if(node->extra->tlink->target==masternode)
  continue;
    }
     /* <<< debugging 
     if(strcmp(node->name,"")==0)
       printf("");
     */

   ret= Fsnode_get_data(node,&namept,&node_level,&node_volume,0);
   if(ret<=0)
     {ret= -1; goto ex;}
   Fsnode_get_type(node,&file_type,0);
   if(file_type!=2 && !skip_next_and_match) {
     if(accept_only_volume && node_volume!=volume)
 continue;
     if(accept_only_level && node_level!=level)
 continue;
     if(askme->show_outdated==2 && (!Fstraversal_is_outdated(trav,0)))
 continue;
   }
   Fsnode_get_mctimes(node,&mtime,&ctime,0);
   to_match= source;
   source_valid= 0;
   target_valid= 0;

   if(askme->resume_at_source[0]!=0) { /* for fine granulated -resume */
     ret= Fsnode_get_adr(node,dirnode,source,target,&dummy_depth,4);
     if(ret<=0)
       {ret= -1; goto ex;}
     if(strcmp(askme->resume_at_source,source)!=0) {
       if(file_type!=2)
 continue;
     } else
       askme->resume_at_source[0]= 0;
   }

   if(askme->search_mode==4) {
     to_match= namept;
   } else if(resume_dir_search && dirnode==resume_dirnode &&
             structured_shortcut && !skip_next_and_match) {
     to_match= namept;
     depth= resume_dir_depth;
   } else {
     resume_dir_search= 0;
     if(!search_target)
       adr_flag= 4;
     else if (in_target_tree)
       adr_flag= 4;
     else
       adr_flag= 1;
     ret= Fsnode_get_adr(node,dirnode,source,target,&depth,adr_flag);
     if(ret<=0)
       {ret= -1; goto ex;}
     source_valid= 1;
     if(search_target) {
       if(in_target_tree)
         to_match= source;
       else
         to_match= target;
       target_valid= !(adr_flag&4);
     }
     if(structured_shortcut && !skip_next_and_match)
       if(askme->re_fill<depth) /* will not match */
 continue;
   }
   if(skip_next_and_match) {
     match_ret= 1;
     no_dive= 1;
     skip_next_and_match= 0; /* to avoid endless loops */
   } else {
     match_ret= Askme_match_file(askme,to_match,adr,
                                 1|2|(resume_dir_search<<2));
     if(match_ret<0)
       {ret= -1; goto ex;}
   }
   if(match_ret==0) {
     if(structured_shortcut) {
       if(askme->re_failed_at < depth) /* sub items will not match */
         no_dive= 1;
       if(askme->re_failed_at == depth-1) { /* allow retry with same dirnode */
         resume_dir_search= 1; 
         resume_dir_depth= depth;
         resume_dirnode= dirnode;
       } else
         resume_dir_search= 0;
     }
 continue;
   }
   if(!(source_valid && target_valid)) { 
     /* make up for skipped absolute address */
     ret= Fsnode_get_adr(node,dirnode,source,target,&depth,0);
     if(ret<=0)
       {ret= -1; goto ex;}
   }
   if(file_type==2 && structured_search!=2 && !unique_find) { 
     /* this is a directory which was looked for explicitely */
     /* now report representatives of each volume and/or level */
     ret= Askme_fsm_report_sub(askme,adr,node,dirnode,trav,level,volume,
             ((structured_search==3)<<3) | ((found_on_level>0)<<1) | (flag&1));
     if(ret<=0)
       {goto ex;}
     if(ret==2) {
       /* nothing was reported */
       if(node_level>=0 && node_volume>0) { 
         /* the directory was introduced as leaf on this level */
         if(!((accept_only_level && node_level!=level)
              || (accept_only_volume && node_volume!=volume))) {
           if(structured_search!=3 && Fsnode_is_empty_dir(node,0)) {
             l= strlen(target);
             if(l>0)
               if(target[l-1]=='/')
                 l= -1;
             if(l>=0)
               strcat(target,"/");
             ret= Askme_found(askme,trav,node_level,node_volume-1,
                              source,target,adr,mtime,ctime,
                        ((file_type==2)<<2) | (found_on_level>0 || !(flag&1)));
             if(ret<=0)
               goto ex;
             found= found_on_level= 1;
           }
         }
       } else {
         /* the directory was introduced only implicitely */

         /* >>> ??? find representation of empty implicit directory */;

       }
     } else
       found= found_on_level= 1;
     if(askme->search_mode!=4)
       no_dive= 1; /* continue tree traversal after node */
   } else {
     if(file_type==2 && (structured_search==2 || unique_find))
       if(askme->show_outdated==2 && !Fstraversal_is_outdated(trav,0)) {
         /* find out wether there are outdated files below */
         ret= Askme_fsm_report_sub(askme,adr,node,dirnode,trav,level,volume,
                                   (flag&1)|4|16); /* do not report */
         if(ret<=0)
           goto ex;
         if(ret==2) /* no files found */
           goto skip_found;
       }
     if(askme->look_for_dir) {
       if(file_type==2) {
         if(askme->search_target && !in_target_tree) {
           strcpy(askme->result_line,target);
         } else
           strcpy(askme->result_line,source);
         askme->request_to_abort= 1;
         found= found_on_level= 1;
 break;
       }
     } else {
       sourcept= source;
       targetpt= target;
       if(((structured_search>=2 && structured_search<=3)
           && askme->prepended_wd) || unique_find ) {
         l= strlen(askme->wd);
         if(askme->search_target) {
           origpt= targetpt;
           if(strncmp(targetpt,askme->wd,l)==0 && targetpt[l]=='/') {
             targetpt= targetpt+strlen(askme->wd);
             for(;*targetpt=='/';targetpt++);
           }
           if(targetpt[0]==0)
             targetpt= origpt;
         } else {
           origpt= sourcept;
           if(strncmp(sourcept,askme->wd,l)==0 && sourcept[l]=='/') {
             sourcept= sourcept+strlen(askme->wd);
             for(;*sourcept=='/';sourcept++);
           }
           if(sourcept[0]==0)
             sourcept= origpt;
         }
       }
       ret= Askme_found(askme,trav,node_level,node_volume-1,
                        sourcept,targetpt,adr,mtime,ctime,
                        ((file_type==2)<<2) | (found_on_level>0 || !(flag&1)));
       if(ret<=0)
         goto ex;
       found= found_on_level= 1;
     }
skip_found:;
     if(structured_search==2)
       no_dive= 1|32;
     else if(structured_search==3)
       no_dive= 2;
   }
 }
 if(askme->request_to_skip_volume>0)
   askme->request_to_skip_volume--;
 if(askme->request_to_abort)
   {ret= found; goto ex;}
 if(found && askme->levelall==0)
   {ret= found; goto ex;}
 if(askme->nolevel)
   {ret= found; goto ex;}

 if(askme->resume_at_source[0]!=0) { /* resume item was not found */

   /* >>> ??? This is not normal. How to react ? */;

   askme->resume_at_source[0]= 0;
 }
 if((flag&16) && level>=0) {
   /* switch up to next populated volume */
   for(volume++;volume<=askme->highest_volno;volume++)
     if(Bitmap2d_is_set(askme->lev_vol_map,volume,level,0)>0)
   break;
   if(volume<=askme->highest_volno)
     goto once_again_same_level;
   volume= 1;
 }
 if(flag&32) {
   if(level<askme->highest_level) {
     level++;
     goto once_again;
   }
 } else if(level>0 && (flag&1)) {
   level--;
   goto once_again;
 }
 ret= found;
 if((flag&16) && (flag&32)) {
                    /* keep completed search from resuming without arguments */
   Askme_record_resume_state(askme,-1,-1,"",askme->resume_search_text,
                             askme->restore_mode&1,1); /* only set state */
 }
ex:;
 if((flag&16) && (flag&32))
   Askme_record_resume_state(askme,-1,-1,"","",-1,2|8);
                                             /* only write to position files */
 Fstraversal_destroy(&master_trav,0);
 if(flag&8) {
   Fstraversal_destroy(&(askme->found_trav),0);
   askme->found_trav= trav;
 } else
   Fstraversal_destroy(&trav,0);
 return(ret);
}


int Askme_search_file(askme,adr,level,volume,flag)
struct AskmE *askme;
char *adr;
int level;
int volume;
int flag;
/*
 bit1: for fsm_search_file: only search below current working directory
>>> bit4= process volumes one by one
>>> bit5= process levels in ascending order up to askme->highest_level
*/
{
 int ret,i,j,found= 0,match_ret,levno= -1,structured_search;

 if(adr[0]==0)
   return(0);

 if(askme->search_mode==4 && strchr(adr,'/')) {
   Askme_de_eN(
     sprintf(askme->info_text,
         "=== Hinweis: '/' macht im Modus -search_shname meist wenig Sinn\n");
   ,
     sprintf(askme->info_text,
         "=== Note: Letter '/' is hardly of any use in mode -search_shname\n");
   )
   Askme_info(askme,0);
 }

 Askme_reset_counters(askme,0);
 if(askme->root_node!=NULL) {
   if(askme->highest_level<=0)
     level= -1;
   else
     level= askme->highest_level;
   ret= Askme_fsm_search_file(askme,adr,level,volume,
                              (level>=0) | (flag&(2|16|32)));
   return(ret);
 }

 ret= Askme_prepare_regex(askme,adr,0);
 if(ret<=0)
   return(ret);
 structured_search= askme->structured_search * (askme->re_count>0);
 if(structured_search && askme->re_fill<=0) {
   if(askme->look_for_dir) {
     strcpy(askme->result_line,"/");
     askme->request_to_abort= 1;
     found= askme->found= 1;
     goto ex;
   }
   if(structured_search==2) {
     Askme_found(askme,NULL,-1,-1,"/","",(time_t) 0,(time_t) 0,0);
     found= askme->found= 1;
     goto ex;
   }
 }

 for(i=0;i<askme->num_volumes;i++) {
   if(askme->volumes[i]==NULL)
 continue;
   for(j=0;j<askme->count_in_volume[i];j++) {
     if(askme->volumes[i][j]==NULL)
   continue;
     match_ret= Askme_match_file(askme,askme->volumes[i][j],adr,0);
     if(match_ret<0)
       return(match_ret);
     if(match_ret==0)
   continue;
     if(level>=0)
       levno= level;
     else if(askme->num_sublevels>0)
       levno= askme->num_sublevels;
     found= 1;
     if(askme->look_for_dir) {
       strcpy(askme->result_line,askme->volumes[i][j]);
       askme->request_to_abort= 1;
       goto ex;
     }
     Askme_found(askme,NULL,levno,i,askme->volumes[i][j],"",
                 (time_t) 0,(time_t) 0,adr,found>0);
     if(askme->request_to_abort)
       goto ex;
     if(match_ret==2)
   break;
   }
 }
ex:;
 if(askme->request_to_abort)
   return(found);
 if(found && (askme->levelall==0 || askme->look_for_dir || 
              askme->search_mode==4))
   return(found);
 if(askme->nolevel)
   return(found);
 for(i= askme->num_sublevels-1;i>=0;i--) {
   if(askme->sublevels[i]==NULL)
 continue;
   Askme_clone_options(askme->sublevels[i],askme,0);
   askme->sublevels[i]->nolevel= 1;
   ret= Askme_search_file(askme->sublevels[i],adr,i,-1,0);
   if(ret<0)
     return(ret);
   found|= ret;
   askme->found+= askme->sublevels[i]->found;
   strcpy(askme->result_line,askme->sublevels[i]->result_line);
   if(found && (askme->levelall==0 || askme->look_for_dir || 
                askme->search_mode==4))
 break;
 }
 if(found==0)
   askme->search_failed++;
 return(found);
}


int Askme_resume(askme,level,volume,option,flag)
struct AskmE *askme;
int level;
int volume;
char *option;
int flag;
/* return: see Askme_interpreter() */
{
 char *adr= "",shellsafe[5*4096];
 int ret,hflag= 1|2,volumewise,was_option= 0;

 volumewise= askme->volumewise;
 if(option[0]!=0) {
   if(strncmp(option,"-restore:",9)==0 || strncmp(option,"-restore ",9)==0) {
     adr= option+9;
     hflag|= 4;
     was_option= 1;
   } else if(strncmp(option,"-search:",8)==0
             || strncmp(option,"-search ",8)==0) {
     adr= option+8;
     was_option= 1;
   } else {
     Askme_de_eN(
       sprintf(askme->info_text,
 "--- Option mit -resume nicht erlaubt (nur -restore oder -search):\n    %s\n",
              option);
     ,
       sprintf(askme->info_text,
 "--- Option not permitted with -resume (only -restore or -search):\n    %s\n",
              option);
     )
     Askme_info(askme,0);
     {ret= 0; goto ex;}
   }
 }
 if(adr[0]==0) {
   adr= askme->resume_search_text,adr;
   if(askme->resume_was_restore && !was_option)
     hflag|= 4;
 }
 if(strlen(adr)==0) {
   Askme_de_eN(
     sprintf(askme->info_text,
             "-resume: Da war kein vorheriges Rueckspeichern\n         %s\n",
             "und auch keine vorherige Suche mit -volumewise:on");
   ,
     sprintf(askme->info_text,
             "-resume: No previous restore run recorded\n         %s\n",
             "and no previous search with -volumewise:on either");
   )
   Askme_info(askme,0);
   {ret= 2; goto ex;}
 }
 if(level>=0)
   askme->resume_at_level= level;
 else if(askme->resume_last_level>=0)
   askme->resume_at_level= askme->resume_last_level;
 else {
   Askme_de_eN(
     sprintf(askme->info_text,
             "-resume: Keine Levelnummer angegeben\n         %s\n",
             "und kein abgebrochener Lauf anhaengig");
   ,
     sprintf(askme->info_text,
             "-resume: No level number given\n         %s\n",
             "and no aborted run pending");
   )
   Askme_info(askme,0);
   {ret= 2; goto ex;}
 }
 if(volume>0)
   askme->resume_at_volume= volume;
 else if(askme->resume_last_volume>0)
   askme->resume_at_volume= askme->resume_last_volume;
 else
   askme->resume_at_volume= 1;

 if(askme->resume_last_dirty[0]!=0) {
   sprintf(askme->info_text,
"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
          );
   Askme_de_eN(
     sprintf(askme->info_text+strlen(askme->info_text),
"!!  WARNUNG:  Ein Fileobjekt ist beim letzten Lauf FRAGWUERDIG geblieben  !!\n"
            );
   ,
     sprintf(askme->info_text+strlen(askme->info_text),
"!!!!  WARNING:  The last run left a file object in QUESTIONABLE state   !!!!\n"
            );
   )
   sprintf(askme->info_text+strlen(askme->info_text),
"!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
          );
   Askme_info(askme,1);
   if(askme->request_to_abort)
     goto final_dirty_handling;

   Askme_de_eN(
     sprintf(askme->info_text,
             "Die Adresse des fragwuerdigen Fileobjekts ist:\n");
   ,
     sprintf(askme->info_text,
             "The address of the questionable file object is:\n");
   )
   sprintf(askme->info_text+strlen(askme->info_text),
           "\n   %s\n\n",askme->resume_last_dirty);
   Askme_info(askme,1);
   if(askme->request_to_abort)
     goto final_dirty_handling;

   askme->info_text[0]= 0;
   Askme_de_eN(
     sprintf(askme->info_text+strlen(askme->info_text),
"Die darueberliegenden Directories koennten zu grosszuegige Rechte fuer\n");
     sprintf(askme->info_text+strlen(askme->info_text),
"ihren Besitzer bieten, oder zu knappe fuer jedermann. Pruefen mit:  ls -ld\n"
            );
     sprintf(askme->info_text+strlen(askme->info_text),
"Ggf. wieder geradebiegen mit:  chmod\n");
     sprintf(askme->info_text+strlen(askme->info_text),
"Das Fileobjekt selbst ist komplett fragwuerdig. Ist es ein Directory, dann\n");
     sprintf(askme->info_text+strlen(askme->info_text),
"sollte man die Zugriffsrechte pruefen und ggf. korrigieren. Datenfiles\n");
     sprintf(askme->info_text+strlen(askme->info_text),
"sollte man am besten ganz loeschen, wenn ihre Korrektheit in Zweifel steht.\n"
            );
     if(!askme->dialog)
       sprintf(askme->info_text+strlen(askme->info_text),
"Danach diese Wiederaufnahme des Rueckspeicherns fortsetzen.\n");
   ,
     sprintf(askme->info_text+strlen(askme->info_text),
"The directories above the file might have too generous permissions for\n");
     sprintf(askme->info_text+strlen(askme->info_text),
"their owner or too sparse permissions for anybody. Check that by:  ls -ld\n");
     sprintf(askme->info_text+strlen(askme->info_text),
"Eventually fix it by help of:  chmod\n");
     sprintf(askme->info_text+strlen(askme->info_text),
"The file object itself is completely questionable. Check it. If it is a\n");
     sprintf(askme->info_text+strlen(askme->info_text),
"directory then eventually fix its access permissions. Regular data files\n");
     sprintf(askme->info_text+strlen(askme->info_text),
"should best be removed if their validity is hard to determine.\n");
     if(!askme->dialog)
       sprintf(askme->info_text+strlen(askme->info_text),
"Then continue this resumed restore.\n");
   )
   Askme_info(askme,1);
   if(askme->request_to_abort)
     goto final_dirty_handling;

final_dirty_handling:;
   if(askme->dialog) {
     askme->resume_last_level= askme->resume_at_level;
     askme->resume_last_volume= askme->resume_at_volume;
     if(askme->request_to_abort)
       {ret=2; goto ex;}

     Askme_de_eN(
       sprintf(askme->info_text,
               "\nWenn alles geregelt ist, den Text \"weiter\" eingeben\n");
     ,
       sprintf(askme->info_text,
               "\nIf everything is fixed, enter text \"go on\"\n");
     )
     Askme_info(askme,1);
     askme->result_line_counter= 0;
     askme->result_page_counter++;
     ret= Askme_dialog_input(askme,shellsafe,sizeof(shellsafe),1);   
     if(ret<=0)
       {ret= 2; goto ex;}
     if(strcmp(shellsafe,"weiter")!=0 && strcmp(shellsafe,"\"weiter\"")!=0 &&
        strcmp(shellsafe,"go on")!=0 && strcmp(shellsafe,"\"go on\"")!=0) {
       Askme_de_eN(
         sprintf(askme->info_text,
"\nWiederaufnahme der Operation abgebrochen !\n");
         sprintf(askme->info_text+strlen(askme->info_text),
"Wenn das Problem behoben ist, kann man es erneut versuchen mit:\n   -resume\n"
                );
       ,
         sprintf(askme->info_text,
"\nAttempt to resume operation aborted !\n");
         sprintf(askme->info_text+strlen(askme->info_text),
"After the problem is solved, one may retry by option:\n   -resume\n\n");
       )
       Askme_info(askme,1);
       {ret= 2; goto ex;}
     }
   } else {

     if(askme->resume_state_file[0]!=0 && !askme->request_to_abort) {
       strcpy(askme->info_text,"\n");
       Askme_de_eN(
         sprintf(askme->info_text+strlen(askme->info_text),
"Um diese Sicherheitssperre schliesslich loszuwerden, ausfuehren:\n");
       ,
         sprintf(askme->info_text+strlen(askme->info_text),
"To finally get rid of this safety barrier, execute:\n");
       )
       sprintf(askme->info_text+strlen(askme->info_text),
"  echo > %s_dirty\n",Text_shellsafe(askme->resume_state_file,shellsafe,0));
       strcat(askme->info_text,"\n");
       Askme_info(askme,1);
     }
     {ret= 0; goto ex;}
   }
 }

 Askme_record_resume_dirty(askme,"",1);
 strcpy(askme->resume_at_source,askme->resume_last_source);
 if(!(hflag&4))
   askme->volumewise|= 3;
 sprintf(askme->info_text,"+ resuming at level %d volume %d",
         askme->resume_at_level,askme->resume_at_volume);
 if(askme->resume_at_source[0]!=0)
   sprintf(askme->info_text+strlen(askme->info_text)," file %s\n",
           askme->resume_at_source);
 else
   strcat(askme->info_text,"\n");
 sprintf(askme->info_text+strlen(askme->info_text),"+ %s:%s\n",
         ((hflag&4)?"-restore":"-search"),adr);
 Askme_info(askme,1);
 ret= Askme_execute_option(askme,adr,hflag);
ex:;
 askme->resume_at_volume= askme->resume_at_level= -1;
 askme->volumewise= volumewise;
 askme->resume_at_source[0]= 0;
 return(ret);
}


int Askme_fsm_list(askme,level,volume,flag)
struct AskmE *askme;
int level;
int volume;
int flag;
/*
 bit0= add level number to volume number
 bit5= for debugging: list target tree
*/
{
 int ret= 0,node_level,node_volume,i,depth,new_volume= 0,file_type,upto;
 int list_all= 0;
 char source[4096],target[4096],*namept;
 struct FsnodE *dirnode,*node;
 struct FstraversaL *trav= NULL;

 if(flag&32) {
   upto= 0;
   list_all= 1;
 } else
   upto= askme->highest_volno-1;
 for(i=0;i<=upto;i++) {
   if(volume>0 && i!=volume-1)
 continue;
   Fstraversal_destroy(&trav,0);
   if(flag&32)
     ret= Fstraversal_new(&trav,askme->target_root,
                          32|((!!askme->show_outdated)<<6));
   else {
     new_volume= 1;
     ret= Fstraversal_new(&trav,askme->root_node,(!!askme->show_outdated)<<6);
   }
   if(ret<=0)
     goto ex;
   while(1) {
     ret= Fstraversal_next(trav,NULL,&node,&dirnode,0);
     if(ret<=0)
       goto ex;
     if(node==NULL)
   break;
      if(askme->show_outdated==2 && (!Fstraversal_is_outdated(trav,0)))
 continue;
     ret= Fsnode_get_data(node,&namept,&node_level,&node_volume,0);
     if(ret<=0)
       goto ex;
     Fsnode_get_type(node,&file_type,0);
     if(!list_all) {
       if(file_type==2 || (level>=0 && node_level!=level))
   continue;
       if(node_volume!=i+1)
   continue;
     }
     ret= Fsnode_get_adr(node,dirnode,source,target,&depth,
                         (!askme->search_target)<<2);
     if(ret<=0)
       goto ex;
     if(new_volume) {
       if((flag&1) && level>=0)
         sprintf(askme->result_line,"########## part %d  (level %d)\n",
                 i+1,level);
       else
         sprintf(askme->result_line,"########## part %d\n",i+1);
       Askme_result(askme,0);
       new_volume= 0;
     }
     if(askme->search_target)
       sprintf(askme->result_line,"%s\n",target);
     else
       sprintf(askme->result_line,"%s\n",source);
     Askme_result(askme,0);
     (askme->found)++;
     if(askme->request_to_abort)
       {ret= 1; goto ex;}
   }
 }
 ret= 1;
ex:;
 Fstraversal_destroy(&trav,0);
 return(ret);
}


int Askme_list(askme,volume,flag)
struct AskmE *askme;
int volume;
int flag;
/*
 bit5= for debugging: list target (only with fsm_list)
*/
{
 int i,j,level= -1,ret;

 askme->did_something_useful= 1;
 Askme_reset_counters(askme,0);
 if(askme->root_node!=NULL) {
   if(flag&32) {
     return(Askme_fsm_list(askme,-1,0,32));
   } else if(askme->levelall && volume<=0) {
     for(level= askme->highest_level;level>=0;level--) {
       ret= Askme_fsm_list(askme,level,volume,1|(flag&32));
       if(ret<=0)
         return(ret);
       if(askme->request_to_abort)
         return(1);
     }
     return(1);
   } else 
     return(Askme_fsm_list(askme,askme->highest_level,volume,0));
 }

 for(i=0;i<askme->num_volumes;i++) {
   if(volume>0 && i!=volume-1)
 continue;
   sprintf(askme->result_line,
           "########## part %d  (%s)\n",i+1,askme->volume_info[i]);
   Askme_result(askme,0);
   for(j=0;j<askme->count_in_volume[i];j++) {
     if(askme->volumes[i][j]!=NULL) {
       sprintf(askme->result_line,"%s\n",askme->volumes[i][j]);
       Askme_result(askme,0);
       (askme->found)++;
       if(askme->request_to_abort)
         return(1);
     }
   }
 }
 return(1);
}


int Askme_id(askme,flag)
struct AskmE *askme;
int flag;
{
 int i;

 askme->did_something_useful= 1;
 for(i=0;i<askme->num_level_info;i++) {
   if(askme->level_info[i]==NULL)
 continue;
   sprintf(askme->result_line,"%s\n",askme->level_info[i]);
   Askme_result(askme,0);
   if(i==0) {
     sprintf(askme->result_line,"\n");
     Askme_result(askme,0);
   }
 }
 sprintf(askme->result_line,"\n");
 Askme_result(askme,0);
 return(1);
}


int Askme_get_level_info(askme,level,backup_date,younger_than,num_parts,flag)
struct AskmE *askme;
int level;
char *backup_date;
char *younger_than;
int *num_parts;
int flag;
{
 int i,line_level= -1;
 char *ipt;

 strcpy(younger_than,"unkown");
 strcpy(backup_date,"unknown");
 *num_parts= 0;
 askme->did_something_useful= 1;
 for(i=0;i<askme->num_level_info;i++) {
   ipt= askme->level_info[i];
   if(ipt==NULL)
 continue;
   if(strncmp(ipt,"backup date ",12)==0) {
     if(line_level==level)
 break;
     strncpy(backup_date,ipt+12,80);
     backup_date[80]= 0;
     strcpy(younger_than,"unkown");
     *num_parts= 0;
   } else if(strncmp(ipt,"-level ",7)==0) {
     sscanf(ipt+7,"%d",&line_level);
     ipt= strstr(ipt,"-younger_than ");
     if(ipt==NULL)
 continue;
     strncpy(younger_than,ipt+13,80);
     younger_than[80]= 0;
   } else if(strncmp(ipt,"this backup consists of ",24)==0) {
     sscanf(ipt+24,"%d",num_parts);
   }
 }
 if(*num_parts<=0 || strcmp(backup_date,"unknown")==0)
   return(0);
 return(1);
}


int Askme_status_result(askme,filter,fp,flag)
struct AskmE *askme;
char *filter;
FILE *fp;
int flag;
/* 
bit1= do only report to fp
*/
{
 int ret,l;

 if(filter!=NULL)
   if(filter[0]=='-') {
     l= strlen(filter);
     if(strncmp(filter,askme->result_line,l)!=0)
       return(2);
   }
 if(!(flag&2))
   Askme_result(askme,0);
 if(fp!=NULL) {
   ret= fwrite(askme->result_line,strlen(askme->result_line),1,fp);
   if(ret<=0)
     return(ret);
 }   
 return(1);
}


int Askme_status(askme,filter,fp,flag)
struct AskmE *askme;
char *filter;
FILE *fp;
int flag;
/*
 bit0= do only report non-default settings
 bit1= do only report to fp
 bit2= report current -resume status even if bit0 is set, but only if valid
 bit3= report readline history
 bit4= report -resume options indirectly as 
              -options_from_file:${resume_state_file}_pos
*/
{
 int is_default,no_defaults,is_shadow,pass,ret;
 struct FprotectoR *pr,*pr_prev;
 char *adr,*redirection,*line;

 no_defaults= flag&1;
 line= askme->result_line;

 sprintf(line,"-info_source:%s\n",askme->filename);
 Askme_status_result(askme,filter,fp,flag&2);

 if(askme->no_dir_list) {
   sprintf(line,"-no_dir_list\n");
   Askme_status_result(askme,filter,fp,flag&2);
 }

 is_default= 0;
 if(askme->read_pickiness==0)
   sprintf(line,"-read_pickiness:zero\n");
 else if(askme->read_pickiness==1)
   sprintf(line,"-read_pickiness:low\n");
 else if(askme->read_pickiness==2) {
   sprintf(line,"-read_pickiness:high\n");
   is_default= 1;
 } else
   sprintf(line,"-read_pickiness:overdone\n");
 if(!(is_default && no_defaults))
   Askme_status_result(askme,filter,fp,flag&2);

 if(askme->levelall)
   sprintf(line,"-levelall\n");
 else if(askme->nolevel)
   sprintf(line,"-nolevel\n");
 else
   sprintf(line,"-level_reset\n");
 Askme_status_result(askme,filter,fp,flag&2);

 is_default= 0;
 if(askme->volumewise==0) {
   sprintf(line,"-volumewise:off\n");
   is_default= 1;
 } else if(askme->volumewise==1)
   sprintf(line,"-volumewise:volume_ascending\n");
 else if(askme->volumewise==2)
   sprintf(line,"-volumewise:level_ascending\n");
 else if(askme->volumewise==3)
   sprintf(line,"-volumewise:on\n");
 if(!(is_default && no_defaults))
   Askme_status_result(askme,filter,fp,flag&2);

 is_default= 0;
 if(askme->sub==0)
   sprintf(line,"-fast\n");
 else if(askme->suball)
   sprintf(line,"-suball\n");
 else {
   sprintf(line,"-sub_reset\n");
   is_default= 1;
 }
 if(!(is_default && no_defaults))
   Askme_status_result(askme,filter,fp,flag&2);

 is_default= 0;
 if(askme->show_outdated==0) {
   sprintf(line,"-outdated:hide\n");
   is_default= 1;
 } else if(askme->show_outdated==2)
   sprintf(line,"-outdated:only\n");
 else
   sprintf(line,"-outdated:show\n");
 if(!(is_default && no_defaults))
   Askme_status_result(askme,filter,fp,flag&2);

 is_default= 0;
 if(askme->dialog)
   sprintf(line,"-dialog\n");
 else {
   sprintf(line,"-dialog_reset\n");
   is_default= 1;
 }
 if(!(is_default && no_defaults))
   Askme_status_result(askme,filter,fp,flag&2);

 is_default= 0;
 if(askme->german_dialog)
   sprintf(line,"-deutsch\n");
 else {
   sprintf(line,"-english\n");
   is_default= 1;
 }
 if(!(is_default && no_defaults))
   Askme_status_result(askme,filter,fp,flag&2);

 is_default= (askme->result_page_length==0 && askme->result_page_width==80);
 sprintf(line,"-page:%d:%d\n",
              (askme->result_page_length>=0?askme->result_page_length
                                           :-askme->result_page_length),
              askme->result_page_width);
 if(!(is_default && no_defaults))
   Askme_status_result(askme,filter,fp,flag&2);

 is_default= 0;
 if(askme->use_stdin)
   sprintf(line,"-use_stdin\n");
 else {
   sprintf(line,"-use_readline\n");
   is_default= 1;
 }
 if(!(is_default && no_defaults))
   Askme_status_result(askme,filter,fp,flag&2);

 is_default= !askme->packet_output;
 sprintf(line,"-pkt_output:%s\n",(askme->packet_output?"on":"off"));
 if(!(is_default && no_defaults))
   Askme_status_result(askme,filter,fp,flag&2);

 is_default= (askme->mark_text[0]==0);
 sprintf(line,"-mark:%s\n",askme->mark_text);
 if(!(is_default && no_defaults))
   Askme_status_result(askme,filter,fp,flag&2);

 sprintf(line,"-prog:%s\n",askme->progname);
 Askme_status_result(askme,filter,fp,flag&2);

 sprintf(line,"-result_format");
 if(askme->result_format&1)
   sprintf(line+strlen(line),":levhdr");
 if(askme->result_format&2)
   sprintf(line+strlen(line),":levno");
 if(askme->result_format&4)
   sprintf(line+strlen(line),":volno");
 if(askme->result_format&8) {
   if(askme->result_format&(1<<19))
     sprintf(line+strlen(line),":target_esc");
   else
     sprintf(line+strlen(line),":target");
 }
 if(askme->result_format&16) {
   if(askme->result_format&(1<<20))
     sprintf(line+strlen(line),":source_d");
   else
     sprintf(line+strlen(line),":source");
 }
 if(askme->result_format&32)
   sprintf(line+strlen(line),":expr");
 if(askme->result_format&64)
   sprintf(line+strlen(line),":mtime");
 if(askme->result_format&128)
   sprintf(line+strlen(line),":ctime");
 if(strcmp(line,"-result_format")==0)
   strcat(line,":");
 sprintf(line+strlen(line),"\n");
 Askme_status_result(askme,filter,fp,flag&2);

 if(askme->search_target)
   sprintf(line,"-search_target\n");
 else
   sprintf(line,"-search_source\n");
 Askme_status_result(askme,filter,fp,flag&2);

 if(askme->search_mode==0) {
   sprintf(line,"-search_reset\n");
 } else if(askme->search_mode==1)
   sprintf(line,"-search_fgrep\n");
 else if(askme->search_mode==2)
   sprintf(line,"-search_grep\n");
 else if(askme->search_mode==3) {
   if(askme->structured_search)
     sprintf(line,"-search_sh\n");
   else
     sprintf(line,"-search_shpat\n");
 } else if(askme->search_mode==4)
   sprintf(line,"-search_shname\n");
 Askme_status_result(askme,filter,fp,flag&2);

 sprintf(line,"-cd:%s\n",(askme->wd[0]?askme->wd:"/"));
 Askme_status_result(askme,filter,fp,flag&2);

 if(askme->media_mode!=0 || !no_defaults) {
   sprintf(line,"-media_mode");
   if(askme->media_mode&1)
     strcat(line,":on_abort_try");
   if(askme->media_mode&4)
     strcat(line,":mount_umount");
   if(askme->media_mode&16)
     strcat(line,":load_eject");
   if(strcmp(line,"-media_mode")==0)
     strcat(line,":");
   strcat(line,"\n");
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(strcmp(askme->media_load_cmd,Askme_media_load_cmD)!=0 ||
    !no_defaults) {
   sprintf(line,"-media_cmd_load:%s\n",askme->media_load_cmd);
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(strcmp(askme->media_mount_cmd,Askme_media_mount_cmD)!=0 ||
    !no_defaults) {
   sprintf(line,"-media_cmd_mount:%s\n",askme->media_mount_cmd);
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(strcmp(askme->media_umount_cmd,Askme_media_umount_cmD)!=0 ||
    !no_defaults) {
   sprintf(line,"-media_cmd_umount:%s\n",askme->media_umount_cmd);
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(strcmp(askme->media_eject_cmd,Askme_media_eject_cmD)!=0 ||
    !no_defaults) {
   sprintf(line,"-media_cmd_eject:%s\n",askme->media_eject_cmd);
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(askme->restore_mode!=Askme_restore_mode_startuP || !no_defaults) {
   sprintf(line,"-restore_mode");
   if(askme->restore_mode&8)
     strcat(line,":on_search");
   if(askme->restore_mode&2)
     strcat(line,":use_target");
   if(askme->restore_mode&32)
     strcat(line,":simulate_existence");
   if(askme->restore_mode&64)
     strcat(line,":simulate_action");
   if(askme->restore_mode&512)
     strcat(line,":simulate_mount");
   if((askme->restore_mode&384)==0)
     strcat(line,":silent_file");
   else if((askme->restore_mode&384)==256)
     strcat(line,":verbous_volume");
   else if((askme->restore_mode&384)==384)
     strcat(line,":verbous_action");
   if(askme->restore_mode&1024)
     strcat(line,":open_dir_for_owner");
   if(((askme->restore_mode>>11)&15)==0)
     strcat(line,":overwrite_off");
   if(((askme->restore_mode>>11)&15)==1)
     strcat(line,":overwrite_by_unlink");
   if(((askme->restore_mode>>11)&15)==2)
     strcat(line,":overwrite_by_truncation");
   if(((askme->restore_mode>>11)&15)==3)
     strcat(line,":overwrite_by_move");
   if(askme->restore_mode&(1<<15))
     strcat(line,":keep_ownership");
   if(((askme->restore_mode>>16)&7)==1)
     strcat(line,":on_error_ignore");
   else if(((askme->restore_mode>>16)&7)==2)
     strcat(line,":on_error_exit");
   else if(((askme->restore_mode>>16)&7)==3)
     strcat(line,":after_error_exit");
   else if(((askme->restore_mode>>16)&7)==4)
     strcat(line,":on_error_prompt");
   if(askme->restore_mode&(1<<19))
     strcat(line,":fresh_timestamps");
   if(askme->restore_mode&(1<<20))
     strcat(line,":update_state_sparse");
   if(askme->restore_mode&(1<<21))
     strcat(line,":update_state_dsync");
   if(askme->restore_mode&(1<<22))
     strcat(line,":on_error_keep");
   if(askme->restore_mode&(1<<23))
     strcat(line,":obey_permissions");
   if(askme->restore_mode&(1<<30))
     strcat(line,":disable_restore");
   if(askme->restore_mode&(1<<31))
     strcat(line,":ban_restore");
   if(strcmp(line,"-restore_mode")==0)
     strcat(line,":");
   strcat(line,"\n");
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(askme->restore_from[0]!=0 || !no_defaults) {
   sprintf(line,"-restore_from:%s\n",askme->restore_from);
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(askme->restore_to[0]!=0 || !no_defaults) {
   sprintf(line,"-restore_to:%s\n",askme->restore_to);
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(askme->restore_skip[0]!=0 || !no_defaults) {
   sprintf(line,"-restore_skip:%s\n",askme->restore_skip);
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(askme->restore_asylum[0]!=0 || !no_defaults) {
   sprintf(line,"-restore_asylum:%s\n",askme->restore_asylum);
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(askme->restore_pre_prompt[0]!=0 || !no_defaults) {
   sprintf(line,"-restore_pre_prompt:%s\n",askme->restore_pre_prompt);
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(askme->restore_post_prompt[0]!=0 || !no_defaults) {
   sprintf(line,"-restore_post_prompt:%s\n",askme->restore_post_prompt);
   Askme_status_result(askme,filter,fp,flag&2);
 }

 for(pass=0; pass<1+!no_defaults; pass++) {
   Fprotector_chain_end(askme->protections,&pr,0);
   for(;pr!=NULL;pr= pr_prev) {
     ret= Fprotector_get_par(pr,&adr,&redirection,&is_shadow,&pr_prev,1);
     if(adr==NULL || ret!=1)
   continue;
     if(redirection!=NULL) {

       /* >>> option for redirection is still undecided */;

   continue;
     } else if(is_shadow) {
       if(pass!=1)
   continue;
       sprintf(line,"-shadow:%s\n",adr);
     } else {
       if(pass!=0)
   continue;
       sprintf(line,"-protect:%s\n",adr);
     }
     Askme_status_result(askme,filter,fp,flag&2);
   }
 }

 if(askme->status_history_max!=Askme_status_history_maX || !no_defaults) {
   sprintf(line,"-status_history_max:%d\n",askme->status_history_max);
   Askme_status_result(askme,filter,fp,flag&2);
 }

 if(askme->abort_handler!=Askme_abort_handler_defaulT || !no_defaults) {
   sprintf(line,"-abort_handler");
   if((askme->abort_handler&15)==0)
     strcat(line,":off");
   else if((askme->abort_handler&15)==1)
     strcat(line,":on");
   else if((askme->abort_handler&15)==2)
     strcat(line,":ignore");
   if(((askme->abort_handler>>4)&15)==0)
     strcat(line,":handler_abort");
   else if(((askme->abort_handler>>4)&15)==1)
     strcat(line,":handler_repeat");
   else if(askme->abort_handler&256)
     strcat(line,":extra_verbous");
   if(strcmp(line,"-abort_handler")==0)
     strcat(line,":");
   strcat(line,"\n");
   Askme_status_result(askme,filter,fp,flag&2);
 }

#ifdef Askme_with_readlinE

 if((flag&8) && askme->status_history_max>0) {
   HIST_ENTRY **hl;
   int hc,i;

   hl= history_list();
   if(hl!=NULL) {
     for(hc= 0;hl[hc]!=NULL;hc++);
     if(hc>0)
       if(strcmp(hl[hc-1]->line,"-end")==0)
         hc--;
     if(hc>=askme->status_history_max)
       i= hc-askme->status_history_max;
     else
       i= 0;
     for(;i<hc;i++) {
       sprintf(line,"-history:%s\n",hl[i]->line);
       Askme_status_result(askme,filter,fp,flag&2);
     }   
   }
 }

#endif /* Askme_with_readlinE */


 /* This should stay the last output lines */
 if(askme->resume_state_file[0]!=0 || !no_defaults) {
   sprintf(line,"-resume_state_files:%s\n",askme->resume_state_file);
   Askme_status_result(askme,filter,fp,flag&2);
 }
 if(((flag&4) && askme->resume_last_level>=0 && askme->resume_last_volume>0)
    || !(no_defaults || (flag&4))) {
   if(flag&16) {
     sprintf(line,"-options_from_file:%s_opt\n",askme->resume_state_file);
     Askme_status_result(askme,filter,fp,flag&2);
     sprintf(line,"-options_from_file:%s_src\n",askme->resume_state_file);
     Askme_status_result(askme,filter,fp,flag&2);
     sprintf(line,"-options_from_file:%s_dirty\n",askme->resume_state_file);
     Askme_status_result(askme,filter,fp,flag&2);
     sprintf(line,"-options_from_file:%s_lv\n",askme->resume_state_file);
     Askme_status_result(askme,filter,fp,flag&2);
   } else {
     sprintf(line,"-resume_opt_text:%s:%s\n",
             (askme->resume_was_restore?"-restore":"-search"),
             askme->resume_search_text);
     Askme_status_result(askme,filter,fp,flag&2);
     sprintf(line,"-resume_last_source:%s\n",askme->resume_last_source);
     Askme_status_result(askme,filter,fp,flag&2);
     sprintf(line,"-resume_last_dirty:%s\n",askme->resume_last_dirty);
     Askme_status_result(askme,filter,fp,flag&2);
     sprintf(line,"-resume_last_lv:%d:%d\n",
             askme->resume_last_level,askme->resume_last_volume);
     Askme_status_result(askme,filter,fp,flag&2);
   }
 }
 return(1);
}


int Askme_exec(askme,cmd,flag)
struct AskmE *askme;
char *cmd;
int flag;
{
 char *spt,*cpt,**argv= NULL,errmsg[2*4096];
 int is_done= 0,argc= 0,widx= 0,cmd_l,pass,ret;

 cmd_l= strlen(cmd);
 for(pass=0;pass<2;pass++) {
   is_done= 0;
   widx= 0;
   for(spt= cmd;!is_done;spt= cpt+1) {
     for(cpt= spt;*cpt!=0 && *cpt!=':' && *cpt!=' ';cpt++);
     if(*cpt==0)
       is_done= 1;
     if(pass==0) {
       argc++;
     } else {
       *cpt= 0;
       if(Sregex_string(&(argv[widx]),spt,0)<=0)
         {ret= -1; goto ex;}
       widx++;
     }
   }
   if(pass==0) {
     if(argc==0)
       {ret= 2; goto ex;}
     argv= TSOB_FELD(char *,argc+1);
     if(argv==NULL)
       {ret= -1; goto ex;}
     for(widx= 0;widx<argc+1;widx++)
       argv[widx]= NULL;
   }
 }
 ret= Exec_cmd(askme->progname,cmd,NULL,argv,errmsg,!!(askme->packet_output));
 if(ret<0)
   goto ex;
 if(ret>0) {
   for(widx=0; widx<cmd_l; widx++)
     if(cmd[widx]==0)
       cmd[widx]= ' ';
   sprintf(askme->info_text,"--- failed : %s\n",cmd);
   Askme_info(askme,0);
   if(errmsg[0]!=0) {
     sprintf(askme->info_text,"--- message:  %s\n",errmsg);
     Askme_info(askme,1);
   }
   {ret= 0; goto ex;}
 }
 ret= 1;
ex:
 if(argv!=NULL) {
   for(widx=0;widx<argc;widx++)
     Sregex_string(&(argv[widx]),NULL,0);
   free((char *) argv);
 }
 return(ret);
} 


int Askme_options_from_file(askme,adr,flag)
struct AskmE *askme;
char *adr;
int flag;
/*
 bit0= called from Askme_prescan_args, therefore execute via that same function
*/
/*
return:
 <=0 error , 1 = success , 3 = end program run
*/
{
 int ret,linecount= 0;
 FILE *fp= NULL;
 char line[4*4096],*argv[3],shellsafe[5*4096];

 if(adr[0]==0) {
   Askme_de_eN(
     sprintf(askme->info_text,"--- Leerer Filename bei -options_from_file\n");
   ,
     sprintf(askme->info_text,"--- Empty file name with -options_from_file\n");
   )
   Askme_info(askme,0);
   return(0);
 }
 Text_shellsafe(adr,shellsafe,0);
 argv[0]= askme->progname;
 argv[1]= askme->filename;
 argv[2]= line;
 fp= Afile_fopen(adr,"rb",((!!askme->packet_output)<<6));
 if(fp==NULL)
   return(0);
 while(1) {
   if(Sfile_fgets(line,sizeof(line),fp)==NULL) {
     ret= 1;
     if(ferror(fp))
       ret= 0;
 break;
   }
   linecount++;
   if(line[0]==0 || line[0]=='#')
 continue;
   if(flag&1) {
     ret= Askme_prescan_args(askme,3,argv,1);
     if(ret==0)
       {ret= 3; goto ex;}
     if(ret<0)
       goto ex;
   } else {
     if(askme->is_dialog) {
       sprintf(askme->info_text,"+ line %d of file %s :\n+ %s\n",
                                linecount,shellsafe,line);
       Askme_info(askme,1);
     }
     ret= Askme_execute_option(askme,line,1|(1<<16));
     if(ret<=0 || ret==3)
       goto ex;
   }
 }
ex:;
 Askme_reset_counters(askme,1);
 if(fp!=NULL)
   fclose(fp);
 if(ret<=0) {
   Askme_de_eN(
     sprintf(askme->info_text,
             "--- Fehler ausgeloest durch Zeile %d in File:\n    %s\n",
             linecount,shellsafe);
   ,
     sprintf(askme->info_text,
             "--- error triggered by line %d of file:\n    %s\n",
             linecount,shellsafe);
   )
   Askme_info(askme,1);
 }
 return(ret);
}


int Askme_help(askme,flag)
struct AskmE *askme;
int flag;
{
 static char text[][80]={

#ifdef Askme_no_helP

"This binary program does not contain a help text.",
"Read doc/scdbackup_askme_help if available. Else install scdbackup",
"without executing ./DIET afterwards.",

#else

"This program reads information about a backup created by scdbackup. It then",
"allows to query those informations and eventually assists in restoring",
"files from backup to disk.",
"Information source may be one of the following:",
"* a backup configuration directory containing files ./highest_level ,",
"  ./level_*/content_file_list.gz , ./level_*/vanished_file_list.gz",
"  and ./level_*/complete_dir_list.gz .",
"* a set of media containing directory ./added_by_scdbackup with files",
"  content_file_list.gz , vanished_file_list.gz and complete_dir_list.gz .",
"  The media get prompted.",
"* a file created by scdbackup option -content_list_adr",
"  or -complete_list_adr.",
"* an ASKME script together with eventually available ASKME_level_* from the",
"  same directory which represent older levels of the backup.",
"",
"General options (search and restore options are listed separately):",
"  -help       prints this text",
"  -hilfe      gibt diesen Text in deutscher Sprache aus",
"              (prints this text in german language)",
"  -id         prints a message about this backup",
"  -count      prints the number of volumes",
"  -list       prints the parts and their items",
"  -list:#     prints a single part (e.g. -list:3)",
"  -levelall   lookup all levels (else: show only youngest match)",
"  -nolevel    do not look in older levels",
"  -level_reset  revoke -levelall or -nolevel",
"  -volumewise[:mode]  controls output order of volumes and levels. Modes:",
"                volume_ascending ... report volumes in ascending order", 
"                levels_ascending ... report levels in ascending order",
"                on  [or no mode] ... volumes and levels in ascending order", 
"                off ...... levels in descending order, no volume sorting",
"  -suball     lookup all sub items and levels (else: show one per part)",
"  -fast       do not look for sub items",
"  -sub_reset  revoke -suball or -fast",
"  -outdated:code  how to deal with file objects which are still on some",
"              some backup media but removed or replaced on disk:",
"                hide ... do not deliver outdated file objects",
"                show ... deliver valid and outdated file objects",
"                only ... do not deliver valid file objects",
"  -dialog     after all arguments are processed, enter dialog mode.",
"              In this mode you may enter searchtexts or any of the options",
"              described here. One per line.",
"  -dialog_reset  revoke -dialog (works only if given as argument)",
"  -deutsch    issue dialog messages in german language",
"  -english    issue dialog messages in english language (default)",
"  -page:len[:width]  prompt user after len output lines (0=no prompt).",
"              width (default 80) can adjust line number computation",
"              to the output terminal's line width.",
#ifdef Askme_with_readlinE
"  -use_stdin  use raw standard input even if libreadline is available",
"  -use_readline  use libreadline for dialog if available",
"  -history:textline  copy textline into libreadline history. This command",
"              itself is not copied to the history list.",
#endif /* Askme_with_readlinE */
"  -pkt_output[:on|:off]  direct output to stdout and prefix each line by a",
"              short header which tells channel id and a mode number.",
"              Each such output packet is finalized by a newline.",
"              Channel ids are 'R:' for result lines, 'I:' for notes",
"              and error messages, 'M:' for -mark texts. Bit 0 of the",
"              mode number tells wether the newline is also part of the",
"              packet payload. Example of a info message with newline:",
"                I:1: enter option or search text :",
"              -pkt_output:on is intended for use by frontend programs.",
"  -logfile:channel:fileaddress  copy output of a channel to the given file.",
"              channel may be 'R','I','M' as with -pkt_output or '.'",
"              for the consolidated -pkt_output stream.",
"  -mark:text  if text is not empty it will get put out each time an",
"              option or search is completed.",
"  -prog:text  use text as this program's name in subsequent messages",
"  -prog_help:text  use text as this program's name and perform -help",
"  -prog_hilfe:text use text as this program's name and perform -hilfe",
"  -result_format:code:...:code  compose format of result output from",
"                levhdr ... level headline",
"                levno .... level number in result lines",
"                volno .... volume number",
#ifdef Askme_fsnode_with_timeS
"                mtime .... content modification time (if available)",
"                ctime .... inode modification time (if available)",
#endif /* Askme_fsnode_with_timeS */
"                target ... target address (content list only)",
"                target_esc  target address with escaped '=' and '\\'",
"                source ... source address",
"                expr ..... search text resp. expression ",
"              Defaults: ASKME  -result_format:levhdr:volno:source:expr",
"                Content lists  -result_format:levhdr:volno:target:source",
"  -read_pickiness:code  behavior with problems when reading infosource:",
"                 zero ... try to keep on reading as long as possible",
"                 low .... ignore minor inconsistencies in the infosource",
"                 high ... abort on the first inconsistency in infosource",
"              Only the first -read_pickiness argument gets effective.",
"  -no_dir_list  Do not read  complete_dir_list.gz",
"  -media_mode:[+-]:option:...  set rules how to handle media when the",
"              program demands a particular volume. Options are:",
"                 mount_umount   use external commands to unmount the",
"                                previous volume and to mount the new one.",
"                 load_eject     use external commands to unload the",
"                                previous volume and to load the new one.",
"                 on_abort_try   perform unload even in abort situations",
"  -media_cmd_mount:command  a command and eventual fixed arguments to be",
"              executed rather than command mount. It will get the mount",
"              point directory as additional argument.",
"  -media_cmd_umount:command  like -media_cmd_mount but replacing command",
"              umount",
"  -media_cmd_load:command  like -media_cmd_mount but replacing command",
"              eject -t",
"  -media_cmd_eject:command  like -media_cmd_mount but replacing command",
"              eject",
"  -status[:mode|:filter]  report the current settings of persistent options.",
"              Modes:",
"                 short... print only important or altered options",
"                 long ... print options even if they have default settings",
"                 long_history  like long plus -history: lines",
"              Filters begin with '-' and are compared literally against the",
"              output lines of -status:long_history. A line is put out only",
"              its start matches the filter. Try these example filters",
"                 -status:-res  ,  -status:-restore  ,  -status:-restore_mode",
"              If neither mode nor filter is given, mode 'short' is assumed.",
"  -status_history_max:number  maximum number of history lines to be reported",
"              with -status:long_history",
"  -options_from_file:fileaddress  liest Zeilen aus dem angegebenen File und",
"              fuehrt sie als Programmoptionen aus.",
"  -abort_handler:option:...  controls handling of program abort by system",
"              signals. Anything other than \"on\" or \"on:extra_verbous\"",
"              should only be used in case of severe system deficiencies:",
"                on .... signals get caught, unclean file states get cleared",
"                        as far as possible and then the program ends",
"                off ... signals lead to immediate end of the program",
"                ignore  signals are rejected as far as possible and the",
"                        program continues. This may cause single files",
"                        to have corrupted content.",
"                handler_abort  abort immediately on further signals which",
"                        arrive after start of abort handling",
"                handler_repeat  restart abort handling on further signals",
"                extra_verbous  report about the particular actions and",
"                        decisions of abort handling",
"  # any text  is ignored. In dialog mode the input line will be stored in",
"              the eventual readline history, nevertheless.", 
"  -info_source:addresse  gets only into effect if given as first argument.",
"              Usually the prefix -info_source: is not needed.",
"  -version    nennt Programm und Versionsnummer",
"  -end        end program immediately",
"",
"The options -help, -hilfe, -prog_help, -prog_hilfe are recognized even if",
"given as first argument instead of an infosource. Options with parameters",
"accept not only ':' but also ' ' as separator. ' ' might be convenient in",
"dialog but has to be protected from the shell parser if given as argument.",
"",
"Option -page causes a user prompt after the given number of result lines.",
"Empty input resumes output until the next prompt. Other input may be:",
"  @     suppresses pageing until the current option or search is done",
"  @@    suppresses further result output but continues the search ",
"  @@@   aborts the current search or option",
"  @skip suppresses -volumewise matching and -restore of the current volume",
"  other aborts the current search or option and executes input as new",
"        search or option",
"",
"                                  Searching",
"",
"Depending on the information source the search capabilities differ:",
"* With ASKME, filenames resp. directories are looked up in the volume",
"  item lists. If the searchtext or a directory above is found then a result",
"  line according to -result_format is printed. If filename is a directory",
"  and an item below this address is found, then the action depends on",
"  -suball and -fast.",
"  The search functions only apply to the lines which are visible with -list.",
"  Therefore one cannot search in ASKME for a file name of which only a",
"  parent directory is listed.",
#ifdef Askme_with_regeX
"  Pattern matching search modes are available but of limited use.",
#endif /* Askme_with_regeX */
"* Information read from content-vanish-list files forms a detailed tree",
"  model in which any file object can be found that is covered by the lists.",
"  Additionally this model can tell the target addresses on the backup media",
"  and not only the source addresses on hard disk.",
#ifdef Askme_with_regeX
"  This model is well suited for pattern matching search modes. Therefore",
"  -search_sh is the default mode after loading it.",
#else /* ! Askme_with_regeX */
"  This model is well suited for structured search. Therefore -search_sh",
"  is the default mode after loading it.",
#endif /* Askme_with_regeX */
"  If a directory address is matched it will be reported for any volume that",
"  contains one of its non-directory fileobjects. With option -suball",
"  every single non-directory below the found one will be reported.",
"If nothing is found on the newest level then usually the older levels are",
"searched. This behavior depends on -levelall and -nolevel .",
"",
"The search is done on either the source addresses as they were given at",
"backup time or the target adresses as they are in the backup volumes.",
"  -search_source search on source addresses (default)",
"                 This option also adds :source: to -result_format.",
"  -search_target search on target addresses. Targets may get listed multiple",
"                 times if they come from different sources.",
"                 This option also adds :target: to -result_format.",
"  -search:searchtext  prevent searchtext from being understood as option",
"                 or comment (usually the prefix -search: is not needed)",
"Search comparison modes:",
"  -search_fgrep  try to find searchtext as arbitrary piece of source",
#ifdef Askme_with_regeX
"  -search_grep   searchtext is regular expression  (^\\/vids\\/.*)",
"  -search_sh     searchtext is shell parser pattern    (/vids/*)",
"  -search_shpat  like -search_sh but without the special handling of '/'",
#else /* Askme_with_regeX */
"  -search_sh     searchtext is a structured pattern    (/vids/*)",
"                 Due to the lack of regular expressions in this program",
"                 version, only complete filenames or a single '*' can be",
"                 used. The latter matches any filename.",
#endif /* ! Askme_with_regeX */
"  -search_shname like -search_sh but for leafname, not for absolute path.",
"                 This mode searches further below matched directories.",
"  -search_reset  search by comparing start pieces of searchtext and items",
"Navigation options:",
"  -cd:searchtext defines the working directory for -search_sh by the first",
"                 directory match of searchtext. It is prepended to any",
"                 searchtext which does not begin with '/'. -search_sh and",
"                 related options handle file names '..' and '.' as usual",
"                 in unix-style filesystems.",
"  -cd            resets working directory to the read-in working directory",
"                 resp. in -search_target mode to '/'",
"  -pwd           tells the current working directory",
"  -ls:searchtext gives an overview of known addresses. searchtext is like",
"                 with -search_sh. All levels and volumes are reported but",
"                 any match is reported only once. Matched directories list",
"                 their complete content.",
"  -ls            same as -ls:.",
"  -ls_d:searchtext like -ls but with no directory content reported. Empty",
"                 directories get reported (unlike -ls or normal search).",
"  -ls_d          same as -ls_d:*  (and nearly the same as -ls)",
"  -find:searchtext  lists matching files below current working directory.",
"                 searchtext is as with -search_shname. Empty directories",
"                 get listed.",
"  -find          same as -find:*",

#endif /* ! Askme_no_helP */

"@ENDE_OF_HELPTEXT_(HOPEFULLY_UNIQUELY_SILLY_TEXT)@"
 };
 static char restore_text[][80]={

#ifndef Askme_no_helP

"",
"                                Restoring",
"",
"--------------------------------------------------------------------------",
"Important: Restoring will eventually temporarily widen the owner's access",
"           permissions of existing files. You cannot protect an *own* file",
"           from overwriting by taking away any permissions of the file",
"           itself or of your *own* directories above. Since the superuser",
"           usually has ownership privileges with all files, no permission",
"           settings can protect a file from being overwritten by the",
"           superuser (but -restore_mode:+:overwrite_off or -protect will do)",
"--------------------------------------------------------------------------",
"Restore is currently suitable only for runs based on a tree model and only",
"for mountable ISO-9660 backups. It needs to be enabled by",
"  -restore_mode:-:disable_restore",
"or by any -restore_mode: that does not start with '+' or '-'.",
"",
"Restore operations copy from backup address to restore address.",
"This text refers \"source\" and \"target\" addresses in quotation marks",
"because these terms stem from the backup situation. Now in the restore",
"situation their meaning gets inverted. \"target\" is actually part of the",
"backup address (i.e. the origin of the copy). By default \"source\" is a",
"decisive part of the restore address (i.e. the destination of the copy).",
"Restore options:",
"  -restore_from:path  defines the directory path which is prepended to the",
"                 \"target\" address to form the backup address.",
"                 path is also used for eventual load-mount-umount-eject",
"                 operations.",
"  -restore_skip:path  defines the starting part of the restore address",
"                 which is to be cut off before -restore_to does its work.",
"                 The restore address is either derived from \"source\" or",
"                 from \"target\", depending on -restore_mode:use_target.",
"                 If the skip path does not match the start of the restore",
"                 address then this considered an error as if the file",
"                 would not exist on the backup media.",
"  -restore_to:path    defines the directory path which is prepended to the",
"                 emerging restore address after eventual -restore_skip",
"                 has cut off its part.",
"  -restore_asylum:path  defines the location where restore mode",
"                 overwrite_by_move shall save the overwritten files. The",
"                 asylum address is obtained by prepending path to the full",
"                 absolute address of the overwritten file object.",
"  -restore:searchtext  locate matching file objects and copy them to the",
"                 according restore address. Matching directories get",
"                 restored including all their sub files.",
"                 Matching is done by the same rules as with -search",
"  -restore_mode:[+-]:option:...  sets properties of restore operations.",
"                 If the first character after -restore_mode: is '+' then",
"                 all mentioned options get added to the existing properties.",
"                 If that first character is '-' then _all_ mentioned options",
"                 get reset to default. Else the whole -restore_mode is reset", 
"                 before the mentioned options get added. Options are:",
"                   on_search      perform restore operations with any search",
"                                  (else only with -restore:...)",
"                   use_target     compose restore address from \"target\"",
"                   silent_file     do not report about single files",
"                   verbous_volume  report about mount_umount,load_eject",
"                   verbous_action  report in detail about actions",
"                   open_dir_for_owner  rwx-permission for directory owner",
"                   overwrite_off           do not overwrite existing files",
"                   overwrite_by_unlink     overwrite existing files, unlink",
"                   overwrite_by_truncation truncate and overwrite data files",
"                   overwrite_by_move  move existing files to -restore_asylum",
"                   obey_permissions do not widen owner permissions of parent",
"                                    directories. This can hamper restore !",
"                   disable_restore  disable restore operations temporarily",
"                   ban_restore      disable restore operations forever",
"                   on_error_ignore  continue operation after partial error",
"                   on_error_prompt  ask for instructions after partial error",
"                   on_error_exit    abort program after partial error",
"                   after_error_exit continue operation, abort when done",
"                   on_error_keep    do not remove incompletely restored file",
"                                    after error",
"                   keep_ownership   superuser shall not adjust owner,group",
"                   fresh_timestamps do not copy atime and mtime from backup",
"                   update_state_sparse  avoid to update -resume_state_file",
"                                    with each file. Update at dialog time.",
"                                    This devalues _dirty file tracing.",
"                   update_state_dsync  push the _dirty info to disk by fsync",
"                                    This is slow and puts load on the disk.",
"                   simulate_existence  debugging option",
"                   simulate_action     debugging option",
"                   simulate_mount      debugging option",
"  -restore_pre_prompt:command  a program and eventual fixed arguments",
"                 to be executed before the prompt for a new backup volume.",
"                 Four arguments will get appended: level, volume,",
"                 backup address of next match, and its restore address.",
"                 If command begins with '@' then it will be exceuted even",
"                 if -restore_mode:+:simulate_mount is set.",
"  -restore_post_prompt:command  like -restore_pre_prompt but to be executed",
"                 after the user input to the prompt for a new backup volume.",
"  -resume[:level[:volume[:restore_or_search_option]]]  resumes a previous",
"                 -restore or volumewise -search at the level and volume",
"                 where it was aborted. If level is given, the operation is",
"                 resumed at volume 1 of this level, unless a volume is given",
"                 as second argument. If the next component begins with",
"                 \"-restore:\" or \"-search:\" then this operation is",
"                 resumed instead of the most recent operation.",
"  -resume_last_source:address  sets the absolute source address of the file",
"                 object where exactly to resume within the given level and",
"                 volume. If the address is not matched in the first volume",
"                 of a resumed operation, then this first volume gets skipped",
"                 and the following volumes get restored normally.",
"  -resume_last_dirty:address  sets the absolute restore address of an",
"                 eventual file object that is neither in its pre-restore",
"                 state nor restored properly. The file object itself is",
"                 completely questionable. The directories above might have",
"                 too generous permissions for their owner (nobody else) or",
"                 (if created newly) too sparse permissions for anybody.",
"                 On next -resume, a non-empty last dirty address causes a",
"                 warning. In dialog mode a confirmation request is made.",
"                 In non-dialog mode the program aborts.",
"  -resume_last_lv:level[:volume]   sets the level and eventually volume for",
"                 the next -resume if this has no own level given.",
"  -resume_opt_text:restore_or_search_option  sets the default operation to",
"                 be performed by the next -resume if this has no own",
"                 restore_or_search_option given.",
"  -resume_state_files:fileaddress  defines a family of files which will get",
"                 updated at various stages of -restore or volumwise search.",
"                 fileaddress is later suitable for -resume_from_file.",
"                 filedaddress itself will hold a -status:long_history report",
"                 which defines the search environment and calls the other",
"                 files of the family, which get updated more often:",
"                   fileaddress\"_opt\" contains current -resume_opt_text",
"                   fileaddress\"_src\" contains current -resume_last_source",
"                   fileaddress\"_lv\"  contains -resume with current level",
"                                     and volume.",
"                   fileaddress\"_dirty\" contains current -resume_last_dirty",
"                 The reason for this scattering is to simplify manipulation",
"                 of the resume parameters.",
"                 Note: the commands s*backup_askme use an own file family",
"                 tmp/askme_resume_state* and will get confused if you use",
"                 -resume_state_files yourself.",
"  -resume_from_file:fileaddress  much like -options_from_file but allowed",
"                 to be the first argument of the program and able to set",
"                 the info source to be loaded. fileaddress should point to",
"                 a family of files created by  -resume_state_files .",
"  -resume_state_to_file  write current (possibly artificial) resume state",
"                 to state files.", 
"  -protect:path  prevent restore address path and any restore below path even",
"                 if overwriting in general is allowed. If a file object with",
"                 address path exists, then not only the literal path is",
"                 protected but also any other path that leads to this",
"                 file object's inode and device number. If path leads to",
"                 a symbolic link, the link target is protected, too.",
"                 The protection of link target and inode is recorded when",
"                 the option is executed. Subsequently changed link targets",
"                 or inodes on disk will not be protected by that aspect of",
"                 protection. Literal path, the recorded link targets, and",
"                 the recorded inodes are still protected then.",
"                 Protections given as program arguments get recorded before",
"                 any restore or search operation is performed. The number of",
"                 protections is not limited. Nevertheless: use this option",
"                 sparsely because it is expensive at restore time.",
"  -shadow:path   like -protect but not causing an error condition if a",
"                 restore operation gets rejected. Thus a restore will",
"                 continue even if in -restore_mode:-:on_error_ignore .",
"                 -shadow is intended for applications as protection from",
"                 unwantedly overwriting themselves.",
"                 -shadow options are reported only with -status:long",

#endif /* ! Askme_no_helP */

"@ENDE_OF_HELPTEXT_(HOPEFULLY_UNIQUELY_SILLY_TEXT)@",
""
 };
 static char no_restore_text[][80]={

#ifndef Askme_no_helP
"",
"--------------------------------------------------------------------------",
"Note: Restore operations have been irrevocably blocked by option",
"        -restore_mode:ban_restore",
"--------------------------------------------------------------------------",
#endif /* ! Askme_no_helP */

"@ENDE_OF_HELPTEXT_(HOPEFULLY_UNIQUELY_SILLY_TEXT)@",
""
 };
 char *tpt;
 int i,pass;

/*
 if(askme->level_info!=NULL) 
   Askme_id(askme,0);
*/
 Askme_restxt(askme,"\n"); 
 sprintf(askme->result_line,"usage: %s infosource [options|searchtexts]\n",
         askme->progname); 
 Askme_result(askme,0);
 Askme_restxt(askme,"\n"); 
 for(pass=0;pass<2;pass++) {
   for(i=0;1;i++) {
     if(pass==0)
       tpt= text[i];
     else if(askme->restore_mode&(1<<31))
       tpt= no_restore_text[i];
     else
       tpt= restore_text[i];
     
     if(strcmp(tpt,"@ENDE_OF_HELPTEXT_(HOPEFULLY_UNIQUELY_SILLY_TEXT)@")==0)
   break;
     sprintf(askme->result_line,"%s\n",tpt);
     Askme_result(askme,0);
     if(askme->request_to_abort)
       return(1);
   }
 }
 Askme_restxt(askme,"\n"); 

#ifndef Askme_no_helP

 Askme_restxt(askme,"Example (search oak.jpg on all backup levels) :\n");
 sprintf(askme->result_line,
         "  %s /cdrom/ASKME -levelall /home/pictures/trees/oak.jpg\n",
         askme->progname);
 Askme_result(askme,0);
 Askme_restxt(askme,"\n");
 Askme_restxt(askme,
         "Example (load incremental configuration, start dialog):\n");
 sprintf(askme->result_line,
         "  %s $HOME/backup_configuration -levelall -dialog\n",
         askme->progname);
 Askme_result(askme,0);
 Askme_restxt(askme,"Dialog input:\n");

#ifdef Askme_with_regeX

 Askme_restxt(askme,"  /*/*/trees/oak*\n");
 Askme_restxt(askme,"  -search_shpat\n");
 Askme_restxt(askme,"  *trees/oak*\n");
 Askme_restxt(askme,"  -search_shname\n");
 Askme_restxt(askme,"  oak*\n");
 Askme_restxt(askme,
 "More complex patterns: Filenames which begin with A,B,C,K,W, or Z either\n"); 
 Askme_restxt(askme,
 "in capital or in small letters, the second character may not be a digit,\n");
 Askme_restxt(askme,"the suffix must be 3 characters:\n");
 Askme_restxt(askme,"  [A-CKWZa-ckwz][!0-9]*.???\n");
 Askme_restxt(askme,
  "Filenamess which consist of \"DSC\", 5 digits and a 3 character suffix:\n");
 Askme_restxt(askme,"  DSC[0-9][0-9][0-9][0-9][0-9].???\n");

#else /* Askme_with_regeX */

 Askme_restxt(askme,"  /*/*/trees/oak.jpg\n");
 Askme_restxt(askme,"  -search_shname\n");
 Askme_restxt(askme,"  oak.jpg\n");

#endif /* ! Askme_with_regeX */

 if(!(askme->restore_mode&(1<<31))) {
   Askme_restxt(askme,"\n");
   Askme_restxt(askme,
 "Restore directory with original source address  /home/pictures/tree\n");
   Askme_restxt(askme,
 "together with all its files from backup volumes back to disk:\n");
   Askme_restxt(askme,"  -restore_from:/cdrom\n");
   Askme_restxt(askme,"  -restore_skip:/home/pictures\n");
   Askme_restxt(askme,"  -restore_to:/home/test/restored_pictures\n");
   Askme_restxt(askme,"  -media_mode:load_eject:mount_umount\n");
   Askme_restxt(askme,"  -restore_mode:overwrite_off\n");
   Askme_restxt(askme,"  -restore:/home/pictures/trees\n");
   Askme_restxt(askme,
 "This will create directory  /home/test/restored_pictures/trees .\n");
   Askme_restxt(askme,
  "You will get prompted for backup volumes to be inserted in /cdrom .\n");
   Askme_restxt(askme,
  "It will not be possible to overwrite existing files.\n");
 }

 Askme_restxt(askme,"\n");
 Askme_restxt(askme,
         "Example (load incremental information from backup media):\n");
 sprintf(askme->result_line,
         "  %s /cdrom/added_by_scdbackup -levelall -dialog\n",
         askme->progname);
 Askme_result(askme,0);

#endif /* ! Askme_no_helP */

 return(1);
}


int Askme_hilfe(askme,flag)
struct AskmE *askme;
int flag;
{
 static char text[][80]={

#ifdef Askme_no_german_helP

#ifdef Askme_no_helP
"In diesem Binaerprogramm ist kein Hilfetext enthalten.",
"Lesen Sie den englischsprachigen Hilfetext doc/scdbackup_askme_help oder",
"installieren Sie scdbackup ohne es hinterher mit ./DIET abzuspecken.",
#else
"In diesem Binaerprogramm ist keine deutschsprachige Hilfe enthalten.",
"Lesen Sie den englischsprachigen Hilfetext mit Option -help oder",
"installieren Sie scdbackup ohne es hinterher mit ./DIET abzuspecken.",
#endif

#else /* Askme_no_german_helP */

"Dieses Programm liest Informationen ueber einen Backup, erlaubt dann diese",
"zu durchsuchen und hilft ggf. auch beim Rueckspeichern vom Backup auf die",
"Platte.",
"Als Informationsquelle kommt eine der folgenden in Frage:",
"* ein Backup Konfigurationsdirectory mit den Files ./highest_level ,",
"  ./level_*/content_file_list.gz , ./level_*/vanished_file_list.gz",
"  und ./level_*/complete_dir_list.gz .",
"* ein Satz Datentraeger mit Directory ./added_by_scdbackup und darin",
"  content_file_list.gz , vanished_file_list.gz und complete_dir_list.gz .",
"  Die Datentraeger werden im Dialog angefordert.",
"* ein File, erzeugt mit scdbackup-Option -content_list_adr",
"  oder -complete_list_adr.",
"* ein ASKME-Skript zusammen mit evtl. vorhandenen ASKME_level_* aus dem",
"  selben Directory, die dann die aelteren Levels eines Backups darstellen.",
"",
"Allgemeine Optionen (Such- und Rueckspeicheroptionen siehe weiter unten):",
"  -hilfe      gibt diesen Text aus",
"  -help       prints this text in english language",
"              (gibt diesen Text in englischer Sprache aus)",
"  -id         gibt eine Meldung ueber diesen Backup aus",
"  -count      gibt die Zahl der Backupstuecke aus",
"  -list       gibt die Backupstuecke und ihre Filelisten aus",
"  -list:#     gibt nur ein einzelnes Stueck aus (zB. -list:4)",
"  -levelall   alle Level zeigen (sonst: nur den juengsten Fund)",
"  -nolevel    keine aelteren Levels durchsuchen", 
"  -level_reset  Widerruf von -levelall oder -nolevel",
"  -volumewise[:modus]  Ausgabesortierung fuer Levels und Backupstuecke:",
"                volume_ascending ... Stuecke getrennt, Nummern aufsteigend", 
"                levels_ascending ... Levelnummern aufsteigend",
"                on  [oder leer]  ... Stuecke und Levels aufsteigend", 
"                off ...... Levels absteigend, Stuecke nicht sortiert",
"  -suball     alle Unteradressen zeigen (sonst: eine pro Stueck)",
"  -fast       nicht nach Unteradressen suchen",
"  -sub_reset  Widerruf von -suball oder -fast  ",
"  -outdated:code  Filteranweisung fuer Fileobjekte, die noch im Backup",
"              vorkommen, aber auf der Platte geloescht oder ersetzt wurden:",
"                hide ... keine veralteten Fileobjekte ausgeben",
"                show ... aktuelle und veraltete Fileobjekte ausgeben",
"                only ... keine aktuellen Fileobjekte ausgeben",
"  -dialog     nachdem alle Argumente verarbeitet sind, geht das Programm in",
"              den Dialogmodus. Man kann dann Suchtexte oder die hier",
"              beschriebenen Optionen eingeben (nur eine pro Zeile).",
"  -dialog_reset  Widerruf von -dialog (wirkt nur als Argument)",
"  -deutsch    Dialogmeldungen auf Deutsch (manches bleibt englisch)",
"  -english    Dialogmeldungen auf Englisch (Grundeinstellung)",
"  -page:len[:width]  nach len Ergebniszeilen anhalten und fragen (0=aus).",
"              width kann die Zeilenzahlberechnung an die Zeilenbreite",
"              des Ausgabeterminals anpassen (Grundeinstellung 80).",
#ifdef Askme_with_readlinE
"  -use_stdin  Standardeingabe verwenden auch wenn libreadline moeglich ist",
"  -use_readline  libreadline verwenden wenn moeglich",
"  -history:textzeile  bringt textzeile in die libreadline history. Dieses",
"              selbst wird nicht in die history Liste eingetragen.",
#endif /* Askme_with_readlinE */
"  -pkt_output[:on|:off]  leitet alle Ausgaben auf stdout und setzt vor jede",
"              Zeile eine Kennung, die Kanal-Id und eine Zahl angibt.",
"              Jedes Ausgabepaket hat ein Newline-Zeichen am Ende.",
"              Kanal-Ids sind: 'R:' fuer Ergebniszeilen, 'I:' fuer",
"              Hinweise und Fehlermeldungem, 'M:' fuer -mark Texte.",
"              Das Bit 0 der Zahl zeigt an, dass das Newlinezeichen auch",
"              zum Datenteil des Pakets gehoert. Beispiel:",
"                I:1: Option oder Suchtext eingeben :",
"              Frontend-Programme sollten -pkt_output:on verwenden.",
"  -logfile:kanal:adresse  leitet die Ausgabe eines Kanals in File adresse.",
"              kanal kann sein: 'R','I','M' wie bei -pkt_output oder '.'",
"              fuer den zusammengefassten -pkt_output Datenfluss.",
"  -mark:text  wenn text nicht leer ist wird er von nun an nach dem Ende",
"              jeder Suche oder Option ausgegeben",
"  -prog:text  text als Programmname in allen weiteren Meldungen verwenden.",
"  -prog_help:text  text als Programmname verwenden und -help aufrufen",
"  -prog_help:text  text als Programmname verwenden und -hilfe aufrufen",
"  -result_format:code:...:code  Format der Ergebnisausgabe zusammenstellen",
"                levhdr ... Ueberschrift fuer Levels",
"                levno .... Level Nummer in Ergebniszeilen",
"                volno .... Backupstuecknummer",
#ifdef Askme_fsnode_with_timeS
"                mtime .... Zeit der letzten Inhaltsaenderung (falls zu haben)",
"                ctime .... Zeit der letzten inode-Aenderung (falls zu haben)",
#endif /* Askme_fsnode_with_timeS */
"                target ... Zieladresse (nur bei Inhaltslisten)",
"                target_esc  Zieladresse mit '=' und '\\' als '\\=' und '\\\\'",
"                source ... Quelladresse",
"                expr ..... Suchtext bzw. -ausdruck",
"              Grundeinst.: ASKME:  -result_format:levhdr:volno:source:expr",
"                   Inhaltslisten:  -result_format:levhdr:volno:target:source",
"  -read_pickiness:code  Verhalten bei Problemen mit dem Lesen von Infoquelle:",
"                zero ... Einlesen fortsetzen so weit es irgendwie geht",
"                low .... Kleinere Ungereimtheiten ignorieren",
"                high ... Bei der ersten Ungereimtheit sofort abbrechen",
"              Nur das erste -read_pickiness Argument tritt in Kraft.",
"  -no_dir_list  Nicht  complete_dir_list.gz  einlesen.",
"  -media_mode:[+-]:option:...  bestimmt die Behandlung von Datentraegern,",
"              wenn das Programm ein Backupstueck anfordert. Optionen sind:",
"                 mount_umount  externe Kommandos verwenden, um das vorherige",
"                               Stueck abzumounten und das neue zu mounten.",
"                 load_eject    externe Kommandos verwenden, um das vorherige",
"                               Stueck auszuwerfen und das neue zu laden.",
"                 on_abort_try  Auswerfen auch bei Programmabbruch probieren.",
"  -media_cmd_mount:command  ein Kommando und seine eventuellen Argumente,",
"              das anstelle des Kommandos mount ausgefuehrt wird.",
"              Es erhaelt die -media_from Adresse als weiteres Argument.",
"  -media_cmd_umount:command  wie -media_cmd_mount jedoch als Ersatz fuer",
"              Kommando umount",
"  -media_cmd_load:command  wie -media_cmd_mount jedoch als Ersatz fuer",
"              Kommando eject -t",
"  -media_cmd_eject:command  wie -media_cmd_mount jedoch als Ersatz fuer",
"              Kommando eject",
"  -status[:modus|:filter] Gegenwaertigen Stand der Einstellungsoptionen",
"              melden. Modi:",
"                 short... nur wichtige oder geaenderte Optionen melden",
"                 long ... auch Optionen melden, die in Grundeinstellung sind",
"                 long_history  wie long plus -history Zeilen",
"              filter beginnen mit dem Zeichen '-'. Sie erden auf die Ausgabe",
"              von -status:long_history angewendet, Eine Zeile wird nur dann",
"              ausgegeben, wenn ihr Anfang mit diesem Text uebereinstimmt.",
"              Z.B.: -status:-res , status:-restore , -status:restore_mode",
"              Wenn weder modus noch filter angegeben ist, gilt 'short'.",
"  -status_history_max:number  Maximalzahl von -history Zeilen, die bei",
"              -status:long_history gemeldet werden sollen.",
"  -options_from_file:fileaddress  liest Zeilen aus dem angegebenen File und",
"              fuehrt sie als Programmoptionen aus.",
"  -abort_handler:option:...  steuert die Behandlung von Programmabbruechen",
"              durch Systemsignale. Alles andere als on oder on:extra_verbous",
"              ist nur bei krassen Systemdefiziten zu verwenden:",
"                on .... Signale werden abgefangen, unsaubere Filezustaende",
"                        soweit moeglich bereinigt und das Programm beendet",
"                off ... Signale fuehren zum sofortigen Abbruch des Programms",
"                ignore  Signale werden so weit es geht abgewiesen und das",
"                        Programm fortgefuehrt. Das kann zu einzelnen Files",
"                        mit kaputten Inhalten fuehren.",
"                handler_abort nach Eintritt in die Abbruchbehandlung bei",
"                        weiteren Signalen sofort das Programm abbrechen",
"                handler_repeat nach Eintritt in die Abbruchbehandlung bei",
"                        weiteren Signale erneut Behandlung starten",
"                extra_verbous ueber die einzelnen Aktionen der Abbruch-",
"                        behandlung (auf englisch) informieren",
"  # text      wird ignoriert. Im Dialogmodus wird die Eingabzeile aber in",
"              die eventuelle Readline-History uebernommen.", 
"  -info_source:addresse  wird nur wirksam, wenn erstes Programmargument.",
"              Normalerweise braucht man -info_source: nicht voranzustellen.",
"  -version    tells program an version number",
"  -end        Programm sofort beenden",
"",
"Die Optionen -help, -hilfe, -prog_help, -prog_hilfe werden auch dann erkannt",
"wenn sie anstelle der Infoquelle als erstes Argument angegeben werden.",
"Optionen mit Parametern erkennen ausser ':' auch ' ' als Trennzeichen.",
"' ' ist bequem im Dialog, muss aber vor dem Shellparser geschuetzt werden,",
"falls die Option als Argument uebergeben wird.",
"",
"Die Option -page veranlasst eine Abfrage nach einer gegebenen Anzahl von",
"Ergebniszeilen. Leere Eingaben setzen die Ausgabe bis zur naechsten Abfrage",
"fort. Andere Eingaben koennen sein:",
"  @     verhindert weitere Abfragen bis der Suchlauf beendet ist",
"  @@    verhindert weitere Ergbnisausgabe aber setzt Suche fort",
"  @@@   bricht den aktuellen Suchlauf (bzw. Option) ab",
"  @skip unterdrueckt -volumewise Finden oder Rueckspeichern im laufenden",
"        bzw. gerade angeforderten Backupstueck",
"  sonst bricht den aktuellen Suchlauf ab und fuehrt die Eingabe als neuen",
"        Suchtext oder Option aus",
"",
"                                    Suchen",
"",
"Je nach Informationsquelle sind die Suchmoeglichkeiten unterschiedlich:",
"* Bei ASKME-Skripten wird in den Filelisten der Backupstuecke gesucht.",
"  Wenn der Suchtext oder ein darueberliegendes Directory gefunden wird, dann",
"  wird eine Ergebniszeile entsprechend -result_format ausgegeben.",
"  Wenn der Suchtext ein Directory bezeichnet und wenn eine Adresse unter",
"  diesem Directory in der Liste gefunden wird, dann haengt die Reaktion von",
"  -suball und -fast ab.",
"  Die Suchfunktionen wirken nur auf die bei -list sichtbaren Listenzeilen.",
"  Man kann also nicht nach Filenamen suchen, von denen nur ein Oberdirectory",
"  in der Liste steht.",
#ifdef Askme_with_regeX
"  -search_grep, -search_sh etc. sind moeglich aber nur beschraenkt nuetzlich.",
#endif /* Askme_with_regeX */
"* Informationen, die von content-vanish-Listenfiles stammen, bilden ein",
"  detailliertes Baummodell, in dem jedes Filobjekt vorkommt, das von den",
"  Listen abgedeckt wird. Zusaetzlich kann dieses Modell auch von jedem",
"  Fileobjekt die Adresse auf dem Datentraeger und nicht nur die auf der",
"  Festplatte nennen.",
#ifdef Askme_with_regeX
"  Dieses Modell eignet sich gut fuer Suchmuster. Deshalb ist nach dem Laden",
"  eines solchen Modells der Suchmodus -search_sh voreingestellt.",
#else  /* Askme_with_regeX */
"  Dieses Modell eignet sich gut fuer strukturierten Zugriff. Deshalb ist nach",
"  dem Laden eines solchen Modells der Suchmodus -search_sh voreingestellt.",
#endif /* ! Askme_with_regeX */
"  Wenn eine Directoryadresse passt, dann wird sie fuer jedes Backupstueck",
"  gemeldet, auf dem einer ihrer Insassen vorkommt, der kein Directory ist.",
"  Bei Option -suball werden alle Insassen gemeldet.",
"Wird nichts gefunden, dann werden normalerweise auch eventuell vorhandene",
"aeltere Levels durchsucht. Das haengt von -levelall und -nolevel ab.",
"",
"Die Suche wird entweder auf den Quelladdressen durchgefuehrt, so wie sie zur",
"Zeit des Backups gegeben waren, oder auf den Zieladressen wie sie in den",
"Backupstuecken vorliegen.",
"  -search_source suchen in Quelladressen (Grundeinstellung)",
"                 Schaltet auch :source: in -result_format an.",
"  -search_target suchen in Zieladressen. Es kann vorkommen, dass Ziele",
"                 mehrfach als Ergebnis auftauchen, weil sie von",
"                 unterschiedlichen Quellen stammen.",
"                 Schaltet auch :target: in -result_format an.",
"  -search:suchtext  Suchtext vor Verwechselung mit Optionen schuetzen",
"                    (Normalerweise geht es auch ohne -search: davor)",
"Suchverfahren:",
"  -search_fgrep  Suchtext kommt als Teiltext an beliebiger Stelle vor",
#ifdef Askme_with_regeX
"  -search_grep   Suchtext ist regular expression ( ^\\/videos\\/.* )",
"  -search_sh     Suchtext ist Shellparserausdruck    ( /videos/* )",
"  -search_shpat  wie -search_sh aber ohne die besondere Behandlung von '/'",
#else  /* Askme_with_regeX */
"  -search_sh     Suchtext ist struktierter Pfad      ( /videos/* )",
"                 Weil in dieser Programmversion keine regex-Unterstuetzung",
"                 vorhanden ist, werden nur komplette Filenamen oder ein",
"                 einzelnes '*' erkannt. Letzteres passt auf jeden Namen.",
#endif /* ! Askme_with_regeX */
"  -search_shname wie -search_sh, mit dem letzten Namen in der Quelladresse.",
"                 Dieser Modus sucht unter gefundenen Directories weiter.",
"  -search_reset  Suchtext ist ein Anfangsstueck der gesuchten Adresse",
"Navigationsoptionen:",
"  -cd:suchtext   legt das Arbeitsdirectory fuer -search_sh fest. Das erste",
"                 passende Directory wird dann vor jeden weiteren suchtext",
"                 gesetzt, der nicht mit '/' beginnt.  -search_sh und die",
"                 verwandten Optionen behandeln die Filenamen '..' und '.'",
"                 so, wie es in unix-artigen Filesystemen ueblich ist.",
"  -cd            stellt Arbeitsdirectory auf das eingelesene zurueck bzw.",
"                 auf / im Modus -search_target",
"  -pwd           gibt das aktuelle Arbeitsdirectory aus",
"  -ls:suchtext   gibt Uebersicht der Adressen. suchtext wie bei -search_sh.",
"                 Es wird auf allen Levels gesucht. Jeder Treffer wird aber",
"                 nur einmal gemeldet. Passende Directories melden ihren",
"                 gesamten Inhalt.",
"  -ls            Abkuerzung fuer -ls:.",
"  -ls_d:suchtext wie -ls aber ohne Directoryinhalte zu melden. Auch leere",
"                 Directories werden gemeldet.",
"  -ls_d          Abkuerzung fuer -ls_d:*  (nicht ganz das selbe wie -ls)",
"  -find:suchtext  listet passende Files unter dem Arbeitsdirectory.",
"                 suchtext filtert ggf. wie bei -search_shname. Auch leere",
"                 Directories werden gemeldet.",
"  -find          Abkuerzung fuer -find:*",

#endif /* ! Askme_no_german_helP */

"@ENDE_OF_HELPTEXT_(HOPEFULLY_UNIQUELY_SILLY_TEXT)@"
 };
 static char restore_text[][80]={

#ifndef Askme_no_german_helP

"",
"                               Rueckspeichern",
"",
"--------------------------------------------------------------------------",
"Wichtig: Rueckspeichern verschafft sich noetigenfalls (und vorruebergehend)",
"         Zugriffsrechte auf Files, die dem ausfuehrenden User gehoeren.",
"         Man kann also keine *eigenen* Files auf der Platte vor Ueberscheiben",
"         schuetzen, indem man ihnen oder den *eigenen* Directories darueber",
"         Rechte wegnimmt. Weil der Superuser normalerweise Besitzrechte an",
"         allen Files hat, kann man ihn durch Zugriffsbeschraenkungen an",
"         garnichts hindern. (Schutz bieten jedoch die Optionen",
"          -restore_mode:+:overwrite_off  oder  -protect )",
"----------------------------------------------------------------------------",
"Rueckspeichern geht zZt. nur auf Basis eines Baummodells und nur fuer",
"Backups in einem Format, das gemountet werden kann (also zB. ISO-9660).",
"Man muss es erst freischalten mit",
"  -restore_mode:-:disable_restore",
"oder einer anderen -restore_mode Option ohne [+-].",
"",
"Rueckspeicher-Operationen kopieren von Backupadresse nach Restoreadresse.",
"Dieser Text erwaehnt \"Quelle\" und \"Ziel\" in Anfuehrungszeichen, weil",
"ihre Benennung von der Backupsituation stammt und sich ihre Bedeutung in",
"der Rueckspeichersituation weitgehend umkehrt. Beim Rueckspeichern ist",
"\"Ziel\" Teil der Backupadresse (also der Kopierquelle) und \"Quelle\"",
"bestimmt meist die Restoreadresse (also das Kopierziel).",
"Rueckspeicheroptionen:",
"  -restore_from:pfad  legt den Directorypfad fest, der vor die \"Ziel\"-",
"                 Adresse gesetzt wird, um die Backupadresse zu bilden.",
"                 pfad wird ggf. auch fuer Lade-, Mount-, Umount- oder",
"                 Auswerf-Operationen benutzt.",
"  -restore_skip:pfad  legt das Anfangsstueck der Restoreadresse fest, das",
"                 entfernt wird, bevor -restore_to seine Arbeit tut.",
"                 Abhaengig von -restore_mode:use_target wird die Restore-",
"                 adresse entweder von \"Quelle\" oder von \"Ziel\" abgeleitet.",
"                 Wenn pfad nicht zum Anfang der Restoreadresse passt, dann",
"                 wird das als Fehler angesehen, so als ob der File im Backup",
"                 nicht vorhanden waere.",
"  -restore_to:pfad  legt den Directorypfad fest, der vor die entstehende",
"                 Restoreadresse gesetzt wird, ggf. nachdem -restore_skip",
"                 ein Anfangsstueck davon entfernt hat.",
"  -restore_asylum:pfad  legt fest, wo -restore_mode:+:overwrite_by_move",
"                 die geretteten Files unterbringen soll. Die Asyladresse",
"                 entsteht, indem pfad vor die volle Absolutadresse des zu",
"                 ueberschreibenden Fileobjekts gesetzt wird.",
"  -restore:suchtext  sucht passende Fileobjekte und kopiert sie auf die",
"                 entsprechende Restoreadresse. Gefundene Directories werden",
"                 mitsamt all ihren Files und Unterdirectories kopiert.",
"                 Die Suche unterliegt den selben Regeln wie bei -search",
"  -restore_mode:[+-]:option:...  legt Eigenschaften des Rueckkopierens fest.",
"                 Wenn das erste Zeichen nach -restore_mode: ein '+' ist,",
"                 dann werden alle genannten Optionen der aktuellen",
"                 Einstellung hinzugefuegt. Ist das erste Zeichen ein '-',",
"                 dann werden _alle_ genannten Optionen auf ihre Grund-",
"                 einstellung zurueckgesetzt. Sonst wird die gesamte",
"                 Einstellung zurueckgesetzt und nur die genannten Optionen",
"                 werden neu aktiviert. Optionen sind:",
"                   on_search      bei jeder normalen Suche rueckspeichern",
"                                  (sonst nur bei -restore:...)",
"                   use_target     Restoreadresse von \"Ziel\" ableiten",
"                   silent_file      nicht ueber einzelne Files berichten",
"                   verbous_volume   berichte ueber mount_umount,load_eject",
"                   verbous_action   berichte ueber alle Teilaktionen",
"                   open_dir_for_owner fuer Besitzer eines rueckgespeicherten",
"                                    Directory rwx-Rechte einrichten.",
"                   overwrite_off    keine existierenden Files ueberschreiben",
"                   overwrite_by_unlink      ueberschreiben, vorher loeschen",
"                   overwrite_by_truncation  ueberschreiben, nicht loeschen",
"                   overwrite_by_move        existierende Files nach",
"                                            -restore_asylum verschieben.",
"                   obey_permissions nicht die Rechte des Besitzers an",
"                                    Oberdirectories ausweiten. Das kann aber",
"                                    den Rueckspeichererfolg verhindern.",
"                   disable_restore  Rueckspeichern zeitweilig verbieten",
"                   ban_restore      Rueckspeichern dauerhaft verbieten",
"                   keep_ownership   Superuser: nicht owner,group kopieren",
"                   fresh_timestamps nicht atime und mtime von Backup holen",
"                   on_error_ignore  Operation nach Fehler fortsetzen",
"                   on_error_prompt  Wenn Fehler auftritt, User befragen",
"                   on_error_exit    Programm bei Fehler abbrechen",
"                   after_error_exit Operation fortsetzen, aber danach das",
"                                    Programm abbrechen",
"                   on_error_keep    unvollstaendig rueckgespeicherte Files",
"                                    nicht wieder loeschen",
"                   update_state_sparse  nicht bei jedem einzelnen File",
"                                    -resume_state_file auffrischen, sondern",
"                                    nur vor Eingabeaufforderungen. Das laehmt",
"                                    aber die _dirty File-Sicherung.",
"                   update_state_dsync  _dirty info per fsync auf die Platte",
"                                    schubsen. Das bremst die Platte sehr.",
"                   simulate_existence  debugging option",
"                   simulate_action     debugging option",
"                   simulate_mount      debugging option",
"  -restore_pre_prompt:command ein Programm und seine eventuellen Argumente,",
"                 das vor der Anforderung des naechsten Datentraegers",
"                 ausgefuehrt wird. Es werden zusaetzlich vier Argumente",
"                 angehaengt: Level, Backupstuecknummer, Backupadresse des",
"                 naechsten Suchtreffers, seine Restoreadresse.",
"                 Wenn command mit '@' beginnt, wird es auch dann ausgefuehrt,",
"                 wenn -restore_mode:+:simulate_mount angeschaltet ist.",
"  -restore_post_prompt:command  wie -restore_pre_prompt, jedoch auszufuehren",
"                 nach der Antwort des Users auf die Anforderung.",
"  -resume[:level[:stueck[:restore_oder_such_option]]]  nimmt ein vorheriges",
"                 -restore oder -volumewise -search an dem Level und Stueck",
"                 wieder auf, wo es zuvor abgebrochen wurde. Wenn level",
"                 angegeben ist, wird bei Stueck 1 dieses Levels begonnen.",
"                 Ist stueck gegeben, dann wird bei dieser Stuecknummer",
"                 begonnen. Wenn der Rest mit \"-restore:\" oder \"-search:\"",
"                 beginnt, dann wird diese Operation anstelle der zuletzt",
"                 ausgefuehrten Operation wiederaufgenommen.",
"  -resume_last_source:address  legt die absolute Quelladresse des Fileobjekts",
"                 fest, an dem innerhalb des vorgegebenen Levels und Stuecks",
"                 weitergemacht werden soll. Wenn die Adresse nicht im ersten",
"                 untersuchten Stueck gefunden wird, dann wird dieses Stueck",
"                 ganz uebersprungen. Die folgenden Stuecke werden dann",
"                 normal abgearbeitet.",
"  -resume_last_dirty:address  legt die absolute Restoreadresse eines evtl.",
"                 Fileobjekts fest, das weder in seinem Zustand vor dem",
"                 Rueckspeichern ist, noch korrekt rueckgespeichert wurde.",
"                 Dieses Fileobjekt selbst ist komplett fragwuerdig. Die",
"                 Directories darueber koennen zu grosszuegige Rechte fuer",
"                 ihren Besitzer bieten (aber sonst fuer keinen) oder",
"                 (falls neu erzeugt) zu knappe Rechte fuer jedermann.",
"                 Beim naechsten -resume, loest eine nichtleere last_dirty",
"                 Adresse eine Fehlermeldung aus. Im Dialog erfolgt dann eine",
"                 Bestaetigungsabfrage. Im Nicht-Dialogmodus bricht das",
"                 Programm dann bei -resume ab.",
"  -resume_last_lv:level[:stueck]   legt Level und Backupstueck fuer das",
"                 naechste -resume fest, wenn es keine eigenen mitbringt.", 
"  -resume_opt_text:restore_or_search_option  legt die Option fest, die vom",
"                 naechsten -resume ausgefuehrt werden soll, wenn dieses",
"                 keine eigene Operation mitbringt.",
"  -resume_state_files:fileadresse  legt eine Familie von Files fest, die",
"                 waehrend -restore oder -volumwise -search bei mehreren",
"                 Gelegenheiten aufgefrischt werden. In fileadresse selbst",
"                 wird ein -status:long_history report geschrieben. Darin",
"                 werden einige andere Files verwendet:",
"                   fileaddress\"_opt\" enthaelt aktuellen -resume_opt_text",
"                   fileaddress\"_src\" enthaelt aktuelle -resume_last_source",
"                   fileaddress\"_lv\"  enthaelt -resume mit aktuellem Level",
"                                     und Stuecknummer.",
"                   fileaddress\"_dirty\"  aktuelles -resume_last_dirty",
"                 Durch diese Aufteilung kann man die -resume-Parameter vor",
"                 einem Start mit -resume_from_file leichter manipulieren.",
"                 Beachten Sie, dass die Kommandos s*backup_askme eine",
"                 eigene Filefamilie tmp/askme_resume_state* verwenden",
"                 und von der Option -resume_state_files verwirrt wuerden.",
"  -resume_from_file:fileaddress  fast genauso wie -options_from_file, darf ",
"                 aber das erste Programmargument sein und ist dann faehig",
"                 eine Infoquelle zu laden. fileaddress sollte auf eine",
"                 Filefamilie zeigen, die zuvor von -resume_state_files",
"                 erzeugt wurde.",
"  -resume_state_to_file  aktuellen (evtl. kuenstlichen) resume-Zustand in",
"                 die -resume_state_files schreiben.",
"  -protect:path  Rueckspeicheradresse path abblocken und auch alle Adressen",
"                 unterhalb von path. Wenn ein Fileobjekt mit Adresse path",
"                 existiert, dann wird nicht nur die wortwoertliche Adresse",
"                 geschuetzt, sondern auch jede andere, die zur selben Inode-",
"                 und Device-Kombination fuehrt. Wenn path zu einem Softlink",
"                 fuehrt, wird auch das Linkziel so geschuetzt. Linkziele und",
"                 Inodes werden aufgezeichnet, wenn die -protect Option",
"                 ausgefuehrt wird. Spaeter geaenderte Linkziele oder Inodes",
"                 sind nicht geschuetzt. Die wortwoertliche Adresse, erfasste",
"                 Linkziele und Inodes bleiben auch nach solchen Aenderungen",
"                 auf der Platte geschuetzt. -protect-Optionen, die dem",
"                 Programm als Argumente vorliegen werden ausgefuehrt, bevor",
"                 irgendeine Rueckspeicherung stattfinden kann.",
"                 Die Zahl der -protect-Optionen ist nicht begrenzt. Dennoch",
"                 sollte man diese Option sparsam benutzen, weil sie beim",
"                 Rueckspeichern recht viel Aufwand verursacht.",
"  -shadow:path   wie -protect aber ohne als Fehler angesehen zu werden, wenn",
"                 eine Rueckspeicherung blockiert wird. Danach laeuft ein",
"                 -restore weiter, auch wenn -restore_mode:-:on_error_ignore",
"                 eingestellt ist. -shadow soll von Anwendungen genutzt",
"                 werden, die einen Schutz vor versehentlichem Ueberschreiben",
"                 durch sich selbst brauchen.",
"                 Aktive -shadow Optionen werden von -status:long angezeigt.",

#endif /* ! Askme_no_german_helP */

"@ENDE_OF_HELPTEXT_(HOPEFULLY_UNIQUELY_SILLY_TEXT)@",
""
 };
 static char no_restore_text[][80]={

#ifndef Askme_no_german_helP
"",
"--------------------------------------------------------------------------",
"Hinweis: Rueckspeichern ist unwiderruflich gesperrt durch Option",
"           -restore_mode:ban_restore",
"--------------------------------------------------------------------------",
#endif /* ! Askme_no_german_helP */

"@ENDE_OF_HELPTEXT_(HOPEFULLY_UNIQUELY_SILLY_TEXT)@",
""
 };
 char *tpt;
 int i,pass;

/*
 if(askme->level_info!=NULL)
   Askme_id(askme,0);
*/
 Askme_restxt(askme,"\n"); 
 sprintf(askme->result_line,"Anwendung: %s Infoquelle [Optionen|Suchtexte]\n",
         askme->progname); 
 Askme_result(askme,0);
 Askme_restxt(askme,"\n"); 
 for(pass=0;pass<2;pass++) {
   for(i=0;1;i++) {
     if(pass==0)
       tpt= text[i];
     else if(askme->restore_mode&(1<<31))
       tpt= no_restore_text[i];
     else
       tpt= restore_text[i];
     if(strcmp(tpt,"@ENDE_OF_HELPTEXT_(HOPEFULLY_UNIQUELY_SILLY_TEXT)@")==0)
   break;
     sprintf(askme->result_line,"%s\n",tpt);
     Askme_result(askme,0);
     if(askme->request_to_abort)
       return(1);
   }
 }
 Askme_restxt(askme,"\n"); 

#ifndef Askme_no_german_helP

 Askme_restxt(askme,"Beispiel (suche eiche.jpg auf allen Backup Levels):\n");
 sprintf(askme->result_line,
         "  %s /cdrom/ASKME -levelall /home/bilder/baeume/eiche.jpg\n",
         askme->progname);
 Askme_result(askme,0);
 Askme_restxt(askme,"\n");
 Askme_restxt(askme,
           "Beispiel (Lade inkrementelle Konfiguration und starte Dialog):\n");
 sprintf(askme->result_line,
         "  %s $HOME/backup_configuration -levelall -dialog\n",
         askme->progname);
 Askme_result(askme,0);
 Askme_restxt(askme,"Eingaben im Dialog:\n");

#ifdef Askme_with_regeX
 Askme_restxt(askme,"  /*/*/baeume/eiche*\n");
 Askme_restxt(askme,"  -search_shpat\n");
 Askme_restxt(askme,"  *baeume/eiche*\n");
 Askme_restxt(askme,"  -search_shname\n");
 Askme_restxt(askme,"  eiche*\n");
 Askme_restxt(askme,
   "Kompliziertere Suchtexte: Namen die mit A,B,C,K,W, or Z anfangen, egal\n");
 Askme_restxt(askme,
   "ob gross oder kleingeschrieben. Das zweite Zeichen darf keine Ziffer\n");
 Askme_restxt(askme,
   "sein und der Suffix muss aus 3 Zeichen bestehen:\n");
 Askme_restxt(askme,"  [A-CKWZa-ckwz][!0-9]*.???\n");
 Askme_restxt(askme,
        "Namen der Form \"DSC\", 5 Ziffern, 3-stelliger Suffix:\n");
 Askme_restxt(askme,"  DSC[0-9][0-9][0-9][0-9][0-9].???\n");

#else /* Askme_with_regeX */

 Askme_restxt(askme,"  /*/*/baeume/eiche.jpg\n");
 Askme_restxt(askme,"  -search_shname\n");
 Askme_restxt(askme,"  eiche.jpg\n");
#endif /* ! Askme_with_regeX */

 if(!(askme->restore_mode&(1<<31))) {
   Askme_restxt(askme,"\n");
   Askme_restxt(askme,
 "Directory mit der urspruenglichend Adresse /home/bilder/baeume/eiche.jpg\n");
   Askme_restxt(askme,
 "zusammen mit all seinen Files vom Backup auf die Platte rueckspeichern:\n");
   Askme_restxt(askme,"  -restore_from:/cdrom\n");
   Askme_restxt(askme,"  -restore_skip:/home/bilder\n");
   Askme_restxt(askme,"  -restore_to:/home/test/backup_von_bilder\n");
   Askme_restxt(askme,"  -media_mode:load_eject:mount_umount\n");
   Askme_restxt(askme,"  -restore_mode:overwrite_off\n");
   Askme_restxt(askme,"  -restore:/home/bilder/baeume\n");
   Askme_restxt(askme,
     "Es entsteht das Directory  /home/test/backup_von_bilder/baeume .\n");
   Askme_restxt(askme,
   "Man wird aufgefordert die noetigen Backupstuecke in /cdrom einzulegen.\n");
   Askme_restxt(askme,
    "Existierende Files werden nicht ueberschrieben.\n");
 }
 Askme_restxt(askme,"\n");
 Askme_restxt(askme,
         "Beispiel (Lade inkrementelle Information von Datentraegern):\n");
 sprintf(askme->result_line,
         "  %s /cdrom/added_by_scdbackup -levelall -dialog\n",
         askme->progname);
 Askme_result(askme,0);

#endif /* ! Askme_no_german_helP */

 return(1);
}


int Askme_interpreter(askme,in_cmd,flag)
struct AskmE *askme;
char *in_cmd;
int flag;
/*
return:
 <=0 error , 1 = success , 2 = nothing found , 3 = end program run
*/
/*
 bit0= recursion
 bit1= this is surely a search text and not an option
 bit2= with bit1: this is a restore operation
*/
{
 int ret,volno= -1,no_sublevels= 1,i,fmt,l,search_mode,structured_search;
 int result_format,volumewise,restore_mode,do_restore= 0,negative_mode= 0;
 int level,volume,search_ret,show_outdated;
 double num;
 char cmd[2*4096],*cpt,*opt;
 
 if(askme==NULL)
   return(0);
 if(askme->is_dialog) {
   askme->result_line_counter= askme->result_page_counter= 0;
   if(askme->result_page_length<0)
     askme->result_page_length= -askme->result_page_length;
 }
 askme->request_to_abort= askme->request_to_skip_volume= 0;
 askme->suppress_result= 0;
 askme->prepended_wd= askme->look_for_dir= 0;
 strcpy(cmd,in_cmd);

 if((flag&2)&&(flag&4))
   do_restore= 1;
 if((flag&2) || (cmd[0]!='-' && cmd[0]!='#' && cmd[0]!=0)) {
   goto search_text;

 } else if(cmd[0]=='#' || cmd[0]==0) {
   /* ignore comment line and empty option */; 

 } else if(strncmp(cmd,"-abort_handler",14)==0 &&
           (cmd[14]==':' || cmd[14]==' ')) {
   askme->abort_handler= Askme_abort_handler_defaulT;
   if(strstr(cmd+15,"off")!=NULL)
     askme->abort_handler= ((askme->abort_handler)&~15);
   if(strstr(cmd+15,"ignore")!=NULL)
     askme->abort_handler= ((askme->abort_handler)&~15)|2;
   if(strstr(cmd+15,"on")!=NULL)
     askme->abort_handler= ((askme->abort_handler)&~15)|1;
   if(strstr(cmd+15,"handler_abort")!=NULL)
     askme->abort_handler= ((askme->abort_handler)&~(15<<4));
   if(strstr(cmd+15,"handler_repeat")!=NULL)
     askme->abort_handler= ((askme->abort_handler)&~(15<<4))|(1<<4);
   if(strstr(cmd+15,"extra_verbous")!=NULL)
     askme->abort_handler= (askme->abort_handler)|256;
   
 } else if(strcmp(cmd,"-cd")==0) {
   sprintf(askme->info_text,"previous working directory:\n");
   Askme_info(askme,0);
   sprintf(askme->result_line,"%s/\n",askme->wd);
   Askme_result(askme,0);
   if(askme->search_target)
     strcpy(askme->wd,"");
   else
     strcpy(askme->wd,askme->initial_wd);
   goto pwd;

 } else if(strncmp(cmd,"-cd",3)==0 && (cmd[3]==':' || cmd[3]==' ')) {
   if(strlen(cmd)>sizeof(askme->wd)) {
/* cd_too_long:; */
     Askme_de_eN(
       sprintf(askme->info_text,"%s: -cd: zu lang\n",askme->progname);
     ,
       sprintf(askme->info_text,"%s: -cd: too long\n",askme->progname);
     )
     Askme_info(askme,0);
     {ret= 0; goto ex;}
   }
   sprintf(askme->info_text,"previous working directory:\n");
   Askme_info(askme,0);
   sprintf(askme->result_line,"%s/\n",askme->wd);
   Askme_result(askme,0);
   search_mode= askme->search_mode;
   structured_search= askme->structured_search;
   askme->search_mode= 3;
   askme->structured_search= 2;
   askme->look_for_dir= 1;
   ret= Askme_search_file(askme,cmd+4,-1,-1,0);
   askme->request_to_abort= askme->suppress_result= 0;
   askme->look_for_dir= 0;
   askme->search_mode= search_mode;
   askme->structured_search= structured_search;
   askme->found= -1;
   if(ret<0)
     goto ex;
   if(ret==0) {
     if(askme->root_node!=NULL) {
       Askme_de_eN(
         sprintf(askme->info_text,
                 "-cd: nicht gefunden oder kein Directory :\n%s\n",cmd+4);
       ,
         sprintf(askme->info_text,
                 "-cd: not found or not a directory :\n%s\n",cmd+4);
       )
       Askme_info(askme,0);
       {ret= 0; goto ex;} /* in the tree model there is no doubt */
     }
     strcpy(askme->wd,cmd+4);
   } else
     strcpy(askme->wd,askme->result_line);
   l= strlen(askme->wd);
   while(l>0)
     if(askme->wd[l-1]=='/')
       askme->wd[--l]= 0;
     else
   break;
   goto pwd;

 } else if(strcmp(cmd,"-count")==0) {
   if(askme->root_node!=NULL && !askme->levelall)
     sprintf(askme->result_line,"%d\n",askme->num_highest_level_volumes);
   else
     sprintf(askme->result_line,"%d\n",askme->num_volumes);
   Askme_result(askme,0);
   if(askme->levelall)
     no_sublevels= 0;

 } else if(strcmp(cmd,"-dialog")==0) {
   askme->dialog= 1;
   no_sublevels= 0;

 } else if(strcmp(cmd,"-dialog_reset")==0) {
   askme->dialog= 0;
   no_sublevels= 0;

 } else if(strcmp(cmd,"-deutsch")==0) {
   askme->german_dialog= 1;

 } else if(strcmp(cmd,"-end")==0) {
   {ret= 3; goto ex;}

 } else if(strcmp(cmd,"-english")==0) {
   askme->german_dialog= 0;
   no_sublevels= 0;

 } else if(strncmp(cmd,"-exec",5)==0 && (cmd[5]==':' || cmd[5]==' ')) {
   Askme_exec(askme,cmd+6,0);

 } else if(strcmp(cmd,"-fast")==0) {
   askme->sub= 0;
   no_sublevels= 0;

 } else if(strncmp(cmd,"-find",5)==0 && 
           (cmd[5]==':' || cmd[5]==' ' || cmd[5]==0)) {
   if(cmd[5]==0)
     strcpy(cmd,"*");
   else {
     l= strlen(cmd);
     for(i= 6;i<=l;i++)
       cmd[i-6]= cmd[i];
   }
   search_mode= askme->search_mode;
   structured_search= askme->structured_search;
   result_format= askme->result_format;
   askme->search_mode= 4;
   askme->structured_search= 4;
   if(askme->search_target)
     askme->result_format= 8|(1<<20);
   else
     askme->result_format= 16|(1<<20);
   ret= Askme_search_file(askme,cmd,-1,-1,2);
   askme->search_mode= search_mode;
   askme->structured_search= structured_search;
   askme->result_format= result_format;
   goto ex;

 } else if(strcmp(cmd,"-help")==0) {
   Askme_help(askme,0);

 } else if(strcmp(cmd,"-hilfe")==0) {
   Askme_hilfe(askme,0);

 } else if(strncmp(cmd,"-history",8)==0 && (cmd[8]==':' || cmd[8]==' ')) {
   /* add to readline history */
   Askme_dialog_input(askme,cmd+9,strlen(cmd+9)+1,2);

 } else if(strcmp(cmd,"-id")==0) {
   Askme_id(askme,0);
   if(askme->levelall)
     no_sublevels= 0;

 } else if(strncmp(cmd,"-info_source",12)==0 
           && (cmd[12]==':' || cmd[12]==' ')) {
   /* is already handled in Askme_prescan_args() */;

 } else if(strcmp(cmd,"-level_reset")==0) {
   askme->levelall= 0;
   askme->nolevel= 0;

 } else if(strcmp(cmd,"-levelall")==0) {
   askme->levelall= 1;

 } else if(strncmp(cmd,"-list",5)==0 && 
          (cmd[5]==':' || cmd[5]==' ' || cmd[5]==0)) {
   if(cmd[5]!=0) {
     volno= -1;
     sscanf(cmd+6,"%d",&volno);
     if(volno<=0 || volno>askme->num_volumes)
       volno= -1;
   }
   ret= Askme_list(askme,volno,0);
   if(ret<=0)
     goto ex;
   if(askme->levelall && volno<=0)
     if(askme->sublevels!=NULL)
       for(i=askme->num_sublevels-1;i>=0;i--) {
         sprintf(askme->result_line,
                 "\n\n--------------------------------------------------\n\n");
         Askme_result(askme,0);
         Askme_id(askme->sublevels[i],0);
         ret= Askme_list(askme->sublevels[i],-1,0);
         if(ret<=0)
           goto ex;
       }

 } else if(strcmp(cmd,"-list_target")==0) { /* <<< inofficial */
   ret= Askme_list(askme,0,32);
   if(ret<=0)
     goto ex;
   {ret= 1; goto ex;}

 } else if(strncmp(cmd,"-ls",3)==0 && 
           (cmd[3]==':' || cmd[3]==' ' || cmd[3]==0)) {
   if(cmd[3]==0)
     strcpy(cmd,".");
   else {
     l= strlen(cmd);
     for(i= 4;i<=l;i++)
       cmd[i-4]= cmd[i];
   }
   search_mode= askme->search_mode;
   structured_search= askme->structured_search;
   result_format= askme->result_format;
   volumewise= askme->volumewise;
   restore_mode= askme->restore_mode;
   askme->search_mode= 3;
   askme->structured_search= 3;
   if(askme->search_target)
     askme->result_format= 8|(1<<20);
   else
     askme->result_format= 16|(1<<20);
   askme->volumewise= 0;
   askme->restore_mode&= ~(1|8);
   ret= Askme_interpreter(askme,cmd,2|1);
   askme->search_mode= search_mode;
   askme->structured_search= structured_search;
   askme->result_format= result_format;
   askme->volumewise= volumewise;
   askme->restore_mode= restore_mode;
   goto ex;

 } else if(strncmp(cmd,"-ls_d",5)==0 && 
           (cmd[5]==':' || cmd[5]==' ' || cmd[5]==0)) {
   if(cmd[5]==0)
     strcpy(cmd,"*");
   else {
     l= strlen(cmd);
     for(i= 6;i<=l;i++)
       cmd[i-6]= cmd[i];
   }
   search_mode= askme->search_mode;
   structured_search= askme->structured_search;
   result_format= askme->result_format;
   volumewise= askme->volumewise;
   restore_mode= askme->restore_mode;
   askme->search_mode= 3;
   askme->structured_search= 2;
   if(askme->search_target)
     askme->result_format= 8|(1<<20);
   else
     askme->result_format= 16|(1<<20);
   askme->volumewise= 0;
   askme->restore_mode&= ~(1|8);
   ret= Askme_interpreter(askme,cmd,2|1);
   askme->search_mode= search_mode;
   askme->structured_search= structured_search;
   askme->result_format= result_format;
   askme->volumewise= volumewise;
   askme->restore_mode= restore_mode;
   goto ex;

 } else if(strncmp(cmd,"-logfile",8)==0 && (cmd[8]==':' || cmd[8]==' ')) {
   int hflag,channel_no= 0;
   
   if(strlen(cmd)<12) {
logfile_wrong_form:;
     Askme_de_eN(
       sprintf(askme->info_text,
               "--- Falsche Form. Richtig waere: -logfile:[.RIM]:adresse\n");
     ,
       sprintf(askme->info_text,
               "--- Wrong form. Correct would be: -logfile:[.RIM]:address\n");
     )
     Askme_info(askme,0);
     {ret= 0; goto ex;}
   }
   hflag= 2;
   if(cmd[9]=='R')
     channel_no= 1;
   else if(cmd[9]=='I')
     channel_no= 2;
   else if(cmd[9]=='M')
     channel_no= 3;
   else if(cmd[9]=='.')
     hflag= 4;
   else
     goto logfile_wrong_form;
   if(strcmp(cmd+11,"-")==0)
     hflag|= (1<<15);
   ret= Write_to_channel(cmd+11,channel_no,hflag);
   if(ret<=0) {
     Askme_de_eN(
       sprintf(askme->info_text,
               "--- Kann Logfile nicht oeffnen:  %s\n   %s\n",
               cmd+11,(errno>0?strerror(errno):"-unbekannter fehler-"));
     ,
       sprintf(askme->info_text,
               "--- Cannot open logfile:  %s\n   %s\n",
               cmd+11,(errno>0?strerror(errno):"-unbekannter fehler-"));
     )
     Askme_info(askme,0);
     {ret= 0; goto ex;}
   }
   
 } else if(strncmp(cmd,"-mark",5)==0 && 
           (cmd[5]==':' || cmd[5]==' ' || cmd[5]==0)) {
   if(cmd[5]==0)
     askme->mark_text[0]= 0;
   else
     strncpy(askme->mark_text,cmd+6,sizeof(askme->mark_text)-1);
   askme->mark_text[sizeof(askme->mark_text)-1]= 0;

 } else if(strncmp(cmd,"-media_cmd_eject",16)==0 &&
           (cmd[16]==':' || cmd[16]==' ')) {
   strcpy(askme->media_eject_cmd,cmd+17);

 } else if(strncmp(cmd,"-media_cmd_mount",16)==0 &&
           (cmd[16]==':' || cmd[16]==' ')) {
   strcpy(askme->media_mount_cmd,cmd+17);

 } else if(strncmp(cmd,"-media_cmd_umount",17)==0 &&
           (cmd[17]==':' || cmd[17]==' ')) {
   strcpy(askme->media_umount_cmd,cmd+18);

 } else if(strncmp(cmd,"-media_cmd_load",15)==0 &&
           (cmd[15]==':' || cmd[15]==' ')) {
   strcpy(askme->media_load_cmd,cmd+16);

 } else if(strncmp(cmd,"-media_mode",11)==0 &&
           (cmd[11]==':' || cmd[11]==' ')) {
   if(cmd[12]=='-')
     negative_mode= 1;
   else if(cmd[12]!='+') {
     askme->media_mode= 0;
   }
   if(strstr(cmd+12,"on_abort_try")!=NULL)
     Set_bit(&(askme->media_mode),0,negative_mode);
   if(strstr(cmd+12,"mount_umount")!=NULL)
     Set_bit(&(askme->media_mode),2,negative_mode);
   if(strstr(cmd+12,"load_eject")!=NULL)
     Set_bit(&(askme->media_mode),4,negative_mode);

 } else if(strcmp(cmd,"-no_dir_list")==0) {
   /* is already handled in Askme_prescan_args() */;

 } else if(strcmp(cmd,"-nolevel")==0) {
   askme->nolevel= 1;

 } else if(strncmp(cmd,"-options_from_file",18)==0 &&
           (cmd[18]==':' || cmd[18]==' ')) {
   ret= Askme_options_from_file(askme,cmd+19,0);
   if(ret<=0 || ret==3)
     goto ex;

 } else if(strncmp(cmd,"-outdated",9)==0 &&
           (cmd[9]==':' || cmd[9]==' ' || cmd[9]==0)) {
   if(cmd[9]==0)
     askme->show_outdated= 1;
   else if(strcmp(cmd+10,"only")==0)
     askme->show_outdated= 2;
   else if(strcmp(cmd+10,"show")==0)
     askme->show_outdated= 1;
   else if(strcmp(cmd+10,"hide")==0)
     askme->show_outdated= 0;
   else {
     Askme_de_eN(
       sprintf(askme->info_text,"%s: -outdated: unbekannter Modus : %s\n",
               askme->progname,cmd+10);
     ,
       sprintf(askme->info_text,"%s: -outdated: unknown mode : %s\n",
               askme->progname,cmd+12);
     )
     Askme_info(askme,0);
     {ret= 0; goto ex;}
   }

 } else if(strncmp(cmd,"-page",5)==0 && (cmd[5]==':' || cmd[5]==' ')) {
   sscanf(cmd+6,"%d",&(askme->result_page_length));
   cpt= strchr(cmd+6,':');
   if(cpt==NULL)
     cpt= strchr(cmd+6,' ');
   if(cpt!=NULL) {
     if(cpt[1]!=0) {
       sscanf(cpt+1,"%d",&(askme->result_page_width));
       if(askme->result_page_width<=0)
         askme->result_page_width= 1;
     }
   }
   no_sublevels= 0;

 } else if(strncmp(cmd,"-pkt_output",11)==0 &&
           (cmd[11]==':' || cmd[11]==' ' || cmd[11]==0)) {
   if(cmd[11]==0)
     askme->packet_output= 1;
   else if(strcmp(cmd+12,"off")==0)
     askme->packet_output= 0;
   else if(strcmp(cmd+12,"on")==0)
     askme->packet_output= 1;
   else {
     Askme_de_eN(
       sprintf(askme->info_text,"%s: -pkt_output: unbekannter Modus : %s\n",
               askme->progname,cmd+12);
     ,
       sprintf(askme->info_text,"%s: -pkt_output: unknown mode : %s\n",
               askme->progname,cmd+12);
     )
     Askme_info(askme,0);
     {ret= 0; goto ex;}
   }

 } else if(strncmp(cmd,"-prog",5)==0 && (cmd[5]==':' || cmd[5]==' ')) {
   strcpy(askme->progname,cmd+6);
   no_sublevels= 0;

 } else if(strncmp(cmd,"-protect",8)==0 && (cmd[8]==':' || cmd[8]==' ')) {
   ret= Fprotector_add(askme->protections,cmd+9,NULL,&(askme->protections),
                       NULL,0);
   if(ret<=0)
     goto ex;

 } else if(strcmp(cmd,"-pwd")==0) {
pwd:;
   sprintf(askme->info_text,"current working directory:\n");
   Askme_info(askme,0);
   sprintf(askme->result_line,"%s/\n",askme->wd);
   Askme_result(askme,0);

 } else if(strncmp(cmd,"-read_pickiness",15)==0 &&
           (cmd[15]==':' || cmd[15]==' ')) {
   /* is already handled in Askme_prescan_args() */;

 } else if(strncmp(cmd,"-restore",8)==0 && (cmd[8]==':' || cmd[8]==' ')) {
   do_restore= 1;
   l= strlen(cmd);
   for(i=9;i<=l;i++)
     cmd[i-9]= cmd[i];
   flag|= 1;
   goto search_text;

 } else if(strncmp(cmd,"-resume",7)==0 
           && (cmd[7]==':' || cmd[7]==' ' || cmd[7]==0)) {
   level= volume= -1;
   cpt= opt= NULL;
   if(cmd[7]!=0) {
     for(cpt= cmd+8;*cpt!=0;cpt++)
       if(*cpt==':' || *cpt==' ')
     break;
     if(*cpt==0)
       cpt= NULL;
     else {
       for(opt= cpt+1;*opt!=0;opt++)
         if(*opt==':' || *opt==' ')
       break;
       if(*opt==0)
         opt= NULL;
     }
     if(cpt!=NULL) {
       if(opt!=NULL) {
         *opt= 0;
         opt++;
       }
       *cpt= 0;
       num= -1.0;
       sscanf(cpt+1,"%lf",&num);
       if(num>=0 && num<=1e9)
         volume= num;
     }
     num= -1.0;
     sscanf(cmd+8,"%lf",&num);
     if(num>=0 && num<=1e9) {
       level= num;
       if(volume<=0) 
         volume= 1;
     }
   }
   if(opt==NULL)
     opt= "";
   ret= Askme_resume(askme,level,volume,opt,0);
   goto ex;

 } else if(strncmp(cmd,"-resume_from_file",17)==0 
           && (cmd[17]==':' || cmd[17]==' ')) {
   ret= Askme_options_from_file(askme,cmd+18,0);
   if(ret<=0)
     return(-1);
   if(ret==3)
     return(0);

 } else if(strncmp(cmd,"-resume_last_lv",15)==0 
           && (cmd[15]==':' || cmd[15]==' ')) {
   level= volume= -1;
   cpt= NULL;
   for(cpt= cmd+16;*cpt!=0;cpt++)
      if(*cpt==':' || *cpt==' ')
   break;
   if(*cpt==0)
     cpt= NULL;
   if(cpt!=NULL) {
     *cpt= 0;
     num= -1.0;
     sscanf(cpt+1,"%lf",&num);
     if(num>=0 && num<=1e9)
       volume= num;
   }
   num= -1.0;
   sscanf(cmd+16,"%lf",&num);
   if(num>=0 && num<=1e9) {
     level= num;
     if(volume<=0) 
       volume= 1;
   }
   askme->resume_last_level= level;
   askme->resume_last_volume= volume;

 } else if(strncmp(cmd,"-resume_last_source",19)==0 
           && (cmd[19]==':' || cmd[19]==' ')) {
   Askme_record_resume_source(askme,cmd+20,2|4); 

 } else if(strncmp(cmd,"-resume_last_dirty",18)==0 
           && (cmd[18]==':' || cmd[18]==' ')) {
   Askme_record_resume_dirty(askme,cmd+19,4); 

 } else if(strncmp(cmd,"-resume_opt_text",16)==0 
           && (cmd[16]==':' || cmd[16]==' ')) {
   if(strncmp(cmd+17,"-restore:",9)==0) {
     askme->resume_was_restore= 1;
     strcpy(askme->resume_search_text,cmd+17+9);
   } else if(strncmp(cmd+17,"-search:",8)==0) {
     askme->resume_was_restore= 0;
     strcpy(askme->resume_search_text,cmd+17+8);
   } else {
     Askme_de_eN(
       sprintf(askme->info_text,
 "--- Option mit -resume nicht erlaubt (nur -restore oder -search):\n    %s\n",
               cmd+17);
     ,
       sprintf(askme->info_text,
 "--- Option not permitted with -resume (only -restore or -search):\n    %s\n",
               cmd+17);
     )
     Askme_info(askme,0);
     {ret= 0; goto ex;}
   }

 } else if(strncmp(cmd,"-resume_state_files",19)==0 
           && (cmd[19]==':' || cmd[19]==' ')) {
   strcpy(askme->resume_state_file,cmd+20);

 } else if(strcmp(cmd,"-resume_state_to_file")==0 ) {
   Askme_record_resume_state(askme,-1,-1,"","",-1,2); /* only write to file */

 } else if(strncmp(cmd,"-restore_from",13)==0 &&
           (cmd[13]==':' || cmd[13]==' ')) {
   strcpy(askme->restore_from,cmd+14);
   askme->restore_from_compcount= 
     Sfile_count_components(askme->restore_from,0);

 } else if(strncmp(cmd,"-restore_mode",13)==0 &&
           (cmd[13]==':' || cmd[13]==' ')) {
   if(cmd[14]=='-')
     negative_mode= 1;
   else if(cmd[14]!='+') {
     askme->restore_mode&= (1<<31);
     askme->restore_mode|= (1<<7);
   }
   if(strstr(cmd+14,"use_target")!=NULL)
     Set_bit(&(askme->restore_mode),1,negative_mode);
   if(strstr(cmd+14,"on_search")!=NULL) {
     if((askme->restore_mode&(1<<31)) && !negative_mode) {
       Askme_de_eN(
         sprintf(askme->info_text,
           "Hinweis:  -restore_mode:ban_restore  verbietet jedes -restore:\n");
       ,
         sprintf(askme->info_text,
                 "Note:  -restore_mode:ban_restore  cannot be revoked !\n");
       )
       Askme_info(askme,0);
     } else
       Set_bit(&(askme->restore_mode),3,negative_mode);
   }
   if(strstr(cmd+14,"simulate_existence")!=NULL)
     Set_bit(&(askme->restore_mode),5,negative_mode);
   if(strstr(cmd+14,"simulate_action")!=NULL)
     Set_bit(&(askme->restore_mode),6,negative_mode);
   if(strstr(cmd+14,"simulate_mount")!=NULL)
     Set_bit(&(askme->restore_mode),9,negative_mode);
   if(strstr(cmd+14,"silent_file")!=NULL) {
     if(negative_mode)
       askme->restore_mode= (askme->restore_mode&~384)|128;
     else
       askme->restore_mode= (askme->restore_mode&~384)|0;
   }
   if(strstr(cmd+14,"verbous_volume")!=NULL) {
     if(negative_mode)
       askme->restore_mode= (askme->restore_mode&~384)|128;
     else
       askme->restore_mode= (askme->restore_mode&~384)|256;
   }
   if(strstr(cmd+14,"verbous_action")!=NULL) {
     if(negative_mode)
       askme->restore_mode= (askme->restore_mode&~384)|128;
     else
       askme->restore_mode= (askme->restore_mode&~384)|384;
   }
   if(strstr(cmd+14,"open_dir_for_owner")!=NULL)
     Set_bit(&(askme->restore_mode),10,negative_mode);
   if(strstr(cmd+14,"overwrite_off")!=NULL) {
     if(negative_mode)
       askme->restore_mode= (askme->restore_mode&~(15<<11))|0;
     else
       askme->restore_mode= (askme->restore_mode&~(15<<11))|0;
   }
   if(strstr(cmd+14,"overwrite_by_unlink")!=NULL) {
     if(negative_mode)
       askme->restore_mode= (askme->restore_mode&~(15<<11))|0;
     else
       askme->restore_mode= (askme->restore_mode&~(15<<11))|(1<<11);
   }
   if(strstr(cmd+14,"overwrite_by_truncation")!=NULL) {
     if(negative_mode)
       askme->restore_mode= (askme->restore_mode&~(15<<11))|0;
     else
       askme->restore_mode= (askme->restore_mode&~(15<<11))|(2<<11);
   }
   if(strstr(cmd+14,"overwrite_by_move")!=NULL) {
     if(negative_mode)
       askme->restore_mode= (askme->restore_mode&~(15<<11))|0;
     else
       askme->restore_mode= (askme->restore_mode&~(15<<11))|(3<<11);
   }
   if(strstr(cmd+14,"keep_ownership")!=NULL)
     Set_bit(&(askme->restore_mode),15,negative_mode);
   if(strstr(cmd+14,"on_error_ignore")!=NULL) {
     if(negative_mode)
       askme->restore_mode= (askme->restore_mode&~(7<<16))|0;
     else
       askme->restore_mode= (askme->restore_mode&~(7<<16))|(1<<16);
   }
   if(strstr(cmd+14,"on_error_exit")!=NULL) {
     if(negative_mode)
       askme->restore_mode= (askme->restore_mode&~(7<<16))|0;
     else
       askme->restore_mode= (askme->restore_mode&~(7<<16))|(2<<16);
   }
   if(strstr(cmd+14,"after_error_exit")!=NULL) {
     if(negative_mode)
       askme->restore_mode= (askme->restore_mode&~(7<<16))|0;
     else
       askme->restore_mode= (askme->restore_mode&~(7<<16))|(3<<16);
   }
   if(strstr(cmd+14,"on_error_prompt")!=NULL) {
     if(negative_mode)
       askme->restore_mode= (askme->restore_mode&~(7<<16))|0;
     else
       askme->restore_mode= (askme->restore_mode&~(7<<16))|(4<<16);
   }
   if(strstr(cmd+14,"fresh_timestamps")!=NULL)
     Set_bit(&(askme->restore_mode),19,negative_mode);
   if(strstr(cmd+14,"update_state_sparse")!=NULL)
     Set_bit(&(askme->restore_mode),20,negative_mode);
   if(strstr(cmd+14,"update_state_dsync")!=NULL)
     Set_bit(&(askme->restore_mode),21,negative_mode);
   if(strstr(cmd+14,"on_error_keep")!=NULL)
     Set_bit(&(askme->restore_mode),22,negative_mode);
   if(strstr(cmd+14,"obey_permissions")!=NULL)
     Set_bit(&(askme->restore_mode),23,negative_mode);
   if(strstr(cmd+14,"disable_restore")!=NULL)
     Set_bit(&(askme->restore_mode),30,negative_mode);
   if(strstr(cmd+14,"ban_restore")!=NULL) {
     if(negative_mode) {
       Askme_de_eN(
         sprintf(askme->info_text,
               "Hinweis:  -restore_mode:ban_restore  ist unwiderruflich !\n");
       ,
         sprintf(askme->info_text,
                 "Note:  -restore_mode:ban_restore  cannot be revoked !\n");
       )
       Askme_info(askme,0);
     } else {
       Set_bit(&(askme->restore_mode),3,1);
       Set_bit(&(askme->restore_mode),31,0);
     }
   }

 } else if(strncmp(cmd,"-restore_post_prompt",20)==0 &&
           (cmd[20]==':' || cmd[20]==' ')) {
   strcpy(askme->restore_post_prompt,cmd+21);

 } else if(strncmp(cmd,"-restore_pre_prompt",19)==0 &&
           (cmd[19]==':' || cmd[19]==' ')) {
   strcpy(askme->restore_pre_prompt,cmd+20);

 } else if(strncmp(cmd,"-restore_skip",13)==0 &&
           (cmd[13]==':' || cmd[13]==' ')) {
   strcpy(askme->restore_skip,cmd+14);

 } else if(strncmp(cmd,"-restore_asylum",15)==0 &&
           (cmd[15]==':' || cmd[15]==' ')) {
   strcpy(askme->restore_asylum,cmd+16);

 } else if(strncmp(cmd,"-restore_to",11)==0 &&
           (cmd[11]==':' || cmd[11]==' ')) {
   strcpy(askme->restore_to,cmd+12);

 } else if(strncmp(cmd,"-result_format",14)==0 &&
           (cmd[14]==':' || cmd[14]==' ' || cmd[14]==0)) {
   fmt= 0;
   if(strstr(cmd+14,"levhdr")!=NULL)
     fmt|= 1;
   if(strstr(cmd+14,"levno")!=NULL)
     fmt|= 2;
   if(strstr(cmd+14,"volno")!=NULL)
     fmt|= 4;
#ifdef Askme_fsnode_with_timeS
   if(strstr(cmd+14,"mtime")!=NULL)
     fmt|= 64;
   if(strstr(cmd+14,"ctime")!=NULL)
     fmt|= 128;
#endif /* Askme_fsnode_with_timeS */
   if(strstr(cmd+14,"target")!=NULL)
     fmt|= 8;
   if(strstr(cmd+14,"target_esc")!=NULL)
     fmt|= (1<<19);
   if(strstr(cmd+14,"source")!=NULL) 
     fmt|= 16;
   if(strstr(cmd+14,"source_d")!=NULL) 
     fmt|= (1<<20);
   if(strstr(cmd+14,"expr")!=NULL)
     fmt|= 32;
   if(fmt==0) {
     Askme_de_eN(
       sprintf(askme->info_text,
               "Hinweis:  %s  bestellt leere Ausgabezeilen\n",cmd);
     ,
       sprintf(askme->info_text,
               "Note:  %s  will produce empty output lines\n",cmd);
     )
     Askme_info(askme,0);
   }
   askme->result_format= fmt;
   no_sublevels= 0;

 } else if(strncmp(cmd,"-search",7)==0 && (cmd[7]==':' || cmd[7]==' ')) {
   l= strlen(cmd);
   for(i=8;i<=l;i++)
     cmd[i-8]= cmd[i];
   flag|= 1;
   goto search_text;

 } else if(strcmp(cmd,"-search_reset")==0) {
   askme->search_mode= 0;

 } else if(strcmp(cmd,"-search_fgrep")==0) {
   askme->search_mode= 1;

 } else if(strcmp(cmd,"-search_source")==0) {
   sprintf(askme->info_text,"Search mode: -search_source\n");
   Askme_info(askme,0);
   askme->result_format|= 16;
   if(askme->search_target==1) {
     askme->search_target= 0;
     strcpy(askme->wd,askme->initial_wd);
     goto pwd;
   }  
   askme->search_target= 0;

 } else if(strcmp(cmd,"-search_target")==0) {
   sprintf(askme->info_text,"Search mode: -search_target\n");
   Askme_info(askme,0);
   askme->result_format|= 8;
   if(askme->search_target==0) {
     askme->search_target= 1;
     strcpy(askme->wd,"");
     goto pwd;
   }  
   askme->search_target= 1;

#ifdef Askme_with_regeX
 } else if(strcmp(cmd,"-search_grep")==0) {
   askme->search_mode= 2;
#endif /* Askme_with_regeX */

 } else if(strcmp(cmd,"-search_sh")==0) {
   askme->search_mode= 3;
   askme->structured_search= 1;

 } else if(strcmp(cmd,"-search_shname")==0) {
   askme->search_mode= 4;

#ifdef Askme_with_regeX
 } else if(strcmp(cmd,"-search_shpat")==0) {
   askme->search_mode= 3;
   askme->structured_search= 0;
#endif /* Askme_with_regeX */

 } else if(strncmp(cmd,"-shadow",7)==0 && (cmd[7]==':' || cmd[7]==' ')) {
   ret= Fprotector_add(askme->protections,cmd+8,NULL,&(askme->protections),
                       NULL,2);
   if(ret<=0)
     goto ex;

 } else if(strncmp(cmd,"-status",7)==0 &&
           (cmd[7]==':' || cmd[7]==' ' || cmd[7]==0)) {
   if(cmd[7]==0)
     Askme_status(askme,NULL,NULL,1);
   else if(strcmp(cmd+8,"short")==0)
     Askme_status(askme,NULL,NULL,1);
   else if(strcmp(cmd+8,"long")==0)
     Askme_status(askme,NULL,NULL,0);
   else if(strcmp(cmd+8,"long_history")==0)
     Askme_status(askme,NULL,NULL,8);
   else if(cmd[8]=='-')
     Askme_status(askme,cmd+8,NULL,8);
   else {
     Askme_de_eN(
       sprintf(askme->info_text,"%s: -status: unbekannter Modus : %s\n",
               askme->progname,cmd+8);
     ,
       sprintf(askme->info_text,"%s: -status: unknown mode : %s\n",
               askme->progname,cmd+8);
     )
     Askme_info(askme,0);
   }

 } else if(strncmp(cmd,"-status_history_max",19)==0 &&
           (cmd[19]==':' || cmd[19]==' ')) {
   sscanf(cmd+20,"%lf",&num);
   if(num>=0 && num<1e6)
     askme->status_history_max= num;

 } else if(strcmp(cmd,"-sub_reset")==0) {
   askme->sub= 1;
   askme->suball= 0;

 } else if(strcmp(cmd,"-suball")==0) {
   askme->sub= 1;
   askme->suball= 1;
   Askme_interpreter(askme,"-levelall",1);

 } else if(strncmp(cmd,"-test_lreader",13)==0 &&
           (cmd[13]==':' || cmd[13]==' ')) {
   /* <<< inofficial */
   ret= Askme_test_lreader(askme,cmd+14,0);
   if(ret<=0)
     goto ex;

 } else if(strcmp(cmd,"-use_readline")==0) {
   askme->use_stdin= 0;

 } else if(strcmp(cmd,"-use_stdin")==0) {
   askme->use_stdin= 1;

 }else if(strcmp(cmd,"-version")==0){
   printf("scdbackup program : saskme %s\n",PROG_VERSION);

 } else if(strncmp(cmd,"-volumewise",11)==0 && 
           (cmd[11]==':' || cmd[11]==' ' || cmd[11]==0)) {
   askme->volumewise= 0;
   if(cmd[11]==0)
     askme->volumewise|= 3;
   if(strcmp(cmd+12,"on")==0)
     askme->volumewise|= 3;
   if(strstr(cmd+12,"volume_ascending")!=NULL)
     askme->volumewise|= 1;
   if(strstr(cmd+12,"level_ascending")!=NULL)
     askme->volumewise|= 2;

 } else if(cmd[0]=='-') {
   Askme_de_eN(
     sprintf(askme->info_text,
      "=== Eingabe beginnt mit '-', ist aber keine bekannte Programmoption\n");
     sprintf(askme->info_text+strlen(askme->info_text),
          "===  '%s'\n",cmd);
     Askme_info(askme,0);
     sprintf(askme->info_text,
 "=== Hinweis: -search:... akzeptiert auch Suchtexte, die mit '-' beginnen\n");
   ,
     sprintf(askme->info_text,
          "=== Input line beginning with '-' is not a known option\n");
     sprintf(askme->info_text+strlen(askme->info_text),
          "===  '%s'\n",cmd);
     Askme_info(askme,0);
     sprintf(askme->info_text,
          "=== Note: -search:... accepts search texts which begin with '-'\n");
   )
   Askme_info(askme,0);
   {ret= 0; goto ex;}

 } else {
search_text:;
   restore_mode= askme->restore_mode;
   volumewise= askme->volumewise;
   show_outdated= askme->show_outdated;
   if(do_restore || (askme->restore_mode&(1|8))) {
     if(askme->root_node==NULL) {
       Askme_de_eN(
         sprintf(askme->info_text,
     "--- Kann zZt. nur auf Basis von content-vanish-Listen rueckspeichern\n");
       ,
         sprintf(askme->info_text,
   "--- Currently restore is only possible by help of content-vanish lists\n");
       )
       Askme_info(askme,0); 
       {ret= 0; goto ex;}
     }
     if(askme->restore_mode&(3<<30)) {
       Askme_de_eN(
         sprintf(askme->info_text,
              "\nVerweigert:  -restore_mode:%s  verbietet jedes -restore:\n\n",
              (askme->restore_mode&(1<<31)?"ban_restore":"disable_restore"));
       ,
         sprintf(askme->info_text,
          "\nRejected:  -restore_mode:%s  does not allow any -restore !\n\n",
              (askme->restore_mode&(1<<31)?"ban_restore":"disable_restore"));
       )
       Askme_info(askme,0);
       {ret= 0; goto ex;}
     }
     Askme_record_resume_state(askme,0,1,"",cmd,1,0); /* write whole setup */
     askme->restore_mode|= 1;
     askme->volumewise= 3;
     askme->mounted_level= -1;
     askme->mounted_volume= -1;
     askme->show_outdated= 0;
   } else if((askme->volumewise&3)==3)
     Askme_record_resume_state(askme,0,1,"",cmd,0,0); /* write whole setup */

   search_ret= Askme_search_file(askme,cmd,-1,-1,(askme->volumewise&3)<<4);

   if(askme->restore_mode&1) {
     Askme_mount_for_restore(askme,-1,-1,"","",1<<15); /* umount */
                                           /* check for delayed program exit */
     ret= Askme_error_on_restore(askme,NULL,1);
     if(ret==-1)
       goto ex;
   }
   askme->show_outdated= show_outdated;
   askme->restore_mode= restore_mode;
   askme->volumewise= volumewise;
   if(cmd[0]=='-' && !(flag&1)) {
     Askme_de_eN(
       sprintf(askme->info_text,
      "=== Hinweis: Suchtext beginnt mit '-'. Falsch geschriebene Option ?\n");
     ,
       sprintf(askme->info_text,
              "=== Note: Search text began with '-'. A misspelled option ?\n");
     )
     Askme_info(askme,0);
   }
   ret= search_ret;
   if(ret<0)
     goto ex;
   if(ret==0)
     {ret= 2; goto ex;}
 }
 if(!no_sublevels)
   if(askme->sublevels!=NULL)
     for(i=0;i<askme->num_sublevels;i++)
       Askme_interpreter(askme->sublevels[i],cmd,flag|1);
 ret= 1;
ex:;
 if((!askme->is_dialog) && !(flag&1))
   Askme_mark(askme,0);
 fflush(stdout);
 return(ret);
}


int Askme_execute_option(askme,line,flag)
struct AskmE *askme;
char *line;
int flag;
/*
 bit0-bit15 are forwarded to Askme_interpreter
 
 bit16= no pageing of info lines
 bit17= print === bar even if askme->found<0
*/
{
 int ret;
 char count_text[81],bytes_text[81];
 double tdiff;
 struct timeval tv;
 struct timezone tz;

 gettimeofday(&tv,&tz);
 Askme_reset_counters(askme,1);
 askme->idle_time= 0.0;
 tdiff= tv.tv_sec+(1.e-6*(double) tv.tv_usec);
 ret= Askme_interpreter(askme,line,flag&0xffff);
 if(ret<0)
   return(ret);
 gettimeofday(&tv,&tz);
 tdiff= tv.tv_sec+(1.e-6*(double) tv.tv_usec)-tdiff-askme->idle_time;
 if(tdiff<0.001)
   tdiff= 0.001;
 if(askme->restore_count>0) {
   Sfile_scale(askme->restore_count,count_text,7,1e6,1);
   Sfile_scale(askme->restore_bytes,bytes_text,6,1e4,1);
   Askme_de_eN(
     sprintf(askme->info_text,
          "Rueckgespeichert:               %s Fileobjekte, %s Datenbytes\n",
             count_text,bytes_text);
   ,
     sprintf(askme->info_text,
          "Restored:                     %s file objects, %s data bytes\n",
             count_text,bytes_text);
   )
   Askme_info(askme,!(flag&(1<<16)));
 }
 if(askme->error_count>0) {
   Askme_de_eN(
     sprintf(askme->info_text,
             "------------------------------- %7.lf Fehler aufgetreten\n",
             askme->error_count);
   ,
     sprintf(askme->info_text,
             "----------------------------- %7.lf errors encountered\n",
             askme->error_count);
   )
   Askme_info(askme,!(flag&(1<<16)));
 }
 if(askme->found>=0) {
   Askme_de_eN(
     sprintf(askme->info_text,
             "=============================== %7d Treffer in %8.3f Sekunden\n",
             askme->found,tdiff);
   ,
     sprintf(askme->info_text,
             "============================= %7d matches in %9.3f seconds\n",
             askme->found,tdiff);
   )
   Askme_info(askme,0);
   askme->bar_is_fresh= 1;
 } else if((flag&(1<<17)) && !askme->bar_is_fresh) {
   Askme_de_eN(
     sprintf(askme->info_text,"===============================\n");
   ,
     sprintf(askme->info_text,"=============================\n");
   )
   Askme_info(askme,0);
   askme->bar_is_fresh= 1;
 }
 Askme_reset_counters(askme,1);
 return(ret);
}


int Askme_dialog(askme,flag)
struct AskmE *askme;
int flag;
{
 int ret,first_round;
 char line[2*4096];

 askme->is_dialog= 1;
 for(first_round= 1;1;first_round= 0) {
   if(askme->pending_option[0]!=0) {
     Askme_mark(askme,0);
     strcpy(line,askme->pending_option);
     askme->pending_option[0]= 0;
   } else {
     if(!askme->bar_is_fresh) {
       Askme_de_eN(
         sprintf(askme->info_text,"===============================\n");
       ,
         sprintf(askme->info_text,"=============================\n");
       )
       Askme_info(askme,0);
       askme->bar_is_fresh= 1;
     }
     Askme_de_eN(
       sprintf(askme->info_text,"Option oder Suchtext eingeben :\n");
     ,
       sprintf(askme->info_text,"enter option or search text :\n");
     )
     Askme_info(askme,0);
     Askme_mark(askme,0);
     ret= Askme_dialog_input(askme,line,sizeof(line),4);
     if(ret<=0)
 break;
   }
   sprintf(askme->info_text,
           "==============================================================\n");
   Askme_info(askme,0);

   ret= Askme_execute_option(askme,line,1<<17);
   if(ret<0)
     goto ex;
   if(ret==3)
 break;
   askme->did_something_useful= 1;
 }
 ret= 1;
ex:;
 askme->is_dialog= 0;
 return(ret);
}


int Askme_prescan_args(askme,argc,argv,flag)
struct AskmE *askme;
int argc;
char **argv;
int flag;
/*
 bit0= do not interpret argv[1]
*/
/*
 return:
  <0  error
   0  end program
   1  ok, go on
*/
{
 int i,ret,negative_mode= 0;
 char *cmd;

 for(i=1+(flag&1);i<argc;i++) {
   cmd= argv[i];
   if(i>1)
     askme->did_something_useful= 1;
   if(i==1) {
     if(strncmp(cmd,"-prog_help",10)==0 &&
        (cmd[10]==':' || cmd[10]==' ' || cmd[10]==0)) {
       strcpy(askme->progname,cmd+11);
       goto help;
     } else if(strcmp(cmd,"-help")==0) {
help:;
       Askme_help(askme,0);
       askme->did_something_useful= 1;
       return(0);
     } else if(strncmp(cmd,"-prog_hilfe",11)==0 && 
        (cmd[11]==':' || cmd[11]==' ' || cmd[11]==0)) {
       strcpy(askme->progname,cmd+12);
       goto hilfe;

     } else if(strcmp(cmd,"-hilfe")==0) {
hilfe:;
       Askme_hilfe(askme,0);
       askme->did_something_useful= 1;
       return(0);

     } else if(strncmp(cmd,"-resume_from_file",17)==0 && 
               (cmd[17]==':' || cmd[17]==' ')) {
       askme->prescan_state->info_source_enabled= 1;
       ret= Askme_options_from_file(askme,cmd+18,1);
       askme->prescan_state->info_source_enabled= 0;
       askme->prescan_state->execute_first_argument= 1;
       askme->prescan_state->is_resume_from_file= 1;
       if(ret<=0)
         return(-1);
       if(ret==3)
         return(0);

     } else if(strcmp(cmd,"-version")==0) {
       printf("scdbackup program : saskme %s\n",PROG_VERSION);
       askme->did_something_useful= 1;
       return(0);

     } else {
       /* do not interpret infosource address */;

     }
   } else if((!askme->prescan_state->language_set) 
             && strcmp(cmd,"-deutsch")==0) {
     askme->german_dialog= 1;
     askme->prescan_state->language_set= 1;

   } else if((!askme->prescan_state->language_set) 
             && strcmp(cmd,"-english")==0) {
     askme->german_dialog= 0;
     askme->prescan_state->language_set= 1;

   } else if(strncmp(cmd,"-info_source",12)==0 
             && (cmd[12]==':' || cmd[12]==' ')) {
      if(askme->prescan_state->info_source_enabled)
        strcpy(askme->filename,cmd+13);

   } else if(strncmp(cmd,"-logfile",8)==0 && (cmd[8]==':' || cmd[8]==' ')) {
     int hflag,channel_no= 0;
   
     if(strlen(cmd)<12) {
logfile_wrong_form:;
 continue;
     }
     hflag= 2;
     if(cmd[9]=='R')
       channel_no= 1;
     else if(cmd[9]=='I')
       channel_no= 2;
     else if(cmd[9]=='M')
       channel_no= 3;
     else if(cmd[9]=='.')
       hflag= 4;
     else
       goto logfile_wrong_form;
     if(hflag&2) {
       if(askme->prescan_state->logfile_set[channel_no])
 continue;
       askme->prescan_state->logfile_set[channel_no]= 1;
     }
     if(hflag&4) {
       if(askme->prescan_state->logfile_pkt_set)
 continue;
       askme->prescan_state->logfile_pkt_set= 1;
     }
     if(strcmp(cmd+11,"-")==0)
 continue;
     ret= Write_to_channel(cmd+11,channel_no,hflag);
     if(ret<=0)
 continue;

   } else if((!askme->prescan_state->media_mode_set)
             && strncmp(cmd,"-media_mode",11)==0 &&
             (cmd[11]==':' || cmd[11]==' ')) {
     if(cmd[12]=='-')
       negative_mode= 1;
     else if(cmd[12]!='+')
       askme->media_mode= 0;
     if(strstr(cmd+12,"mount_umount")!=NULL)
       Set_bit(&(askme->media_mode),2,negative_mode);
     if(strstr(cmd+12,"load_eject")!=NULL)
       Set_bit(&(askme->media_mode),4,negative_mode);
     askme->prescan_state->media_mode_set= 1;
   } else if(strcmp(cmd,"-no_dir_list")==0) {
     askme->no_dir_list= 1;

   } else if(strncmp(cmd,"-options_from_file",18)==0 &&
             (cmd[18]==':' || cmd[18]==' ')) {
     ret= Askme_options_from_file(askme,cmd+19,1);
     if(ret<=0)
       return(-1);
     if(ret==3)
       return(0);

   } else if((!askme->prescan_state->pkt_set)
             && strncmp(cmd,"-pkt_output",11)==0 &&
             (cmd[11]==':' || cmd[11]==' ' || cmd[11]==0)) {
     if(cmd[11]==0)
       askme->packet_output= 1;
     else if(strcmp(cmd+12,"off")==0)
       askme->packet_output= 0;
     else if(strcmp(cmd+12,"on")==0)
       askme->packet_output= 1;
     askme->prescan_state->pkt_set= 1;

   } else if(strncmp(cmd,"-protect",8)==0 && (cmd[8]==':' || cmd[8]==' ')) {
     ret= Fprotector_add(askme->protections,cmd+9,NULL,&(askme->protections),
                         NULL,0);
     if(ret<=0)
       return(ret);

   } else if((!askme->prescan_state->picky_set)
             && strncmp(cmd,"-read_pickiness",15)==0 &&
           (cmd[15]==':' || cmd[15]==' ')) {
     if(strcmp(cmd+16,"zero")==0)
       askme->read_pickiness= 0;
     else if(strcmp(cmd+16,"low")==0)
       askme->read_pickiness= 1;
     else if(strcmp(cmd+16,"high")==0)
       askme->read_pickiness= 2;
     else if(strcmp(cmd+16,"overdone")==0)
       askme->read_pickiness= 3;
     else {
       Askme_de_eN(
         sprintf(askme->info_text,
                 "%s: -read_pickiness: unbekannter Modus : %s\n",
                 askme->progname,cmd+16);
       ,
         sprintf(askme->info_text,"%s: -read_pickiness: unknown mode : %s\n",
                 askme->progname,cmd+16);
       )
       Askme_info(askme,0);
     }
     askme->prescan_state->picky_set= 1;

   } else if(strncmp(cmd,"-resume_from_file",17)==0 
             && (cmd[17]==':' || cmd[17]==' ')) {
     Askme_de_eN(
       sprintf(askme->info_text,
             "%s: -resume_from_file ist nur als wirksam als erstes Argument\n",
               askme->progname);
     ,
       sprintf(askme->info_text,
    "%s: -resume_from_file gets only into effect if given as first argument\n",
               askme->progname);
     )
     Askme_info(askme,0);

   } else if(strncmp(cmd,"-shadow",7)==0 && (cmd[7]==':' || cmd[7]==' ')) {
     ret= Fprotector_add(askme->protections,cmd+8,NULL,&(askme->protections),
                         NULL,2);
     if(ret<=0)
       return(ret);

   }
 }
 return(1);
}


#ifdef Askme_with_maiN

int main(argc,argv)
int argc;
char **argv;
{
 int ret,i,search_failed,found;
 char *cpt;
 struct AskmE *askme= NULL;

 if(argc<2) {
   fprintf(stderr,"usage : %s sourcefile|-help [options|filenames]\n",
           argv[0]);
   exit(2);
 }
 cpt= argv[1];
 if(strncmp(cpt,"-info_source:",13)==0)
   cpt+= 13;
 ret= Askme_new(&askme,argv[0],cpt,0);
 if(ret<=0) {
   fprintf(stderr,"Creation of AskmE object failed. (not enough memory ?)\n");
   exit(3);
 }
 i= 1;
 ret= Askme_prescan_args(askme,argc,argv,0);
 if(ret==0)
   goto end_sucessfully;
 if(ret<0)
   exit(5);

 ret= Askme_read_all(askme,0);
 if(ret<=0) {
   sprintf(askme->info_text,"Could not successfully read given infosource.\n");
   Askme_info(askme,0);
   sprintf(askme->info_text,"Infosource: %s\n",argv[1]);
   if(errno>0) {
     sprintf(askme->info_text+strlen(askme->info_text),
"Last system error encountered (but not necessarily related to the failure):\n"
            );
     sprintf(askme->info_text+strlen(askme->info_text),
             "   %s\n",strerror(errno));
   }
   Askme_info(askme,0);
   exit(4);
 }
 for(i=2-!!(askme->prescan_state->execute_first_argument);i<argc;i++) {
   ret= Askme_execute_option(askme,argv[i],0);
   if(ret<=0)
     exit(5);
   if(ret==3)
     goto end_sucessfully;
   
 }
 if(askme->dialog) {
   ret= Askme_dialog(askme,0);
   if(ret<=0)
     exit(6);
 }
end_sucessfully:;
 if(argc==2 && !(askme->prescan_state->is_resume_from_file)
    && !askme->did_something_useful) {
   fprintf(stderr,
        "Note: Program ended without any action after reading info source.\n");
   fprintf(stderr,
        "      Option  -dialog  would bring you into dialog mode.\n");
 }
 found= askme->found;
 search_failed= askme->search_failed;
 Askme_destroy(&askme,0);
 exit(search_failed>0 && found<=0);
}

#endif /* Askme_with_maiN */

